#ifndef MSIPDECOMP_H
#define MSIPDECOMP_H
// ----------------------------------------------------------------------------
// Class for decomposition based on marking structurally implicit places(MSIPs).
// MSIPs are places that can be added without changing the nets firing behavior.
// Here they are used to replace and approximate the original net.
// Rough explanation: 
// 1. The net is partitioned into modules defined by the subtransitions
//    on the top level page and their subnets
// 2. For each module, equivalence classes are computed consisting of 
//    all places and transitions (no interface transitions) that are weakly
//    connected.
// 3. For each equivalence class, one ore more MSIPs are computed via external
//    programs callinvars (invariants calculation, incidence vector of MSIP)
//    and lp_solve (linear program solver, marking of MSIP).
// 4. New nets are created and placed in a directory: A basic skeleton net, 
//    where all modules are aggregated (=replaced by MSIPs), one net per 
//    module Mi where all modules are aggregated except Mi, which is extended
//    (=MSIPs added to original subnet). 
// When the decomposition has been computed, you can doubleclick on the     
// subtransitions to see aggregated or extended subnets.You are allowed to move
// objects, other actions are blocked. Clicking the save button when viewing
// a subnet updates that subnet in all generated files where it occurs.
// The decomposition button pops up the window "Decomposition Options" 
// where you can choose aggregated or extended view. Clicking "Analyze" in the 
// window starts iterative approximate analysis: A file ".delays" with initial
// delays and some program flags is generated, and the external program
// masterrg is started.
// Please view "Strucured Solution of Asynchronously Communicating Stochastic 
// Modules, Javier Campos, Susanna Donatelli and Manuel Silva" for details on
// MSIP computation, and "On state space decomposition for the numerical
// analysis of stochastic Petri nets, Carlos Perez-Jimenez and Javier Campos"
// for details on iterative throughput approximation methods.  
// ----------------------------------------------------------------------------

#include "decomposition.h"
// ----------------------------------------------------------------------------
// defines
// ----------------------------------------------------------------------------
#define NETFILENAME   ".net"
#define TABLEFILENAME ".table"
#define OUTFILENAME   ".decomp"
#define DIREXTENSION "DIRM"
// ----------------------------------------------------------------------------
// class declarations
// ----------------------------------------------------------------------------
//list of edspnnodes(=unfolded transitions and places)
typedef LinkedList<EDSPNNode*> NodeList_T; 
//array of edspnnodes
typedef Array<int, EDSPNNode*> NodeArray_T;
// ----------------------------------------------------------------------------
//incidence matrix type
class Inc_Matrix_T
{
 public:
  Inc_Matrix_T(){ 
    places = NodeArray_T();
    transitions = NodeArray_T();
    rows = columns = 0;
    V=0;};
  ~Inc_Matrix_T(){
    if (V)
      {
        for (unsigned i = 0; i < rows; i++)
	    delete V[i];
	delete V;
      }
  };
  NodeArray_T places;      /*pointer to place node
			     for each matrix row*/
  NodeArray_T transitions; /*pointer to a transition node
			     for each matrix column*/
  unsigned int rows; /*number of rows*/
  unsigned int columns; /*number of columns*/
  int **V; /*integer array*/
}; 
// ----------------------------------------------------------------------------
// equivalence class type
class EClass_T
{
 public:
  EClass_T(){list = 0;
  list = new NodeList_T();
  matrix = new Inc_Matrix_T();
  new_nodes = new NodeList_T();
  };
  ~EClass_T(){
    delete list;
    delete matrix;
    delete new_nodes;
  }
  NodeList_T   *list; /*list of nodes that are in the eq class*/
  Inc_Matrix_T *matrix; /*small incidence matrix for the nodes*/
  NodeList_T *new_nodes; /*msips computed for this class*/
};

//list of equivalence classes
typedef LinkedList<EClass_T*> EClasslist_T;
//array of eqclasslist that is indexed by SubTrans
typedef Array<SubTrans*, EClasslist_T*> EClassArray_T;
// ----------------------------------------------------------------------------
// Main class
// "Marking structurally implicit places"-decomposition
// ----------------------------------------------------------------------------
class MSIPDecomp: virtual public Decomposition
{ 
  EClassArray_T eclassarray; /*for each subtransition
			       a linked list of equivalence classes,
			       each of which is a
			       linked list of EDSPNNodes
			     */
 protected:
  void StartDecomposition();
  void EndDecomposition();
  void OpenOptionsMenu();
  void OnDialogButton(Ident& button);
  void OnLogwinButton(Ident& button);

  void LoadDecomposition();
  void ComputeDecomposition();

  Text AnalysisParameters();
  void AnalyzeDecomposition(Struct* params);

  Net *CreateBasicSkeleton(Text bsname);
  Net *CreateLowLevelSystem(Text lowname, Text path);
  void CreateAggregation(SubTrans* S, EClasslist_T &sclasses,  Page *cprime);
  void CreateExtension(SubTrans* S, EClasslist_T &sclasses, Page *cprime);
  Position CalcMSIPPosition(EDSPNPlace *newnode);
  void AddMSIPS(EClasslist_T &s_classes, Page *subpage);
  void PutInputOutput(SubTrans *S, Page *subpage, bool bufferonly);
  void PutTPResults(SubTrans* S, Page *cprime);

  void FillIncMatrix(EClass_T *class_i);
  unsigned int GetMSIPMarking(int* postcol, int* precol, Text basefilename);
  void GetInvariants(EClass_T *class_i, Inc_Matrix_T* &invmatrix, Text basefilename);
  EDSPNNode* ComputeMSIP(SubTrans *S, EClass_T *class_i, int i,
			 Inc_Matrix_T *invmatrix, unsigned int invcol);
  void ComputeSIPS(SubTrans *S, EClasslist_T *s_classes);
 
  void FindPlace(unsigned int trans_nb, EClass_T *curr_class,
		 SubTrans* S);
  void FindTransition(unsigned int place_nb, EClass_T *curr_class,
		      SubTrans* S);
  bool IsInClasses(EClasslist_T &sclasses, EDSPNNode* node);
  void PrintClass(EClass_T *eclass);
  void PrintClasses(EClasslist_T &sclasses);
  void CalcClasses(SubTrans* S, EClasslist_T &sclasses);
 public:
  MSIPDecomp(Net& n): 
    Decomposition(n),
    CommModule(n,"MSIPDecomp")
    {
      dirname = netname + DIREXTENSION;
    }
  
  ~MSIPDecomp() {}
  bool Init() {return (net.Prime()->Class()->Name() == "HCPN");};   
};
#endif
