#include "object.h"
#include "node.h"
#include "connect.h"
#include "conncl.h"
#include "attrib.h"
#include "page.h"

#include "constants.h"

Connector::Connector(ConnectorClass* cc, Page* p, char* name):
	Object(p,cc,name,new Struct(cc->Attribute(),&p->GetNet()->Attribute())),
	conncl(cc),
	curr(0),
	input(0),
	output(0)
{
}


Connector::Connector(Connector& c, Page* parent):
	Object(c,parent,new Struct(*c.attr,&parent->GetNet()->Attribute())),
	conncl(c.conncl),
	input(0),
	output(0)
{
	if (parent)
	{
		if (c.input)
		{
			if ((input = parent->GetNode(c.input->ID())))
				input->AddOutput(this);
			else
				SetConnections(c.input->Copy(parent),output);
		}
		
		if (c.output)
		{
			if ((output = parent->GetNode(c.output->ID())))
				output->AddInput(this);
			else
				SetConnections(input,c.output->Copy(parent));
		}
	}

	Iterator<Position*> iter(c.pos);
	for (Position* p=iter; p; p=++iter)
		pos.Append(new Position(*p));
}


void Connector::SetClass(ConnectorClass* cl)
{
	if (conncl!=cl)
	{
		conncl = cl;
		Struct* oldattr = attr;
		attr = new Struct(cl->Attribute(),&page->GetNet()->Attribute());
		if (oldattr)
		{
			*attr = *oldattr;
			delete oldattr;
		}
		
		Object::SetClass(cl);
	}	
}


Connector::~Connector()
{
	if (attr) delete attr;
	
	Iterator<Position*> iter(pos);
	for (Position* mp=iter; mp; mp=++iter) delete mp; 
	SetConnections(0,0);
}


Connector* Connector::Copy(Page* parent)
{
	return new Connector(*this,parent);
}


void Connector::SetConnections(Node* in, Node* out)
{
	if (input==in && output==out) return;
	
	if (input) input->RemoveOutput(this);
	if (output) output->RemoveInput(this);
	input = in; output = out;
	
	if (input) input->AddOutput(this);
	if (output) output->AddInput(this);
  
  if (input && pos.Length()<1) InsertPos(0, input->Mid());
  if (output && pos.Length()<2) InsertPos(1, output->Mid());
}


Position* Connector::InsertPos(int p, const Position psn) 
{
	Position* np;
	pos.Insert(np = new Position(psn),p);
	
	return np;
}


void Connector::RemovePos(const Position psn)
{
	Iterator<Position*> iter(pos);
	for (Position* p=iter; p; p=++iter)
		if (*p==psn)
		{
			pos.Remove(p);
			delete p;
			if (int(pos)<2)
				delete this;
			else
				page->Update();
			return;
		}
}


void Connector::Move(const Position rel)
{
	Iterator<Position*> iter(pos);
	
	for (Position* p=iter; p; p=++iter)
		if ((p!=iter.First() || input->Selected())
		&& (p!=iter.Last() || output->Selected()))
			*p = *p+rel;
}


Position Connector::LocateLink(Position& npos, bool inp)
{
	Position nearest,best; double dist,bestd = 1e300;
	Node* bestob = 0;
	Iterator<Node*> iter = page->Nodes();
	
	for (Node* ob=iter; ob; ob=++iter)
		if ((dist = ob->Nearest(npos,nearest))<bestd
		&& (!output || // Connector wird gerade erzeugt!
		(inp && conncl->AllowedConnection(ob,output))
		|| !inp && conncl->AllowedConnection(input,ob)))
		{
			bestd = dist; best = nearest; bestob = ob;
		}
	
	if (!bestob) return best;
	if (inp)
		SetConnections(bestob,output);
	 else
	 	SetConnections(input,bestob);
	return best;
}


void Connector::MoveInput(const Position rel)
{
	Position* p = *pos.First();
	*p = *p+rel;
}


void Connector::Rearrange()
{
	Iterator<Position*> iter(pos);

  if (input) *iter.First() = input->Mid();
  if (output) *iter.Last() = output->Mid();
  
	for (Position* p=iter; p; p=++iter)
	{
		if (iter.Index()==1 && input)
			*iter.First() = input->RearrangePoint(*p);
	
		if (iter.Index()==int(iter)-2 && output)
			*iter.Last() = output->RearrangePoint(*p);
	}
}



void Connector::MoveOutput(const Position rel)
{
	Position* p = *pos.Last();
	*p = *p+rel;
}


bool Connector::Event(Display* dpy, EventInfo& ev)
{
	Position *first = *pos.First(),*last = *pos.Last();
	
	switch (ev.event)
	{
	case EventDouble:
	case EventInit:
		if (ev.button==ButtonSelect || ev.event==EventDouble)
		{
		  //AH Changed: Remove unnecessary test
		  //if (IsUnder(dpy,ev.pos))
			{
				double bd = 1e300,d;
		
				curr = 0;
				Iterator<Position*> iter(pos);
				for (Position* b=iter; b; b=++iter)
					if ((d = b->Distance(ev.pos)) <= bd)
					{
						bd = d;
						curr = b;
					}
	
				if (!curr) break;
				
				if (ev.event==EventDouble && curr!=first && curr!=last)
				{
					RemovePos(*curr);
					return false;
				}
	
				return true;
			}
		}
		else // ev.button == ButtonCreate
		{
			if (IsUnder(dpy,ev.pos))
			{
				curr = Insert(ev.pos);
				return true;
			} 
			else
			{
				curr = 0;
				break;
			}
		}
		break;
	
	case EventMotion:
		if (curr)
		{
			*curr = ev.pos;
      if (*curr!=*first && *curr!=*last) Rearrange();
			return true;
		}
		else break;

	case EventQuickExit:
	case EventExit:
		if (!curr) break;
		if (curr==first)
			*curr = LocateLink(ev.pos,true);
		else if (curr==last)
			*curr = LocateLink(ev.pos,false);
		else
			*curr = ev.pos;
		
		curr = 0;
		return true;
	}
	
	return Object::Event(dpy, ev);
}


Position* Connector::Insert(const Position psn)
{
	double bestd = 1e300,d;
	Iterator<Position*> iter(pos);
	Position* p1 = iter,*p2;
	int cnt = 1,bestc = 1;
		
	while ((p2 = ++iter))
	{
                if ((d = p1->Distance(psn)+p2->Distance(psn)-p1->Distance(*p2))<bestd)
		{	bestd = d; bestc = cnt; }
		cnt++;
		p1 = p2;
	}
	
	return InsertPos(bestc,psn);
}


Position Connector::Mid()
{
	double length = 0;
	Iterator<Position*> iter(pos);
	Position* p1 = iter,* p2 = ++iter,* last = *pos.Last();
	for (; p2; p1=p2,p2=++iter) length += p1->Distance(*p2); 

	double dist = 0, clen = length/2;
	
	for (p1=iter,p2=++iter; p2; p1=p2,p2=++iter)
		if ((clen -= (dist = p1->Distance(*p2)))<=0.0 || p2==last)
			break;
	
	if (dist==0) dist = 1;
	
	return *p1 + (*p2-*p1)*((clen+dist)/dist);
}


istream& Connector::Input(istream& is)
{
	Ident id(false);
	unsigned nid;
	Scope sn(is);
	
	while (is && sn.GetId(id))
	{
		if (id=="input")
		{
			if (is>>nid && (input = page->GetNode(nid)))
				input->AddOutput(this);
		}
		else if (id=="output")
		{
			if (is>>nid && (output = page->GetNode(nid)))
				output->AddInput(this);
		}
		else if (id=="positions")
		{
			Scope sp(is);
			while (is && sp.Get())
			{
				Position* p = new Position;
				if (is>>*p) pos.Append(p);
			}
		}
		else Object::Input(is, id);
	}
	
	return is;
}


ostream& Connector::Output(ostream& os)
{
	Object::Output(os);
	
	if (input) os<<"; input = "<<input->ID();
	if (output) os<<"; output = "<<output->ID();
	
	os<<"; positions = (";
	Iterator<Position*> iter(pos);
	for (Position *p=iter; p; )
	{
		os<<*p;
		if ((p = ++iter)) os<<',';
	}
	os<<')';
	
	return os<<')';
}


ShapeData Connector::GetShapeData(double zoom, bool)
{
	Iterator<Position*> riter(relatives);
	Iterator<Position*> piter(pos);
	return ShapeData(zoom,Mid(),scale,0,piter,riter,false);
}


ExprData Connector::GetExprData(bool mono, bool draft)
{
	return ExprData(*attr,name,selected,!attr,draft,mono);
}


