#include "expr.h"
#include "value.h"
#include "text.h"
#include "float.h"
#include "struct.h"
#include "context.h"

extern "C" {
#include <string.h>
#include <ctype.h>
}

/*
the BNF (jede Regel eine Funktion):
qop = or [? qop : or]
or = and {| and}
and = comp {& comp} 
comp = expr [>|< |=  expr]
expr = term | {+|- term}
term = factor | {*|/ factor}
factor = [!|-] ref | const | ( qop )
ref = (@ ident  [qop] | {. ident-const}) | # objref
objref = name | selected | virtual | inputs | outputs ...
const = "text-const" | ident-const | fp-number
ident = a-zA-Z_ {a-zA-Z0-9_}
*/

class ExpressionNode
{
public:
	virtual ~ExpressionNode() {}
	virtual Value* Evaluate(ExprData&) = 0;
	virtual ostream& Output(ostream&) = 0;	
};

typedef ExpressionNode* ExpressionNodeP;

inline ostream& operator <<(ostream& os, ExpressionNode& e)
{
	return e.Output(os);
}

	
class Constant: public ExpressionNode
{
	Value* val;
public:
	Constant(Value* v): val(v) {}
	~Constant() { if (val) delete val; }
	Value* Evaluate(ExprData&) { return val; }
	ostream& Output(ostream& os) { return val?(os<<*val):os; }
};
	
class Reference: public ExpressionNode
{
	Ident ref;
	bool objref;
	Value* result;
public:
	Reference(Ident& r, bool obref): ref(r),objref(obref),result(0) {}
	~Reference() { if (result) delete(result); }
	Value* Evaluate(ExprData&);
	ostream& Output(ostream& os) { return objref ? (os<<'#'<<ref) : (os<<ref);	}
};

class Operation: public ExpressionNode
{
	ExpressionNode* left;
	ExpressionNode* right;
	ExpressionNode* middle;
	char op;
	Value* result;
public:
	Operation(ExpressionNodeP l,char o,ExpressionNodeP r,ExpressionNodeP m = 0): 
		left(l),right(r),middle(m),op(o),result(0) {}
	~Operation();
	Value* Evaluate(ExprData&);
	ostream& Output(ostream& os);
};


Operation::~Operation()
{
	if (left) delete left;
	if (right) delete right;
	if (middle) delete middle;
	if (result) delete result;
}

Value* Operation::Evaluate(ExprData& para)
{
	Value* r = 0;
	if (result) delete result;
	result = 0;

	Value* lval = 0,* rval = 0;
	if (left) lval = left->Evaluate(para);

	if (op!='?')
	{
		if (right)	rval = right->Evaluate(para);
		if (!rval) return 0;
	}

	if (!lval) // Unaere Operation ?
		switch (op)
		{
		case '-': return result = rval->Neg();
		case '!': return result = rval->Not();
		case '.': r = para.attr.Idx(*rval); break;
		}
	else
		switch (op)
		{
 		//AH Changed Start => .to_int()
 		case '?': return ((!lval || (*lval).to_int())?middle:right)->Evaluate(para);
 		//AH Changed End
		case '=': return result = lval->Eq(*rval);
		case '>': return result = lval->Gt(*rval);
		case '<': return result = lval->Lt(*rval);
		case '+': return result = lval->Add(*rval);
		case '-': return result = lval->Sub(*rval);
		case '*': return result = lval->Mul(*rval);
		case '/': return result = lval->Div(*rval);
		case '&': return result = lval->And(*rval);
		case '|': return result = lval->Or(*rval);
		case '[': 
		//AH Changed Start => .to_int()
 		case '.': r = lval->Idx((*rval).to_int()); break;
 	        //AH Changed End
		}
	if (r) result = r->Cast();
	if (result==r) result = 0; else r = result;
	return r;
}	

ostream& Operation::Output(ostream& os)
{
	os<<'(';
	if (left) os<<*left;
	os<<op;
	if (right) os<<*right;
	if (op=='[') os<<']';
	return os<<')';
}

Value* Reference::Evaluate(ExprData& para)
{
	if (result) delete result;
	result = 0;
	
	if (objref)
	{
		if (ref=="name") return result = new Text(para.name);
		else if (ref=="selected") return result = new Float(para.sel);
		else if (ref=="virtual") return result = new Float(para.virt);
		else if (ref=="draft") return result = new Float(para.draft);
		else if (ref=="mono") return result = new Float(para.mono);
 		//AH Changed Start =>.to_charPtr()
 		else return para.context?para.context->GetVar(ref.to_charPtr()):0;
 		//AH Changed End
	}
	else
		return &ref;
}

// BNF Rules:

ExpressionNode* Expression::QOp(istream& is)
{
	char op; ExpressionNode* left = Or(is);
	if (!left) return 0;
	if (is>>op && op=='?')
	{
		ExpressionNode* middle = QOp(is);
		if (!middle) { delete left; return 0; }
		if (is>>op && op==':')
		{
			ExpressionNode* right = Or(is);
			if (!right) { delete left; delete middle; return 0; }
			return new Operation(left,'?',right,middle);
		}
		delete middle;
	}
	is.putback(op);
	return left;
}

ExpressionNode* Expression::Or(istream& is)
{
	char op; ExpressionNode* curr = And(is);
	if (!curr) return 0;
	while (is>>op && op=='|')
	{
		ExpressionNode* next = And(is);
		if (!next) { delete curr; return 0; }
		curr = new Operation(curr,op,next);
	}
	is.putback(op);
	return curr;
}

ExpressionNode* Expression::And(istream& is)
{
	char op; ExpressionNode* curr = Comp(is);
	if (!curr) return 0;
	while (is>>op && op=='&')
	{
		ExpressionNode* next = Comp(is);
		if (!next) { delete curr; return 0; }
		curr = new Operation(curr,op,next);
	}
	is.putback(op);
	is.clear();
	return curr;
}

ExpressionNode* Expression::Comp(istream& is)
{
	char op; ExpressionNode* curr = Expr(is);
	if (!curr) return 0;
	if (is>>op && (op=='<' || op=='=' || op=='>'))
	{
		ExpressionNode* next = Expr(is);
		if (!next) { delete curr; return 0; }
		curr = new Operation(curr,op,next);
	}
	else is.putback(op);
	return curr;
}

ExpressionNode* Expression::Expr(istream& is)
{
	char op; ExpressionNode* curr = Term(is);
	if (!curr) return 0;
	while (is>>op && (op=='+' || op=='-'))
	{
		ExpressionNode* next = Term(is);
		if (!next) { delete curr; return 0; }
		curr = new Operation(curr,op,next);
	}
	is.putback(op);
	is.clear();
	return curr;
}

ExpressionNode* Expression::Term(istream& is)
{
	char op; ExpressionNode* curr = Factor(is);
	if (!curr) return 0;
	while (is>>op && (op=='*' || op=='/'))
	{
		ExpressionNode* next = Factor(is);
		if (!next) { delete curr; return 0; }
		curr = new Operation(curr,op,next);
	}
	is.putback(op);
	is.clear();
	return curr;
}

ExpressionNode* Expression::Factor(istream& is)
{
	char uop = 0,ch = 0; ExpressionNode* curr = 0;
	if (is>>uop && (uop=='-' || uop=='!')) // unaere Operatoren
	{}
	else
	{
		is.putback(uop);
		uop = 0;
	}

	if (is>>ch && ch=='(')
	{
		if (!(curr = QOp(is)) || !(is>>ch) || ch!=')')
		{
			delete curr;
			is.putback(ch);
			return 0;
		}
	}
	else
	{
		is.putback(ch);
		curr = (ch=='@' || ch=='#') ? Ref(is) : Const(is);
	}

	if (uop) curr = new Operation(ExpressionNodeP(0),uop,curr);
	return curr;
}


ExpressionNode* Expression::Const(istream& is)
{
	Value* val = 0; char c;
	if (is>>c)
	{
		is.putback(c);
		if (isdigit(c))
		{
			double d;
			if (is>>d) val = new Float(d);
		}
		else if (isalpha(c))
		{
			Ident id(false); is>>id;
			if (id=="true") val = new Float(double(1));
			else if (id=="false") val = new Float(double(0));
 			else val = new Text(id.to_charPtr());
		}
		else if (c=='"')
		{
			val = new Text;
			if (!(is>>*val)) delete val;
		}		

		if (!is || !val) return 0;
	}
	return new Constant(val);
}


ExpressionNode* Expression::Ref(istream& is)
{
	ExpressionNode* curr = 0;
	char c;
	if (is>>c)
	{
		if (c=='@')
		{
			Ident attref;
			if (is>>attref)
			{
				curr = new Operation(ExpressionNodeP(0),'.',new Reference(attref,false));
				is>>c;
				switch (c)
				{
					case '[':
						ExpressionNode* index;
						if ((index = QOp(is)))
						 	if (is>>c && c==']')
								return new Operation(curr,'[',index);
							else
							{
								delete curr;
								is.putback(c);
							}
						return 0;
					case '.':
						while (c=='.')
						{
							if (is>>attref)
								curr = new Operation(curr,'.',new Reference(attref,false));
							else break;
							is>>c;
						}
					default:
						is.putback(c);
				}
			}
		}
		else // if (c=='#')
		{
			Ident objref(false);
			if (is>>objref)
				curr = new Reference(objref,true);
		}
	}
	return curr;
}


Expression::~Expression()
{
	if (root) delete root;
}


Value* Expression::Evaluate(ExprData& ob)
{
  Value* r;
  if(result)
    delete result;
  result = 0;
  r = root?(root->Evaluate(ob)):0;
  if(r)
    result = r->Cast();
  return result;
}


istream& Expression::Input(istream& is)
{
	if (root) delete root;
	root = QOp(is);
	return is;
}


ostream& Expression::Output(ostream& os)
{
	return root?(os<<*root):os;
}
