#include "page.h"
#include "node.h"
#include "connect.h"
#include "object.h"
#include "position.h"
#include "constants.h"
 
Node::Node(NodeClass* const nc, Page* pg, char* name, Position p, double ang):
	Object(pg,nc,name,new Struct(nc->Attribute(),&pg->GetNet()->Attribute())),
	nodecl(nc),
	pos(p),
	angle(ang),
	real(0),
	subpage(0)
{
	if (nc->SubstitutionNode())
		subpage = new Page(pg->GetNet(),nc->SubClass(),this);
}


void Node::SetReal(Node* r)
{
	if (real)
	{
		real->RemoveAlias(this);
		attr = 0;
	}
	else
		if (attr && r)
			delete attr;
	
	if ((real = r))
	{
		real->AddAlias(this);
		attr = real->attr;
		id = real->id;
		
  	if (real->nodecl!=nodecl)
			Object::SetClass((nodecl = real->nodecl));
	}
	else
	{
		if (!attr)
			attr = new Struct(nodecl->Attribute());	
	}

	//AH Changed: use index-notation instead of iterator
	for (int i=0; i<aliases.Length(); i++)
	  {
	    (*aliases[i])->SetReal(this);
	  }
	//AH Changed end
}


Node::Node(Node& n, Page* parent, bool portcopy):
	Object(n,parent, 
		portcopy ? 0 : new Struct(*n.attr,&parent->GetNet()->Attribute())),
	nodecl(n.nodecl),
	pos(n.pos),
	angle(n.angle),
	real(0),
	subpage(0)
{
	if (portcopy)
		SetReal(&n);
	else
		// Kopie eines Portnode
		if (n.real && parent->SuperPage())
			SetReal(parent->SuperPage()->GetNode(id));
	
	if (!n.real)
		if (nodecl->SubstitutionNode())
			subpage = new Page(parent->GetNet(),nodecl->SubClass(),this);
}

//AH New: remove node and all virtual aliases
void Node::RemoveNodeAndAliases()
{//same as destructor
	Node *n;
	Iterator<Node*> iter(aliases);
	while ((n = iter)) 
	  delete n; //different
	delete this;
}


Node::~Node()
{
	if (real)
		real->RemoveAlias(this);

	Node *n;
	Iterator<Node*> iter(aliases);
	while ((n = iter)) {	
	  n->SetReal(0);
	  *n->attr=*attr;	  
	}

  if (!real && attr) delete attr;
	
	if (subpage)
		delete subpage;

	// Connectoren entfernen sich dabei autom. aus der inputs/outputs
	// Liste dieses (und anderer) Nodes!
	Connector* c;
	
	Iterator<Connector*> citer(inputs);
	while ((c = citer)) delete c;
	citer = outputs;
	while ((c = citer)) delete c;
}


void Node::SetClass(NodeClass* cl)
{
	if (real)
		real->SetClass(cl);
	else // der reale Knoten
		if (nodecl!=cl)
		{
			nodecl = cl;
			// ...wechselt seine Attribute
			Struct* oldattr = attr;
			attr = new Struct(cl->Attribute(),&page->GetNet()->Attribute());
			
			// ubernimmt ggf vorhandene mit gleichem namen
			if (oldattr)
			{
				*attr = *oldattr;
				delete oldattr;
			}
      
			// ndert die Unternetzklasse
      if (subpage)
      {
        if (cl->SubClass()) subpage->SetClass(cl->SubClass());
        else delete subpage; 
      }
			// .. gibt alles an die Portnodes weiter
			Iterator<Node*> iter(aliases);
			for (Node* n=iter; n; n=++iter)
				n->SetReal(this);
			
			Object::SetClass(cl);
		}	
}


void Node::Renumber()
{
	id = real?real->id:NewID();
}


Node* Node::Copy(Page* parent)
{
	return new Node(*this,parent);
}


Node* Node::CreateAlias(Page* parent)
{
	Node* v = new Node(*this,parent,true);
	return v;
}

Node* Node::GetAlias(Page* parent)
{
	Iterator<Node*> iter(aliases);
	for (Node* n=iter; n; n=++iter)
		if (n->GetPage()==parent) return n;
	
	return 0;
}


void Node::AddInput(Connector* c)
{
	inputs.Append(c);
}


void Node::RemoveInput(Connector* c)
{	
	inputs.Remove(c);
}


void Node::AddOutput(Connector* c)
{
	outputs.Append(c);
}


void Node::RemoveOutput(Connector* c)
{
	outputs.Remove(c);
}


void Node::AddAlias(Node* n)
{
	aliases.Append(n);
}


void Node::RemoveAlias(Node* n)
{
	aliases.Remove(n);
}


bool Node::IsConnected(Node* n)
{
	Iterator<Connector*> iter(inputs);
	Connector* c;
	
	for (c=iter; c; c=++iter)
		if (c->GetInput()==n) return true;

	for (iter=outputs,c=iter; c; c=++iter)
		if (c->GetOutput()==n) return true;

	return false;
}


void Node::UpdateHierarchy()
{
	// Nur auf Substitution-Nodes anwendbar
	if (subpage && subpage->Class()==page->Class())
	{
		// 1. Verbundene Knoten ggf. in der Unterseite einblenden
		//    (nach dem Ziehen einer virt. Kante)
		
		Node* n,* r;
		Iterator<Connector*> iter(inputs);
		Connector* c;
		
		for (c=iter; c; c=++iter)
		{
			if ((n = c->GetInput()))
				if (!n->GetAlias(subpage))
					n->CreateAlias(subpage);
		}
		for (iter=outputs,c=iter; c; c=++iter)
		{
			if ((n = c->GetOutput()))
				if (!n->GetAlias(subpage))
					n->CreateAlias(subpage);
		}

		// 2. Unverbundene Port-Knoten ausblenden
		//    (nach dem Loeschen einer virt. Kante)
		Iterator<Node*> niter = subpage->Nodes();
		for (n=niter; n; )
			if ((r = n->Real()) && !IsConnected(r))
			{
				delete n;
				n = niter;
			}
			else n = ++niter;
	}
}

		
void Node::UpdateVirtualConnections()
{
	if (subpage && subpage->Class()==page->Class())
	{
		// 3. virt. Kanten in der Ueberseite entfernen.
		//    (nach dem Loeschen eines Portnodes)
		
		Node* n;
		Iterator<Connector*> iter(inputs);
		Connector* c;
		
		for (c=iter; c; )
			if ((n = c->GetInput()) && !n->GetAlias(subpage))
			{
				delete c;
				c = iter;
			}
			else c = ++iter;

		for (iter=outputs,c=iter; c; )
			if ((n = c->GetOutput()) && !n->GetAlias(subpage))
			{
				delete c;
				c = iter;
			}
			else c = ++iter;
	}
}


void Node::RearrangeConnections()
{
	Iterator<Connector*> iter(inputs);
	Connector* c;
	
	for (c=iter; c; c=++iter)
		if (!c->Selected()) c->Rearrange();
	for (iter=outputs,c=iter; c; c=++iter)
		if (!c->Selected()) c->Rearrange();
}


void Node::Move(const Position rel)
{
	pos = pos+rel;
  RearrangeConnections();
}


bool Node::Meltable(Node* n)
{
	return n->page==page && n->nodecl==nodecl 
		&& (!real || !n->real) && !subpage && !n->subpage;
}


void Node::Melt(Node* n) // this-Node ueberlebt
{
	if (!Meltable(n))
		return;

	// Verbindungen umhaengen
	Iterator<Connector*> iter(n->inputs);
	Connector* c;
	
	while ((c = iter))
	{
		c->SetConnections(c->GetInput(),this);
		c->Rearrange();	
	}
	
	iter = n->outputs;
	while ((c = iter))
	{
		c->SetConnections(this,c->GetOutput());
		c->Rearrange();	
	}
	
	// Aliase auf this umbiegen
	Iterator<Node*> aiter(n->aliases);
	Node *a;
	
	while ((a = aiter))
		a->SetReal(this);
	
	delete n;
}


istream& Node::Input(istream& is)
{
	Ident iid(false);
	Scope sn(is);
	while (is && sn.GetId(iid))
	{
		if (iid=="position")
			is>>pos;
		else if (iid=="orientation")
		{
			if (!(is>>angle)) cerr<<"angle expected!"<<endl;
		}
		else if (iid=="portnode")
		{
			is>>iid;
			if (iid=="true")
				SetReal(page->SuperPage()->GetNode(id));
		}
		else Object::Input(is, iid);
	}
	return is;
}

ostream& Node::Output(ostream& os)
{
	Object::Output(os)<<"; position = "<<pos<<"; orientation = "<<angle;
	if (real) os<<"; portnode = true";
	return os<<')';
}

ShapeData Node::GetShapeData(double zoom, bool)
{
	Iterator<Position*> riter(relatives);
	return ShapeData(zoom,pos,scale,angle,riter,riter,true);
}

ExprData Node::GetExprData(bool mono, bool draft)
{
	return ExprData(*attr,name,selected,real!=0,draft,mono);
}


double Node::Nearest(Position& ps, Position& nearest) 
{
	ShapeData sd = GetShapeData(1,true);
	return nodecl->Nearest(sd,ps,nearest);
}


Position Node::RearrangePoint(Position& ps)
{
	ShapeData sd = GetShapeData(1,true);
	if (nodecl->Align()) return nodecl->PointToMid(sd,ps);
	Position nearest;
	nodecl->Nearest(sd,ps,nearest);
	return nearest;
}

