/*************************************************************************
*
*  Datei         : load_net.c
*
*  Aufgabe       : Parsiert das Dateiformat von .TN und erstellt das interne
*                  Datenformat. Applikationen koennen dann entweder direkt mit
*                  diesen Daten arbeiten, oder eine Konvertierung in die eigenen
*                  Datenstrukturen vornehmen.
*                                                                          
*  Programmierer : Ulrich Hertlein
*
*  Datum         : 17. Mai 1994
*
*  Modifiziert   : 21. Juni 1994   Joachim Schneider
*                  show_net( NetDescr *net );
*
**************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#include <sys/file.h>
#include <string.h>
#include <malloc.h>

#include "netio.h"
#include "load_funcs.h"


/*
 * General definitions
 */
#define FILE_BUFLEN 2048
#define BUFLEN      2048


/*
 * KeyWord typedef
 */
typedef struct
{
   char *name;
   int (*func)();
   int call_cnt;
} KeyWord;


/*
 * List of keywords to scan for
 */
KeyWord keylist[] = 
{
   { "NET_TYPE:",           f_nettype },
   { "DESCRIPTION:",        f_descr },
   { "PLACES:",             f_nplaces },
   { "TRANSITIONS:",        f_ntrans },
   { "DELAY_PARAMETERS:",   f_ndelaypar },
   { "MARKING_PARAMETERS:", f_nmarkpar },
   { "REWARD_MEASURES:",    f_nrewards },

   { "DELAYPAR_DEF",        f_delaypar_def }, /* Muss vor DELAYPAR! */

   { "MARKPAR",             f_markpar },
   { "PLACE",               f_place },
   { "DELAYPAR",            f_delaypar },
   { "TRANSITION",          f_trans },

   { "WEIGHT",              f_delay },
   { "EXP_DELAY",           f_delay },
   { "DET_DELAY",           f_delay },
   { "GEN_DELAY",           f_delay },
   { "FUNCTION",            f_md_enable },
   { "CARDINALITY",         f_md_arc_mult },
   { "MEASURE",             f_reward }
};
int nkeywords = sizeof(keylist) / sizeof(KeyWord);


/*
 * DefaultError:
 */
int DefaultError(msg, fatal)
char *msg;
int fatal;
{
   if( fatal )
   {
      fprintf(stderr,"Fatal: %s\n", msg);
      exit( fatal );
   }
   else
      fprintf(stderr,"Warning: %s\n", msg);

   return( 0 );
}


/*
 * Free: free with NULL-check
 */
void Free(ptr)
void *ptr;
{
   if( ptr != NULL )
      free(ptr);
}


/*
 * nMalloc: allocate a given size of bytes, return pointer to that area;
 *          exit() on error
 */
void *nMalloc(size)
unsigned size;
{
   register void *m;

   if( !size )
      return( NULL );

   if( (m = (void *)malloc(size)) == NULL )
   {
      fprintf(stderr,"Unable to malloc %u bytes\n", size);
      exit( -1 );
   }
   memset((char *)m,0,size);

   return( m );
}


/*
 * nRealloc: reallocate a given pointer to a given size, return pointer to that
 *          new area; if pointer is null, allocate fresh memory;
 *          exit() on error
 */
void *nRealloc(ptr, size)
char *ptr;
unsigned size;
{
   register void *np;

   if( ptr )
   {
      if( !(np = (void *)realloc(ptr,size)) )
      {
         fprintf(stderr,"Unable to realloc %u bytes\n", size);
         exit( -1 );
      }
   }
   else
      np = (void *)nMalloc(size);

   return( np );
}


/*
 * nStrdup: duplicates a string; exit()s on error
 */
char *nStrdup(str)
char *str;
{
   char *m;

   if( (m = (char *)strdup(str)) == NULL )
   {
      fprintf(stderr,"Unable to duplicate string (%s)\n", str);
      exit( -1 );
   }

   return( m );
}


/*
 * find_place: find place with a given name; returns pointer to that place,
 *             else NULL
 */
Place *find_place(net, name)
NetDescr *net;
String name;
{
   register int i;

   for(i=0;i<net->num_places;i++)
      if( net->place[i].name && !strcmp(net->place[i].name,name) )
         return &net->place[i];

net->error_func("Place not found",TRUE);
   if(i == net->num_places) /* keine Stelle gefunden */
   {
      net->error_func("Place not found",TRUE);
      return NULL;
   }
}


/*
 * find_trans: find transition with a given name; returns pointer to that
 *             transition, else NULL
 */
Transition *find_trans(net, name)
NetDescr *net;
String name;
{
   register int i;

   for(i=0;i<net->num_trans;i++)
      if( net->trans[i].name && !strcmp(net->trans[i].name,name) )
         return &net->trans[i];

   if(i == net->num_trans) /* keine Transition gefunden */
   {
      net->error_func("Transition not found",TRUE);
      return NULL;
   }
}


/*
 * find_delaypar:
 */
Delaypar *find_delaypar(net, name)
NetDescr *net;
String name;
{
   register int i;

   for(i=0;i<net->num_delaypars;i++)
      if( net->delaypar[i].name && !strcmp(net->delaypar[i].name,name) )
         return &net->delaypar[i];

   if(i == net->num_delaypars) /* kein Delayparameter gefunden */
   {
      net->error_func("Delaypar not found",TRUE);
      return NULL;
   }
}


/*
 * find_markpar:
 */
Markpar *find_markpar(net, name)
NetDescr *net;
String name;
{
   register int i;

   for(i=0;i<net->num_markpars;i++)
      if( net->markpar[i].name && !strcmp(net->markpar[i].name,name) )
         return &net->markpar[i];

   if(i == net->num_markpars) /* kein Markierungsparameter gefunden */
   {
      net->error_func("Markpar not found",TRUE);
      return NULL;
   }
}


/*
 * skip_white: skip over spaces and/or tabs; returns pointer to first non-white
 */
char *skip_white(buf)
char *buf;
{
   /* skip over spaces and tabs */
   for( ; *buf == ' ' || *buf == '\t'; buf++ )
      ;

   return( buf );
}



/*
 * find_file: return a filepointer to the given filename
 */
static FILE *find_file(net, filename)
NetDescr *net;
char *filename;
{
   char filepath[FILE_BUFLEN];
   FILE *fp;

   char *env, *pathcpy;

   /* append .TN if necessary */
   if( !strstr(filename, ".TN") )
      sprintf(filepath,"%s.TN", filename);
   else
      strcpy(filepath, filename);

   /* open file to see if it's a absolute filename or in the current directory */
   if( fp = fopen(filepath,"r") )
      return( fp );

   /* assume it's in MODELDIR/filename */
   if( !(env = getenv("MODELDIR")) )
   {
      net->error_func("'MODELDIR' not set", FALSE);
      return( NULL );
   }
   pathcpy = nStrdup(filepath);
   sprintf(filepath,"%s/%s", env,pathcpy);
   Free(pathcpy);

   /* open file with complete MODELDIR/filename */
   if( fp = fopen(filepath,"r") )
      return( fp );

   /* we *really* did all we could ... */

   return( NULL );
}

/*
 * fgetline: get one line from a stream, strip newline, returns non-zero if
 *           end-of-file is reached; buf must be of size BUFLEN
 */
static int fgetline(fp, buf)
FILE *fp;
char *buf;
{
   register int len;

   /* read line, strip newline */
   fgets(buf,BUFLEN,fp);

   if( (len = strlen(buf)) )
      buf[--len] = '\0';

   return( feof(fp) );
}


/*
 * make_bullet_proof: check net for unresolved references, unused entries etc.
 */
static NetDescr *make_bullet_proof(net)
NetDescr *net;
{
   int i, used;
   Transition *t;

   char buf[256];

   /* check number of places & transitions */
   if( net->num_places == -1 || net->num_trans == -1 )
   {
      net->error_func("Invalid number of places or transitions", TRUE);

      free_net(net);
      return( NULL );
   }

   /* check places */
   for( used = 0, i = 0; i < net->num_places; i++ )
      if( net->place[i].name )
         used++;
   if( used != net->num_places )
   {  
      sprintf(buf,"%d places announced, but only %d used", net->num_places,used);
   	  net->error_func(buf, FALSE);
      net->num_places = used;
   }
   
   /* check transition references */
   for( used = 0, i = 0; i < net->num_trans; i++ )
   {
      t = &net->trans[i];

      if( t->name )
         used++;

      switch( t->delay_type )
      {
      case MARKDEP:
      case PMFDEF:
         if( !t->delay.md_delay )
         {
            sprintf(buf,"Unresolved MD/PMF-reference in transition '%s'", t->name);
            net->error_func(buf, TRUE);
			return net;
         }
         break;
      }
   }
   if( used != net->num_trans )
   {
      sprintf(buf,"%d transitions announced, but only %d used", net->num_trans,used);
      net->error_func(buf, FALSE);
	  net->num_trans = used;
   }

   /* check delay parameters */
   for( used = 0, i = 0; i < net->num_delaypars; i++ )
      if( net->delaypar[i].name )
         used++;
   if( used != net->num_delaypars )
   {
      sprintf(buf,"%d delay parameters announced, but only %d used", net->num_delaypars,used);
      net->error_func(buf, FALSE);
	  net->num_delaypars = used;
   }

   /* check marking parameters */
   for( used = 0, i = 0; i < net->num_markpars; i++ )
      if( net->markpar[i].name )
         used++;
   if( used != net->num_markpars )
   {
      sprintf(buf,"%d marking parameters announced, but only %d used", net->num_markpars,used);
      net->error_func(buf, FALSE);
	  net->num_markpars = used;
   }

   /* check rewards */
   for( used = 0, i = 0; i < net->num_rewards; i++ )
      if( net->reward[i].name )
         used++;
   if( used != net->num_rewards )
   {
      sprintf(buf,"%d rewards announced, but only %d used", net->num_rewards,used);
      net->error_func(buf, FALSE);
	  net->num_rewards = used;
   }

   return( net );
}


/*
 * load_net: parses .TN file and builds internal data structures
 */
NetDescr *load_net(netname, error_func)
char *netname;
int (*error_func)();
{
   FILE *fp;
   int done = 0;

   NetDescr *net;
   char linebuf[BUFLEN], *buffer = linebuf, funcbuf[102400];
   int i, key = -1;

   /* allocate net object */
   net = (NetDescr *)nMalloc(sizeof(NetDescr));
   net->netname = (char *)nStrdup(netname);
   if( error_func )
      net->error_func = error_func;
   else
      net->error_func = DefaultError;

   /* set defaults */
   net->num_places = -1;
   net->num_trans = -1;
   net->num_delaypars = -1;
   net->num_markpars = -1;
   net->num_rewards = -1;
   
   /* initialize call-counters */
   for( i = 0; i < nkeywords; keylist[i++].call_cnt = 0)
      ;

   /* if no netname is given, read STDIN */
   if( netname )
   {
      /* find file */
      if( (fp = find_file(net,netname)) == (FILE *)NULL )
      {
         sprintf(linebuf,"Unable to open file: %s\n", netname);
         net->error_func(linebuf, TRUE);
         free_net(net);

         return( NULL );
      }
   }
   else
      fp = stdin;

   /* read entire file */
   done = feof(fp);
   while( !done )
   {
      /* break loop on eof */
      if( !(done = fgetline(fp,linebuf)) )
      {
         /* skip empty lines & comments */
         buffer = (char *)skip_white(linebuf);
         if( *buffer == '\0' || !strncmp(buffer,"--", 2) )
            continue;

         /* match keywords */
         for( i = 0; i < nkeywords; i++ )
            if( !strncmp(buffer,keylist[i].name,strlen(keylist[i].name)) )
               break;
      }

      /* keyword found? */
      if( i != nkeywords || done )
      {
         if( key != -1 )
         {
            /* call keyword function */
            keylist[key].func( net,
                               funcbuf + strlen(keylist[key].name),
                               keylist[key].call_cnt++);
         }
         key = i;
         strcpy(funcbuf,buffer);
      }
      else
      {
         if( key == -1 )
            fprintf(stderr,"Syntax error: %s\n", buffer);
         else
         {
            strcat(funcbuf,"\n");
            strcat(funcbuf,buffer);
         }
      }
   }

   /* close file */
   if( netname )
      fclose(fp);

   /* check result of read */
   return( make_bullet_proof(net) );
}


/*
 * show_net: shows all data associated with the given network description
 */
void show_net(net)
NetDescr *net;
{
  int i;
  
  printf( "netname:\t\t\"%s\"\n",
    net->netname);
  printf( "nettype:\t\t\"%sSPN\"\n",
    net->type==GSPN?"G":
    net->type==DSPN?"D":
    net->type==EDSPN?"ED":
    net->type==CDSPN?"CD":
    "???");
  printf( "comment:\t\t\"%s\"\n",
    net->comment);
  printf( "error_func:\t\t\"%s\"\n",
    net->error_func != DefaultError ? "OWN" : "DEFAULT");
net->error_func( "I'm error_func.\n", FALSE );  


  printf( "\n%d markpars:\n", net->num_markpars );
  for(i=0;i<net->num_markpars;i++) {
    printf( "\t%2d.name\t\t\t\"%s\"\n",
      i, net->markpar[i].name);
    printf( "\t  .pos\t\t\t(%lg;%lg)\n",
      net->markpar[i].pos.x, net->markpar[i].pos.y);
    printf( "\t  .value\t\t\t<%d>\n",
      net->markpar[i].value);
  } /* for */



  printf( "\n%d places:\n", net->num_places );
  for(i=0;i<net->num_places;i++) {
    printf( "\t%2d.name\t\t\t\"%s\"\n",
      i, net->place[i].name);
    printf( "\t  .pos\t\t\t(%lg;%lg)\n",
      net->place[i].pos.x, net->place[i].pos.y);
    printf( "\t  .name_pos\t\t(%lg;%lg)\n",
      net->place[i].name_pos.x, net->place[i].name_pos.y);
    if( net->place[i].is_value )
      printf( "\t  .marking.value\t<%d>\n",
        net->place[i].marking.value);
    else
      printf( "\t  .marking.mark_par\t\"%s\"\n",
        net->place[i].marking.markpar->name);
  } /* for */
  


  printf( "\n%d delaypars:\n", net->num_delaypars );
  for(i=0;i<net->num_delaypars;i++) {
    printf( "\t%2d.name\t\t\t\"%s\"\n",
      i, net->delaypar[i].name);
    printf( "\t  .pos\t\t\t(%lg;%lg)\n",
      net->delaypar[i].pos.x, net->delaypar[i].pos.y);
    if( net->delaypar[i].is_value )
      printf( "\t  .delay.value\t\t<%lg>\n",
        net->delaypar[i].delay.value);
    else
      printf( "\t  .delay.md_delay\t\"%s\"\n",
        net->delaypar[i].delay.md_delay);
  } /* for */



  printf( "\n%d transitions:\n", net->num_trans );
  for(i=0;i<net->num_trans;i++) {
    int a;
    Transition *t=&net->trans[i];
    
    printf( "\n\t%2d.name\t\t\t\"%s\"\n",
      i, t->name);
    printf( "\t  .pos\t\t\t(%lg;%lg)\n",
      t->pos.x, t->pos.y);
    printf( "\t  .name_pos\t\t(%lg;%lg)\n",
      t->name_pos.x, t->name_pos.y);
    printf( "\t  .delay_pos\t\t(%lg;%lg)\n",
      t->delay_pos.x, t->delay_pos.y);
    printf( "\t  .priority\t\t(%d)\n",
      t->priority);
    printf( "\t  .type\t\t\t\"%s\"\n",
      t->type==IMM?"IMM":
      t->type==EXP?"EXP":
      t->type==DET?"DET":
      t->type==GEN?"GEN":
      "???");
    printf( "\t  .delay_type\t\t\"%s\"\n",
      t->delay_type==DELAY?"DELAY":
      t->delay_type==DELAYPAR?"DELAYPAR":
      t->delay_type==MARKDEP?"MARKDEP":
      t->delay_type==PMFDEF?"PMFDEF":
      "???");
    printf( "\t  .orient\t\t\"ORIENT_%s\"\n",
      t->orient==ORIENT_0?"0":
      t->orient==ORIENT_45?"45":
      t->orient==ORIENT_90?"90":
      t->orient==ORIENT_M45?"M45":
      "???");
    printf( "\t  .inf_server\t\t\"%s\"\n",
      t->inf_server?"YES":"NO");
    printf( "\t  .f_policy\t\t\"R%s\"\n",
      t->f_policy==RS?"S":
      t->f_policy==RA?"A":
      t->f_policy==RE?"E":
      "???");

    if( t->type == IMM )
      printf( "\t  .md_enable\t\t\"%s\"\n",
        t->md_enable);
    
    if( t->delay_type == DELAY )
      printf( "\t  .delay.value\t\t\"%lg\"\n",
        t->delay.value);
    else if( t->delay_type == DELAYPAR )
      printf( "\t  .delay.delay_par\t\"%s\"\n",
        t->delay.delay_par->name);
    else
      printf( "\t  .delay.md_delay\t\t\"%s\"\n",
        t->delay.md_delay);

    printf( "\t   %d inarcs:\n", t->num_inarcs );
    for(a=0;a<t->num_inarcs;a++) {
      int c;
      printf( "\t\t%2d.place\t\t\"%s\"\n",
        a, t->in_arc[a].place->name);
      printf( "\t\t  .point (%d)\t\t",
        t->in_arc[a].num_pts );
      for(c=0;c<t->in_arc[a].num_pts;c++) {
        printf( "(%lg;%lg) ",
          t->in_arc[a].point[c].x, t->in_arc[a].point[c].y);
      } /* for */
      if(t->in_arc[a].is_value)
        printf( "\n\t\t  .multipl.value\t<%d>\n",
          t->in_arc[a].multipl.value);
      else
        printf( "\t\t  .multipl.md_mult\t\"%s\"\n",
          t->in_arc[a].multipl.md_mult);
    } /* for */
    
    printf( "\t   %d outarcs:\n", t->num_outarcs );
    for(a=0;a<t->num_outarcs;a++) {
      int c;
      printf( "\t\t%2d.place\t\t\"%s\"\n",
        a, t->out_arc[a].place->name);
      printf( "\t\t  .point (%d)\t\t",
        t->out_arc[a].num_pts );
      for(c=0;c<t->out_arc[a].num_pts;c++) {
        printf( "(%lg;%lg) ",
          t->out_arc[a].point[c].x, t->out_arc[a].point[c].y);
      } /* for */
      if(t->out_arc[a].is_value)
        printf( "\n\t\t  .multipl.value\t<%d>\n",
          t->out_arc[a].multipl.value);
      else
        printf( "\t\t  .multipl.md_mult\t\"%s\"\n",
          t->out_arc[a].multipl.md_mult);
    } /* for */
    
    printf( "\t   %d inharcs:\n", t->num_inharcs );
    for(a=0;a<t->num_inharcs;a++) {
      int c;
      printf( "\t\t%2d.place\t\t\"%s\"\n",
        a, t->inh_arc[a].place->name);
      printf( "\t\t  .point (%d)\t\t",
        t->inh_arc[a].num_pts );
      for(c=0;c<t->inh_arc[a].num_pts;c++) {
        printf( "(%lg;%lg) ",
          t->inh_arc[a].point[c].x, t->inh_arc[a].point[c].y);
      } /* for */
      if(t->inh_arc[a].is_value)
        printf( "\n\t\t  .multipl.value\t<%d>\n",
          t->inh_arc[a].multipl.value);
      else
        printf( "\t\t  .multipl.md_mult\t\"%s\"\n",
          t->inh_arc[a].multipl.md_mult);
    } /* for */
  } /* for */
    

  printf( "%d rewards:\n", net->num_rewards );
  for(i=0;i<net->num_rewards;i++) {
    printf( "\t%2d.name\t\t\t\"%s\"\n",
      i, net->reward[i].name);
    printf( "\t  .def\t\t\t\"%s\"\n",
      net->reward[i].def);
  } /* for */
}


/*
 * free_net: frees all data associated with the given network description
 */
void free_net(net)
NetDescr *net;
{
   int i;
  
   Free(net->netname);
   Free(net->comment);
  
   for(i=0;i<net->num_markpars;i++)
      Free(net->markpar[i].name);
   Free(net->markpar);  
  
   for(i=0;i<net->num_places;i++)
      Free(net->place[i].name);
   Free(net->place);
  
   for(i=0;i<net->num_delaypars;i++)
   {
      Free(net->delaypar[i].name);
      if(!net->delaypar[i].is_value)
         Free(net->delaypar[i].delay.md_delay);
   }
   Free(net->delaypar);  

   for(i=0;i<net->num_trans;i++)
   {
      int a;
      Transition *t=&net->trans[i];
    
      Free(t->name);
    
      if(t->delay_type==MARKDEP)
         Free(t->delay.md_delay);

      if(t->type==IMM)
         Free(t->md_enable);

      for(a=0;a<t->num_inarcs;a++)
      {
         if(!t->in_arc[a].is_value)
            Free(t->in_arc[a].multipl.md_mult);
         if(t->in_arc[a].num_pts)
            Free(t->in_arc[a].point);
      }
      Free(t->in_arc);
    
      for(a=0;a<t->num_outarcs;a++)
      {
         if(!t->out_arc[a].is_value)
            Free(t->out_arc[a].multipl.md_mult);
         if(t->out_arc[a].num_pts)
            Free(t->out_arc[a].point);
      }
      Free(t->out_arc);
    
      for(a=0;a<t->num_inharcs;a++)
      {
         if(!t->inh_arc[a].is_value)
            Free(t->inh_arc[a].multipl.md_mult);
         if(t->inh_arc[a].num_pts)
            Free(t->inh_arc[a].point);
      }
      Free(t->inh_arc);
   }
    
   for(i=0;i<net->num_rewards;i++)
   {
      Free(net->reward[i].name);
      Free(net->reward[i].def);
   }
   Free(net->reward);  
  
   Free(net);
}


#ifdef  MAIN

/*
 * ef: own error_func
 * shows poss. errors, 'cause arguments does not match
 */
void  ef(str)
char *str;
{
  fprintf( stderr, "OWN_error_func: %s", str );
  exit( 1 );
} /* ef */


/*
 * main: provide dummy interface for test purposes
 */
int main(argc, argv)
int argc;
char **argv;
{
  NetDescr  *net = NULL;

  if( argc <= 1 ) {
    fprintf( stderr, "Usage: %s <filename>\n", argv[0] );
    return( 1 );
  } /* if */

  printf( "\nTEST: load_net( %s, NULL )\n\n", argv[1] );
  net = load_net(argv[1],NULL);
  show_net( net );
  free_net( net );


  printf( "\n\nnext test with <ENTER> ... " );
  getchar();
  printf( "\n\n\nTEST: load_net( %s, ef )\n\n", argv[1] );
  net = load_net(argv[1],ef);
  show_net( net );
  free_net( net );

  return( 0 );
}
#endif MAIN
