
/***************************************************************************

 File:		expoly.c

 Purpose:	routines for manipulating expolynomial distributions
		and for computing the factors needed by the
		randomization procedure

 Programmer:	Reinhard German

 History:	Release 1: June 2, 1993	
                Release ?: October 1997 Heiko Bergmann
***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "pmfExpoly.h"

/* malloc with error-check: */
char *Malloc(size)
unsigned size;
{
  char *p;
  if (!(p = malloc(size))) {
    fprintf(stderr, "ERROR: can't malloc %d bytes!\n",size);
    exit(1);
  };
  return p;
} /* Malloc */

/*****************************************************************************/
/*avoid 0^0 -Problem                                                         */
/*****************************************************************************/
double hb_pow(a,b)
     double a,b;
{
  if ((a==0.0) && (b==0.0))
    return (1.0);
  return pow(a,b);
}
/*****************************************************************************/
/*Romberg Integration of Normal Distribution; Numerical Recipes in C p.140   */
/*****************************************************************************/
#define EPS 1.0e-07
#define JMAX 20
#define JMAXP (JMAX+1)
#define K 5
#define NR_END 1
#define FREE_ARG char*
/* #define PI 3.141592654 */
/* #define NEG_INFINITY -999.0 */
/* #define CDF_NORMAL 1 */
/* #define EXPECT_NORMAL 2 */
/* #define EXPECT_WEIBULL 3 */

double *vector(long nl, long nh)
{
  double *v;
  
  v=(double *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(double)));
  if (!v) fprintf(stdout,"allocation failure in vector()");
  return v-nl+NR_END;
}

void free_vector(double *v, long nl, long nh)
{
  free((FREE_ARG) (v+nl-NR_END));
}

void polint(double xa[], double ya[], int n, double x, double *y, double *dy)
{
  int i,m,ns=1;
  double den,dif,dift,ho,hp,w;
  double *c,*d;
  dif=fabs(x-xa[1]);
  c=vector(1,n);
  d=vector(1,n);
  for (i=1; i<=n; i++) {
    if ( (dift=fabs(x-xa[i])) < dif) {
      ns=i;
      dif=dift;
    }
    c[i]=ya[i];
    d[i]=ya[i];
  }
  *y=ya[ns--];
  for (m=1;m<n;m++) {
    for (i=1; i<=n-m;i++) {
      ho=xa[i]-x;
      hp=xa[i+m]-x;
      w=c[i+1]-d[i];
      if ( (den=ho-hp) == 0.0) fprintf(stdout,"Error in routine polint");
      den=w/den;
      d[i]=hp*den;
      c[i]=ho*den;
    }
    *y += (*dy=(2*ns < (n-m) ? c[ns+1] : d[ns-1]));
  }
  free_vector(d,1,n);
  free_vector(c,1,n);
}

double funk_wert(double x, double sigma, double mu, int mode)
{
#ifdef NORDEBUG
  fprintf(stdout,"\nNormalwert(%lg)=%lg",x,1.0/(sigma*sqrt(2.0*PI))*exp(-hb_pow(x-mu,2.0)/(2.0*hb_pow(sigma,2.0))) );
#endif
  switch(mode)
  {
    case CDF_NORMAL : 
      return  1.0/(sigma*sqrt(2.0*PI))*exp(-hb_pow(x-mu,2.0)/(2.0*hb_pow(sigma,2.0))); 
      break;
    case EXPECT_NORMAL :
      return x*( 1.0/(sigma*sqrt(2.0*PI))*exp(-hb_pow(x-mu,2.0)/(2.0*hb_pow(sigma,2.0)))); 
      break;
    case EXPECT_WEIBULL :
      return x*((mu/sigma)*hb_pow(x/sigma,mu-1)*exp(-hb_pow(x/sigma,mu)));
	break;
  }
  return -1.0; /*error*/
}

double trapzd(double a, double b, int n, double sigma, double mu, int mode)
{
  double x,tnm,sum,del;
  static double s;
  int it,j;
  
  if (n == 1) {
    return (s=0.5*(b-a)*(funk_wert(a,sigma,mu,mode)+funk_wert(b,sigma,mu,mode)));
  } else {
    for (it=1, j=1; j<n-1; j++) it <<= 1;
    tnm=it;
    del=(b-a)/tnm;
    x=a+0.5*del;
    for (sum=0.0,j=1;j<=it;j++,x+=del) sum += funk_wert(x,sigma,mu,mode);
    s=0.5*(s+(b-a)*sum/tnm);
    return s;
  }
}

double qromb( double a, double b , double sig, double m, int mode)
{
  double ss,dss;
  double s[JMAXP+1],h[JMAXP+1];
  int j;
  
  h[1]=1.0;
  for (j=1;j<=JMAX;j++) {
    s[j]=trapzd(a,b,j,sig,m,mode);
#ifdef INTDEBUG
      fprintf(stdout,"\ns(%d)=%lg",j,s[j] );
#endif
    if (j>= K) {
      polint(&h[j-K],&s[j-K],K,0.0,&ss,&dss);
     if (fabs(dss) <=EPS*fabs(ss)) return ss;
    }
    s[j+1]=s[j];
    h[j+1]=0.25*h[j];
  }
  fprintf(stdout,"To many steps in Integration routine. MODE : %s\n",mode==1?"CDF_NORMAL":mode==2?"EXPECT_NORMAL":"EXPECT_WEIBULL");
  exit(1);
  return 0.0;
}



/* set deterministic distribution */
time_l *set_det_expoly(tau)
double tau;
{
  time_l *expoly, *return_expoly;
  power_m *power;
  expo_n *expo;

  return_expoly=(time_l *) Malloc(sizeof(time_l));
  expoly = return_expoly;

  if(tau>0)
  {
      expoly->type = DETERMINISTIC;
      expoly->x1 = 0.0;
      expoly->x2 = tau;
      expoly->p = 0.0;
      expoly->power = (power_m *) Malloc(sizeof(power_m));
        power = expoly->power;
        power->m = 0;
        power->expo = (expo_n *) Malloc(sizeof(expo_n));
          expo = power->expo;
          expo->lambda = 0.0;
          expo->b = 1.0;
          expo->c = 0.0;
          expo->next = NULL;
        power->next = NULL;
      expoly->next = (time_l *) Malloc(sizeof(time_l));
      expoly = expoly->next;
  }
  expoly->type = DETERMINISTIC;
  expoly->x1 = tau;
  expoly->p = 1.0;
  expoly->power = (power_m *) Malloc(sizeof(power_m));
    power = expoly->power;
    power->m = 0;
    power->expo = (expo_n *) Malloc(sizeof(expo_n));
      expo = power->expo;
      expo->lambda = 0.0;
      expo->b = 0.0;
      expo->c = 0.0;
      expo->next = NULL;
    power->next = NULL;
  expoly->next = NULL;

  return return_expoly;
} /* set_det_expoly */


/* set uniform distribution */
time_l *set_uniform_expoly(t1, t2)
double t1, t2;
{
  time_l *expoly, *return_expoly;
  power_m *power;
  expo_n *expo;

  return_expoly=(time_l *) Malloc(sizeof(time_l));
  expoly = return_expoly;

  if(t1>0)
  {
      expoly->type = UNIFORM;
      expoly->x1 = 0.0; 
      expoly->x2 = t1; 
      expoly->p = 0.0;
      expoly->power = (power_m *) Malloc(sizeof(power_m));
        power = expoly->power;
        power->m = 0;
        power->expo = (expo_n *) Malloc(sizeof(expo_n));
          expo = power->expo; 
          expo->lambda = 0.0;
          expo->b = 1.0;
          expo->c = 0.0;
          expo->next = NULL;
        power->next = NULL;
      expoly->next = (time_l *) Malloc(sizeof(time_l));
      expoly = expoly->next;
  }
  expoly->type = UNIFORM;
  expoly->x1 = t1;
  expoly->x2 = t2;
  expoly->p = 0.0;
  expoly->power = (power_m *) Malloc(sizeof(power_m));
    power = expoly->power;
    power->m = 0;
    power->expo = (expo_n *) Malloc(sizeof(expo_n));
      expo = power->expo;
      expo->lambda = 0.0;
      expo->b = 1+t1/(t2-t1); 
      expo->c = 1/(t2-t1);
      expo->next = NULL;
    power->next = (power_m *) Malloc(sizeof(power_m));
    power = power->next;
    power->m = 1;
    power->expo = (expo_n *) Malloc(sizeof(expo_n));
      expo = power->expo;
      expo->lambda = 0.0;
      expo->b = -1/(t2-t1);
      expo->c = 0.0;
      expo->next = NULL;
    power->next = NULL;
  expoly->next = (time_l *) Malloc(sizeof(time_l));
  expoly = expoly->next;

  expoly->type = UNIFORM;
  expoly->x1 = t2;
  expoly->p = 0.0;
  expoly->power = (power_m *) Malloc(sizeof(power_m));
    power = expoly->power;
    power->m = 0;
    power->expo = (expo_n *) Malloc(sizeof(expo_n));
      expo = power->expo;
      expo->lambda = 0.0;
      expo->b = 0.0;
      expo->c = 0.0;
      expo->next = NULL;
    power->next = NULL;
  expoly->next = NULL;

  return return_expoly;
} /* set_uniform_expoly */


/* set weibull distribution */
time_l *set_weibull(t1, t2, sigma, theta)
double t1, t2, sigma, theta;
{
  time_l *expoly, *return_expoly;
  
  

  return_expoly=(time_l *) Malloc(sizeof(time_l));
  expoly = return_expoly;

  if(t1>0)
  {
    expoly->type = WEIBULL;
    expoly->x1 = 0.0; 
    expoly->x2 = t1; 
    expoly->p = 0.0;
    expoly->sigma = 0.0;
    expoly->mu = 0.0;
    expoly->power = NULL;
    expoly->next = (time_l *) Malloc(sizeof(time_l));
    expoly = expoly->next;
  }
  expoly->type = WEIBULL;
  expoly->x1 = t1;
  expoly->x2 = t2;
  expoly->p = 1-exp(-hb_pow(t1/sigma,theta)); /*Vorlaufwahrscheinlichkeit*/
  expoly->sigma = sigma;
  expoly->mu = theta;
  expoly->power =  NULL;
  expoly->next = (time_l *) Malloc(sizeof(time_l));
  expoly = expoly->next;

  expoly->type = WEIBULL;
  expoly->x1 = t2;
  expoly->p = exp(-hb_pow(t2/sigma,theta)); /*Restwahrscheinlichkeit*/   
  expoly->sigma = 0.0;
  expoly->mu = 0.0;
  expoly->power = NULL;
  expoly->next = NULL;

  return return_expoly;
} /* set_weibull */

/* set normal distribution */
time_l *set_normal(t1, t2, sigma, mu)
double t1, t2, sigma, mu;
{
  time_l *expoly, *return_expoly;
  
  

  return_expoly=(time_l *) Malloc(sizeof(time_l));
  expoly = return_expoly;

  if(t1>0)
  {
    expoly->type = NORMAL;
    expoly->x1 = 0.0; 
    expoly->x2 = t1; 
    expoly->p = 0.0;
    expoly->sigma = 0.0;
    expoly->mu = 0.0;
    expoly->power = NULL;
    expoly->next = (time_l *) Malloc(sizeof(time_l));
    expoly = expoly->next;
  }
  expoly->type = NORMAL;
  expoly->x1 = t1;
  expoly->x2 = t2; /*Vorlaufwahrscheinlichkeit*/
  expoly->p = qromb(NEG_INFINITY, t1, sigma, mu, CDF_NORMAL);
  expoly->sigma = sigma;
  expoly->mu = mu;
  expoly->power =  NULL;
  expoly->next = (time_l *) Malloc(sizeof(time_l));
  expoly = expoly->next;

  expoly->type = NORMAL;
  expoly->x1 = t2;        /*Restwahrscheinlichkeit*/
  expoly->p = 1.0-qromb(NEG_INFINITY, t2, sigma, mu, CDF_NORMAL); 
  expoly->sigma = 0.0;
  expoly->mu = 0.0;
  expoly->power = NULL;
  expoly->next = NULL;

  return return_expoly;
} /* set_normal */


/* print an expolynomial distribution: */
void print_expoly(expoly)
  time_l *expoly;
{
  power_m *power;
  expo_n *expo;
  unsigned l;

  for (l = 0; expoly; expoly = expoly->next, l++) 
  {
    printf("---------------------------------------------------------------\n"); 
    printf("interval %d: ",l);
    printf("\tx1=%.16g, ", expoly->x1);
    if (expoly->next)
      printf("x2=%.16g, ", expoly->x2);
    else 
      printf("x2=infinity, ");
    printf("p=%.16g\n", expoly->p);
    for (power = expoly->power; power; power = power->next)
    {
      printf("power m=%d:", power->m); 
      for (expo = power->expo; expo; expo = expo->next)
      {
	  if(expo == power->expo)
	      printf("\tlambda=%.16g, b=%.16g, c=%.16g\n",
		     expo->lambda, expo->b, expo->c);
	  else
	      printf("          \tlambda=%.16g, b=%.16g, c=%.16g\n",
		     expo->lambda, expo->b, expo->c);
    
      }
    }
  }
  printf("---------------------------------------------------------------\n"); 
} /* print_expoly */



/****************************************************************************

function:	scan_expoly

 input:		fname: name of file which contains expoly distribution
		return_expoly: pointer to expoly distribution
 
 output:	return_expoly points to the expoly distrinution 

 purpose:	scans an expolynomial distribution from an ascii file
		and converts it to the internal data structure
		The expected file format is the format of the output of
		print_expoly.

***************************************************************************/
int scan_expoly(fname, return_expoly)
char *fname;
time_l **return_expoly;
{
  FILE *fp;
  time_l *expoly;
  power_m *power=0;
  expo_n *expo=0;
  char line[80];
  double x1=0.0, x2=0.0, p=0.0, lambda=0.0, b=0.0, c=0.0;
  unsigned l, m;
  enum {initial, read_int, read_expo} status = initial;

  if (!(fp=fopen(fname,"r"))) {
    fprintf(stderr, "ERROR: can't open file %s!\n", fname);
    return 0; 
  }; /* if */
  
  *return_expoly = (time_l *) Malloc(sizeof(time_l));
  expoly = *return_expoly;

  /* read lines of input file and scan them: */
  while (fgets(line, 80, fp)) {
    if (sscanf(line, "interval %d:\tx1=%lf, x2=%lf, p=%lf", &l, &x1, &x2, &p)) {
      if (status==read_int)
	return 0;
      else if (status==read_expo) {
        expoly->next = (time_l *) Malloc(sizeof(time_l));
	expoly = expoly->next;
      }; /* else if status */
      expoly->x1 = (double) x1;
      expoly->x2 = (double) x2;
      expoly->p = (double) p;
      expoly->next = NULL;
      status = read_int;
    } /* if scanf */
    else if (sscanf(line, "power m=%d:\tlambda=%lf, b=%lf, c=%lf", 
                    &m, &lambda, &b, &c)==4) {
      if (status==read_int) { 
        expoly->power = (power_m *) Malloc(sizeof(power_m));
        power = expoly->power;
      } /* if status */
      else if (status==read_expo) {
        power->next = (power_m *) Malloc(sizeof(power_m));
        power = power->next;
      } /* else if status */
      else return 0;
      power->m = m;
      power->expo = (expo_n *) Malloc(sizeof(expo_n));
      expo = power->expo;
      expo->lambda = (double) lambda;
      expo->b = (double) b;
      expo->c = (double) c;
      expo->next = NULL;
      power->next = NULL;
      status = read_expo;
    } /* else if scanf */
    else if (sscanf(line, "\tlambda=%lf, b=%lf, c=%lf", &lambda, &b, &c)) {
      if (status==read_expo) {
        expo->next = (expo_n *) Malloc(sizeof(expo_n));
        expo = expo->next;
      }
      else return 0;
      expo->lambda = (double) lambda;
      expo->b = (double) b;
      expo->c = (double) c;
      expo->next = NULL;
    }; /* else if scanf */
  }; /* while */
  
  fclose(fp);
  if (status!=read_expo)
    return 0;
  else
    return 1;

} /* scan_expoly */



/* print the computed factors: */
void print_factors(omega_factors, psi_factors, R, omega_error, psi_error)
  double *omega_factors,
	 *psi_factors,
	 omega_error,
	 psi_error;
  unsigned R;
{
  double x=0,y=0;
  unsigned k;
  printf("\n");
  printf("---------------------------------------------------------------\n"); 
  printf("*** print omega and psi factors, R=%d: ***\n", R);
  printf("omega_error=%.16f, psi_error=%.16f\n", omega_error, psi_error);
  printf("---------------------------------------------------------------\n"); 
  for(k=0; k <= R; k++) {
    x+=omega_factors[k];
    y+=psi_factors[k];
    printf("k=%d, omega_factor=%.16f, psi_factor=%.16f\n", 
	   k, omega_factors[k], psi_factors[k]);
  };
  printf("---------------------------------------------------------------\n"); 
  printf("omega_approx=%.16f, psi_approx=%.16f\n",x,y);
  printf("---------------------------------------------------------------\n"); 
} /* print_factors */



void poisson_prob(k, qx, beta)
unsigned k;
double qx, **beta;
{
  if (!k) {
    *beta = (double *) Malloc(sizeof(double));
    **beta = exp(-qx);
  }
  else
    **beta *= qx/k;
} /* poisson_prob */



/**************************************************************************

 function: 	calculate_factors

 input:		expoly: internal representation of expolynomial distribution
		q: 1.02* max. abs. diag. entry of generator matrix
		omega_factors: vector of factors for transient probs 
		psi_factors: vector of factors for sojourn times 
    		R : right truncation point

 output:	values in omega_factors, psi_factors, and R 

 purpose:	calculation of the factors needed for each iteration of
		the randomization procedure

**************************************************************************/
void calculate_factors(expoly, q, omega_factors, psi_factors, R,
		       omega_error, psi_error)
time_l *expoly;
double q;
double **omega_factors;
double **psi_factors;
unsigned *R;
double *omega_error;
double *psi_error;
{
  power_m *power;
  expo_n *expo;
  double x1, x2=0, p, lambda, b, c,
         qx1, qpl, qplx1, qplx2,
         alpha1, alpha2,
         *beta1, *beta2,
         gamma,
         kappa1, kappa2,
	 temp, diff,
         exact1, exact2,
         error1, error2,
         psi_exact = 0;
  unsigned h, k, m;

  if (!(*omega_factors=(double *) calloc((MAXINDEX+1), sizeof(double)))) {
    fprintf(stderr, "ERROR: can't alloc for omega_factors!\n");
    exit(1);
  };
  if (!(*psi_factors=(double *) calloc((MAXINDEX+1), sizeof(double)))) {
    fprintf(stderr, "ERROR: can't alloc for psi_factors!\n");
    exit(1);
  };
  
  /* for all time intervals of the expolynomial distribution: */
  for ( ; expoly; expoly=expoly->next) {
    x1 = expoly->x1;
    if (expoly->next) x2 = expoly->x2;
    p = expoly->p;

    /* calculate factors corresponding to point mass: */
    if (p>0.0) {
      qx1 = q * x1;
      kappa1 = 1.0;
      for (k=0; (kappa1>EPSILON) && (k<=MAXINDEX); k++) {
	poisson_prob(k, qx1, &beta1);
	kappa1 -= *beta1;
	(*omega_factors)[k] += p*(*beta1);
      }; /* for k */	
      if (k-1>*R) *R = k-1;
    }; /* if p>0.0 */
    if (!expoly->next) break;
     
    /* for all power terms: */
    for (power=expoly->power; power; power=power->next) {
      m = power->m;
      /* for all exponential terms: */
      for (expo=power->expo; expo; expo=expo->next) {
	lambda = expo->lambda;
	b = expo->b;
	c = expo->c;

	/* compute exact sums of alpha-factors: */
	if (lambda==0.0) {
          exact1 = pow(x1, (double) (m+1)) / (m+1);
	  exact2 = pow(x2, (double) (m+1)) / (m+1);
        } /* if lambda==0.0 */
	else {
	  kappa1 = 1.0; 
	  kappa2 = 1.0;
	  temp = 1.0/lambda;
          for (h=0; h<=m; h++) {
	    poisson_prob(h, lambda*x1, &beta1);
	    poisson_prob(h, lambda*x2, &beta2);
	    kappa1 -= *beta1;
	    kappa2 -= *beta2;
	    if (h) temp *= h/lambda;
          }; /* for h */
	  exact1 = temp*kappa1;
	  exact2 = temp*kappa2;
        }; /* else */
	psi_exact += b*exact2 - b*exact1; 

	/* compute alpha-factors: */
	qpl = q+lambda;
	qplx1 = qpl*x1;
	qplx2 = qpl*x2;
	gamma = 1.0/qpl;
	kappa1 = 1.0;
	kappa2 = 1.0;
	for (h=0; h<=m; h++) {
	  poisson_prob(h, qplx1, &beta1);
	  poisson_prob(h, qplx2, &beta2);
	  if (h) gamma *= h/qpl;
	  kappa1 -= *beta1;
	  kappa2 -= *beta2;
	}; /* for h */
	alpha1 = gamma*kappa1;
	alpha2 = gamma*kappa2;
	diff = alpha2-alpha1;
	(*omega_factors)[0] += c*diff;
        (*psi_factors)[0] += b*diff;
	error1 = exact1-alpha1;
	error2 = exact2-alpha2;
	for (k=1; ((error1>EPSILON) || (error2>EPSILON)) && (k<=MAXINDEX);
	     k++) {
          poisson_prob(k+m, qplx1, &beta1);
	  poisson_prob(k+m, qplx2, &beta2);
	  gamma *= ((double) (k+m))/k;
	  kappa1 -= *beta1;
	  kappa2 -= *beta2;
	  alpha1 = gamma*kappa1;
	  alpha2 = gamma*kappa2;
	  diff = alpha2-alpha1;
	  (*omega_factors)[k] += c*diff;
	  (*psi_factors)[k] += b*diff;
	  error1 -= alpha1;
	  error2 -= alpha2;
        }; /* for k */
	if (k-1>*R) *R = k-1;

      }; /* for expo */	
    }; /* for power */
  }; /* for expoly */

  /* compute resulting error bounds: */
  *omega_error = 1.0;
  *psi_error = psi_exact;
  for (k=0; k<=*R; k++) {
    *omega_error -= (*omega_factors)[k];
    *psi_error -= (*psi_factors)[k];
  }; /* for k */

} /* calculate_factors */

double expected_value2(p1)
     time_l *p1; 
{
    time_l  *expoly;
    double retu;

    expoly = p1;
    retu = 0.0;
    if (expoly->type == WEIBULL) {
      while((expoly->mu == 0.0)&&(expoly->next))
	expoly = expoly->next; /*definition intervall*/
      retu = (expoly->x1*expoly->p); /* 1. Punktmass */
      retu += qromb(expoly->x1,expoly->x2,expoly->sigma,expoly->mu,EXPECT_WEIBULL);     /* Int(x*f(x)) "uber x1 ...x2*/
      expoly = expoly->next; 
      retu += (expoly->x1*expoly->p);/* 2. Punktmass */
     }
    else if (expoly->type == NORMAL) { 
      while((expoly->sigma == 0.0)&&(expoly->next))
	expoly = expoly->next; /*definition intervall*/
      retu = (expoly->x1*expoly->p); /* 1. Punktmass */
      retu += qromb(expoly->x1,expoly->x2,expoly->sigma,expoly->mu,EXPECT_NORMAL);     /* Int(x*f(x)) "uber x1 ...x2*/
      expoly = expoly->next; 
      retu += (expoly->x1*expoly->p);/* 2. Punktmass */
    }
    else retu = -1.0;     
    return(retu);
} /* expected_value2 */
