/**************************************************************************
 File:		expoly.h

 Purpose:	include-file for expoly.c and rspnrand.c
		contains data structure of an expolynomial distribution
		and headers of functions to manipulate expolynomial
		distributions

 Programmer:	Reinhard German

 History:	Release 1: June 2, 1993
**************************************************************************/

#define MAXINDEX 100           /* size of array of factors */   
#define EPSILON 1.0e-12        /* error bound */

#define DETERMINISTIC           1
#define UNIFORM     	        2
#define EXPOLY		        3
/*new*/
#define WEIBULL                 4
#define NORMAL                  5

#define PI 3.141592654
#define NEG_INFINITY -999.0
#define CDF_NORMAL 1
#define EXPECT_NORMAL 2
#define EXPECT_WEIBULL 3

/**************************************************************************
  internal representation of an expolynomial distribution:

  recursive structure consisting of three hierarchy levels
  intervals on the real time axis: 0 =< l =< L
  powers of x with exponent m    : 0 =< m =< M
  exponentials with rate lambda_n: n =< n =< N
**************************************************************************/

typedef struct time_l {        /* time interval l */
  double x1;                   /* beginning of the interval */
  double x2;                   /* end of the interval (if < infinity) */
  double p;                    /* point mass */
  int type;                    /* type is det, uni or expoly */
  struct power_m *power;       /* list of powers */
  struct time_l *next;         /* next interval */
  /*new*/
  double sigma;                 /* parameter sigma for Weibull and Normal */
  double mu;                    /* parameter theta for Weibull and mu for 
				  Normal */
} time_l;

typedef struct power_m {       /* power with exponent m */
  unsigned m;                  /* exponent */  
  struct expo_n *expo;         /* list of exponentials */
  struct power_m *next;        /* next power */
} power_m;

typedef struct expo_n {        /* exponential with rate lambda_n */
  double lambda;               /* rate */ 
  double b;                    /* constant corresponding to 1-PDF */
  double c;                    /* constant corresponding to pdf */
  struct expo_n *next;         /* next exponential */
} expo_n;

/**************************************************************************
 function headers:
**************************************************************************/

void print_expoly();
int scan_expoly();
time_l *set_det_expoly();
time_l *set_uniform_expoly();
time_l *set_weibull();
time_l *set_normal();
void calculate_factors();
void print_factors();
double expected_value2();   /*Erwartungswert der Normal-/Weibullverteilung*/
double hb_pow();
double qromb();
