/******************************************************************************
 File:              xlib.c

 Purpose:           Handle the Pixmap to draw on, expose events, resizes (still
                    somewhat buggy), and draw the graph itself. This is done on
                    a virtual drawing area which is then scaled onto the phys.
                    drawing area.

 Project Director:  Reinhard German

 Programmer:        Ulrich Hertlein & Joerg W. Mitzlaff

 History:           Release 1:
******************************************************************************/

#include <stdio.h>

#include "pmfXlib.h"
#include "pmfEdit.h"
#include "pmfGraph.h"
#include <math.h>

extern int func_mode;
extern double erwartungswert;

/*
 * Local variables
 */
static Display *disp;
static Widget wid;
static int depth;

static Colormap cmap;
static GC gc;
static Pixmap pix;

/* physical size of pixmap */
static Dimension width = WIDTH, height = HEIGHT;

/* virtual size of drawing area */
static double vres_x, vres_y;


/*
 * real_x: calculate pixmap x from virtual position
 */
static int real_x(vx)
double vx;
{
   int x;

   x = XBORDER + (int )(vx * (width - F_BORDER - H_BORDER) / vres_x);

   return( x );
}

/*
 * real_y: calculate pixmap y from virtual position
 */
static int real_y(vy)
double vy;
{
   int y;

   y = height - YBORDER - (int )(vy * (height - F_BORDER - H_BORDER) / vres_y);

   return( y );
}

/*
 * arrow: draw a filled arrow-head
 */
static void arrow(x,y, dir)
int x,y, dir;
{
   int xs,ys;
   int xe,ye;
   XPoint p[3];

   /* determine direction */
   switch( dir )
   {
   case NORTH:
      xs = x - ARROW_WIDTH;
      xe = x + ARROW_WIDTH;
      ys = ye = y + ARROW_HEIGHT;
      break;
   case SOUTH:
      xs = x - ARROW_WIDTH;
      xe = x + ARROW_WIDTH;
      ys = ye = y - ARROW_HEIGHT;
      break;
   case EAST:
      xs = xe = x - ARROW_HEIGHT;
      ys = y - ARROW_WIDTH;
      ye = y + ARROW_WIDTH;
      break;
   case WEST:
      xs = xe = x + ARROW_HEIGHT;
      ys = y - ARROW_WIDTH;
      ye = y + ARROW_WIDTH;
      break;
   }

   /* draw polygon */
   p[0].x = xs; p[0].y = ys;
   p[1].x = x; p[1].y = y;
   p[2].x = xe; p[2].y = ye;
   XFillPolygon(disp,pix,gc,p,3,Convex,CoordModeOrigin);
}


/*
 * scale: draw an appropriate scale
 */
static void scale()
{
   static double scale_tab [11][7] =
   {
       {0,   0,   0,   0,   0,   0,   0},
       {0.2, 0.4, 0.6, 0.8, 1,   0,   0}, /* ziffer 1 genau */
       {0.5, 1,   1.5, 2,   0,   0,   0},
       {1,   2,   3,   0,   0,   0,   0},
       {1,   2,   3,   4,   0,   0,   0},
       {1,   2,   3,   4,   5,   0,   0},
       {1,   2,   3,   4,   5,   6,   0},
       {1,   2,   3,   4,   5,   6,   7},
       {2,   4,   6,   8,   0,   0,   0},
       {1.5, 3,   4.5, 6,   7.5, 9,   0},
       {2,   4,   6,   8,   10,  0,   0}  /* ziffer 9 aufgerundet */
   };

   int i,y;
   char buf[256];
   double faktor;   /* groessenordnung (exponent) + oder - */
   double manta;    /* 1. zahl der mantisse +1 */
   double skawert;

   /* draw axis */
   XSetLineAttributes(disp,gc,1,LineSolid,0,0);
   XSetForeground(disp,gc,BlackPixel(disp,0));
   XDrawLine(disp,pix,gc,XBORDER,H_BORDER,XBORDER,height - YBORDER + H_BORDER);
   XDrawLine(disp,pix,gc,H_BORDER,height - YBORDER,width - H_BORDER,height - YBORDER);

   /* draw arrows */
   arrow(XBORDER,H_BORDER,NORTH);
   arrow(width - H_BORDER,height - YBORDER,EAST);

   /* gerades y-maximum */
   faktor = pow(10.0,floor(log10(vres_y)));  /* groessenordn. */
   manta = ceil(vres_y/faktor); /* aufrunden */
   vres_y = manta * faktor;

   /* draw y-scales */
   i=0;
   while(i<7 && scale_tab[(int)manta][i]!=0)
   {
       skawert = scale_tab[(int)manta][i] * faktor;
       y = real_y(skawert);
       sprintf(buf,"%g",skawert);
       XDrawLine(disp,pix,gc,H_BORDER,y,H_BORDER + XBORDER,y);
       XDrawImageString(disp,pix,gc,XBORDER + XBORDER,y + 5,buf,strlen(buf));
       i++;
   }
}

/*
 * line: draw a line from (xs,ys) to (xe,ye) in virtual space
 */
static void line(vxs,vys, vxe,vye)
double vxs,vys;
double vxe,vye;
{
   int xs,ys, xe,ye;

   /* calculate pixmap position */
   xs = real_x(vxs);
   ys = real_y(vys);
   xe = real_x(vxe);
   ye = real_y(vye);

   /* draw line */
   XDrawLine(disp,pix,gc,xs,ys,xe,ye);
}


/*
 * create_pix: if we already have a pixmap, destroy that one; allocate a new
 *             pixmap with the current width and height
 */
static void create_pix(w)
Widget w;
{
   /* destroy old one */
   if( pix )
      XFreePixmap(disp,pix);

   /* create new one */
   pix = XCreatePixmap(disp,
		RootWindow(XtDisplay(w), DefaultScreen(disp)),width,height,depth);
   wid = w;
}

/*
 * clear_pix: clear drawing pixmap
 */
void clear_pix()
{
   /* clear pixmap */
   XSetForeground(disp,gc,WhitePixel(disp,0));
   XFillRectangle(disp,pix,gc,0,0,width,height);
}


/*
 * xlib_setup: store some vital parameters and create a pixmap
 */
void xlib_setup(w)
Widget w;
{
   /* save some variables */
   disp = XtDisplay(w);
   depth = DefaultDepth(disp,0);
   cmap = DefaultColormap(disp,0);
   gc = DefaultGC(disp,0);

   /* create X stuff for graphics */
   create_pix(w);
}

/*
 * xlib_refresh: copy pixmap to window, every time an expose event is
 *               received. If the window has been resized, change the size of
 *               the pixmap and redraw the graph. Unfortunately, expose
 *               events are only generated when the window gets bigger but not
 *               when it gets smaller ... have to work around that ...
 */
void xlib_refresh(w, event, params, num_params)
Widget w;
XEvent *event;
String *params;
Cardinal *num_params;
{
   Dimension nwidth,nheight;

   /* only process last event */

   if( event==NULL || event->xexpose.count == 0 )  /* achtung bus error provoziert */
   {
      /* get present size of window */
      XtVaGetValues(w,
         XtNwidth,&nwidth,
         XtNheight,&nheight,
         NULL);

      /* resize if changed */
      if( (nwidth != width) || (nheight != height) )
      {
         /* make it new size */
         width = nwidth;
         height = nheight;

	 /* make new pixmap */
	 create_pix(w);
	 
	 /* redraw graph */
	 xlib_draw_graph();
      }

      /* copy pixmap to Core widget */
      if( depth == 1 )
         XCopyArea(disp,pix,XtWindow(w),gc,0,0,width,height,0,0);
      else
         XCopyPlane(disp,pix,XtWindow(w),gc,0,0,width,height,0,0,1);
   }
}

/*
 * xlib_draw_graph: clear the pixmap, calculate the graph, draw a scale, and
 *                  draw the graph
 */
void xlib_draw_graph()
{
   struct g_data *gd, *this;
   double old_x, old_y;
   int x,y;
   char buf[256];

   /* clear pixmap */
   clear_pix();

   /* determine virtual resolution & pre-calculate graph */
   gd = graph_build(&vres_x,&vres_y);

   /* draw scale */
   scale();
      
   y = real_y(vres_y);

   /* write status-line */
   switch(func_mode)
     {
     case MODE_PMF: 
       strcpy(buf,"pmf function");
       break;
     case MODE_PDF:
       strcpy(buf,"PDF function");
       break;
     case MODE_1MPDF:
       strcpy(buf,"1-PDF function");
       break;
     }

   XDrawImageString(disp,pix,gc, width-100, y, buf, strlen(buf));

   sprintf(buf,"mean value: %g",erwartungswert);
   XDrawImageString(disp,pix,gc, width/2-100, y, buf, strlen(buf));


   /* draw some graphics */
   XSetLineAttributes(disp,gc,2,LineSolid,0,0);
   old_x = old_y = 0.0;
   for( this = gd; this; this = this->next )
      switch( this->type )
      {
      case CONNECT:
	 XSetLineAttributes(disp,gc,2,LineSolid,0,0);
         line(old_x,old_y,this->x,this->y);
         old_x = this->x;
         old_y = this->y;
         break;
      case INTERVAL:
	 switch(func_mode)
	   {
	   case MODE_PMF: 
	     if(this->y>0)
	       {
		 XSetLineAttributes(disp,gc,3,LineSolid,0,0);
		 XDrawLine(disp,pix,gc,real_x(old_x),
			               real_y((double)0),
			               real_x(old_x),
			               real_y(this->y)+ARROW_HEIGHT);
		 arrow(real_x(old_x), real_y(this->y), NORTH);
	       }
	     break;
	   case MODE_PDF:
	     XSetLineAttributes(disp,gc,2,LineSolid,0,0);
	     line(old_x, old_y, old_x, old_y + this->y);
	     old_y += this->y;
	     break;
	   }

	 /* x-achen beschriftung */
	 XSetLineAttributes(disp,gc,1,LineSolid,0,0);
	 x = real_x(old_x);

	 sprintf(buf,"%g",old_x);
	 XDrawLine(disp,pix,gc,x,height - YBORDER+H_BORDER,x,height - YBORDER-H_BORDER);
	 XDrawImageString(disp,pix,gc,x-H_BORDER,height-H_BORDER,buf,strlen(buf));

         break;
     }

   XSetLineAttributes(disp,gc,2,LineSolid,0,0);
   if(func_mode == MODE_PDF)
     {
       XDrawLine(disp,pix,gc,real_x(old_x), real_y(old_y), width - H_BORDER, real_y(old_y));
     }
   else
     {
       line(old_x,old_y,old_x,0);  /* this->y == 0 !!! */
       XDrawLine(disp,pix,gc,real_x(old_x), height - YBORDER, width - H_BORDER, height - YBORDER);
     }

   /* free graph data */
   gd = graph_free(gd);

   xlib_refresh((Widget)wid,NULL); 
}
