/******************************************************************************
 Program:           convert_results.c

 Parameters:        Model name with complete path name

 Purpose:           Convert the .sol and .tmark files to obtain the data
                    expected by the DSPNexpress interface. This is useful when
                    using DSPNexpress create and solve nets.
                    This file is compiled and linked to the measures-dependent
		    file created by proc_MDF. This executable can then be run
                    and will create the .PROBTOK and .RESULTS files.

 Project Director:  Reinhard German

 Programmer:        Ulrich Hertlein

 History:           Release 1: June 24, 1993
                    Release 2: February 2, 1994
		    Release 3: June 21, 1994
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>

#include "netio.h"


#define BUF_SIZE    100
#define COMMENT     '#'


/* place definition */
typedef struct
{
   char *name;
   int number;
} place_t;


/* local variables */
static int rflag = 0;

static unsigned nmarking = 0;
static double *prob_list;

static unsigned nplaces = 0;
static place_t *place;

static unsigned ntoken = 0;
static int *tok_list;

static double *tok_prob;

static int this_mark;

/* exported variables */
int digits;


/*
 * act_marking: return number of token per place per marking; this function is
 *              used by the result measures solver (therefore non-static)
 */
int act_marking(p)
int p;
{
   register int i,addr;

   /* check bounds */
   if( p < 0 || p >= nplaces )
      return( 0 );

   /* find place number 'place' in place list */
   for( i = 0; i < nplaces && (p != place[i].number); i++ );
   if( i == nplaces )
   {
      fprintf(stderr,"Unable to find place %d\n", p);
      exit( 1 );
   }

   /* get number of token */
   addr = this_mark * nplaces + i;

   return( tok_list[addr] );
}


/*
 * usage: print program usage and exit
 */
static int usage()
{
   fprintf(stderr,"Usage: convert_results [options] <netname>\n");
   fprintf(stderr,"  -p DOUBLE precision for file output\n");
   fprintf(stderr,"  -r        write data in readable form to stdout\n");
   fprintf(stderr,"  -h        print this help message\n");

   exit( 1 );
}


/*
 * spool: read lines from a file until a non-comment line is found
 */
static int spool(fp, buf)
FILE *fp;
char *buf;
{
   do
   {
      fgets(buf,BUF_SIZE,fp);
   } while( buf[0] == COMMENT );

   if( strlen(buf) > 1 )
      buf[strlen(buf) - 1] = '\0';

   return( 0 );
}


/*
 * read_prob: read number of markings from file, build list of
 *            probabilities, read probabilities from file;
 *            reads .sol file
 */
static int read_prob(fp)
FILE *fp;
{
   char buffer[BUF_SIZE];
   int i;

   /* read number of markings */
   spool(fp,buffer);
   sscanf(buffer,"%d",&nmarking);
   if( rflag )
      fprintf(stdout,"Number of tangible markings: %d\n", nmarking);

   /* allocate probabilities list */
   if( (prob_list = (double *)malloc(nmarking * sizeof(double))) == NULL )
   {
      fprintf(stderr,"Unable to allocate probabilities list\n");
      exit( 1 );
   }

   /* read probabilities */
   for( i = 0; i < nmarking; i++ )
   {
      spool(fp,buffer);
      sscanf(buffer,"%lf",&prob_list[i]);
   }

   return( 0 );
}


/*
 * read_marking: read number of places, allocate matrix for number of token
 *               per marking per place, read data from file;
 *               reads .tmark
 */
static int read_marking(fp)
FILE *fp;
{
   char buffer[BUF_SIZE], buf_name[256], *buf;
   int i,j;
   int mark,tok;

   /* read number of places */
   spool(fp,buffer);
   sscanf(buffer,"%d", &nplaces);
   if( rflag )
      fprintf(stdout,"Number of places: %d\n", nplaces);

   /* allocate name & number list */
   if( (place = (place_t *)malloc(nplaces * sizeof(place_t))) == NULL )
   {
      fprintf(stderr,"Unable to allocate places name list\n");
      exit( 1 );
   }

   /* read place info */
   for( i = 0; i < nplaces; i++ )
   {
      /* read place info */
      spool(fp,buffer);
      sscanf(buffer,"%s %d", buf_name,&place[i].number);

      /* copy place name */
      if( (place[i].name = (char *)strdup(buf_name)) == NULL )
      {
         fprintf(stderr,"Unable to allocate name for place\n");
         exit( 1 );
      }
   }

   /* read number of markings */
   spool(fp,buffer);
   sscanf(buffer,"%d", &i);
   if( i != nmarking )
   {
      fprintf(stderr,"Number of markings in .sol and .tmark differ\n");
      exit( 1 );
   }

   /* allocate token list */
   if( (tok_list = (int *)malloc(nmarking * nplaces * sizeof(int))) == NULL )
   {
      fprintf(stderr,"Unable to allocate token list\n");
      exit( 1 );
   }

   /* read token distribution */
   for( i = 0; i < nmarking; i++ )
   {
      /* read marking number */
      spool(fp,buffer);
      buf = buffer;
      sscanf(buf,"%d", &mark);

      /* read number of token per place */
      for( j = 0; j < nplaces; j++ )
      {
         /* this spools buf to the next char after a the first space */
         while( *buf++ != ' ' )
            ;

         /* scan number of token */
         sscanf(buf,"%d", &tok);
         tok_list[mark * nplaces + j] = tok;

         /* determine maximum number of tokens */
         if( tok > ntoken )
            ntoken = tok;
      }
   }

   return( 0 );
}


/*
 * compute_prob: compute the token probability in each place;
 *               writes .PROBTOK
 */
static int compute_prob(net,fd)
NetDescr* net;
int fd;
{
   int read = 0;

   int i,j,k;
   char *pname;

   int min,max;
   double dummy;

   /* allocate token probability list; ntoken is the maximum number of token
    * found in one place at a time, since we also might have no token at all
    * in a place, we actually allocate room for ntoken + 1 probabilities
    */
   if( (tok_prob = (double *)malloc((ntoken + 1) * sizeof(double))) == NULL )
   {
      fprintf(stderr,"Unable to allocate token probability list\n");
      exit( 1 );
   }

   if( net->num_places != nplaces )
   {
      fprintf(stderr,"Number of places in .TN and .tmark differ\n");
      exit( 1 );
   }

   /* we can now read 'nplaces' definitions */
   while( read < nplaces )
   {
      /* fetch place name, determine index */
      pname = net->place[read].name;
      for( i = 0; i < nplaces && strcmp(pname,place[i].name) != 0; i++ )
         ;
      if( i == nplaces )
      {
         fprintf(stderr,"Place %s not in .tmark file\n", pname);
         exit( 1 );
      }


      /* we have found place 'pname' at position 'i'; since the order of
       * appearance in .TN and .tmark is not neccessarily the same, we now
       * record the index in the .TN file in the place description.
       */
      place[i].number = read;

      /* initialize data; we must initialize ntoken + 1 values since we might
       * have no token at all in one place (see above)
       */
      for( j = 0; j <= ntoken; j++ )
         tok_prob[j] = (double )0.0;
      min = ntoken;
      max = 0;

      for( j = 0; j < nmarking; j++ )
      {
         /* fetch number of tokens per marking per place */
         k = tok_list[j * nplaces + i];
         if( k < min )
            min = k;
         if( k > max )
            max = k;

         /* add probability of marking to token probability */
         tok_prob[k] += prob_list[j];
      }

      /* save min and max values */
      dummy = (double )min;
      write(fd,&dummy,sizeof(double));
      dummy = (double )max;
      write(fd,&dummy,sizeof(double));

      if( rflag )
         fprintf(stdout,"\nPlace #%d '%s': min = %d, max = %d\n", place[i].number,place[i].name,min,max);

      /* save probabilities */
      for( j = min; j <= max; j++ )
      {
         dummy = tok_prob[j];
         write(fd,&dummy,sizeof(double));
         if( rflag )
            fprintf(stdout,"%d: %.*g\n", j,digits,tok_prob[j]);
      }

      read++;
   }

   return( 0 );
}


/*
 * compute_results: compute network dependent result measures;
 *                  functions sum_prob() and calc_results() are created by
 *                  a lex/yacc analysis of the .DEF file (derive_results);
 *                  writes .RESULTS file
 */
static int compute_results(fp)
FILE *fp;
{
   /* calculate result measures */
   for( this_mark = 0; this_mark < nmarking; this_mark++ )
      sum_prob(prob_list[this_mark]);

   /* write results */
   calc_results(fp,digits);

   /* flush buffer */
   fflush(fp);

   return( 0 );
}


/*
 * free_mem: free allocated memory
 */
static int free_mem()
{
   int i;

   free(prob_list);

   for( i = 0; i < nplaces; i++ )
      free(place[i].name);
   free(place);
   free(tok_list);

   free(tok_prob);

   return( 0 );
}


/*
 * main: check arguments, open files, process, close files
 */
int main(argc, argv)
int argc;
char **argv;
{
   int c;
   extern int optind;
   extern char *optarg;

   double p = (double )10.0e-9, prec;
   int errflag = 0;

   char *filename;
   char sol_file[256], tmark_file[256];
   char ptok_file[256], res_file[256];

   NetDescr* net;

   FILE *sol_fp, *tmark_fp;
   FILE *res_fp;
   int ptok_fd;

   /* check parameters */
   while( (c = getopt(argc,argv,"p:rh?")) != -1 )
      switch( c )
      {
      case 'p': p = (double )atof(optarg);
                break;
      case 'r': if( rflag++ ) /* produce readable output */
                   errflag++;
                break;
      case 'h':
      case '?':
      default:  errflag++;
      }
   if( errflag || (argc - optind) != 1 )
      usage();

   filename = argv[optind];

   /* calculate # of digits */
   prec = p;
   for( digits = 0; p < 1.0; digits++, p *= 10.0 )
      ;
   if( rflag )
      fprintf(stderr,"Precision: %g (%d digits)\n", prec,digits);

   /* build filenames */
   sprintf(sol_file,"%s.sol", filename);
   sprintf(tmark_file,"%s.tmark", filename);

   sprintf(ptok_file,"%s.PROBTOK", filename);
   sprintf(res_file,"%s.RESULTS", filename);

   /* open files */
   if( (sol_fp = fopen(sol_file,"r")) == NULL )
   {
      fprintf(stderr,"Unable to open file '%s'\n", sol_file);
      exit( 1 );
   }
   if( (tmark_fp = fopen(tmark_file,"r")) == NULL )
   {
      fprintf(stderr,"Unable to open file '%s'\n", tmark_file);
      exit( 1 );
   }
   if( (ptok_fd = open(ptok_file,O_WRONLY | O_CREAT,0644)) < 0 )
   {
      fprintf(stderr,"Unable to open file '%s'\n", ptok_file);
      exit( 1 );
   }
   if( (res_fp = fopen(res_file,"w")) == NULL )
   {
      fprintf(stderr,"Unable to open file '%s'\n", res_file);
      exit( 1 );
   }

   /* parse .sol and .tmark files */
   read_prob(sol_fp);
   read_marking(tmark_fp);

   /* compute & save probabilities */
   net = load_net(filename,NULL);
   compute_prob(net,ptok_fd);

   /* calculate result measures */
   compute_results(res_fp);

   /* free allocated memory */
   free_mem();
   free_net(net);

   /* close files */
   fclose(sol_fp);
   fclose(tmark_fp);

   close(ptok_fd);
   fclose(res_fp);

   exit( 0 );
}
