// File    HostConfig.cpp  (C++ Implementation)
// Module  HostConfig      (Administration of PVM host configuration)
// Library PVM3CPP         (C++ Extension to PVM3.4)
/**
 *@author  Philipp Berndt
 *@version $Id: HostConfig.cpp,v 1.2 2000/01/12 16:02:00 pumphutt Exp $
 */

#include <iostream>
#include <iomanip>
#include <functional>
#include <algorithm>
#include <unistd.h> // gethostname
#include <netdb.h>  // MAXHOSTNAMELEN
#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 256
#endif
#ifndef RSTAT_CPU_IDLE
#define RSTAT_CPU_IDLE 3
#endif
// From $TNETHOME/Shared/src/sysdep:
#include "Time.h"
#include "rstat.h"

// From current dir:
#include "pvm3wrap.h"
using namespace P3PP;
#include "pvm_util.h"
#include "HostConfig.h"

using HostConfig::Host;

/** The host configuration */
vector<Host> HostConfig::hosts;

extern "C" int rstat2(const char *host, struct statsswtch *st);
extern "C" int pvm_start_pvmd2(int argc, char **argv, int block);


/************************** INTERNAL HELP FUNCTIONS **************************/

/** Returns hostname of localhost. */
string HostConfig::gethostname()
{
    char name[MAXHOSTNAMELEN];
    ::gethostname(name, MAXHOSTNAMELEN);
    return name;
}

/** Attribute: Has host 'H' the hostname 'n' ? */
struct hostname : public binary_function<Host, string, bool> {
    bool operator() (const Host& H, string n) const
        { return H.name == n; } };

/** Prints one host structure. */
ostream& operator<<(ostream &os, const Host& h)
{
    const char* const statename[] = { "unreach.", "offline ", " online " };
    cout << h.name << ", "<< statename[h.state] << ", " << h.speed << ", "
         << ((h.mine) ? "owned" : "borrowed") << ", " 
         << ((h.reserve) ? "reserve" : "not res") << endl;  
    return os;
}

bool Host::operator<(const Host& h) const
{ return getRating()>h.getRating(); }

double Host::getRating() const
{
    double eff_load
        = 1.0                                // my hypothetic load share
        + realload                           // current load
        + virtload * 1.0                     // penalty for double use 
        + ((state == OFFLINE) ? 10.0 : 0.0); // penalty for offline hosts

    return speed/eff_load;
}


/***************************** EXPORTED FUNCTIONS ****************************/
/********************** see header file for description **********************/

void HostConfig::addhosts(vector<Host>& hostlist)
{
    vector<string> hosts_to_add;
    vector<Host*> hostrefs;
    for (vector<Host>::iterator i=hostlist.begin(); i!=hostlist.end(); ++i)
        if (i->state == Host::OFFLINE && !i->reserve)
            {
            hosts_to_add.push_back(i->name);
            hostrefs.push_back(&(*i));
            }
    vector<int> infos = pvm_addhosts(hosts_to_add);

    for (int i=0; i<hosts_to_add.size(); ++i)
        {
        cout << "Adding " << hosts_to_add[i] << "...";
        if (infos[i] < 0)
            if (infos[i] == PvmDupHost)
                {
                cout << "Already in machine" << endl;
                hostrefs[i]->state = Host::ONLINE;
                }
            else
                {
                cout << "Error " << infos[i] << endl;
                hostrefs[i]->state = Host::UNREACHABLE;
                }
        else
            {
            cout << "OK" << endl;
            hostrefs[i]->state = Host::ONLINE;
            hostrefs[i]->mine = true;
            hostrefs[i]->virtload = 0.0;
            P3PP::pvm_notify(PvmHostDelete, TAG_HOST_DELETED,
                             1, &infos[i]);
            }
        }
    cout << endl;
    sort(hosts.begin(),hosts.end());
}

void HostConfig::determineLoad()
{
    struct statsswtch rstatinfo;
    int n = hosts.size();
    vector<Time> sampletime(n);
    vector<int> cpu_idle_time(n);
    vector<bool> fail(n);

    for (int i=0; i< hosts.size(); ++i)
        if (!rstat2(hosts[i].name.c_str(), &rstatinfo))
            {
            sampletime[i] = Time::getTimeOfDay();
            cpu_idle_time[i] = rstatinfo.cp_time[RSTAT_CPU_IDLE];
            }
        else
            fail[i] = true;
            
    sleep(1);
    for (int i=0; i< hosts.size(); ++i)
        if (fail[i] || rstat2(hosts[i].name.c_str(), &rstatinfo))
            {
            hosts[i].realload += 1.0;
            cerr << "Cannot determine load for " << hosts[i].name
                 <<". Assuming " << hosts[i].realload << " ." << endl;
            if (hosts[i].state == Host::OFFLINE)
                hosts[i].state = Host::UNREACHABLE;
            }
        else
            {
            Time Delta_t = Time::getTimeOfDay() - sampletime[i];
            int Delta_idle = rstatinfo.cp_time[RSTAT_CPU_IDLE] 
                - cpu_idle_time[i];
            // idle percentage = delta_idle / delta_time
            double load = 1.0 - (Delta_idle/Delta_t.to_secs())/100.0; 

            // If any rstat loadaverage ist greater use that instead.
            for(int j=0; j<3; ++j)
                {
                double avg = rstatinfo.avenrun[j]/256.0;
                if (avg>load)
                    load = avg;
                }
            hosts[i].realload = load;
            if (hosts[i].state == Host::UNREACHABLE)
                hosts[i].state = Host::OFFLINE; // Host OK.
            }
    sort(hosts.begin(),hosts.end());
}

void HostConfig::update()
{
    vector<Host>::iterator host;
    int nhost, narch;
    pvmhostinfo *hostinfos;
    P3PP::pvm_config(&nhost, &narch, &hostinfos);

    // Check for added hosts
    for (int i=0; i<nhost; ++i)
        if ( (host = find_if(hosts.begin(), 
                             hosts.end(),
                             bind2nd(hostname(), hostinfos[i].hi_name)))
             != hosts.end() )
            {
            if (host->state == Host::OFFLINE)
                {
                host->state = Host::ONLINE;
                host->virtload = 0.0;
                P3PP::pvm_notify(PvmHostDelete, TAG_HOST_DELETED,
                                 1, &hostinfos[i].hi_tid);
                cout << "Grabbing host " << host->name << endl;
                }
            }
        else // Found new host not yet in list
            {
            Host newHost(hostinfos[i].hi_name, Host::DEFAULT_SPEED,
                         Host::ONLINE, true, false);
            newHost.virtload = 0.0;
            hosts.push_back(newHost);
            P3PP::pvm_notify(PvmHostDelete, TAG_HOST_DELETED,
                             1, &hostinfos[i].hi_tid);
            cout << "Grabbing host " << hostinfos[i].hi_name
                 << ". Nothing known about. Assuming default speed "
                 << newHost.speed << "." << endl;
            }

    // Check for removed hosts
    for (host = hosts.begin(); host != hosts.end(); ++host)
        if (host->state == Host::ONLINE)
            {
            bool found = false;
            for (int i=0; i<nhost; ++i)
                if (host->name == hostinfos[i].hi_name)
                    {
                    found = true;
                    break;
                    }
            if (!found)
                {
                host->state = Host::OFFLINE;
                cout << "Host " << host->name 
                     << " was removed from host config." << endl;
                }
            }

    determineLoad();
}

void HostConfig::hostAdded()
{ update(); }

void HostConfig::hostDeleted()
{ update(); }

void HostConfig::init(vector<Host> hostlist)
{
    // Add localhost to hostlist if not already contained
    hosts = hostlist;
    string localhost = gethostname();
    vector<Host>::iterator host;
    if ( (host = find_if(hosts.begin(), 
                         hosts.end(),
                         bind2nd(hostname(), localhost)))
         == hosts.end() )
        {
        hosts.push_back(Host(localhost, Host::DEFAULT_SPEED,
                             Host::OFFLINE, true, false));
        host = &hosts.back();
        }

    // Start PVM
    cout << "Starting up PVM..." << flush;
    switch(pvm_start_pvmd2(0, NULL, 0))
        {
        case PvmOk:
            cout << "OK" << endl;
            host->state = Host::ONLINE;
            host->mine = true;
            break;
        case PvmDupHost:
            cout << "pvmd already running." << endl;
            host->state = Host::ONLINE;
            host->mine = false;
            break;
        default:
            cout << "failed" << endl;
            throw PvmException("Failed to start pvmd");
        }

    // Enroll in PVM
    pvm_mytid();

    // Take over hosts that are already part of the machine 
    update();

    // Add all hosts in hostlist that are not reserve
    addhosts(hosts);

    // Request notification about user-added hosts
    pvm_notify(PvmHostAdd, TAG_HOST_ADDED, -1, 0);

    print(cout);
}

void HostConfig::print(ostream &os)
{
    const char* const statename[] = { "unreach.", "offline ", " online " };

    os << "Host          Status     Speed  VirtLoad "
       << "RealLoad   Owned   Res   Rating\n"
       << "-----------------------------------------"
       << "-------------------------------\n";
    os.setf(ios::fixed, ios::floatfield);
    os.fill(' ');
    int oldprec = os.precision();
    for (vector<Host>::iterator i=hosts.begin(); i!=hosts.end(); ++i)
        {
        os << setiosflags(ios::left) << setw(12) << i->name.c_str() << " "
           << resetiosflags(ios::left) << statename[i->state] << "   " 
           << setprecision(6) << setw(6) << i->speed << "   "
           << setprecision(2) << setw(5) << i->virtload << "    "
           << setw(5) << i->realload << "       " 
           << (i->mine ? "yes" : "no ") << "   "
           << (i->reserve ? "yes" : "no ") << "  "
           << setiosflags(ios::right) << setw(7) << i->getRating() << '\n';
        }
    os.precision(oldprec);
    os << "-----------------------------------------"
       << "-------------------------------\n" << endl;
}

bool HostConfig::addHost()
{
    sort(hosts.begin(),hosts.end());
    for (vector<Host>::iterator host = hosts.begin();
         host != hosts.end(); ++host)
        if (host->state == Host::OFFLINE)
            if (pvm_addhost(host->name)>0)
                {
                host->state = Host::ONLINE;
		sort(hosts.begin(),hosts.end());
                return true;
                }
    sort(hosts.begin(),hosts.end());
    return false;
}

string HostConfig::getHost()
{
    sort(hosts.begin(),hosts.end());

    for (vector<Host>::iterator host = hosts.begin();
         host != hosts.end(); ++host)
        if (host->state == Host::ONLINE)
            return host->name;
    return string(); // This should not happen since localhost is online!
}

bool HostConfig::idleHosts()
{
    for (vector<Host>::iterator host = hosts.begin();
         host != hosts.end(); ++host)
        if (host->state == Host::ONLINE && host->virtload == 0.0)
            return true;
    return false;
}

void HostConfig::addVirtLoad(string name, double virtload)
{
    vector<Host>::iterator host;
    if ( (host = find_if(hosts.begin(), 
                         hosts.end(),
                         bind2nd(hostname(), name)))
         != hosts.end() )
        {
        host->virtload += virtload;
        if (host->virtload < 0)
            host->virtload = 0;
        }

    sort(hosts.begin(),hosts.end());
}

void HostConfig::removehosts()
{
    int mytid = pvm_mytid();
    int nhost, narch;
    pvmhostinfo *hostinfos;
    P3PP::pvm_config(&nhost, &narch, &hostinfos);
    for (int i=0; i<nhost; ++i)
        {
        vector<Host>::iterator host;
        if ( (host = find_if(hosts.begin(), 
                             hosts.end(),
                             bind2nd(hostname(), hostinfos[i].hi_name)))
             != hosts.end() )
            {
            if (!host->mine) // Don't remove hosts that we didn't add
                continue;

            int ntask;       // Look for tasks running on host
            pvmtaskinfo *taskinfos;
            P3PP::pvm_tasks(hostinfos[i].hi_tid, &ntask, &taskinfos);
            bool occupied = false;
            for (int j=0; j<ntask; ++j)
                {
                int tid = taskinfos[j].ti_tid;
                if (tid != 0 && tid != mytid)
                    {
                    cout << "Task "<<taskinfos[j].ti_a_out << endl;
                    occupied = true;
                    break;
                    }
                }
          
            if (occupied)
                {
                cout << "Host " << host->name << " that we spawned "
                     << "is now occupied. Leaving it in machine." << endl;
                }
            else
                {
                cout << "Removing host " <<host->name<< " from machine."<<endl;
                int info;
                pvm_delhosts(&hostinfos[i].hi_name, 1, &info);
                }
            }
        }
    sort(hosts.begin(),hosts.end());
}

// void HostConfig::setHostStatus(string name, Host::status state)
// {
//     vector<Host>::iterator host;
//     if ( (host = find_if(hosts.begin(), 
//                       hosts.end(),
//                       bind2nd(hostname(), name)))
//       != hosts.end() )
//      host->state = state;
// }

// Host::status HostConfig::getHostStatus(string name)
// {
//     vector<Host>::iterator host;
//     host = find_if(hosts.begin(), 
//                 hosts.end(),
//                 bind2nd(hostname(), name));
//     return (host != hosts.end()) ? host->state : Host::OFFLINE;
// }

// void HostConfig::hostAdded()
// {
//   int nhost, narch;
//   pvmhostinfo *hostinfos;
//   P3PP::pvm_config(&nhost, &narch, &hostinfos);
//   int newhosts = pvm_upkint();
//   for(int i = 0; i<newhosts; ++i)
//     {
//       int newhosttid = pvm_upkint();
//       for (int j=0; j<nhost; ++j)
//      if (hostinfos[j].hi_tid == newhosttid)
//        idlehosts.push_back(hostinfos[j].hi_name);
//     }
// }

// void HostConfig::hostDeleted()
// {
//   for (map<int,string>::iterator it = hosts.begin();
//       it != hosts.end();
//       ++it)
//     if (pvm_mstat(it->second.c_str())<0)
//       hosts.erase(it);
//   for (list<string>::iterator it = idlehosts.begin();
//       it != idlehosts.end();
//       ++it)
//     if (pvm_mstat(it->c_str())<0)
//       idlehosts.erase(it);
// }
