// File    HostConfig.h    (C++ Interface)
// Module  HostConfig      (Administration of PVM host configuration)
// Library PVM3CPP         (C++ Extension to PVM3.4)
/**
 *@author  Philipp Berndt
 *@version $Id: HostConfig.h,v 1.2 2000/01/12 16:02:00 pumphutt Exp $
 */

#ifndef HOSTCONFIG_H
#define HOSTCONFIG_H

#include <vector>
#include <string>

enum {
    TAG_HOST_ADDED = 20000,
    TAG_HOST_DELETED,
    TAG_TERMINATE
};

namespace HostConfig
{

/**
 *         Holds resource management data for one host.
 *
 *@author  Philipp Berndt
 *@version $Id: HostConfig.h,v 1.2 2000/01/12 16:02:00 pumphutt Exp $
 */
    struct Host
    {
        enum status { UNREACHABLE, OFFLINE, ONLINE }; // Possible states

        /** Constructor. */
        Host(string Name, int Speed,status State, bool Reserve, bool Mine) 
            : name(Name), state(State), speed(Speed),
              reserve(Reserve), mine(Mine), virtload(0.0), realload(0.5) {}

        string name;     // Hostname.
        status state;    // Current state.
        int    speed;    // Relative speed.
        bool   reserve;  // == not USE_ALWAYS == do not add at startup.
        bool   mine;     // Was added by me.
        double virtload; // Virtual load assigned to host
        double realload; // Load average determined with rstat(3)

        enum { DEFAULT_SPEED = 1000 };
    
        /** Computes merit function for this host. */
        double getRating() const;

        /**
         * Comparison operator for sorting hostlist
         * according to speed and availability.
         */
        bool operator<(const Host& h) const;
    };

/** The host configuration. Use with caution. */
    extern vector<Host> hosts;

/**
 * Sets up a virtual machine.
 * localhost is added to hostlist if not already contained.
 * PVM is started on localhost if not already running.
 * Hosts that are already in the machine are taken over (Resistance is futile).
 * Hosts in the hostlist that do not have the reserve flag set are added to
 * the machine.
 */
    void init(vector<Host>);

/**
 * Adds the hosts in hostlist to the machine
 * that are offline and not marked as reserve.  
 */
    void addhosts(vector<Host>& hostlist);

/**
 * Removes all hosts that belong to us and have no processes running.
 */
    void removehosts();

/**
 * Determines the physical load for each host in the config
 * (including OFFLINE hosts) using the rstat-protocol.
 */
    void determineLoad();

/** Prints host configuration to given ostream. */
    void print(ostream &os);

/**
 * Adds (for negative arguments subtracts) an amount to/from the
 * virtual CPU load of that host.
 */
    void addVirtLoad(string hostname, double virtload);

/**
 * Tries to add one offline host from the hostlist to the virtual machine.
 *@returns true in case of success.
 */
    bool addHost();

/**
 * Returns the hostname of a preferrably idle host.
 * If no host in the machine is idle, a busy one is taken instead.
 * If you really need an idle one first do:
 * <pre>
 *@@ if(!idleHosts())
 *@@     addHost();
 * </pre>
 */
    string getHost();

/** Returns wether there are hosts without work in the virtual machine. */
    bool idleHosts();

/**
 * Updates the host configuration.
 * Adds new hosts that it finds to the hostlist if not there already
 * and sets their status to IDLE.
 * Changes the status of hosts that have been deleted to OFFLINE.
 */ 
    void update();

/**
 * Updates the host configuration when a host has been added.
 * Note: Current implementation simply calls update();
 */
    void hostAdded();

/**
 * Updates the host configuration when a host has been deleted.
 * Note: Current implementation simply calls update();
 */
    void hostDeleted();

/** Returns hostname of localhost. */
    string gethostname();

/**
 * Sets status of host.
 * Of course this does not affect the *real* state, but only what 
 * HostConfig believes.
 * Use this to tell HostConfig that you started a process on that host, etc.
 */
//void setHostStatus(string hostname, Host::status state);

/**
 * Returns status of host with given hostname
 * according to what HostConfig believes.
 */
    Host::status getHostStatus(string hostname);

}; // namespace HostConfig

#endif // HOSTCONFIG_H
