// Library PVM3CPP       (C++ Extension to PVM3.4)
// Module  pvm_util      (New Functions not found in PVM3.4)
// File    pvm_util.h    (C++ Interface and Inline Implementation)
/**
 *@author  Philipp Berndt
 *@version $Id: pvm_util.h,v 1.2 2000/01/12 16:02:04 pumphutt Exp $
 */

#ifndef PVM_UTIL_H
#define PVM_UTIL_H

#include <vector>
#include <string>
#include "pvm3wrap.h"

/**
 * Dummy class for unpacking constructors.
 *
 * This class is to be used as argument to class constructors that
 * construct objects from the PVM receive buffer.
 * The user functions then call the constructor with the object PVM_UNPACK
 * as argument.
 *
 *@author  Philipp Berndt
 *@version $Id: pvm_util.h,v 1.2 2000/01/12 16:02:04 pumphutt Exp $
 */
class PVM_UNPACK_t {};

/**
 * Dummy argument to class constructors 
 * that construct objects from the PVM receive buffer
 */
extern PVM_UNPACK_t PVM_UNPACK;

/** Does the same as PVM3::pvm_halt except that the caller is not terminated.*/
void pvm_halt();

/** Resets the virtual machine by terminating all processes save consoles. */
void pvm_reset();

/**
 * Explicitly starts the PVM group server (pvmgs).
 *
 * The arguments correspond to those in pvm_spawn.
 *@param flag Integer specifying spawn options.
 *@param where Character string specifying where to start the PVM process.
 */
void pvm_start_pvmgs(int flag, const string& where);

/** 
 * Returns the unix process id (pid) 
 * of the task with given PVM task id (tid).
 */
int pvm_tid2pid(int tid);

/**
 * Starts one new PVM process.
 *
 * throws PvmException in case of a PVM system error.
 *
 *@returns the TID of the spawned process 
 *         or 0 if spawning failed because of any (non-system) error.
 */
int pvm_spawn(const char *task, const vector<string> &argv,
              int flag, const string& where);

/**
 * Starts new PVM processes.
 *
 * throws PvmException in case of a PVM system error.
 *
 *@returns a vector containing the TIDs of the spawned processes
 *         or 0 if spawning failed because of any (non-system) error.
 */
vector<int> pvm_spawn(const char *task, const vector<string> &argv,
                      int flag, const string& where, int ntask);

/**
 * Adds one host to the virtual machine.
 *
 * throws PvmException in case of a PVM system error.
 *@returns the TID of the added host 
 *         or 0 if adding failed because of any (non-system) error.
 */
int pvm_addhost(const string& hostname);

/**
 * Add hosts to the virtual machine.
 *
 * throws PvmException in case of a PVM system error.
 *@returns a vector containing the TIDs of the added hosts
 *         or 0 if adding failed because of any (non-system) error.
 */
vector<int> pvm_addhosts(const vector<string>& hosts);

// The following functions (un)pack one value and are simply abbreviations.

inline void pvm_pkbyte(const char& b)
{ P3PP::pvm_pkbyte(&b, 1, 1); }

inline void pvm_pkdouble(const double& d)
{ P3PP::pvm_pkdouble(&d, 1, 1); }

inline void pvm_pkint(const int& i)
{ P3PP::pvm_pkint(&i, 1, 1); }

inline void pvm_pkuint(const unsigned int& i)
{ P3PP::pvm_pkuint(&i, 1, 1); }

inline void pvm_pklong(const long& l)
{ P3PP::pvm_pklong(&l, 1, 1); }

inline char pvm_upkbyte() {
    char b;
    P3PP::pvm_upkbyte(&b, 1, 1);
    return b; }

inline double pvm_upkdouble() {
    double d;
    P3PP::pvm_upkdouble(&d, 1, 1);
    return d; }

inline int pvm_upkint() { 
    int i;
    P3PP::pvm_upkint(&i, 1, 1);
    return i; }

inline unsigned int pvm_upkuint() { 
    unsigned int i;
    P3PP::pvm_upkuint(&i, 1, 1);
    return i; }

inline long pvm_upklong() { 
    long l;
    P3PP::pvm_upklong(&l, 1, 1);
    return l; }


// The following functions add support for the types string and bool.

inline void pvm_pkstring(const string &s)
{
    int len = s.size();
    pvm_pkint(len);
    if(len)
        P3PP::pvm_pkstr(s.c_str());
}

inline string pvm_upkstring()
{
    int len = pvm_upkint();
    if(!len)
        return string();
    char s[len+1];
    P3PP::pvm_upkstr(s);
    string res(s);
    return res;
}

inline void pvm_pkbool(bool b) // Do not use with PvmDataInPlace !!!
{ pvm_pkbyte(b); }

inline bool pvm_upkbool()
{ return pvm_upkbyte(); }

// void notifyHostDeleteAll(int msgtag);

#endif // PVM_UTIL_H
