/*
 *	   PVM version 3.4:  Parallel Virtual Machine System
 *		 University of Tennessee, Knoxville TN.
 *	     Oak Ridge National Laboratory, Oak Ridge TN.
 *		     Emory University, Atlanta GA.
 *	Authors:  J. J. Dongarra, G. E. Fagg, M. Fischer
 *	    G. A. Geist, J. A. Kohl, R. J. Manchek, P. Mucci,
 *	   P. M. Papadopoulos, S. L. Scott, and V. S. Sunderam
 *		     (C) 1997 All Rights Reserved
 *
 *				NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * Neither the Institutions (Emory University, Oak Ridge National
 * Laboratory, and University of Tennessee) nor the Authors make any
 * representations about the suitability of this software for any
 * purpose.  This software is provided ``as is'' without express or
 * implied warranty.
 *
 * PVM version 3 was funded in part by the U.S. Department of Energy,
 * the National Science Foundation and the State of Tennessee.
 */

// Modified version of pvm_startpvmd() that uses vfork() instead of fork().
// Philipp Berndt, Summer 1999

#include <stdio.h>
#include <fcntl.h>
#include <malloc.h>
#include <strings.h>
#include <wait.h>
#include <signal.h>
#include <unistd.h>
#include <sys/stat.h>

// From pvm3/include:
#include "pvm3.h"
#include "pvmtev.h"

// From pvm3/src:
#include "lpvm.h"
#include "tevmac.h"
#include "pvmalloc.h"
#include "bfunc.h"

int   lpvmerr(char *f, int n);
int   pvmbeatask();
char *pvmdsockfile(); 
char *pvmgetpvmd();
int   pvmlogerror(char *s);
int   pvmputenv(char *s);
int   pvmsleep(int secs);
int   tev_begin(int kind, int entry_exit );
int   tev_fin();

/** Starts new PVM daemon using vfork().
 *
 *     Apart from pvm_start_pvmd(), this function uses vfork() instead of
 *     fork(), thus making it possible to to start PVM from within programs
 *     that use a lot of memory.
 *     Don't use this if you have any Posix threads running!
 *
 *@arg argc   number of args to pass to pvmd (>= 0)
 *@arg argv   args for pvmd or null
 *@arg block  if true, don't return until add hosts are started
*/
int pvm_start_pvmd2(int argc, char **argv, int block)
{
	char *sfn;
	struct stat sb;
	int cc;
	char *fn;			/* file to exec */
	char **av;
	int pfd[2];
	int n;
	FILE *ff;
	char buf[128];


	TEV_DECLS;

	if (TEV_EXCLUSIVE) {
		if (pvmmytid != -1
		&& TEV_DO_TRACE(TEV_START_PVMD,TEV_EVENT_ENTRY)) {
		    TEV_PACK_INT( TEV_DID_BF, TEV_DATA_SCALAR, &block, 1, 1 );
		    TEV_PACK_STRING( TEV_DID_AS, TEV_DATA_ARRAY,
				     argv, argc, 1 );
		    TEV_FIN;
		}
	}

	if (argc < 0 || !argv)
		argc = 0;

	if ((pvm_useruid = getuid()) == -1) {
		pvmlogerror("can't getuid()\n");
		cc = PvmSysErr;
		goto bail;
	}

	if (!(sfn = pvmdsockfile())) {
		pvmlogerror("pvm_start_pvmd() pvmdsockfile() failed\n");
		cc = PvmSysErr;
		goto bail;
	}

	if (stat(sfn, &sb) != -1) {
		cc = PvmDupHost;
		goto bail;
	}


#ifdef	IMA_TITN
	if (socketpair(AF_UNIX, SOCK_STREAM, 0, pfd) == -1)
#else
	if (pipe(pfd) == -1)
#endif
	{
		cc = PvmSysErr;
		goto bail;
	}

	fn = pvmgetpvmd();

	av = TALLOC(argc + 2, char *, "argv");
	if (argc > 0)
		BCOPY((char *)&argv[0], (char *)&av[1], argc * sizeof(char*));
	av[0] = fn;
	av[argc + 1] = 0;

	if (!vfork()) {
		(void)close(pfd[0]);
    /* fork again so the pvmd is not the child - won't have to wait() for it */
		if (!vfork()) {
			if (pfd[1] != 1)
				dup2(pfd[1], 1);
			for (n = getdtablesize(); n-- > 0; )
				if (n != 1)
					(void)close(n);
			(void)open("/dev/null", O_RDONLY, 0); /* should be 0 */
			(void)open("/dev/null", O_WRONLY, 0); /* should be 2 */
			(void)signal(SIGINT, SIG_IGN);
			(void)signal(SIGQUIT, SIG_IGN);
			(void)signal(SIGTSTP, SIG_IGN);
			execvp(av[0], av);
		}
		_exit(0);
	}
	(void)close(pfd[1]);
	(void)wait(0);
	PVM_FREE(av);

	if (!(ff = fdopen(pfd[0], "r"))) {
		cc = PvmSysErr;
		(void)close(pfd[0]);
		goto bail;
	}

	strcpy(buf, "PVMSOCK=");
	n = strlen(buf);
	if (!fgets(buf + n, sizeof(buf) - n - 1, ff)) {
		cc = PvmCantStart;
		fclose(ff);
		goto bail;
	}
	fclose(ff);
	if (strlen(buf + n) < 2) {
		cc = PvmCantStart;
		goto bail;
	}
	n = strlen(buf);
	if (buf[n - 1] == '\n')
		buf[n - 1] = 0;
	pvmputenv(STRALLOC(buf));

	/* fprintf(stderr, "pvm_start_pvmd: %s\n", buf); */


	if (cc = BEATASK)
		goto bail;

	if (block) {
		struct pvmhostinfo *hip;
		int t = 1;

		pvm_config((int*)0, (int*)0, &hip);
		while ((cc = pvm_addhosts(&hip[0].hi_name, 1, (int*)0))
		       == PvmAlready) {
			  pvmsleep(t);
			  if (t < 8)
				t *= 2;
		}
		if (cc != PvmDupHost)
			goto bail;
		cc = BEATASK;
	}

bail:
	if (TEV_AMEXCL) {
		if (TEV_DO_TRACE(TEV_START_PVMD,TEV_EVENT_EXIT)) {
			TEV_PACK_INT( TEV_DID_CC, TEV_DATA_SCALAR, &cc, 1, 1 );
			TEV_FIN;
		}
		TEV_ENDEXCL;
	}

	return (cc < 0 ? lpvmerr("pvm_start_pvmd", cc) : 0);
}
