// File    tnet_util.cpp  (C++ Implementation)
// Module  tnet_util      (TimeNET specific functions related to PVM)
// Library PVM3CPP        (C++ Extension to PVM3.4)
/**
 *@author  Philipp Berndt
 *@version $Id: tnet_util.cpp,v 1.2 2000/01/12 16:02:04 pumphutt Exp $
 */

#include <algorithm> // find()
#include <string>
#include <signal.h>
#include <fstream>
#include <strstream>

#include "pvm3wrap.h"
using namespace P3PP;
#include "pvm_util.h"
#include "HostConfig.h"
using namespace HostConfig;
#include "tnet_util.h"

namespace {
    int tnetauth_pid; // unix process id of idle process tnetauth
    const string TNET_RC(".TimeNET"); // name of TimeNET configuration file
}

/************************** INTERNAL HELP FUNCTIONS **************************/

/** Called when SIGTERM or SIGINT is received. */
void TerminateHandler(int sig)
{
    cout << "Caught Signal " << sig <<". Forwarding to tnetauth." << endl;
    kill(tnetauth_pid, SIGTERM);
}

/** Starts idle process tnetauth and installs signal handler. */
void installTerminateHandler()
{
    // Spawn tnetauth and request notification when it dies.
    string path = string(getenv("TNETHOME")) + "/Shared/scripts/pvm_tnet_auth";

    // Don't trace output from tnet_auth
    int otid = pvm_setopt(PvmOutputTid, 0);
    int ttid = pvm_setopt(PvmTraceTid, 0);
    
    int tid = pvm_spawn(path.c_str(), vector<string>(), PvmTaskHost, ".");
    pvm_notify(PvmTaskExit, TAG_TERMINATE, 1, &tid);
    tnetauth_pid = pvm_tid2pid(tid);

    pvm_setopt(PvmOutputTid, otid);
    pvm_setopt(PvmTraceTid, ttid);

    // Block other terminal-generated signals while handler runs.
    sigset_t block_mask;
    sigemptyset(&block_mask);
    sigaddset(&block_mask, SIGINT);
    sigaddset(&block_mask, SIGTERM);

    struct sigaction setup_action;
    setup_action.sa_handler = TerminateHandler;
    setup_action.sa_mask = block_mask;
    setup_action.sa_flags = SA_RESTART;
    sigaction(SIGINT, &setup_action, NULL);  // Install for SIGINT
    sigaction(SIGTERM, &setup_action, NULL); // Install for SIGTERM
}

/** Reads one line from a stream and puts it in a string. */
string readline(istream &is)
{
    const int BUFSIZE = 100;
    char buffer[BUFSIZE];
    string res;
  
    do  {
        is.getline(buffer, BUFSIZE, '\n');
        res += buffer;
        } while (is.gcount() == BUFSIZE);

    return res;
}

/** Reads HOSTLIST from ".TimeNET". */
vector<Host> readHostlist()
{
    vector<Host> hosts;

    ifstream ifs(( string(getenv("HOME")) + "/" + TNET_RC ).c_str());
    
    bool inlist = false;
    int speed = Host::DEFAULT_SPEED;
    do  {
        string line = readline(ifs);
        int p;
        if ((p=line.find("#"))!=string::npos)
            line.erase(p);
        if (line.empty())
            continue;
        istrstream is(line.c_str());
        
        do  {
            string s;
            is >> s;
            if (!inlist)
                {
                if (s == "HOSTLISTSTART")
                    inlist = true;
                }
            else // inlist
                {
                if (s == "HOSTLISTEND")
                    return hosts;

                bool use_always = false;
                if (s == "USE")
                    {
                    is >> s;
                    if (s != "ALWAYS")
                        {
                        cerr << "USE without ALWAYS" << endl;
                        }
                    use_always = true;
                    is >> s;
                    }

                // strip garbage at the end of hostname
                for (string::iterator i = s.begin(); i != s.end(); ++i)
                    if (!isalnum(*i))
                        {
                        s.erase(i);
                        break;
                        }

                hosts.push_back(Host(s, speed, Host::OFFLINE,
                                     !use_always, false));
                break;
                } // if (inlist)
            } while (!is.eof());
        } while (!ifs.eof());
 
    if (inlist)
        cerr << "Unexpected end of HOSTLIST" << endl;
    else
        cerr << "No host list found" << endl;
    return hosts;
}


/***************************** EXPORTED FUNCTIONS ****************************/
/********************** see header file for description **********************/

bool checkSystemMessage(int msgtag)
{
    switch(msgtag)
        {
        case TAG_HOST_ADDED:
            HostConfig::hostAdded();
            return true;
        case TAG_HOST_DELETED:
            HostConfig::hostDeleted();
            return true;
        case TAG_TERMINATE:
            throw InterruptedException();
        }
    return false;
}

void initPvm()
{
    cout.setf(ios::showbase); // Always print 0x before hex values
    
    vector<HostConfig::Host> hostlist = readHostlist();  // Read .TimeNET
    HostConfig::init(hostlist);

    putenv("PVM_EXPORT=TNETHOME:MODELDIR");
    installTerminateHandler(); // Generate Terminate-msg from SIGINT, SIGTERM
    pvm_setopt(PvmAutoErr, 0); // Don't print out warnings
}

void shutdownPvm()
{
    cout << "PVM Shutdown sequence initiated" << endl;
    // Terminate all child processes
    int mytid = pvm_mytid();
    int ntask;
    pvmtaskinfo *taskinfos;
    PVM3::pvm_tasks(0, &ntask, &taskinfos); bool strangers = false;
    for (int i=0; i<ntask; ++i)
        {
        int tid = taskinfos[i].ti_tid;
        if (tid == 0 || tid == mytid)
            continue;  // Spare zombies, myself...

        if (!strcmp(taskinfos[i].ti_a_out,"pvmgs")) // ...and pvmgs.
            {
            cout << "Leaving alone " << taskinfos[i].ti_a_out 
                 << " TID:" << hex << tid << dec << endl;
            continue;
            }

        if (taskinfos[i].ti_ptid == mytid) // Child process
            {
            cout << "Killing " << taskinfos[i].ti_a_out 
                 << " TID:" << hex << tid << dec << endl;
            pvm_kill(tid);
            }
        else
            {
            cout << "Leaving alone " << taskinfos[i].ti_a_out 
                 << " TID:" << hex << tid << dec << endl;
            strangers = true;
            }
        }

    if (strangers)
        HostConfig::removehosts(); // Just remove hosts that we added.
    else
        {
        cout << "No other tasks running. Halting PVM..." << flush;
        pvm_halt();
        cout << "OK" << endl;
        }
}
