/*****************************************************************************

Programm	:	estimse

Komponente	:	estimate.c
                        Zustandsraumabschaetzung auf Basis
			der Zustandsgleichung (state equation)

Programmer	:	Alexander Huck
Date		:	1999

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "polycount.h"
#include "matrix.h"
#include "matrixIO.h"
#include "matrixOps.h"
#include "matrixProcs.h"
#include "hermite2.h"
#include "inequalities.h"
#include "intOps.h"
#include "invars.h"
#include "siphons.h"
#include "traps.h"

#define MAXROWSIZE 100000000   /*Maximum number of matrix-rows allowed during inequation solving*/
#define MIN(a,b) a<b ? a:b


/*#define DEBUG*/
/*
#define PRINTCONSTRAINTS
*/

/*******************************************************************************************
  GCDReduce : compute gcd of all input/ output arc valuations of each place
  and divide initial marking and arc valuations of the place by  this gcd.
*******************************************************************************************/
void GCDReduce(INTMATRIX *IncidenceMat, INTMATRIX *PostMat, INTMATRIX *PreMat, INTMATRIX *M)
{
  INTMATRIX *TransInc, *GCD;
  int i, j;
  int nbplaces, nbtransitions;
  
  /*GCD - REDUCTION */
  
  nbplaces = IncidenceMat->C;
  nbtransitions = IncidenceMat->R;
  
  TransInc = TransMat(IncidenceMat);
  GCD = ZeroMatrix(nbplaces, 1);
  
  for(i=0; i< nbplaces; i++)
    {
      /*compute gcd of all input and output arcs of place i*/
      GCD->V[i][0] = VectorGcd(TransInc->V[i]-1, TransInc->C+1);
    }
  
  /*reduce marking/ arc valuation */
  for(i=0; i< nbplaces; i++)
    {
      if (GCD->V[i][0]>1)
	{
	  M->V[i][0] = M->V[i][0] / GCD->V[i][0]; /*divide marking by gcd*/
	  for(j=0; j<nbtransitions; j++)
	    {
	      IncidenceMat->V[j][i] = IncidenceMat->V[j][i] / GCD->V[i][0];
	      if(PostMat)
		PostMat->V[j][i] = PostMat->V[j][i] / GCD->V[i][0];
	      if(PreMat)
		PreMat->V[j][i] = PreMat->V[j][i] / GCD->V[i][0];  
	    }
	}
    }
  FreeMat(TransInc);
  FreeMat(GCD);
}

/*******************************************************************************************
 EstimateStateEq: estimate state space based on state equation
 Input:initial marking M
       Incidence-matrix,
 Output: size of state-space
*******************************************************************************************/
int EstimateStateEq(INTMATRIX *_M, INTMATRIX *_IncidenceMat)
{
INTMATRIX *CONSTRAINTS;
INTMATRIX  *M, *IncidenceMat; 
USI row;
int i,j;
USI nbtransitions = 0, nbplaces = 0;
USI nbvars = 0, nbconstraints = 0;
int statespace=1;
#ifdef DEBUG
 char filename[50];
 FILE *outfile;
#endif

 M = CopyMat(_M);
 IncidenceMat = CopyMat(_IncidenceMat);

#ifdef DEBUG
 strcpy(filename, "esttest.out");
 printf("filename : %s\n", filename);
 outfile = fopen(filename, "w");
 
 printf("ESTIMATE: initial marking \n");
 PRINTMATI(0,M->R-1,0,M->C-1,M);
 fprintf(outfile,"ESTIMATE: initial marking \n");
 FPRINTMATI(outfile,0,M->R-1,0,M->C-1,M);
 printf("Incidence matrix: transis ------>\n"); 
 PRINTMATI(0,IncidenceMat->C-1,0,IncidenceMat->R-1,TransMat(IncidenceMat));
 fprintf(outfile,"Incidence matrix: transis ------>\n"); 
 FPRINTMATI(outfile,0,IncidenceMat->C-1,0,IncidenceMat->R-1,TransMat(IncidenceMat));
#endif
 
 nbtransitions = IncidenceMat -> R;
 nbplaces = IncidenceMat -> C;
 
#ifdef DEBUG
 printf("The net has %i places and %i transitions\n", nbplaces, nbtransitions);
#endif
 
 if (nbplaces == 0) return 0; /*no places, no marking*/
 if (nbtransitions == 0) return 1; /*no transitions: only initial marking*/
  
 /*GCD - REDUCTION */
 GCDReduce(IncidenceMat, 0, 0, M);

 /*derive the number of variables*/
 nbvars = nbplaces; /*one variable for each place*/
 nbvars += nbtransitions; /*characteristic vector sigma: one var for each transition*/
 
 /*derive the number of constraints*/
 nbconstraints += nbplaces; /*equations for M = M0 +C.Sigma =0*/
 nbconstraints += nbplaces + nbtransitions; /*inequalities for M>=0 and Sigma>=0*/

 /*init matrix for constraints*/
 CONSTRAINTS = ZeroMatrix(nbconstraints, nbvars + 2);
 
 /*equations for M - C.Sigma - M0 =0*/
 row = 0;
 for(i=0; i< nbplaces; i++)
   {
     CONSTRAINTS->V[row][1 + i] = 1; /*mark for mi*/
     for(j=0; j< nbtransitions; j++)
       {
	 CONSTRAINTS->V[row][1 + nbplaces + j] =  -IncidenceMat->V[j][i];
       }
     CONSTRAINTS->V[row][nbvars + 1] =  -M->V[i][0];
     row++;
   }
 
 /*M>=0, Sigma>= 0, firing_vectors >= 0*/
 for(i=0; i< nbplaces+nbtransitions; i++)
   {
     CONSTRAINTS->V[row][0] = 1; /*mark for inequality*/
     CONSTRAINTS->V[row][1 + i] = 1; /*mark for M_i*/
     row++;
   }
 
#ifdef PRINTCONSTRAINTS
 printf("Constraint-matrix, unmodified:\n");
 PRINTMATI(0,CONSTRAINTS->R-1,0,CONSTRAINTS->C-1,CONSTRAINTS);
#endif
 RemoveZeroRows(&CONSTRAINTS);

 /*eliminate identical rows and simple constraints*/
 ElimEqualRows(&CONSTRAINTS);
 ElimSimpleConstraints(&CONSTRAINTS);

 /*sort the constraints matrix*/
 SortRows(CONSTRAINTS);
 
#ifdef DEBUG
 printf("Sortiert und eliminiert:\n ");
 PRINTMATI(0,CONSTRAINTS->R-1,0,CONSTRAINTS->C-1, CONSTRAINTS);
 fprintf(outfile,"Sortiert und eliminiert:\n ");
 FPRINTMATI(outfile,0,CONSTRAINTS->R-1,0,CONSTRAINTS->C-1,CONSTRAINTS);
 fflush(outfile);
#endif
 
 /*derive content of polyhedron defined by the constraints*/
 statespace = PolyEnum(CONSTRAINTS, nbplaces); /*from polycount.c*/
 
 return statespace;
}

int EstimateBackTr(INTMATRIX *_M, INTMATRIX *_IncidenceMat, INTMATRIX *_PostMat, int level)
     /*input: initial marking M0
       Incidence-matrix, Post-matrix (output-arcs),
       backtrace level:
       level 0 = consider only state equation and invariants equation
       level 1 = from the markings defined by state and invariants equation
                 undo the last firing of a transition and check if there exists
		 a predecessor marking
       level 2, 3, 4 = go back 2, 3, 4 ... steps 
       output: size of state-space*/
{
  INTMATRIX *CONSTRAINTS;
  INTMATRIX *M, *IncidenceMat, *PostMat;
  USI row;
  int i,j,k,l;
 
 
  USI nbtransitions = 0, nbplaces = 0;
  USI nbvars = 0, nbconstraints = 0;
  int statespace=1;
#ifdef DEBUG
 char filename[50];
 FILE *outfile;
#endif

  M = CopyMat(_M);
  IncidenceMat = CopyMat(_IncidenceMat);
  PostMat = CopyMat(_PostMat);
  
#ifdef DEBUG
  strcpy(filename, "esttest.out");
  printf("filename : %s\n", filename);
  outfile = fopen(filename, "w");

 printf("ESTIMATE: initial marking \n");
 PRINTMATI(0,M->R-1,0,M->C-1,M);
 fprintf(outfile,"ESTIMATE: initial marking \n");
 FPRINTMATI(outfile,0,M->R-1,0,M->C-1,M);
 printf("Incidence matrix: transis ------>\n"); 
 PRINTMATI(0,IncidenceMat->C-1,0,IncidenceMat->R-1,TransMat(IncidenceMat));
 fprintf(outfile,"Incidence matrix: transis ------>\n"); 
 FPRINTMATI(outfile,0,IncidenceMat->C-1,0,IncidenceMat->R-1,TransMat(IncidenceMat)); 
#endif

  nbtransitions = IncidenceMat -> R;
  nbplaces = IncidenceMat -> C;
  
#ifdef DEBUG
  printf("The net has %i places and %i transitions\n", nbplaces, nbtransitions);
#endif
  
  if (nbplaces == 0) return 0; /*no places, no marking*/
  if (nbtransitions == 0) return 1; /*no transitions: only initial marking*/
  
  /*derive the number of variables*/
  
  nbvars = nbplaces; /*one variable for each place*/
  nbvars += nbtransitions; /*characteristic vector sigma*/
  nbvars += level * nbtransitions; /*one firing vector for each level*/
  
  /*derive the number of constraints*/
  
  nbconstraints += nbplaces; /*equations for M = M0 +C.Sigma =0*/
  nbconstraints += nbplaces + nbtransitions; /*inequations for M>=0 and Sigma>= 0*/
  nbconstraints += level * (nbplaces + nbtransitions); /*inequations for backtracking:*/
  /*M- ... - POST.e>=0 and e>=0*/
  nbconstraints += level; /*one equation for each firing vector: ||e|| = 1*/
  nbconstraints += nbtransitions * (level > 0); /*Sigma - (firing vectors) >=0 */
  
  /*init matrix for constraints*/
  CONSTRAINTS = ZeroMatrix(nbconstraints, nbvars + 2);
  
  /*equations for M - C.Sigma - M0 =0*/
  row = 0;
  for(i=0; i< nbplaces; i++)
    {
      CONSTRAINTS->V[row][1 + i] = 1; /*mark for mi*/
      for(j=0; j< nbtransitions; j++)
	{
	  CONSTRAINTS->V[row][1 + nbplaces + j] =  -IncidenceMat->V[j][i];
	}
      CONSTRAINTS->V[row][nbvars + 1] =  -M->V[i][0];
      row++;
    }
  
  
  /*inequalities for backtracking:*/
  for(k=1; k<= level; k++)
    {
      for(i=0; i< nbplaces; i++)
	{
	  CONSTRAINTS->V[row][0] = 1; /*mark for inequality*/
	  CONSTRAINTS->V[row][1 + i] = 1; /*mark for m_i*/
	  for(l=1; l< k ; l++)
	    {
	      for(j=0; j< nbtransitions; j++)
		{
		  /*-C.firingvector_l*/
		  CONSTRAINTS->V[row][1 + nbplaces + j + l * nbtransitions] = - IncidenceMat->V[j][i];
		}
	    }
	  
	  for(j=0; j< nbtransitions; j++)
	    {
	      /*-PostMat.firingvector_k*/
	      CONSTRAINTS->V[row][1 + nbplaces + j + k * nbtransitions] = - PostMat->V[j][i];
	    }
	  
	  row++;
	}
    }
  /*Sigma - (firing vectors) >=0 */
  if (level > 0)
    {
      for(i=0; i< nbtransitions; i++)
	{
	  CONSTRAINTS->V[row][0] = 1; /*mark for inequality*/
	  CONSTRAINTS->V[row][1 + i + nbplaces] = 1; /*mark for s_i*/
	  for(k=1; k<= level; k++)
	    {
	      CONSTRAINTS->V[row][1 + i + nbplaces + k * nbtransitions] = -1;
	    }
	  row++;
	}
    }
  /*one equation for each firing vector: ||e|| = 1*/
  for(k=1; k<= level; k++)
    {
      for(i=0; i< nbtransitions; i++)
	{
	  CONSTRAINTS->V[row][1 + nbplaces + k * nbtransitions + i] = 1; /*mark for firingvector_k*/
	}
      CONSTRAINTS->V[row][nbvars + 1] = - 1;
      row++;
    }
  /*M>=0, Sigma>= 0, firing_vectors >= 0*/
  
  for(i=0; i< nbplaces; i++)
    {
      CONSTRAINTS->V[row][0] = 1; /*mark for inequality*/
      CONSTRAINTS->V[row][1 + i] = 1; /*mark for M_i*/
      row++;
    }
  
  for(i=0; i< nbtransitions; i++)
    {
      CONSTRAINTS->V[row][0] = 1; /*mark for inequality*/
      CONSTRAINTS->V[row][1 + nbplaces + i] = 1; /*mark for Sigma_i*/
      row++;
    }
  
  for(k=1; k<= level; k++)
    {
      for(i=0; i< nbtransitions; i++)
	{ 
	  CONSTRAINTS->V[row][0] = 1; /*mark for inequality*/
	  CONSTRAINTS->V[row][1 + nbplaces + k * nbtransitions + i] = 1; /*mark for firingvector_k*/
	  row++;
	}
    }
  
  
#ifdef PRINTCONSTRAINTS
  printf("Constraint-matrix:\n");
  PRINTMATI(0,CONSTRAINTS->R-1,0,CONSTRAINTS->C-1,CONSTRAINTS);
#endif
  
  RemoveZeroRows(&CONSTRAINTS);
  
  /*eliminate identical rows and simple constraints*/
  ElimEqualRows(&CONSTRAINTS);
  ElimSimpleConstraints(&CONSTRAINTS);
  
  /*sort the constraints matrix*/
  SortRows(CONSTRAINTS);
  
#ifdef DEBUG
  printf("Sortiert und eliminiert:\n ");
  PRINTMATI(0,CONSTRAINTS->R-1,0,CONSTRAINTS->C-1, CONSTRAINTS);
  fprintf(outfile,"Sortiert und eliminiert:\n ");
  FPRINTMATI(outfile,0,CONSTRAINTS->R-1,0,CONSTRAINTS->C-1,CONSTRAINTS);
  fflush(outfile);
#endif
  
  statespace = PolyEnum(CONSTRAINTS, nbplaces);
  
  return statespace;
}


int EstimateTrapSiphon( INTMATRIX *_M, INTMATRIX *_IncidenceMat, INTMATRIX *_PostMat, INTMATRIX *_PreMat)
     /*input: initial marking M
       Incidence-matrix, Post-matrix (output-arcs), Pre-Matrix (Input arcs)
       this version adds constraints for traps and siphons 
       output: size of state-space*/
{
INTMATRIX *CONSTRAINTS;
INTMATRIX  *M, *IncidenceMat, *PostMat, *PreMat; 
int numtraps, numsiphons;

USI row;
int sum;

int i,j;
USI nbtransitions = 0, nbplaces = 0;
USI  nbvars = 0, nbconstraints = 0;

int statespace=1;

#ifdef DEBUG
 char filename[50];
 FILE *outfile;
#endif

 M = CopyMat(_M);
 IncidenceMat = CopyMat(_IncidenceMat);
 PostMat = CopyMat(_PostMat);
 PreMat = CopyMat(_PreMat);

#ifdef DEBUG
 strcpy(filename, "esttest.out");
 printf("filename : %s\n", filename);
 outfile = fopen(filename, "w");
 
 printf("ESTIMATE: initial marking \n");
 PRINTMATI(0,M->R-1,0,M->C-1,M);
 fprintf(outfile,"ESTIMATE: initial marking \n");
 FPRINTMATI(outfile,0,M->R-1,0,M->C-1,M);
 printf("Incidence matrix: transis ------>\n"); 
 PRINTMATI(0,IncidenceMat->C-1,0,IncidenceMat->R-1,TransMat(IncidenceMat));
 fprintf(outfile,"Incidence matrix: transis ------>\n"); 
 FPRINTMATI(outfile,0,IncidenceMat->C-1,0,IncidenceMat->R-1,TransMat(IncidenceMat));
#endif
 
 nbtransitions = IncidenceMat -> R;
 nbplaces = IncidenceMat -> C;
 
#ifdef DEBUG
 printf("The net has %i places and %i transitions\n", nbplaces, nbtransitions);
#endif
 
 if (nbplaces == 0) return 0; /*no places, no marking*/
 if (nbtransitions == 0) return 1; /*no transitions: only initial marking*/
  
 /*from PRE and POST generate the modified Incidence-matrix NTheta (->traps)
   in NTheta all entrys belonging to the output-arcs of a transition are
   replaced by the sum of its input-arcs*/
 
#ifdef DEBUG
 printf("Incidence matrix: transis ------>\n"); 
 PRINTMATI(0,IncidenceMat->C-1,0,IncidenceMat->R-1,TransMat(IncidenceMat));
 printf("POST matrix: transis ------>\n"); 
 PRINTMATI(0,PostMat->C-1,0, PostMat->R-1,TransMat(PostMat));
 printf("PRE matrix: transis ------>\n"); 
 PRINTMATI(0,PreMat->C-1,0,PreMat->R-1,TransMat(PreMat));
#endif

 /*GCD - REDUCTION */
 GCDReduce(IncidenceMat, PostMat, PreMat, M);
     
 numtraps=Traps(M, IncidenceMat, PostMat, PreMat);
  
if (numtraps)
  {
#ifdef DEBUG
     printf("Traps:\n ");
     PRINTMATI(0,Trap->R-1,0,Trap->C-1, Trap);
#endif
   } 
 else
   Trap = BuildMat(0,0);
 
 numsiphons=Siphons(M, IncidenceMat, PostMat, PreMat); /*from siphons.c*/
  
  
 if (numsiphons)
   {
#ifdef DEBUG
     printf("Siphons:\n ");
     PRINTMATI(0,Siphon->R-1,0,Siphon->C-1, Siphon);
#endif
   }
 else
   Siphon = BuildMat(0,0);
 
 /*derive the number of variables*/
 
 nbvars = nbplaces; /*one variable for each place*/
 nbvars += nbtransitions; /*characteristic vector sigma*/
 
 /*derive the number of constraints*/
 
 nbconstraints += nbplaces; /*equations for M = M0 +C.Sigma =0*/
 nbconstraints += nbplaces + nbtransitions; /*inequations for M>=0 and Sigma>= 0*/
 nbconstraints += Siphon->R; /*one inequation for each siphon*/
 nbconstraints += Trap->R;/*one inequation for each trap*/
 
 /*init matrix for constraints*/
 CONSTRAINTS = ZeroMatrix(nbconstraints, nbvars + 2);
 

 /*equations for M - C.Sigma - M0 =0*/
 row =0;
 for(i=0; i< nbplaces; i++)
   {
     CONSTRAINTS->V[row][1 + i] = 1; /*mark for mi*/
     for(j=0; j< nbtransitions; j++)
       {
	 CONSTRAINTS->V[row][1 + nbplaces + j] =  -IncidenceMat->V[j][i];
       }
     CONSTRAINTS->V[row][nbvars + 1] =  -M->V[i][0];
     row++;
   }
 
 /*M>=0, Sigma>= 0, firing_vectors >= 0*/
 
 for(i=0; i< nbplaces; i++)
   {
     CONSTRAINTS->V[row][0] = 1; /*mark for inequation*/
     CONSTRAINTS->V[row][1 + i] = 1; /*mark for M_i*/
     row++;
   }
 
 for(i=0; i< nbtransitions; i++)
   {
     CONSTRAINTS->V[row][0] = 1; /*mark for inequation*/
     CONSTRAINTS->V[row][1 + nbplaces + i] = 1; /*mark for Sigma_i*/
     row++;
   }
 
 /*inequations for TRAPS:
   Trap.M >= 1*/
 for(i=0; i< Trap->R; i++)
   {
     sum = 0;
      for (j=0; j< nbplaces; j++)
	{
	  sum+= Trap->V[i][j] * M->V[j][0];
	}
      if(!sum)
	{
	  // printf("Trap not marked\n");
	  continue;
	}
      CONSTRAINTS->V[row][0] = 1; /*mark for inequation*/
     for (j=0; j< nbplaces;j++)
       {
	 if(Trap->V[i][j])
	   CONSTRAINTS->V[row][1 + j] = 1; /*mark for M_i*/
	 CONSTRAINTS->V[row][1 + nbvars] = -1; /*>=1*/
       }
     row++;
   }

 /*equations for SIPHONS:
   Siphon.M = 0*/
 for(i=0; i< Siphon->R; i++)
   {
     sum = 0;
      for (j=0; j< nbplaces; j++)
	{
	  sum+= Siphon->V[i][j] * M->V[j][0];
	}
      if(sum)
	{
	  //printf("Siphon marked\n");
	  continue;
	}
     for (j=0; j< nbplaces; j++)
       {
	 if(Siphon->V[i][j])
	   CONSTRAINTS->V[row][1 + j] = 1; /*mark for M_i*/
       }
     row++;
   }

#ifdef PRINTCONSTRAINTS
 printf("Constraint-matrix:\n");
 PRINTMATI(0,CONSTRAINTS->R-1,0,CONSTRAINTS->C-1,CONSTRAINTS);
#endif

 RemoveZeroRows(&CONSTRAINTS);
  
 /*eliminate identical rows and simple constraints*/
 ElimEqualRows(&CONSTRAINTS);
 ElimSimpleConstraints(&CONSTRAINTS);
 
 /*sort the constraints matrix*/
 SortRows(CONSTRAINTS);
 
#ifdef DEBUG
 printf("Sortiert und eliminiert:\n ");
 PRINTMATI(0,CONSTRAINTS->R-1,0,CONSTRAINTS->C-1, CONSTRAINTS);
 fprintf(outfile,"Sortiert und eliminiert:\n ");
 FPRINTMATI(outfile,0,CONSTRAINTS->R-1,0,CONSTRAINTS->C-1,CONSTRAINTS);
 fflush(outfile);
#endif
 
 statespace = PolyEnum(CONSTRAINTS, nbplaces);

 return statespace;
}
