/*****************************************************************************

Programm	:	struct

Komponente	:	hermite2.c

                        Bestimmt die Hermite-Normalform der Eingabe-Matrix
			nach dem Algorithmus von George Havas


Programmer	:	Alexander Huck
Date		:	1998

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "matrix.h"
#include "matrixOps.h"
#include "matrixIO.h"
#include "intOps.h"

unsigned int MLLLVERBOSE=0;
unsigned int HERMITE1VERBOSE=0;
unsigned int keith97;

USI FlagCol(INTMATRIX *Aptr)
/*
 * Returns 1 if the first nonzero row of Aptr contains more than one
 * entry, 0 otherwise.  Only applied to a nonzero matrix with at least 
 * two columns.
 */
{
	USI i, j, k, flag = 0, m, n;

	n = Aptr->C;
	m = Aptr->R;
	for (j = 0; j < n; j++)
	{
		for (i = 0; i < m; i++)
		{
			if ((Aptr->V[i][j])!=0)
			{
				flag = 1;
				break;
			}
		}
		if (flag)
		{
			for (k = i + 1; k < m; k++)
			{
				if ((Aptr->V[k][j])!=0)
					return (1);
			}
			return (0);
		}
	}
	return (0);
}


void Swap1(USI i, USI k, INTMATRIX **Lptr, int D[])
{
	int X1, X2, X3, Y1, Y2;

	X1 = ((*Lptr)->V[k - 2][i - 1]) *  ((*Lptr)->V[k - 2][k - 1]);
	Y1 = ((*Lptr)->V[k - 1][i - 1]) * (D[k - 2]);
	Y1 = Y1 + X1;
	X2 = ((*Lptr)->V[k - 2][i - 1]) * D[k];
	X3 = -((*Lptr)->V[k - 2][k - 1]);
	Y2 = ((*Lptr)->V[k - 1][i - 1]) *  X3;
	Y2 = Y2 + X2;
	(*Lptr)->V[k - 2][i - 1] = INT(Y1, D[k - 1]);
	(*Lptr)->V[k - 1][i - 1] = INT(Y2, D[k - 1]);
	return;
}

void Swap21(USI k, INTMATRIX **Bptr, INTMATRIX **Lptr, INTMATRIX **Aptr)
{
	int T;

	*Aptr = SwapCols(k - 2, k - 1, *Aptr);
	*Bptr = SwapCols(k - 2, k - 1, *Bptr);
	T = ((*Lptr)->V[k - 1][k - 2]);
	*Lptr = SwapCols(k - 2, k - 1, *Lptr);
	(*Lptr)->V[k - 1][k - 2] = T;
	(*Lptr)->V[k - 2][k - 2] = 0;
	if (HERMITE1VERBOSE)
	{
		printf("Swapping Cols %u and %u\n", k-1,k);
		printf("P = \n");
		PRINTMATI(0,(*Bptr)->R-1,0,(*Bptr)->C-1,*Bptr);
		printf("A = \n");
		PRINTMATI(0,(*Aptr)->R-1,0,(*Aptr)->C-1,*Aptr);
		return;
	}
}

void LLLMinus(INTMATRIX **Aptr, INTMATRIX **Pptr, INTMATRIX **Lptr, USI j)
{
	USI r, s, m;
	INTMATRIX *TempMATI;

	m = (*Aptr)->C;
	for (r = 0; r < m; r++)
		for (s = 0; s < r; s++)
		{
			if (r == j || s == j)
			{
				(*Lptr)->V[r][s] = -((*Lptr)->V[r][s]);
			}
		}
	
	TempMATI = *Aptr;
	*Aptr = ScalarCol(j, -1, *Aptr);
	FreeMat(TempMATI);
	TempMATI = *Pptr;
	*Pptr = ScalarCol(j, -1, *Pptr);
	FreeMat(TempMATI);
	if (HERMITE1VERBOSE)
	{
		printf("Col %u -> - Col %u\n", j + 1,j + 1);
		printf("P = \n");
		PRINTMATI(0,(*Pptr)->R-1,0,(*Pptr)->C-1,*Pptr);
		printf("A = \n");
		PRINTMATI(0,(*Aptr)->R-1,0,(*Aptr)->C-1,*Aptr);
	}
	return;
}



void Reduce(USI k, USI l, INTMATRIX **Lptr, INTMATRIX **Bptr, int D[], INTMATRIX **Aptr, USI *row1, USI *row2)
/*
 * updates *Lptr, *Bptr, *Aptr.
 */
{
        int j, n, t;
	int X, Y, Q;
	INTMATRIX *TempMATI;

	n = (*Aptr)->R;
	*row1 =  n;
	for (j = 0; j < n; j++)
	{
		t = ((*Aptr)->V[j][l - 1]);
		if (t!=0)
		{
			*row1 = j;
			if (t <0)
			  {
			    LLLMinus(Aptr, Bptr, Lptr, l - 1);
			  }
			break;
		}
	}
	*row2 =  n;
	for (j = 0; j < n; j++)
	{
		t = ((*Aptr)->V[j][k - 1]);
		if (t!=0)
		{
			*row2 = j;
			break;
		}
	}

	if (*row1 < n)
		Q = INT((*Aptr)->V[*row1][k - 1], (*Aptr)->V[*row1][l - 1]);
	else
	{
		Y = ((*Lptr)->V[k - 1][l - 1]) *  2;
		if (RSV(Y, D[l]) == 1)
			Q = NEAREST_INT((*Lptr)->V[k - 1][l - 1], D[l]);
		else
			Q = 0;
	}
	if (Q == 0)
	{
		return;
	}
	X = -Q;
	TempMATI = *Bptr;
	*Bptr = AddMultCol(l - 1, k - 1, X, *Bptr);
	FreeMat(TempMATI);
	TempMATI = *Aptr;
	*Aptr = AddMultCol(l - 1, k - 1, X, *Aptr);
	FreeMat(TempMATI);
	if (HERMITE1VERBOSE)
	{
		printf("Col %u -> Col %u + ", k,k);printf(" %i ", X);printf(" x Col %u\n", l);
		printf("P = \n");
		PRINTMATI(0,(*Bptr)->R-1,0,(*Bptr)->C-1,*Bptr);
		printf("A = \n");
		PRINTMATI(0,(*Aptr)->R-1,0,(*Aptr)->C-1,*Aptr);
	}
	for (j = 1; j < l; j++)
	{
		X = ((*Lptr)->V[j - 1][l - 1]) * Q;
		(*Lptr)->V[j - 1][k - 1] = ((*Lptr)->V[j - 1][k - 1]) -  X;
	}
	X = D[l] * Q;
	(*Lptr)->V[l - 1][k - 1] = ((*Lptr)->V[l - 1][k - 1]) - X;
}



INTMATRIX *LLLHermite2(INTMATRIX *DD, INTMATRIX **Aptr, USI m1, USI n1, int *rank)
/* 
 * Input: an m x n matrix DD of MPI's.
 * Output: *Aptr = HNF(DD). Also we return a small transformation matrix B
 * if nullity > 0.
 * The limiting form of  LLL applied to [I_m|N^mD_1|N^{m-1}D_2|...] is used.
 * Normally m1/n1 = 3/4.
 * 23/2/97.
 */
{
	unsigned int *ROW=0;
	unsigned int i, j, k, l, m, n, row1, row2, r;
	int t;
	int *D, X, Y, Z, R, **temp1, **temp2;
	INTMATRIX *L, *B;

	keith97 = FlagCol(DD);
	
	m = DD->R;
	n = DD->C;
	B = IdentityMat(n);
	*Aptr = CopyMat(DD);
	if (!keith97){
		(B->V[n - 1][n - 1]) = -(B->V[n - 1][n - 1]);
	for (j = 0; j < m; j++)
	  ((*Aptr)->V[j][n - 1]) = - (((*Aptr)->V[j][n - 1]));
	}
	
	D = (int *)malloc((1 + n) * sizeof(int));
	for (i = 0; i <= n; i++)
		D[i] = 1;

	L = ZeroMatrix(n, n);

	k = 2;
	while (k <= n)
	{
		Reduce(k, k - 1, &L, &B, D, Aptr, &row1, &row2);
		X = D[k - 2] * D[k];
		Y = D[k - 1] * D[k - 1];
		Y = Y * m1;
		R = (L->V[k - 2][k - 1]) * (L->V[k - 2][k - 1]);
		Z = X + R;
		Z = Z * n1;
		if ((RSV(Y, Z) == 1 && row1 == m && row2 == m ) || (row1 <= row2 && row1 < m ))
		{
			for (i = k + 1; i <= n; i++)
			  Swap1(i, k, &L, D);
			Swap21(k, &B, &L, Aptr);
			Y = (L->V[k - 2][k - 1]) * (L->V[k - 2][k - 1]);
			Y = Y + X;
			D[k - 1] = INT(Y, D[k - 1]);
			if (k > 2)
				k--;
		}
		else
		{ 
			for (l = k - 2; l >= 1; l--)
				Reduce(k, l, &L, &B, D, Aptr, &row1, &row2);
			k++;
		}
	}

	FreeMat(L);
	for (i = 0; i <= n; i++)
	ROW = (USI *)malloc((1 + n) * sizeof(USI));
	ROW[0] = 0;
	for (t = n - 1; t >= 0; t--)
	{
		r = RowSeek((*Aptr), t, ROW[n-1-t]);
		if (r == m) 
			break;
		else
			ROW[n-t] = r;
	}
	if (t == -1) 
		*rank = n;
	else
		*rank = n - t - 1;

	if(*rank==0) return (B); /*Input was zero Matrix*/

	temp1 = (int **)malloc(m * sizeof(int *));
	for (i = 0; i < m; i++)
		temp1[i] = (int *)malloc(*rank * sizeof(int));

	temp2 = (int **)malloc(n * sizeof(int *));
	for (i = 0; i < n; i++)
		temp2[i] = (int *)malloc(n * sizeof(int));



	for (i = 0; i < *rank; i++)
	{
	  for(j = 0; j < m; j++)
	    {
		temp1[j][i] = (*Aptr)->V[j][n - 1 - i];
	    }
	  for(j = 0; j < n; j++)
	    {
	      	temp2[j][i] = B->V[j][n - 1 - i];
	    }

	}

	for (i = n - 1; i >= *rank; i--)
	{
	   for(j = 0; j < m; j++)
	    {
		(*Aptr)->V[j][i] = (*Aptr)->V[j][i - *rank];
	
	    }
	  for(j = 0; j < n; j++)
	    {
	      	B->V[j][i] = B->V[j][i - *rank];
	    }
	}

	for (i = 0; i < *rank; i++)
	{
	   for(j = 0; j < m; j++)
	    {
		(*Aptr)->V[j][i] = temp1[j][i];
	    }

	   for(j = 0; j < n; j++)
	    {
	      	B->V[j][i] = temp2[j][i];
	    }
	}

	free(ROW);
	free(temp1);
	free(temp2);
	free(D);
	return (B);
}

