/*****************************************************************************

Programm	:	estimse

Komponente	:	remove_redundants.c
                        Redundante Constraints entdecken und entfernen

Programmer	:	Alexander Huck
Date		:	1999

*****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <malloc.h>

#include "sharedVars.h"
#include "intOps.h"
#include "matrixOps.h"
#include "matrixCombine.h"
#include "matrixIO.h"
#include "matrix.h"

/*#define DEBUG*/

#define exchange(a, b, t)\
{ (t)=(a); (a)=(b); (b)=(t); }


int *newraytemprow, *newsattemprow, *mattemprow, *gausstemprow;

/*
  Combine:
  input:  rays r1 and r2
  output: ray r3 such that r3=lamda1*r1 + lamda2*r2
                      and  r3 saturates the constraint a: r3.a = 0

  Modified version: does not change the status of a ray
*/
static void Combine(int r1, int r2, int r3, int pos, INTMATRIX *Mptr)
{
 int D, lamda1, lamda2, status;

 lamda1 = Mptr->V[r1][pos]; /*first column contains r1.a */
 lamda2 = Mptr->V[r2][pos]; /*first column contains r2.a */
 status = Mptr->V[r3][0];   /*save status*/
 D  = Gcd(abs(lamda1), abs(lamda2));
 lamda1 = lamda1/D;
 lamda2 = lamda2/D;
 CombineRow(r1, r2, r3, lamda2, -lamda1, Mptr); /*write new row r3*/
 Mptr->V[r3][0] = status;   /*restore status*/
} /* Combine */



/* Gauss --
   Computes a minimal system of equations.
   performs Gaussian elimination, detects contridiction.
   Given Mat, a matrix of constraints in which the first NbEq
      constraints are equations,
   Given NbEq, the number of equations in Mat
   Given Dimension, the dimension of the homonogenous system,
   Produce, col_of_rank, the vector of independent variables,
   Uses temp, a temporary vector,
   Returns the rank, if Rank=Dimension ... is there a contradiction
 */
int Gauss(Mat, NbEq, Dimension, temp)
INTMATRIX *Mat;
int     NbEq, Dimension;
int     *temp;
{
  int RowSize = Dimension * sizeof(int);
  int *col_of_rank;
  int i, j, k, pivot, gcd, Rank, *cp;

  col_of_rank=(int *)malloc(RowSize);
  gausstemprow = (int *)malloc(Mat->C  * sizeof(int)); 

  Rank=0;
  for (j=1; j<=Dimension; j++)         	/* for each column (except status) */ 
  { for (i=Rank; i<NbEq; i++)           /* starting at diagonal, look down */
      if (Mat->V[i][j]!=0) break;       /* to find the first non zero */
    if (i!=NbEq)                        /* was one found ? */
    { if (i!=Rank)                      /* was it found below the diagonal? */

      RowExchange(Mat, Rank, i, gausstemprow);
    
    /* Normalize the pivot row */
    if ((gcd=VectorGcd(Mat->V[Rank], Dimension))>=2 )
      { cp=&Mat->V[Rank][1];
      for (k=0; k<Dimension; k++) *cp++/=gcd;
      }
    if (Mat->V[Rank][j]<0)
      { cp = Mat->V[Rank]+1;
      for (k=0; k<Dimension; k++) *cp++ *=-1;
      }
    /* end normalize */
    
    pivot=i;
    for (i=pivot+1; i<NbEq; i++)      /* zero out the rest of the column */
      { if (Mat->V[i][j]!=0)
	Combine(i, Rank, i, j, Mat);
      }
    col_of_rank[Rank]=j;
      Rank++;
    }
  } /* end of Gauss elimination */
  
#ifdef DEBUG
  /*
  printf("end of Gauss elimination: \n");
  PRINTMATI(0,Mat->R-1,0,Mat->C-1,Mat);
  */
#endif 
  

  /* Back Substitution -- normalizes the system of equations */
  for (k=Rank-1; k>=0; k--)
  { j = col_of_rank[k];
    /* Normalize the equations */
    for (i=0; i<k; i++)
    { if (Mat->V[i][j]!=0 )
    Combine(i, k, i, j, Mat);

    }

#ifdef DEBUG
    /*
    printf(" Normalize the equations: \n");
    PRINTMATI(0,Mat->R-1,0,Mat->C-1,Mat);
    */
#endif 

    /* Normalize the inequalities */
    for (i=NbEq; i<Mat->R; i++)
    { if (Mat->V[i][j]!=0 )
    Combine(i, k, i, j, Mat);
    }
  }
  free(col_of_rank);
  free(gausstemprow);
  return Rank;
}

/* Remove_Redundants --
   Given (Mat) a matrix of equations and inequalities, and
   given (Ray) a matrix of rays and lines
   given (Sat) the saturation matrix,
   Return a polyhedron, composed of Mat and Ray after reductions.
   Usually called as a follup to Chernikova to remove redundant
   equations.  (Chernikova ensures that there are no redundant
   lines and rays).
*/

int Remove_Redundant_Constraints(Mat, Ray, Sat)
INTMATRIX **Mat, *Ray, *Sat;
{
 int i, j, k;
  unsigned Dimension, nc, NbRay, NbConstraints, RowSize1,
           *Trace, *bx, *jx, Dim, Status, b;
  unsigned NbBid, NbEq, NbIneq, NbBid2, NbEq2, NbIneq2;
  int Redundant;
  int aux, *p, *temp1;
  INTMATRIX *NewConstraints;
 

  Dimension     = (*Mat)->C-1;     /* homogeneous dimension */

 /*
  printf("*****************************************\n");
  printf("           Remove Redundants             \n");
  printf("*****************************************\n");
  */

  NbRay         = Ray->R;
  nc            = Sat->C;
  NbConstraints = (*Mat)->R;
  RowSize1=(Dimension+1) * sizeof(int);
  temp1=(int *)malloc(RowSize1);

 /*allocate memory for temporary rows*/
  newraytemprow = (int *)malloc(Ray->C  * sizeof(int));
  newsattemprow = (int *)malloc(Sat->C  * sizeof(int)); 
  mattemprow = (int *)malloc((*Mat)->C  * sizeof(int)); 

  /* introduce indirections into Sat matrix to simplify processing
     with Sat and allow easy exchanges of columns */
  bx = (unsigned *)malloc(NbConstraints * sizeof(unsigned));
  jx = (unsigned *)malloc(NbConstraints * sizeof(unsigned));
  i = 0;
  b = MSB;
  for (j=0; j<NbConstraints; j++)
  { jx[j] = i;
    bx[j] = b;
    NEXT(i,b);
  }

  /* Step 1: Compute status counts for both rays and inequalities */
  /* for each equation, count the number of vertices/rays saturated
     by that equation, put the result in the status words.
     At the same time, for each vertex/ray, count the number of
     constraints saturated by it. */
  NbEq=0;
  for (j=0; j<NbConstraints; j++)
  { 
    (*Mat)->V[j][0]=0;

    /* identify and remove the positivity constraint 1>=0 */
    for (i=1, p = &(*Mat)->V[j][1]; i<Dimension; i++) if (*p++) break;
  
    /* count the incidences of saturations */
    for (i=0; i<NbRay; i++) if (!(Sat->V[i][jx[j]]&bx[j]))
    {
      (*Mat)->V[j][0]++;
    }
    if ((*Mat)->V[j][0]==NbRay) NbEq++;    /* all vert/rays are saturated */
  }
  (*Mat)->R = NbConstraints;

 
 NbBid = 0;
  for ( i = 0; i < Ray->R; i++)
    {
      if (!(Ray->V[i][0]))
	NbBid++;
    }
  
 /* Step2: Sort equations to top of (*Mat), inequalities afterward */
  /* Detection of implicit equations such as y>=3; y<=3 */
  /* Keep inequalities in same relative order for SIMPLIFY to work */
  for (i=0; i<NbEq; i++)
    if ((*Mat)->V[i][0]!=NbRay) /* vert/rays are NOT all saturated (inequality)*/
    { for (k=i+1; (*Mat)->V[k][0]!=NbRay && k<NbConstraints; k++)
        /* skip over inequalities -- find an equation */ ;
      if (k==NbConstraints) /* none found--program error (NbEq wrong)*/ break;
      /* slide inequalities up, move equality down (toward begining of array)*/
      memcpy(temp1, (*Mat)->V[k], RowSize1);
      aux = jx[k];
      j   = bx[k];
      for (;k>i;k--)
      {
	memcpy((*Mat)->V[k], (*Mat)->V[k-1], RowSize1);
         jx[k] = jx[k-1];
         bx[k] = bx[k-1];
      }
      memcpy((*Mat)->V[i], temp1, RowSize1);
      jx[i] = aux;
      bx[i] = j;
    }

  /* Step 3: Gaussian elimiation of Equalities and reduction of
             strength on rays */ 
#ifdef DEBUG
  printf("Step3 vorher (Gauss): \n");
  PRINTMATI(0,(*Mat)->R-1,0,(*Mat)->C-1,(*Mat));
#endif 

  NbEq2 = Gauss((*Mat), NbEq, Dimension, temp1);
  
#ifdef DEBUG
  printf("result of Step3 (Gauss): \n");
  PRINTMATI(0,(*Mat)->R-1,0,(*Mat)->C-1,(*Mat));
 printf("NbEq2: %i \n", NbEq2);
#endif 

  NbBid2 = NbBid;

  /* compute dimension of the ray space */
  Dim = Dimension-1-NbEq2-NbBid2;  /* Dim of non-homogen Ray Space */
  
#ifdef DEBUG
  printf("dimension of the ray space: %i \n", Dim);
#endif

  /* Step 6: First pass filter of inequalities and equation identification */
  NbIneq=0;
  for (j=0; j<NbConstraints; j++)
  { /* New positivity constraints may have been created by step 3.
       the Gaussian elimination -- must check for them */
    /* identify and remove the positivity constraint 1>=0 */
    for (i=1, p = &(*Mat)->V[j][1]; i<Dimension; i++) if (*p++) break;
    if (i==Dimension) /* Constraint j is a positivity constraint */
    {   /* or is it 1==0 ??   -- lets check */
      
        /* set the positivity constraint redundant */
        (*Mat)->V[j][0]=2;
        continue;
    }
    if     ((Status=(*Mat)->V[j][0])==0) (*Mat)->V[j][0]=2;  /* redundant */
    else if (Status<Dim)              (*Mat)->V[j][0]=2;  /* redundant */
    else if (Status==NbRay)           (*Mat)->V[j][0]=0;  /* equality */
    else                  { NbIneq++; (*Mat)->V[j][0]=1;} /* inequality */
  }

  /* Allocate the polyhedron (using approximate sizes) */
  /*
  Pol = Polyhedron_Alloc(Dimension-1, NbIneq+1, NbUni+NbBid);
  Pol->NbBid = NbBid;
  Pol->NbEq  = 0;
  */

#ifdef DEBUG
  printf("NbIneq, NbEq2, Dimension,  NbUni,  NbBid2 %i %i %i %i %i \n", NbIneq, NbEq2, Dimension,  NbUni,  NbBid2);
#endif

  NewConstraints = ZeroMatrix( NbIneq + NbEq2 +1, Dimension + 1 );

  /* Partially fill the Polyhedron structure */

 if (NbEq2)
   {
     for ( i = 0; i < NbEq2; i++)
       for( j = 0; j < Dimension +1 ; j++)
	 NewConstraints->V[i][j] = (*Mat)->V[i][j];
   }

#ifdef DEBUG
 printf("result of Step7: \n");
 PRINTMATI(0,NewConstraints->R-1,0,NewConstraints->C-1,NewConstraints);
#endif
  
  /* Step 8: Final pass filter of inequalities */
  /* Detection of redundants inequalities */
  /* redundant inequalitIes will include:
        1. inequalities which are always true, such as 1>=0,
        2. redundant inequalities such as y>=4 given y>=3, or
           x>=2 given x=1, etc.
        3. redundant inequalities such as x+y>=5 given x>=3 and y>=2 */
  /* every 'good' inequality must saturate at least Dimension rays 
     and be unique */
  Trace=(unsigned *)malloc(nc * sizeof(unsigned));

  /* 			j-->
			_________________
			|		|
			|      (*Mat)	|
			|_______________|
				    |
	_________	____________V____
      i	|	|	|	    0	|
      |	|  Ray	|	|      Sat	|
      V	|	|	|	    0	|
	|	|	|	    0	|
	|_______|	|_______________|
			       -OR-
			________|________
			|_____Trace_0___|
  */
  NbIneq2 = 0;
  for (j=NbEq; j<NbConstraints; j++)
  { 
    if ((*Mat)->V[j][0]==1)	/* non-redundant inequality */
    {
      for (k=0; k<nc; k++) Trace[k]=0;		/* init Trace */
      /* compute Trace: the union of all rows of Sat where
			constraint j is saturated */
      for (i=NbBid; i<NbRay; i++) if (Ray->V[i][0]==1)
      { if (!(Sat->V[i][jx[j]]&bx[j])) 
          for (k=0; k<nc; k++) Trace[k] |= Sat->V[i][k];
      }
   

      /* only constraint j should saturate this set of vertices/rays.
	 if another non-redundant constraint also saturates this set,
         then constraint j is redundant */
      Redundant=0;
      for (i=NbEq; i<NbConstraints; i++) {
        if ( ((*Mat)->V[i][0] ==1) && (i!=j) && !(Trace[jx[i]] & bx[i]) )
        {   Redundant=1;
            break;
        }
      }
      if (Redundant) (*Mat)->V[j][0]=2;
      else
      {
	  for( i = 0; i < Dimension +1 ; i++)
	    NewConstraints->V[NbEq2 + NbIneq2][i] = (*Mat)->V[j][i];

	NbIneq2++;
      }
    }
  }
  free(Trace);

#ifdef DEBUG 
  printf("result of Step8: \n");
  PRINTMATI(0,NewConstraints->R-1,0,NewConstraints->C-1,NewConstraints);
  printf("result of Step8: \n");
  PRINTMATI(0,NewRay->R-1,0,NewRay->C-1,NewRay);
#endif  

  free(bx);
  free(jx);
  free(temp1);
  free(newraytemprow);
  free(newsattemprow);
  free(mattemprow);

  PartFreeMat( &NewConstraints, NbEq2 + NbIneq2);
 
  FreeMat((*Mat));

  (*Mat) = NewConstraints;

  return 0;

} /* Remove_Redundants */

int Remove_Redundant_Constraints2(Mat)
INTMATRIX **Mat;
{
 int i, k;
  unsigned Dimension, nc, NbConstraints, RowSize1, RowSize2;
  unsigned NbEq, NbEq2;
  int *temp1;
  NbEq=0;

  Dimension     = (*Mat)->C-1;     /* homogeneous dimension */

 /*
  printf("*****************************************\n");
  printf("           Remove Redundants             \n");
  printf("*****************************************\n");
  */

  NbConstraints = (*Mat)->R;
  RowSize1=(Dimension+1) * sizeof(int);
  RowSize2=nc*sizeof(int);
  temp1=(int *)malloc(RowSize1);

 /*allocate memory for temporary rows*/ 
  mattemprow = (int *)malloc((*Mat)->C  * sizeof(int)); 

  for(i=0; i<NbConstraints; i++)
    if(!(*Mat)->V[i][0]) NbEq++;

 /* Step2: Sort equations to top of (*Mat), inequalities afterward */
  /* Detection of implicit equations such as y>=3; y<=3 */
  /* Keep inequalities in same relative order for SIMPLIFY to work */
  for (i=0; i<NbEq; i++)
    if ((*Mat)->V[i][0]) /* vert/rays are NOT all saturated (inequality)*/
    { for (k=i+1; k<NbConstraints && (*Mat)->V[k][0]; k++)
        /* skip over inequalities -- find an equation */ ;
      /* slide inequalities up, move equality down (toward begining of array)*/
      memcpy(temp1, (*Mat)->V[k], RowSize1);
      for (;k>i;k--)
      {
	memcpy((*Mat)->V[k], (*Mat)->V[k-1], RowSize1);
      }
      memcpy((*Mat)->V[i], temp1, RowSize1);
    }

  /* Step 3: Gaussian elimiation of Equalities and reduction of
             strength on rays */ 
#ifdef DEBUG
  printf("Step3 vorher (Gauss): \n");
  PRINTMATI(0,(*Mat)->R-1,0,(*Mat)->C-1,(*Mat));
#endif 

  NbEq2 = Gauss((*Mat), NbEq, Dimension, temp1);
  
#ifdef DEBUG
  printf("result of Step3 (Gauss): \n");
  PRINTMATI(0,(*Mat)->R-1,0,(*Mat)->C-1,(*Mat));
 printf("NbEq2: %i \n", NbEq2);
#endif 

  free(temp1);
  free(mattemprow);

  return 0;

} /* Remove_Redundants */
