#include <malloc.h>
#include "sharedOps.h"
#include "sharedVars.h"
#include "matrixOps.h"
#include "matrixProcs.h"
#include "matrixCombine.h"
#include "intOps.h"
#include "index.h"
/*******************************************************************************************

*******************************************************************************************/

/*******************************************************************************************
  Combine:
  input:  rays r1 and r2
  output: ray r3 such that r3=lamda1*r1 + lamda2*r2
                      and  r3 saturates the constraint a: r3.a = 0
*******************************************************************************************/
int Combine(int r1, int r2, int r3, INTMATRIX *Mptr)
{
  static int D, lamda1, lamda2, k;
  static USI bitmask;
  
  bitmask = ~(BCount - 1); 
  
  lamda1 = Mptr->V[r1][0]; /*first column contains r1.a */
  lamda2 = Mptr->V[r2][0]; /*first column contains r2.a */
  D  = Gcd(abs(lamda1), abs(lamda2));
  lamda1 = lamda1/D;
  lamda2 = lamda2/D;
  if(CombineRow(r1, r2, r3, lamda2, -lamda1, Mptr)) /*write new row r3*/
    return OVERFLOWERROR; /*overflow error*/
  
  /*update saturation matrix*/
  for (k = 0; k < Sat->C; k++)
    {
      Sat->V[r3][k] = 0;
    }
  
  RowOr(r1, r2, r3, Sat);
  Sat->V[r3][WCount]&=~BCount;
  Sat->V[r3][WCount]&=bitmask;
  
  for (k = WCount + 1; k < Sat->C; k++)
    {
      Sat->V[r3][k] = 0;
    }
  
  return OK;
  
} /* Combine */

/*******************************************************************************************
       SimpleCombine: 
       combine rows without updating the saturation matrix
       used by AugmentR and ConstrainL
*******************************************************************************************/
int SimpleCombine(int r1, int r2, int r3, INTMATRIX *Mptr)
{
  static int D, lamda1, lamda2; 
  lamda1 = Mptr->V[r1][0]; /*first column contains r1.a */
  lamda2 = Mptr->V[r2][0]; /*first column contains r2.a */
  D  = Gcd(abs(lamda1), abs(lamda2));
  lamda1 = lamda1/D;
  lamda2 = lamda2/D;
  if (CombineRow(r1, r2, r3, lamda2, -lamda1, Mptr)) /*write new row r3*/
    return OVERFLOWERROR; /*overflow error*/
  else
    return OK;
} /* SimpleCombine */


/*******************************************************************************************
       CreateLookUpMat:
       a lookup table used for speed optimization
       it contains for each value between 0 .. 65535 (size of a word)
       the number of zeroes contained in its bit representation:
       e.g. ZeroMatrix[0] = 16, ZeroMatrix[12345] = 11 
*******************************************************************************************/
void CreateLookUpMat()
{
 static USI Word, nbzeros, thisbit, i;
  
  LookUpMat = (int *) malloc(TOPHALF * sizeof(int));
  /*TOPHALF = 65535 = 0xffff*/
  for (i = 0 ; i <= TOPHALF; i++)
    {
      nbzeros = 0;
      Word = i;
      /*MSBHALF = 0x8000*/
      for (thisbit=MSBHALF; thisbit!=0; thisbit>>=1)
	{/*scan the bits of each word*/
	  if (!(Word&thisbit)) nbzeros++;
	}
      LookUpMat[i] = nbzeros;
    }
} /*CreateLookUpMat*/

/*******************************************************************************************
 count the number of Rays 
*******************************************************************************************/
int CountRays(INTMATRIX *_Ray)
{
  USI i;
  USI count=0;
  USI lastColumn=_Ray->C-1;
  for(i=0; i<_Ray->R; i++)
    if (_Ray->V[i][lastColumn]==0) count++;
  return count;
} /*CountRays*/

/*******************************************************************************************
  Delete status-information and vertices from Ray,
  leaving only the ray-vectors
*******************************************************************************************/
INTMATRIX* ExtractRays(INTMATRIX *_Ray, int numsol)
{
  USI rowcount;
  USI i,j;
  USI numrows,numcols;
  INTMATRIX *Solution;

  numrows=_Ray->R;
  numcols=_Ray->C;
  rowcount=0;
  
  Solution = ZeroMatrix(numcols-2, numsol);
  
  for (i=0; i<numrows; i++)
    {
      if (_Ray->V[i][numcols-1]==1) { continue;} /*vertex*/
      for(j=1; j<numcols-1; j++)
	{
	  Solution->V[j-1][rowcount]=_Ray->V[i][j];
	}
      rowcount++;
    }
  return Solution;
} /*ExtractRays*/

/*******************************************************************************************
 evaluate constraint:
  Set the status word (column 0) of each ray to ray dot constraint
*******************************************************************************************/
void EvalConstraint(int *constraint)
{
  static USI i, j;
  static int *rayptr, *constaintptr, *statusptr;
  
  for (i=0; i<NbRay; i++)
    {
      /*compute the dot product*/
      rayptr = Ray->V[i]+1;     /*first element of ray*/
      constaintptr = constraint+1;    /*first element of constraint*/
      statusptr = Ray->V[i]; /*pointer to status word (0. element of the ray)*/
      *statusptr = *rayptr++ * *constaintptr++; 
      for (j=1; j<Ray->C - 1; j++)
	{
	  *statusptr += *rayptr++ * *constaintptr++;
	}
    }
}

/*******************************************************************************************
  CheckForStartingVertex:
  find vertex that remains valid for all constraints
  vertex must not be rational
  if found return number of ray, otherwise return -1
*******************************************************************************************/
int CheckForStartingVertex(int startrow, USI Start, USI End)
{
  static USI i, j;
  static int k;
  static int *rayptr, *constaintptr, *p;
  static int status;

  for (i=startrow; i<NbRay; i++)
    {
      for (k = Start; k<End; k++)
	{
	  /*compute the dot product*/
	  constaintptr =  (Constraints->V[k])+1;    /*first element of constraint*/
	  rayptr = Ray->V[i]+1;     /*first element of ray*/
	  status = *rayptr++ * *constaintptr++; 
	  for (j=1; j<Ray->C - 1; j++)
	    {
	      status += *rayptr++ * *constaintptr++;
	    }
	  
	  if (status < 0)
	    {
	      break;
	    }
	}
      if(k==End && Ray->V[i][Ray->C - 1]==1)
	{
	  for (j=1, p = &Ray->V[j][1]; j<Ray->C-1; j++) if (*p++) break;
	  if (j==Ray->C-1) /* Constraint j is a positivity constraint */  
	    continue;
	  /*found vertex that remains valid*/
	  return i; /*return number of ray*/
	}
    }
  return -1;
} /*EvalAllConstraints*/

/*******************************************************************************************
  sort rays by MinF-value 
*******************************************************************************************/
void SortRaysMinF(int sup_bound, int equal_bound)
{
  static int i;
  static USI oldpos;
  
  for (i = equal_bound; --i>=sup_bound; )
    {
      MinF->V[i][1] = i; /*store old position*/
    }
#ifdef HEAVYDEBUG
  printf("Quicksort vorher:\n");
  PRINTMATI(sup_bound, equal_bound -1,0,MinF->C-1,MinF);
  PRINTMATI(sup_bound, equal_bound -1,0,Ray->C-1,Ray);
#endif
  QuickSort(MinF, sup_bound, equal_bound-1, 0, 1);
  /*MinF[i][0] = sorted MinF values */
  /*MinF[i][1] = position of new row i*/
  for (i = equal_bound; --i>=sup_bound; )
    {
      MinF->V[i][2] = MinF->V[i][1]; /*copy*/
      MinF->V[i][3] = i; /*store old position*/
    }
  QuickSort(MinF, sup_bound, equal_bound-1, 2, 3);
  /*MinF[i][3] = new position of row i */
#ifdef HEAVYDEBUG   
  printf("Quicksort nachher:\n");
  PRINTMATI(sup_bound, equal_bound -1,0,MinF->C-1,MinF);
#endif
  for (i = sup_bound; i<equal_bound; i++)
    {
      /*exchange rays, update Index if necessary*/
      oldpos = MinF->V[i][1]; /*read out old position*/
      if(i!=oldpos)
	{
	  RowExchange(Ray, i, oldpos, raytemprow);
	  RowExchange(Sat, i, oldpos, sattemprow);
	  IndexExchange(i, oldpos);
	  MinF->V[MinF->V[i][3]][1] = oldpos; 
	  MinF->V[oldpos][3] = MinF->V[i][3];
#ifdef HEAVYDEBUG     
	  printf("swap row %i and %i\n", i , oldpos);
	  PRINTMATI(sup_bound, equal_bound -1,0,MinF->C-1,MinF);
	  PRINTMATI(sup_bound, equal_bound -1,0,Ray->C-1,Ray);	  
#endif
	}
    }
} /*SortRaysMinF*/


/*******************************************************************************************
    combine adjacent rays
*******************************************************************************************/
int CombineNewRays(LiListElement *ListEl, int thisrow)
{
  static LiListElement *ptr;
  static USI i;
  static USI numels;
  static USI combirow;
  static USI errorflag = OK;
  numels= ListEl->NbElements;
  for (i = 0; i < numels; i++)
    {
      
      /*Check the size of the RaySpace
	allocate new memory if necessary*/
      if(abs(NbRay-OldRays)>RESIZE_ADDROWS)
	{
	  OldRays=NbRay;
#ifdef DEBUG
	  printf("Currently %i Rays\n", NbRay);
#endif
	}
      if (NbRay>=Ray->R)
	{
	  if (NbRay>=NbMaxRays) return MEMORYERROR; /*too many rays, return with error*/
#ifdef DEBUG
	  printf("resize\n");
#endif
	  errorflag |= ResizeRows(&Ray, RESIZE_ADDROWS); /*add rows to Ray Sat MinF*/
	  errorflag |= ResizeRows(&Sat, RESIZE_ADDROWS);
	  errorflag |= ResizeRows(&MinF, RESIZE_ADDROWS);
	  if(errorflag) return MEMORYERROR; /*memory exhausted*/
	}
      
      combirow = IndexLR->V[ListEl->Elements[i]][0];
#ifdef DEBUG
      printf("...%i, \n", combirow);
      PRINTMATI(combirow, combirow, 0,Ray->C-1, Ray);
#endif
      if(Combine(thisrow, combirow, NbRay, Ray))
	return OVERFLOWERROR; /*overflow error*/
      
      /*store the feasible ray next to the new ray in the generator list*/
      if(StorePair(GenList, combirow, NbRay))
	return MEMORYERROR; /*memory exhausted*/
      
      NbRay++;
    }
  ptr = ListEl->next;
  if (ptr)
    return CombineNewRays(ptr, thisrow);
  return OK;
} /*CombineNewRays*/

/*******************************************************************************************

*******************************************************************************************/
int ObtainNewRays(LiList *List)
     /*from the linked list, obtain the new rays by combining all pairs of the list*/
{
  static USI i;
  static USI numrows;
  static int thisrow;
  static USI errorflag = OK;
#ifdef DEBUG
  printf("************ObtainNewRays***********\n");
#endif
  while(List)
    {
      numrows = List->R;
      for (i = 0; i < numrows; i++)
	{
	  thisrow =  IndexLR->V[List->Indices[i]][0];
#ifdef DEBUG
	  printf("Row number %i \n", thisrow);
	  PRINTMATI(thisrow, thisrow, 0,Ray->C-1, Ray);
	  printf("is combined with rows\n");
#endif
	  errorflag = CombineNewRays(List->AdjRows[i], thisrow);
	  if (errorflag) return errorflag;
	}
      List = List ->next;
    }
  return OK; 
}/* ObtainNewRays*/


/*******************************************************************************************
  evaluate all constraints:
  build the saturation matrix for
  all constraints and all rays
  compute the minimum infeasible index
*******************************************************************************************/
void EvalAllConstraints(int startrow, USI Start, USI End)
{
  static USI i, j;
  static int k;
  static int *rayptr, *constaintptr;
  static int status;
  static USI wordcount;
  static USI bitcount;
  
  wordcount = Start /WSIZE;
  bitcount = MSB; bitcount >>= (Start%WSIZE);
  
  for (i = startrow; i< NbRay; i++)
    {
      MinF->V[i][0] = maxMinF;
      /*
	printf("Row %i\n", i);
	printsat(Sat, i , i);
      */
    }
  
  for (i=startrow; i<NbRay; i++)
    {
      wordcount = Start /WSIZE;
      bitcount = MSB; bitcount >>= (Start%WSIZE);
      for (k = Start; k<End; k++)
	{
	  /*compute the dot product*/
	  constaintptr =  (Constraints->V[k])+1;    /*first element of constraint*/
	  rayptr = Ray->V[i]+1;     /*first element of ray*/
	  status = *rayptr++ * *constaintptr++; 
	  for (j=1; j<Ray->C - 1; j++)
	    {
	      status += *rayptr++ * *constaintptr++;
	    }
	  
	  if (status != 0)
	    {
	      Sat->V[i][wordcount]|=bitcount;
	    }
	  if (status < 0)
	    {
	      MinF->V[i][0] = k; /*erster constraint mit A.x <0*/
	      break;
	    }
	  NEXT(wordcount, bitcount); /*increase counter by one bit*/
	}
    }
} /*EvalAllConstraints*/

/*******************************************************************************************

*******************************************************************************************/
int CheckedLater(USI i, USI k, USI rplus, USI rminus)
{
  static int thisword;
  static int WordNr1, WordNr2;
  static USI  BitNr1, BitNr2;

  k--;
  
  if(i>=k) return 0;
  
  WordNr1 = i / WSIZE;
  BitNr1 = MSB; BitNr1 >>= (i %WSIZE);
  
  WordNr2 = k / WSIZE;
  BitNr2 = MSB; BitNr2 >>= (k %WSIZE);
  
  for (thisword = WordNr1+1; thisword<=WordNr2; thisword++) {
    BitRow2[thisword] = -1; /*fill with 1111111111111...*/
  }
  
  BitRow2[WordNr1] = BitNr1 - 1;
  
  for (thisword = WordNr1; thisword< WordNr2; thisword++)
    BitRow3[thisword] = 0;
  
  BitRow3[WordNr2] = BitNr2 - 1;
  
  for (thisword=WordNr1; thisword<=WordNr2; thisword++) 
    /*for (thisword=WordNr2; thisword>=WordNr1; thisword--) */
    { 
      BitRow[thisword] = ~(Sat->V[rplus][thisword] | Sat->V[rminus][thisword]);/*NOT(OR the rows)*/
      BitRow2[thisword] = BitRow3[thisword] ^ BitRow2[thisword];/*XOR the rows*/
      if ( BitRow[thisword] & BitRow2[thisword])/*AND the rows*/
	{
	  checkedlater++;
	  return 1;
	}
    }
  return 0;
} /*CheckedLater*/

/*******************************************************************************************

*******************************************************************************************/
void BuildBorder(USI sup_bound)
     /*store the starting positions of different ck values */
{
  static int i, temp;
  for( i = 0; i<=maxMinF; i++)
    Border[i] = -1;
  
  temp = -1;
  for (i = sup_bound; i<NbRay; i++)
    {
      if ( MinF->V[i][0] != temp)
	{
	  temp = MinF->V[i][0];
	  Border[MinF->V[i][0]] = i;
	}
    }
}/*BuildBorder*/

/*******************************************************************************************

*******************************************************************************************/
USI GetBorder(USI thisBorder)
{
  static USI i;

  for( i = thisBorder + 1; i<=maxMinF; i++)
    if( Border[i] != -1)
      break;
 
  if (i>maxMinF)
    return -1;
  else
    return Border[i];
}/*GetBorder*/

/*******************************************************************************************

*******************************************************************************************/
USI GetUpperBound()
{
  static USI i;
  for( i = maxMinF ; i>=0; i--)
    if( Border[i] != -1)
      break;
  
  return Border[i];
} /*GetUpperBound*/

/*******************************************************************************************
  check if constraint step cuts the lineality space
  modify lineality space if yes, return -1 if no
*******************************************************************************************/
int ConstrainL(USI step)
{
  USI index_non_zero;
  USI l1, l2;
  USI  j;
  int *statusptr, *dummyPtr;

#ifdef DEBUG
  printf("***********ConstrainL******** NbLines: %i\n", NbLines);
#endif
  if (NbLines == 0) return -1;

   /*find first row with status !=0 (does not saturate the constraint)*/
    for (index_non_zero=0; index_non_zero<NbRay; index_non_zero++)
      {
	statusptr = Ray->V[index_non_zero];
	if (*statusptr ) break;
      /*Index of first row with Status!=0*/
    }

#ifdef DEBUG
    if (index_non_zero >= NbLines) printf("constraint does not slice the lineality space\n");
    else printf("first line!=0, NbLines: %i %i\n", index_non_zero, NbLines);
#endif

    if (index_non_zero >= NbLines) return -1;
    /*all lines saturate constraint a, constraint does not slice the lineality space,
      lineality space stays unchanged, return*/


    /*the line Ray->[index_non_zero] is exchanged with line Ray->[Nblines]
      the number of lines (dimension of the lineality space) is decreased by one*/
    NbLines--;
    if (NbLines!=index_non_zero)
      {
	RowExchange(Ray, index_non_zero, NbLines, raytemprow);
      }

    l1 = NbLines;
      
    for (l2 = 0; l2<NbLines; l2++) /*for each line l2 != l1*/
      { 
	if (Ray->V[l2][0]==0) continue;

	SimpleCombine(l2, l1, l2, Ray); /*line l2 is replaced with the combination of line l1 and line l2*/

      }

      /* add the positive part of index_non_zero bidirectional ray to
         the set of unidirectional rays */
      if (Ray->V[l1][0]<0)
	/*status word is negative, invert line l1*/
	/*if the status word is positive, do nothing*/
	{ 
        dummyPtr=Ray->V[l1]; 
        for (j=0; j<Ray->C; j++) { *dummyPtr = -(*dummyPtr); dummyPtr++; }
      }

      return l1; /*return the new row*/
} /*ConstrainL*/

/*******************************************************************************************
   add newlines new lines to _Ray, return extended matrix
*******************************************************************************************/
INTMATRIX *AddLines(INTMATRIX *_Ray, int newlines)
{
  USI i, j;
  INTMATRIX *ExtendedRay;
  /*new ray matrix with newlines lines:*/
  ExtendedRay = ZeroMatrix(_Ray->R + newlines, _Ray->C);

  /*copy old ray matrix*/
  for ( i = 0; i < _Ray->R; i++)
    {
      for(j=0; j<_Ray->C; j++)
	{
	  ExtendedRay->V[i][j] = _Ray->V[i][j];
	}
    }

  /*add lines
    ...
    1 4 2 4 1 
    1 2 2 4 1 
    0 0 0 1 0 <-- Dimension - 2  
    0 0 1 0 0 <-- lines
 */

 j = ExtendedRay->C - 2;
 for ( i =  _Ray->R; i < ExtendedRay->R; i++)
   {
     ExtendedRay->V[i][j] = 1;
     j--;
   }

#ifdef DEBUGCOUNT
 printf("ray matrix with %i lines added:\n", newlines);
 PRINTMATI(0,ExtendedRay->R-1,0,ExtendedRay->C-1,ExtendedRay);
#endif

 return ExtendedRay;
} /*AddLines*/


void WriteStatusConstraints(Constraints, Ray, Sat)
INTMATRIX *Constraints, *Ray, *Sat;
/* compute status words of the Constraints matrix
 from Ray and  Sat*/
{
  int i, j;
  unsigned NbRay, NbConstraints, Dimension;

  USI wordcount;
  USI  bitcount;

  Dimension     = Constraints->C-1;     /* homogeneous dimension */

  NbRay         = Ray->R;
  NbConstraints = Constraints->R;

  /* reset the status numbers */
  for (i=0; i<NbConstraints; i++)
  { 
    Constraints->V[i][0]=0;
  }

  /* count the incidences of saturations */
  for (i=0; i<NbRay; i++) 
    {
      wordcount = 0;
       bitcount = MSB;

      for (j=0; j<NbConstraints; j++)
	{
	  if (!(Sat->V[i][wordcount]&bitcount))
	  { 
	    Constraints->V[j][0]++;
	  }
	NEXT(wordcount, bitcount); /*increase counter by one bit*/
	}
    }


  for (i=0; i<NbConstraints; i++)
    if (Constraints->V[i][0]==NbRay) /*equation*/
      Constraints->V[i][0]=0;
    else
      Constraints->V[i][0]=1; /*inequality*/
  }

void WriteStatusConstraints2(Constraints, Ray, Sat, maxray)
INTMATRIX *Constraints, *Ray, *Sat;
int maxray;
/* compute status words of the Constraints matrix
 from Ray and  Sat*/
{
  int i, j;
  unsigned NbRay, NbConstraints, Dimension;

  USI wordcount;
  USI  bitcount;

  Dimension     = Constraints->C-1;     /* homogeneous dimension */

  NbRay         = Ray->R;
  NbConstraints = Constraints->R;

  /* reset the status numbers */
  for (i=0; i<NbConstraints; i++)
  { 
    Constraints->V[i][0]=0;
  }

  /* count the incidences of saturations */
   wordcount = 0;
       bitcount = MSB;

  for (i=0; i<maxray; i++) 
    {
      for (j=0; j<NbConstraints; j++)
	{
	  if (!(Sat->V[j][wordcount]&bitcount))
	  { 
	    Constraints->V[j][0]++;
	  }
	}
      NEXT(wordcount, bitcount); /*increase counter by one bit*/
    }

  for (i=0; i<NbConstraints; i++)
    if (Constraints->V[i][0]==maxray) /*equation*/
      Constraints->V[i][0]=0;
    else
      Constraints->V[i][0]=1; /*inequality*/
  }




void WriteStatus(Constraints, Ray, Sat)
INTMATRIX *Constraints, *Ray, *Sat;
/* compute status words of the ray matrix
 from Constraints and  Sat*/
{
  int i, j;
  unsigned NbRay, NbConstraints, Dimension;

  USI wordcount;
  USI  bitcount;
  USI NbLines = 0;

  Dimension     = Constraints->C-1;     /* homogeneous dimension */

  NbRay         = Ray->R;
  NbConstraints = Constraints->R;



  /* reset the status numbers */
  for (i=0; i<NbRay; i++)
  { 
    Ray->V[i][0]=0;
  }

  /* count the incidences of saturations */
  for (i=0; i<NbRay; i++) 
    {
      wordcount = 0;
       bitcount = MSB;

      for (j=0; j<NbConstraints; j++)
	{
	  if (!(Sat->V[i][wordcount]&bitcount))
	  { 
	    Ray->V[i][0]++;
	  }
	NEXT(wordcount, bitcount); /*increase counter by one bit*/
	}
    }


  for (i=0; i<NbRay; i++)
  { /* give rays credit for saturating the positivity constraint */
    if (Ray->V[i][Dimension]==0) Ray->V[i][0]++;
    /* test for a line.  Note that a vertex cannot become a line because
       only rays satisfy the pos constr and can status == NbConstraints+1 */
    if (Ray->V[i][0]==NbConstraints+1)
      {/*line*/
	Ray->V[i][0]=0;
	NbLines++;
      }
    else
      {/*ray/vertex*/
	Ray->V[i][0]=1;
      }
  }
//  printf("Number of lines %i\n", NbLines);

}


INTMATRIX *BuildSat(INTMATRIX *Mat, INTMATRIX *_Ray, USI NbCon)
{ 
  INTMATRIX *_Sat;
  int i, j, k, jx, *p1, *p2, *p3;
  unsigned _Dimension, _NbRay, bx, _nc;
   
  _NbRay = _Ray->R;
  _Dimension = Mat->C-1;    
  
  raytemprow = (int *)malloc(_Ray->R  * sizeof(int));

  _nc      = (Mat->R - 1) / (sizeof(int)*8) + 1;
  _Sat     = ZeroMatrix(_NbRay, _nc);
  _Sat->R = _NbRay;

    jx=0; bx=MSB;
    for (k=0; k<NbCon; k++)
      {   for (i=0; i<_NbRay; i++)
        {   p1 = _Ray->V[i]+1;
	p2 = Mat->V[k]+1;
	p3 = raytemprow + i;
	*p3 = *p1++ * *p2++;
	for (j=1; j<_Dimension; j++) 
	  {
	    *p3 += *p1++ * *p2++;
	  }
        }
      for (j=0; j<_NbRay; j++)
        {  
	  if (raytemprow[j]) _Sat->V[j][jx]|=bx;
        }
        NEXT(jx, bx);
      }
    free(raytemprow);
    return _Sat;
}


int Init()
{
  USI i;
  /*Init : rayspace-matrix is filled with 
    a basis for d-dimensional lineality space
    and a point at the origin
    
    Beispiel: Dimension = 4
    0 1 0 0 0 0
    0 0 1 0 0 0
    0 0 0 1 0 0
    0 0 0 0 1 0
    0 0 0 0 0 0
  */

  Ray = ZeroMatrix(Dimension, Dimension+1); /*fill with 0*/

  for (i=0; i<Dimension; i++) Ray->V[i][i+1]=1; /*a Line for each dimension*/
  Ray->V[Dimension-1][0] = 1;		/* vertex at origin */

  /*init saturation matrix Sat
    Sat is a bit matrix with one row for each ray 
    and one bit for each constraint*/
  Sat = ZeroMatrix(Dimension, nc);

  return Dimension - 1; /*number of rays*/
}
