// File    Time.h        (C++ Inline Implementation) 
// Module  Time          (Ressource Monitoring Functions)
/**
 *         Time and memory measuring functions.
 *
 *@author  Philipp Berndt
 *@version $Id: Time.h,v 1.4 2000/08/17 12:45:34 azi Exp $
 */

#ifndef TIME_H
#define TIME_H

// Uncomment to print times as seconds
#define PRINT_SECONDS

 
#include <ctime>
#ifndef _SYS_TIME_H
#include <sys/time.h>   // REALLY include sys/time.h
#endif

#include <string>
#include <strstream>
#include <fstream>
#include <iostream>
#include <iomanip>
#include <sys/resource.h>
#include <cassert>
#include <algorithm>
#include <unistd.h> // getpagesize

#if defined(sun) && defined(__svr4__)
#define SOL2
#include <fstream>
#include <procfs.h>
#endif

#ifdef PVM_SUPPORT
#include "pvm3cpp.h"
#endif

#define MIO 1000000
/**
 * This is a workaround for the not yet existing ios::failure exception.
 */
class failure {};

/**
 * Throws a failure (Exception) if the fail-flag of the stream is set.
 *
 *@see failure
 */
inline istream& throw_if_fail(istream &is)
{
    if(is.fail())
	throw failure(); // Use ios::failure as soon as it becomes available!
    return is;
}


/**
 *         C++ Wrapper for struct timeval.
 *
 *@author  Philipp Berndt
 *@version $Id: Time.h,v 1.4 2000/08/17 12:45:34 azi Exp $
 */
class Time
{
private:
    timeval time;

public:
    /** Constructor */
    Time(double seconds)
	{
	    assert(seconds >= 0.0);
	    time.tv_sec  = static_cast<unsigned long>(seconds);
	    time.tv_usec =
		static_cast<unsigned long>( (seconds-time.tv_sec) * 1.0e6 );
	}

    /** Constructor */
    Time(unsigned long sec=0, unsigned long usec=0)
	{
	    time.tv_sec = sec;
	    time.tv_usec = usec;
	}

    /** Constructor */
    Time(unsigned long hours,   unsigned long minutes, unsigned long seconds, 
	 unsigned long msecs=0, unsigned long usecs=0)
	{
	    time.tv_sec = (((hours * 60) + minutes) * 60) + seconds;
	    time.tv_usec = msecs * 1000 + usecs;
	}

    /** Constructor */
    Time(const timeval& tv)
	: time(tv) {}

    /** hours */
    int hours() const
	{ return time.tv_sec / 3600; }

    /** minutes since hour */
    int mins() const
	{ return (time.tv_sec / 60) % 60; }

    /** seconds since minute */
    int secs() const
	{ return time.tv_sec % 60; }

    /** milliseconds since second */
    int msecs() const
	{ return time.tv_usec / 1000; }

    /** microseconds since millisecond */
    int usecs() const
	{ return time.tv_usec % 1000; }
  
    /** seconds */
    int get_sec() const
	{ return time.tv_sec; }

    /** microseconds since second */
    int get_usec() const
	{ return time.tv_usec; }

    /** seconds with fractions */
    double to_secs() const
	{ return time.tv_sec+time.tv_usec/1.0e6; }

    /** Returns a pointer to the internal representation. */
    const timeval* get_timeval() const
	{ return &time; }

#ifdef PVM_SUPPORT
    /** Packs this Time object into the active PVM sendbuffer. */
    void pvm_pack() const
	{
	    pvm_pklong(time.tv_sec);
	    pvm_pklong(time.tv_usec);
	}

    /** Constructor that unpacks from active PVM receive buffer. */
    Time(PVM_UNPACK_t)
	{
	    time.tv_sec = pvm_upklong();
	    time.tv_usec = pvm_upklong();
	}
#endif

    /** Adds Time. */
    Time& operator+=(const Time& t)
	{
	    time.tv_usec += t.time.tv_usec;
	    time.tv_sec += t.time.tv_sec + time.tv_usec/MIO;
	    time.tv_usec %= MIO;
	    return *this;
	}
  
    /** Subtracts Time. */
    Time& operator-=(const Time& t)
	{
	    time.tv_sec -= t.time.tv_sec;
	    time.tv_usec -= t.time.tv_usec;
	    if(time.tv_usec<0)
		{
		int s = -time.tv_usec/MIO +1;
		time.tv_usec += s * MIO;
		time.tv_sec -= s;
		}
	    assert(time.tv_sec >= 0);
	    return *this;
	}
  
    /** Divide Time. */
    Time operator/(double d) const
	{
	    Time resHi, resLo;
	    assert(d>=0);
	    resHi.time.tv_sec = static_cast<time_t>(time.tv_sec/d);
	    resHi.time.tv_usec = static_cast<long>(((time.tv_sec - resHi.time.tv_sec*d)*MIO)/d);
	    
	    resLo.time.tv_sec = static_cast<time_t>(time.tv_usec/(d*MIO));
	    resLo.time.tv_usec = static_cast<long>((time.tv_usec-resLo.time.tv_sec*(d*MIO))/d);

	    return resHi+resLo;
	}

    /** Adds Times. */
    Time operator+(const Time& t) const
	{ return Time(*this) += t; }

    /** Subtracts Times. */
    Time operator-(const Time& t) const
	{ return Time(*this) -= t; }

    /** Returns consumed CPU-time (user+system) of calling process. */ 
    static Time getCPUTime()
	{
	    rusage r_usage;
	    getrusage(RUSAGE_SELF, &r_usage);
	    return Time(r_usage.ru_utime) + Time(r_usage.ru_stime);
	}

    /** Returns elapsed time since 00:00 UTC (aka GMT) , January 1, 1970. */ 
    static Time getTimeOfDay()
	{
	    Time res;
	    gettimeofday(&res.time, 0);
	    return res;
	}  

    /** Converts this Time to it's string representation. */
    operator string();

    /** Prints this Time in hh:mm:ss:mm format. */
    void print_hmsm(ostream &s) const;

    /** Prints this Time in <seconds>s format. */
    void print_secs(ostream &s) const;
};


/** Stream output of a Time object. */
inline ostream& operator<<(ostream &s, const Time& t)
{
#ifdef PRINT_SECONDS
    t.print_secs(s);
#else
    t.print_hmsm(s);
#endif
    return s;
}

/** Stream input of a Time object. */
istream& operator>>(istream &s, Time &t);

/** Saves a Time object for a TimeNET-Model. */ 
inline void saveTime(const Time time, string netname, string timeType)
{   
    string filename = netname + "." + timeType + ".time";

    ofstream os(filename.c_str());
    time.print_hmsm(os);

    if (os.fail())
	{
        cout << "Warning: saveTime() of " << filename << " failed." << endl;
        throw 1; // use ios::failure when it becomes available !
	}
}

/** Loads a Time object for a TimeNET-Model. */ 
inline Time loadTime(string netname, string timeType)
{   
    Time time;
    string filename = netname + "." + timeType + ".time";

    ifstream is(filename.c_str());
    is >> time;

    if (is.fail())
	{
        cout << "Warning: loadTime() of " << filename << " failed." << endl;
        return (Time)(0.0);
	}
    return time;
}

/** Saves a memory value for a TimeNET-Model. */ 
inline void saveMem(const long mem, string netname, string memType)
{   
    string filename = netname + "." + memType + ".mem";

    ofstream os(filename.c_str());
    os << mem;


    if (os.fail())
	{
        cout << "Warning: saveMem() of " << filename << " failed." << endl;
        throw 1; // use ios::failure when it becomes available !
	}
}

/** Loads a memory value for a TimeNET-Model. */ 
inline long loadMem(string netname, string memType)
{   
    long mem;
    string filename = netname + "." + memType + ".mem";

    ifstream is(filename.c_str());
    is >> mem;

    if (is.fail())
	{
        cout << "Warning: loadMem() of " << filename << " failed." << endl;
        return (long)0;
	}

    return mem;
}

/** Returns the memory usage in KB of this process. */
inline long getUsedMem()
{
#ifdef SOL2
    psinfo_t psinfo;
    ostrstream ost;
    ost << "/proc/" << getpid() << "/psinfo" << ends;
    ifstream is(ost.str());
    ost.freeze(0);
    
    is.read(&psinfo, sizeof(psinfo_t));
    return (is.fail()) ? 0 : psinfo.pr_size;

#else // not SOL2
    
    rusage ru;
    getrusage(RUSAGE_SELF, &ru);
    return (ru.ru_idrss * getpagesize()) >> 10;

#endif 
}

/** Returns the memory usage in bytes of this process. */
inline long getHeapSize()
{
#ifdef SOL2
    ostrstream ost;
    ost << "/proc/" << getpid() << "/status" << ends;
    ifstream is(ost.str());
    ost.freeze(0);
    pstatus_t pstatus;
    is.read(&pstatus, sizeof(pstatus_t));
    return (is.fail()) ? 0 : pstatus.pr_brksize;
    
#else // not SOL2

    return 0;

#endif 
}


/*****************************************************************************\
*                         INLINE FUNCTION DEFINITIONS                         *
*               No classes or user interfaces below this line !               *
\*****************************************************************************/

inline void Time::print_hmsm(ostream &s) const
{
    char oldfill = s.fill();
    int oldflags = s.flags();
    s << setw(2) << setfill(' ') << hours() << ":"
      << setw(2) << setfill('0') << mins()  << ":" 
      << setw(2) << setfill('0') << secs()  << "." 
      << setw(3) << setfill('0') << msecs();
    s.fill(oldfill);
    s.flags(oldflags);
}

inline void Time::print_secs(ostream &s) const
{
    char oldfill = s.fill();
    int oldflags = s.flags();
    int oldprec = s.precision(6);
    s << setw(11) << setfill(' ') << to_secs() << "s";
    s.precision(oldprec);
    s.fill(oldfill);
    s.flags(oldflags);
}

inline Time::operator string()
{
    ostrstream ost;
    ost << *this << ends;
    string res(ost.str());
    ost.freeze(0);
    return res;
}

inline istream& operator>>(istream &s, Time &t)
{
    int hours=0, mins=0, secs, usecs;
    double sec=0.0;
    string word;
    try {
	s >> word >> throw_if_fail;
	int ps=0, pm=0;
	unsigned p;
	if((p = word.rfind(':')) != word.npos)
	    {
	    ps = p+1;
	    if((p = word.rfind(':',p-1)) != word.npos)
		{
		pm = p+1;
		}
	    }
	
	if(pm>0)
	    istrstream(word.substr(0,pm-1).c_str()) >> hours >> throw_if_fail;

	if(ps-pm>0)
	    istrstream(word.substr(pm,ps-pm-1).c_str()) >>mins >>throw_if_fail;

	istrstream(word.substr(ps).c_str()) >> sec >> throw_if_fail;
	secs = static_cast<int>(sec);
	usecs = static_cast<int>((sec - secs) * 1e6 + 5e-7);
	t = Time(hours,mins,secs,0,usecs);
	}
    catch(...)
	{
	}
    return s;
}	

#undef MIO
#endif // not TIME_H

