// File    BitSet.h        (C++ class definition and inline implementation)
// Library VMMATH          (Library for Vector and Matrix Algebra)

#ifndef BITSET_H
#define BITSET_H

/**
 *         Something similar to a vector of bool.
 *
 *@author  Philipp Berndt
 *@version $Id: BitSet.h,v 1.2 2000/01/12 16:02:09 pumphutt Exp $
 */
class BitSet
{
    typedef unsigned long ulong;
public:
    /** Constructor */
    explicit BitSet(size_t size) 
        : Size(size), Data(allocData()) {}

    /** Constructor */
    BitSet(size_t size, bool state) 
        : Size(size), Data(allocData()) { init(state); }

    /** Destructor */
    ~BitSet() { delete Data; }

    /** Copy constructor */
    BitSet(const BitSet& bs);

    /** Assignment operator */
    BitSet& operator=(const BitSet&);

    /** Sets bit n to 1 */
    void setBit(size_t n) { Data[n>>5] |=   1L << (n % 32);  }

    /** Sets bit n to 0 */
    void clrBit(size_t n) { Data[n>>5] &= ~(1L << (n % 32)); }

    /** Toggles bit n */
    void chgBit(size_t n) { Data[n>>5] ^=   1L << (n % 32);  }

    /** Sets bit n to State */
    void setBitTo(size_t n, bool State) 
        {
        if (State)
            setBit(n);
        else
            clrBit(n);
        }

    /* Gets bit n */
    bool getBit(size_t n) const 
        { return (Data[n>>5] & (1L << (n % 32))) ? true : false; }

    // Filling and searching
    /** Returns number of bits of state "1" */
    size_t countTrue() const;

    /** Sets all bits to State */
    void init(bool State);

    /** Finds the next bit with state "1".
     *@param n where to start
     *@returns index. index >= n, index == 0xFFFFFFFFL if not found.
     */
    size_t NextSet(size_t n) const;

    /** Finds the next bit with state "0".
     *@param n where to start
     *@returns index. index >= n, index == 0xFFFFFFFFL if not found.
     */
    size_t NextUnSet(size_t n) const;

protected:
    size_t Size; // The number of bits
    unsigned long int *Data;
    size_t getNoLongs()
        { return (Size+31)>>5; }
    ulong* allocData()
        { return new ulong[getNoLongs()]; }
};


/*****************************************************************************\
*                         INLINE FUNCTION DEFINITIONS                         *
*               No classes or user interfaces below this line !               *
\*****************************************************************************/

inline BitSet::BitSet(const BitSet& bs): Size(bs.Size), Data(allocData())
{ *this = bs; }

inline BitSet& BitSet::operator=(const BitSet &bs)
{
    memcpy(Data, bs.Data, getNoLongs()*sizeof(ulong));
    return *this;
}

inline void BitSet::init(bool State)
{
    ulong fill = State ? 0xFFFFFFFFL : 0L;
    for (size_t i=0; i< (Size+31)>>5; i++)
        Data[i] = fill;
}

inline size_t BitSet::countTrue() const
{
    size_t cnt = 0;
    for (size_t i=0; i<Size; ++i)
        if (getBit(i))
            ++cnt;
    return cnt;
}

inline size_t BitSet::NextSet(size_t i) const
{
    for (; i < Size; ++i)
        if (getBit(i))
            return i;
    return 0xFFFFFFFFL;
}

inline size_t BitSet::NextUnSet(size_t i) const
{
    for (; i < Size; ++i)
        if (!getBit(i))
            return i;
    return 0xFFFFFFFFL;
}

#endif // not BITSET_H
