// File    DenseVector.h   (C++ class definition and inline implementation)
// Library VMMATH          (Library for Vector and Matrix Algebra)

#ifndef DENSEVECTOR_H
#define DENSEVECTOR_H

#include "Vector.h"

/**
 *         A class for normal (densely populated) vectors.
 *
 *         This class is to be used for ordinary, dense vectors that do not 
 *         contain many zero elements or don't need to treat them specially.
 *         The values are stored in an array.
 *
 *@author  Philipp Berndt
 *@version $Id: DenseVector.h,v 1.2 2000/01/12 16:02:09 pumphutt Exp $
 *@see     DVRep
 */
class DenseVector : public Vector
{
    class DVRep;
public:
    /** Constructor */
    explicit DenseVector(int dim);

    /** Copy constructor */
    DenseVector(const DenseVector &v);

    /** Converting constructor */
    DenseVector(const SparseVector&);

    /** Assignment operator */
    DenseVector& operator= (const DenseVector &v);

    /** Converting assignment operator */
    DenseVector& operator= (const Vector&);

    /** Converting assignment operator */
    DenseVector& operator= (const SparseVector&);

    // Vector operations
    DenseVector& operator+= (const SparseVector&);
    DenseVector& operator+= (const DenseVector&);
    DenseVector& operator*= (const FilterMatrix &);
    DenseVector& operator*= (const SparseMatrix &);
    DenseVector& operator*= (const DiagMatrix   &);
    DenseVector& operator*= (double);

    /** Sets all elements of this vector to 0 */
    void clear();

#ifdef PVM_SUPPORT
    /** Constructor that unpacks from active PVM receive buffer. */
    DenseVector(PVM_UNPACK_t);

    /**
     * Sums up several DenseVectors using pvm_reduce().
     *
     * Every group member must call this function with the same arguments.
     *@param msgtag    The message tag that identifies this summation.
     *@param group     The PVM process group in which to operate.
     *@param rootginst The group instance number of the group member that
     *                 gets the result.
     */
    void pvm_reduce_sum(int msgtag, const char *group, int rootginst);

    /**
     * Filters this DenseVector and sends the result.
     *
     * Does the same as
     *@@ pvm_initsend(PvmDataRaw);
     *@@ ((*this) * I).pvm_pack();
     *@@ pvm_send(dest, tag);
     * only faster and using multiple messages for pipelining.
     */
    void filterAndSend(const FilterMatrix &f, int dest, int tag) const;

    /**
     * Unpacks a SparseVector from the receive buffer and adds it to
     * this DenseVector.
     * Does the same as
     *@@ (*this) += SparseVector(PVM_UNPACK);
     * only faster.
     */
    void pvm_add_recv_SparseVector();
#endif

    /** Returns a reference to the DVRep of this DenseVector. */
    DVRep &getDVRep();

    /** Returns a const reference to the DVRep of this DenseVector. */
    const DVRep &getDVRep() const;

    /** Returns pointer to internal double array. Use with caution! */
    double* getData();
};


/** Adds two DenseVectors (optimized) */
DenseVector operator+(const DenseVector &v1, const DenseVector& v2);

/** Subtracts two DenseVectors (optimized) */ 
DenseVector operator-(const DenseVector &v1, const DenseVector& v2);


/**
 *         The representation of a DenseVector.
 *
 *@author  Philipp Berndt
 *@version $Id: DenseVector.h,v 1.2 2000/01/12 16:02:09 pumphutt Exp $
 *@see     DenseVector
 */
class DenseVector::DVRep :  public Vector::VRep
{
public:
    /** Constructor */
    DVRep(int dim) : VRep(dim), data(new double[dim]) { clear(); }

    /** Copy constructor */
    DVRep(const DVRep&);

    /** Destructor */
    ~DVRep() { delete data; }

    /** Sets all elements of this DVRep to 0. */
    void clear();

    /** Returns a new copy of this VRep. */
    DVRep* clone() const { return new DVRep(*this); }

    /** Gets the value of one element. */
    double get(int col) const { return data[col]; }

    /** Sets the value of one element. */
    void set(int col, double d) { data[col] = d; }

#ifdef PVM_SUPPORT
    /** Packs this DVRep into the active PVM send buffer. */
    void pvm_pack() const;

    /** Unpacks from active PVM receive buffer into this DVRep. */
    void pvm_unpack();
#endif

//private:
    double* const data; 
};


/*****************************************************************************\
*                         INLINE FUNCTION DEFINITIONS                         *
*               No classes or user interfaces below this line !               *
\*****************************************************************************/

// Class DenseVector:

inline DenseVector::DenseVector(int dim)
    : Vector(new DenseVector::DVRep(dim)) {}

inline DenseVector::DenseVector(const DenseVector& v) 
    : Vector(v) {}

inline DenseVector::DVRep& DenseVector::getDVRep() 
{ return static_cast<DVRep&>(*rep); }

inline const DenseVector::DVRep& DenseVector::getDVRep() const 
{ return static_cast<DVRep&>(*rep); }

inline double* DenseVector::getData()
{ 
    prepareWrite();
    return getDVRep().data;
}

inline void DenseVector::clear()
{ (*this) = DenseVector(getCols()); }

inline DenseVector& DenseVector::operator= (const SparseVector &v)
{
    (*this) = DenseVector(v);
    return *this;
}

inline DenseVector& DenseVector::operator= (const DenseVector &v) 
{
    Handle::operator= (v);
    return *this;
}

inline DenseVector& DenseVector::operator= (const Vector& v)
{
    if (v.getCols() != getCols())
        throw new TypeMismatchException;
    clear();
    for (int i = 0; i< v.getCols(); ++i)
        (*this)[i] = v[i];
    return *this;
}

#ifdef PVM_SUPPORT
inline DenseVector::DenseVector(PVM_UNPACK_t)
    : Vector(new DenseVector::DVRep(pvm_upkint()))
{ getDVRep().pvm_unpack(); }
#endif

inline DenseVector& DenseVector::operator+= (const DenseVector& v)
{
    int cols = v.getCols();
    double *src = v.getDVRep().data;
    double *dst = getData();
  
    while (cols--)
        *(dst++) += *(src++);

    return *this;
}

inline DenseVector& DenseVector::operator*= (double d )
{
    double *data = getData();
    for (int i = 0; i < getCols(); ++i)
        data[i] *= d;
    return *this;
}

inline DenseVector operator*(double d, const DenseVector& v)
{
    int cols = v.getCols();
    DenseVector res(cols);
    double *src = v.getDVRep().data;
    double *dst = res.getDVRep().data;
  
    while (cols--)
        *(dst++) = *(src++) * d;

    return res;
}


// Class DVRep:

inline DenseVector::DVRep::DVRep(const DVRep& d)
    : VRep(d.getCols()), data(new double[getCols()])
{
    int cols = getCols();  
    double *src = d.data;
    double *dst = data;
  
    while (cols--)
        *(dst++) = *(src++);
}

inline void DenseVector::DVRep::clear()
{
    for (int i=getCols(); i; )
        data[--i] = 0.0;
}

#ifdef PVM_SUPPORT
inline void DenseVector::DVRep::pvm_pack() const
{
    pvm_pkint(getCols());
    pvm_pkdouble(data, getCols(), 1);
}

inline void DenseVector::DVRep::pvm_unpack()
{ pvm_upkdouble(data, getCols(), 1); }

inline void DenseVector::pvm_reduce_sum(int msgtag,
                                        const char *group, int rootginst)
{
    pvm_reduce(PvmSum, getDVRep().data, getCols(), PVM_DOUBLE,
               msgtag, group, rootginst);
}
#endif PVM_SUPPORT

inline DenseVector& DenseVector::operator*= (const DiagMatrix& d)
{
    Vector::operator*= (d);
    return *this;
}

// inline void DenseVector::DVRep::set(int col, double d)
// { data[col] = d; }

// inline double DenseVector::DVRep::get(int col) const 
// { return data[col]; }


// Non-member functions:

inline DenseVector operator+(const DenseVector &v1, const DenseVector& v2)
{
    assertDim(v1.getCols() == v2.getCols());
    DenseVector res(v1.getCols());
    for (int i=0; i< v1.getCols(); i++)
        res[i] = v1[i] + v2[i];
    return res;
}

inline DenseVector operator-(const DenseVector &v1, const DenseVector& v2)
{
    assertDim(v1.getCols() == v2.getCols());
    DenseVector res(v1.getCols());
    for (int i=0; i< v1.getCols(); i++)
        res[i] = v1[i] - v2[i];
    return res;
}

#endif // not DENSEVECTOR_H
