// File    DiagMatrix.h    (C++ class definition and inline implementation)
// Library VMMATH          (Library for Vector and Matrix Algebra)

#ifndef DIAGMATRIX_H
#define DIAGMATRIX_H

#include "Vector.h"
#include "Matrix.h"

/**
 *         An adapter for using a Vector as a diagonal matrix.
 *
 *@author  Philipp Berndt
 *@version $Id: DiagMatrix.h,v 1.2 2000/01/12 16:02:09 pumphutt Exp $
 *@see     DMRep
 */
class DiagMatrix : public Matrix
{
    class DMRep;
public:
    /** Constructor */
    DiagMatrix(const Vector& v);

    /** Copy constructor */
    DiagMatrix(const DiagMatrix&);

    /** Assignment operator */
    DiagMatrix& operator= (const DiagMatrix&); 

    /** Sets all elements of this matrix to 0 */
    void clear();

    /** Returns a reference to the diagonal vector. */
    Vector& getDiagVector();

    /** Returns a const reference to the diagonal vector. */
    const Vector& getDiagVector() const;

    /** Returns value of element (n,n). */
    double getDiagElem(int n) const;

    /** Sets element (n,n) to v. */
    void setDiagElem(int n, double v);   

    /** Returns a reference to the DMRep of this DiagMatrix. */
    DMRep &getDMRep();

    /** Returns a const reference to the SMRep of this DiagMatrix. */
    const DMRep &getDMRep() const;
};


/**
 *         The representation of a DiagMatrix.
 *
 *         The values are stored in a
 *@$       Vector
 *
 *@author  Philipp Berndt
 *@version $Id: DiagMatrix.h,v 1.2 2000/01/12 16:02:09 pumphutt Exp $
 *@see     DiagMatrix
 */
class DiagMatrix::DMRep : public Matrix::MRep, public Vector
{
public:
    /** Constructor */
    DMRep(const Vector& v) : MRep(v.getCols(),v.getCols()), Vector(v) {}

    /** Copy constructor */
    DMRep(const DMRep&);

    /** Returns a new copy of this DMRep. */
    DMRep* clone() const { return new DMRep(*this); }

    /** Gets the value of one element. */
    double get(int row, int col) const;

    /** Sets the value of one element. */
    void set(int row, int col, double d);
#ifdef PVM_SUPPORT
    /** Packs this DMRep into the active PVM send buffer. */
    void pvm_pack() const { Vector::pvm_pack(); }
#endif
private:
    Vector& operator+=(const class SparseVector &) { throw ConstException();} 
    Vector& operator*=(const class FilterMatrix &) { throw ConstException();}
    Vector& operator*=(const class SparseMatrix &) { throw ConstException();}
};


/*****************************************************************************\
*                         INLINE FUNCTION DEFINITIONS                         *
*               No classes or user interfaces below this line !               *
\*****************************************************************************/


// Class DiagMatrix:

inline DiagMatrix::DiagMatrix(const Vector& v)
    : Matrix(new DMRep(v)) {}

inline DiagMatrix::DiagMatrix(const DiagMatrix& dm)
    : Matrix(new DMRep(dm.getDMRep())) {}

inline DiagMatrix& DiagMatrix::operator= (const DiagMatrix &m) 
{
    Handle::operator= (m);
    return *this; 
}

inline void DiagMatrix::clear()
{ getDMRep().clear(); }
    
inline Vector& DiagMatrix::getDiagVector()
{ return getDMRep(); }

inline const Vector& DiagMatrix::getDiagVector() const
{ return getDMRep(); }

inline double DiagMatrix::getDiagElem(int n) const
{ return (getDMRep())[n]; }

inline void DiagMatrix::setDiagElem(int n, double v)
{ (getDMRep())[n] = v; }

inline DiagMatrix::DMRep& DiagMatrix::getDMRep() 
{ return static_cast<DMRep&>(*rep); }

inline const DiagMatrix::DMRep& DiagMatrix::getDMRep() const 
{ return static_cast<DMRep&>(*rep); }


// Class DMRep:

inline DiagMatrix::DMRep::DMRep(const DMRep& r)
    : MRep(r.MRep::getRows(), r.MRep::getCols()), Vector(r) {}

inline double DiagMatrix::DMRep::get(int row, int col) const
{ return (row == col) ? (*this)[row] : 0.0; }

inline void DiagMatrix::DMRep::set(int row, int col, double d) 
{
    if (row == col)
        (*this)[row] = d;
    else
        throw ConstException();
}

#endif // not DIAGMATRIX_H
