// File    FilterMatrix.h  (C++ class definition and inline implementation)
// Library VMMATH          (Library for Vector and Matrix Algebra)

#ifndef FILTERMATRIX_H
#define FILTERMATRIX_H

#include "Matrix.h"
#include "BitSet.h"

/**
 *         Binary diagonal matrix.
 *
 *         This class is to be used for diagonal matrices whose diagonal
 *         elements may be 1 or 0.
 *
 *@author  Philipp Berndt
 *@version $Id: FilterMatrix.h,v 1.2 2000/01/12 16:02:10 pumphutt Exp $
 *@see     SMRep
 */
class FilterMatrix : public Matrix
{
    class FMRep;
public:
    /** Constructor */
    explicit FilterMatrix(int dim);

    /** Copy constructor */
    FilterMatrix(const FilterMatrix&);

    /** Assignment operator */
    FilterMatrix& operator= (const FilterMatrix&); 

    /** Sets all elements of this matrix to 0 */
    void clear();
  
    /** Returns true if element (n,n) equals 1. */
    bool diagElemIsOne(int n) const;

    /** Sets element (n, n) to (v) ? 1 : 0; */ 
    void setDiagElem(int n, bool v);

    /** Returns a FilterMatrix with complemented diagonal elements. */
    FilterMatrix getComplement() const;

    /** Returns the number of non-zero diagonal elements. */
    int getOnesCount() const;

#ifdef PVM_SUPPORT
    /** Constructor that unpacks from active PVM receive buffer. */
    FilterMatrix(PVM_UNPACK_t);
#endif

    /** Returns a reference to the FMRep of this FilterMatrix. */
    FMRep &getFMRep();

    /** Returns a const reference to the FMRep of this FilterMatrix. */
    const FMRep &getFMRep() const;
};


/**
 *         The representation of a FilterMatrix.
 *
 *         The values are stored in a BitSet (similar to an vector<bool>).
 *
 *@author  Philipp Berndt
 *@version $Id: FilterMatrix.h,v 1.2 2000/01/12 16:02:10 pumphutt Exp $
 *@see     SparseMatrix
 */
class FilterMatrix::FMRep : public Matrix::MRep, public BitSet
{
public:
    /** Constructor */
    explicit FMRep(int dim) : MRep(dim,dim), BitSet(dim,false) {}

    /** Copy constructor */
    FMRep(const FMRep&);

    /** Returns a new copy of this FMRep. */
    FMRep* clone() const { return new FMRep(*this); }

    /** Gets the value of one element. */
    double get(int row, int col) const;

    /** Sets the value of one element. */
    void set(int row, int col, double d);

#ifdef PVM_SUPPORT
    /** Packs this SMRep into the active PVM send buffer. */
    void pvm_pack() const;

    /** Unpacks from active PVM receive buffer into this SMRep. */
    void pvm_unpack();
#endif
};


/*****************************************************************************\
*                         INLINE FUNCTION DEFINITIONS                         *
*               No classes or user interfaces below this line !               *
\*****************************************************************************/


// Class FilterMatrix:

inline FilterMatrix::FilterMatrix(int dim)
    : Matrix(new FMRep(dim)) {}

inline FilterMatrix::FilterMatrix(const FilterMatrix &m) 
    : Matrix(m) {}

inline FilterMatrix& FilterMatrix::operator= (const FilterMatrix &m) 
{
    Handle::operator= (m);
    return *this; 
}

inline void FilterMatrix::clear()
{ (*this) = FilterMatrix(getRows()); }

inline bool FilterMatrix::diagElemIsOne(int n) const
{ return getFMRep().getBit(n); }

inline void FilterMatrix::setDiagElem(int n, bool v)
{
    prepareWrite(); 
    getFMRep().setBitTo(n,v);
}

inline int FilterMatrix::getOnesCount() const
{ return getFMRep().countTrue(); }

inline FilterMatrix FilterMatrix::getComplement() const
{
    FilterMatrix res(getRows());
    for (int i=0; i<getRows(); i++)
        res.setDiagElem(i, !diagElemIsOne(i));
    return res;
}

#ifdef PVM_SUPPORT
inline FilterMatrix::FilterMatrix(PVM_UNPACK_t)
    : Matrix(new FMRep(pvm_upkint()))
{ getFMRep().pvm_unpack(); }
#endif

inline FilterMatrix::FMRep& FilterMatrix::getFMRep() 
{ return static_cast<FMRep&>(*rep); }

inline const FilterMatrix::FMRep& FilterMatrix::getFMRep() const 
{ return static_cast<FMRep&>(*rep); }


// Class FMRep:

inline FilterMatrix::FMRep::FMRep(const FMRep& r)
    : MRep(r.getRows(), r.getCols()), BitSet(r) 
{}

inline double FilterMatrix::FMRep::get(int row, int col) const
{ return (row == col && getBit(row)) ? 1.0 : 0.0; }

inline void FilterMatrix::FMRep::set(int row, int col, double d) 
{ setBitTo(row, (d == 1.0)); }

#ifdef PVM_SUPPORT

inline void FilterMatrix::FMRep::pvm_pack() const 
{
    pvm_pkint(getCols());
    pvm_pklong(reinterpret_cast<long *>(BitSet::Data),
               (BitSet::Size+31)>>5, 1);
}

inline void FilterMatrix::FMRep::pvm_unpack()
{ 
    pvm_upklong(reinterpret_cast<long *>(BitSet::Data),
                (BitSet::Size+31)>>5, 1);
}

#endif // PVM_SUPPORT
#endif // not FILTERMATRIX_H
