// File    Handle.h        (C++ class definition and inline implementation)
// Module  Handle          (Handle/Representation concept)
// Library VMMATH          (Library for Vector and Matrix Algebra)

#ifndef HANDLE_H
#define HANDLE_H

//#include <cassert>
class Handle;

/**
 *         Abstract base class of all data representations.
 *
 *         It contains a reference counter which is initially zero
 *         and can only be accessed by 
 *@$       Handle
 *         .
 *         All derived classes must supply a method
 *<code>
 *@@       Rep* clone() const
 *</code>
 *         which generates a copy of the object.
 *
 *@author  Philipp Berndt
 *@version $Id: Handle.h,v 1.2 2000/01/12 16:02:10 pumphutt Exp $
 *@see     Handle
 */
class Rep
{
public:
    /** Copy constructor */
    Rep(const Rep&) : refcnt(0) {}

    /** Default constructor */
    Rep() : refcnt(0) {}

    /** Destructor */
    virtual ~Rep() {}

    /** Generates a copy of this Rep. */
    virtual Rep* clone() const =0;
  
private:
    friend Handle;
    int refcnt;    // Number of Handles that use this Rep
};


/**
 *   Base class of all references.
 *
 *   It contains the complete lifecycle set of 
 *   constructor, copy constructor, destructor and assignment operator
 *   which need to be called by the overriding functions of the derived
 *   classes.
 *
 *@author  Philipp Berndt
 *@version $Id: Handle.h,v 1.2 2000/01/12 16:02:10 pumphutt Exp $
 *@pgk     vmmath
 *@see     Rep
 */

class Handle
{
public:
    /** Constructor */
    Handle(Rep *r)
        : rep(r)
        { rep->refcnt++; }
    
    /** Copy constructor */
    Handle(const Handle& h)
        {
            h.rep->refcnt++;
            rep = h.rep;
        }
    
    /** Destructor */
    ~Handle()
        { if (--rep->refcnt == 0) delete rep; }

    /** Assignment operator */
    Handle& operator= (const Handle& h)
        {
            h.rep->refcnt++;
            if (--rep->refcnt == 0) delete rep;
            rep = h.rep;
            return *this;
        }

    /**
     * Checks wether Rep is used by another Handle 
     * and creates a copy in that case.
     *
     * Derived classes must call this method 
     * before any write access to the representation.
     *
     *@returns true iff the Rep had to be cloned.
     */
    bool prepareWrite();
    
protected:
    /** The representation of this Handle. */
    Rep *rep;
};

inline bool Handle::prepareWrite()
{
//  assert(rep->refcnt > 0);
    if (rep->refcnt == 1) return false;
    rep->refcnt--;
    rep = rep->clone();      
    rep->refcnt++;
    return true;
}

#endif // not HANDLE_H
