// File    IndexList.h     (C++ class definition and inline implementation)
// Module  IndexList       (Single linked list with an index)
// Library VMMATH          (Library for Vector and Matrix Algebra)

#ifndef INDEXLIST_H
#define INDEXLIST_H

#include <climits>
#include <iterator>

// The is what a node of the list looks like.
template <class T>
struct _IndexList_node
{
    _IndexList_node* next;
    int index;
    T data;
};

template<class T> struct _IndexList_const_iterator; // defined below


/**
 *         IndexList::iterator base class.
 *
 *         This iterator is implemented with a pointer to the PREVIOUS node.
 *         That makes it possible to remove the current node 
 *         or insert a new node before the current for the price of
 *         dereferencing a little bit slower than
 *@$       _IndexList_iterator .
 * 
 *@author  Philipp Berndt
 *@version $Id: IndexList.h,v 1.2 2000/01/12 16:02:11 pumphutt Exp $
 *@see     IndexList
 */
template <class T>
struct _IndexList_iterator
{
    // Iterator traits
    typedef forward_iterator_tag iterator_category;
    typedef T                    value_type;
    typedef ptrdiff_t            difference_type;
    typedef T*                   pointer;
    typedef T&                   reference;

    // Abbreviations
    typedef _IndexList_iterator<T>       iterator;
    typedef _IndexList_const_iterator<T> const_iterator;
    typedef _IndexList_node<T>           list_node;

    /** Constructor */
    _IndexList_iterator(list_node* x)
        : prv(x) {}

    /** Default constructor */
    _IndexList_iterator()
        : prv(0) {}

    /** Copy constructor */
    _IndexList_iterator(const iterator& x) 
        : prv(x.prv) {}

    bool operator==(const _IndexList_iterator& x) const
        { return prv->next == x.prv->next; }

    bool operator!=(const _IndexList_iterator& x) const 
        { return prv->next != x.prv->next; }

    bool operator==(const const_iterator& x) const;

    bool operator!=(const const_iterator& x) const;

    reference operator*() const 
        { return prv->next->data; }

    pointer operator->() const 
        { return &(operator*()); }

    iterator& operator++()
        {
            prv = prv->next; 
            return *this;
        }

    iterator operator++(int)
        {
            iterator tmp = *this;
            prv = prv->next; 
            return tmp;
        }

    int getIndex() const 
        { return (prv != 0 && prv->next != 0) ? prv->next->index : INT_MAX; }

    operator int() const 
        { return getIndex(); }

    void invalidate() 
        { prv = 0; }

    list_node* prv; // The pointer to the previous node.
};


/** 
 * IndexList::const_iterator base class.
 *
 * This iterator is implemented with a pointer to the current node.
 * Simple and fast.
 *  
 *@author  Philipp Berndt
 *@version $Id: IndexList.h,v 1.2 2000/01/12 16:02:11 pumphutt Exp $
 *@see     IndexList
 */
template <class T>
struct _IndexList_const_iterator
{
    // Iterator traits
    typedef forward_iterator_tag iterator_category;
    typedef T                    value_type;
    typedef ptrdiff_t            difference_type;
    typedef const T*             pointer;
    typedef const T&             reference;

    // Abbreviations
    typedef _IndexList_iterator<T>       iterator;
    typedef _IndexList_const_iterator<T> const_iterator;
    typedef _IndexList_node<T>           list_node;

    /** Constructor */
    _IndexList_const_iterator(list_node* x)
        : cur(x) {}

    /** Default constructor */
    _IndexList_const_iterator()
        : cur(0) {}

    /** Copy constructor */
    _IndexList_const_iterator(const iterator& x)
        : cur(x.prv->next) {}


    bool operator==(const _IndexList_const_iterator& x) const
        { return cur == x.cur; }

bool operator!=(const _IndexList_const_iterator& x) const 
{ return cur != x.cur; }


const reference operator*() const { return cur->data; }
const pointer operator->() const { return &(operator*()); }

const_iterator& operator++()
{
    cur = cur->next;
    return *this;
}

const_iterator operator++(int)
{
    const_iterator tmp = *this;
    cur = cur->next;
    return tmp;
}

int getIndex() const { return (cur != 0) ? cur->index : -1; }

operator int() const { return getIndex(); }

const list_node* cur; // The pointer to the current node.
};


// Help functions
template<class T>
inline bool _IndexList_iterator<T>::operator==
(const _IndexList_const_iterator<T>& x) const
{ return prv->next == x.cur; }

template<class T>
inline bool _IndexList_iterator<T>::operator!=
(const _IndexList_const_iterator<T>& x) const
{ return prv->next != x.cur; }

template <class T> 
inline size_t _IndexList_size(_IndexList_node<T>* node)
{
    size_t result = 0;
    for ( ; node != 0; node = node->next)
        ++result;
    return result;
}


/**
 * A single linked list whose every node has got an index in addition to
 * the template data.
 *
 * Note that the list will not take care of the index order. It is your
 * responsibility to insert the elements at the correct position.
 *
 *@author  Philipp Berndt
 *@version $Id: IndexList.h,v 1.2 2000/01/12 16:02:11 pumphutt Exp $
 *@see     _Index_List_iterator
 *@see     _Index_List_const_iterator
 */
template <class T>
class IndexList
{
public:
    // Traits
    typedef T                            value_type;
    typedef value_type*                  pointer;
    typedef const value_type*            const_pointer;
    typedef value_type&                  reference;
    typedef const value_type&            const_reference;
    typedef size_t                       size_type;
    typedef ptrdiff_t                    difference_type;

    // Iterators
    typedef _IndexList_iterator<T>       iterator;
    typedef _IndexList_const_iterator<T> const_iterator;
private:
    typedef _IndexList_node<T>           list_node;

    list_node head;

    void insert_after(list_node *prv, list_node *node)
        {
            node->next = prv->next;
            prv->next = node;
        }
    
    void erase_after(list_node *prv)
        {
            list_node* cur = prv->next;
            prv->next = cur->next;
            delete cur;
        }

public:
    /** Default constructor */
    IndexList() { head.next = 0; }

    /** Copy constructor */
    IndexList(const IndexList& L) { head.next = 0; *this = L; }

    /** Assignment operator */
    IndexList& operator= (const IndexList& L);

    /** Destructor */
    ~IndexList()
        { clear(); }

    /** Returns start of iterator sequence. */
    iterator begin()
        { return iterator(&head); }

    /** Returns start of const_iterator sequence. */
    const_iterator begin() const 
        { return const_iterator(head.next);}

    //  const iterator end() { return iterator(0); }
    const_iterator end() const 
        { return const_iterator(0); }

    /** Returns number of elements. Effort: O(n). */
    size_type size() const 
        { return _IndexList_size(head.next); }
  
    /** Returns size limit, that is maximum number of elements. */
    size_type max_size() const 
        { return size_type(-1); }

    /** Is this list empty ? */
    bool empty() const
        { return head.next == 0; }

    /** Swaps contents with other list. */
    void swap(IndexList& L)
        {
            list_node* tmp = head.next;
            head.next = L.head.next;
            L.head.next = tmp;
        }

    /** Tests for equality of all elements. */
    friend bool operator==<>(const IndexList<T>& L1, const IndexList<T>& L2);

    /** Returns reference to first element. */
    reference front()
        { return ((list_node*) head.next)->data; }

    /** Returns const_reference to first element. */
    const_reference front() const
        { return ((list_node*) head.next)->data; }

    /** Inserts default element with index BEFORE pos. */
    iterator insert(iterator pos, int idx)
        {
            list_node *new_node = new list_node;
            new_node->index = idx;
            insert_after(pos.prv, new_node);
            return pos;
        }

    /** Inserts element with value and index BEFORE pos. */ 
    iterator insert(iterator pos, int idx, const value_type& x)
        {
            list_node *new_node = new list_node;
            new_node->index = idx;
            new_node->data = x;
            insert_after(pos.prv, new_node);
            return pos;
        }

/** Deletes element at given position. */
iterator erase(iterator pos)
{
    erase_after(pos.prv);
    return pos;
}

/** Deletes all elements. */
void clear() 
{ while (head.next != 0) erase_after(&head); }
};

/** Assignment operator */
template <class T>
inline IndexList<T>& IndexList<T>::operator=(const IndexList<T>& L)
{
    clear();
    iterator myi = begin();
    for (const_iterator li = L.begin(); li != L.end(); ++li, ++myi)
        {
        insert(myi, li);
        *myi = *li;
        }
    return *this;
} 

/** Tests for equality of all elements. */
template <class T>
bool operator==(const IndexList<T>& L1, const IndexList<T>& L2)
{
    typedef typename IndexList<T>::list_node list_node;
    list_node* n1 = (list_node*) L1.head.next;
    list_node* n2 = (list_node*) L2.head.next;
    while (n1 && n2 && n1->data == n2->data)
        {
        n1 = (list_node*) n1->next;
        n2 = (list_node*) n2->next;
        }
    return n1 == 0 && n2 == 0;
}

/** Swaps contents of two lists. */
template <class T>
inline void swap(IndexList<T>& x, IndexList<T>& y)
{ x.swap(y); }

#endif // not INDEXLIST_H
