// File    Matrix.h        (C++ class definition and inline implementation)
// Library VMMATH          (Library for Vector and Matrix Algebra)

#ifndef MATRIX_H
#define MATRIX_H

#include "Elem.h"
#include "Handle.h"

#ifdef PVM_SUPPORT
#include "pvm3cpp.h"
#endif

/**
 *         Matrix base class.
 *
 *@author  Philipp Berndt
 *@version $Id: Matrix.h,v 1.2 2000/01/12 16:02:11 pumphutt Exp $
 *@see     VRep
 */
class Matrix : public Elem, public Handle
{
public:
    class MRep;

    /** Constructor */
    Matrix(MRep *r);

    /** Copy constructor */
    Matrix(const Matrix&);

    /** Destructor */
    virtual ~Matrix() {}

    /** Assignment operator */
    virtual Matrix& operator= (const Matrix&);

    /** Index operator.
     *  Adresses one element of this matrix for reading. */
    const ConstElemRef operator() (int r, int c) const;
    
    /** Index operator.
     *  Adresses one element of this matrix for reading. */
    ElemRef operator() (int r, int c);

    /** Returns the number of rows of this matrix. */
    int getRows() const;
    
    /** Returns the number of cols of this matrix. */
    int getCols() const;

    /** Returns the number of nonzero elements. */
    virtual size_t size() const;

    /** Returns true if this Matrix contains only zero elements. */
    virtual bool empty() const;

    /** Prints this vector to a stream. */
    void print(ostream &os) const;
  
#ifdef PVM_SUPPORT
    /** Packs this matrix into the active PVM sendbuffer. */
    void pvm_pack() const;
#endif

    /** Sets all elements of this matrix to 0 */
    virtual void clear();

    // Matrix operations
    virtual Matrix& operator+=(const Matrix &);
    virtual Matrix& operator-=(const Matrix &);
    //virtual Matrix& operator*=(const Matrix &) = 0;
    virtual Matrix& operator*=(double);
    virtual Matrix& operator/=(double);

private:
    mutable int rowIndex, colIndex;

    /** Returns a reference to the VRep of this Vector. */
    MRep &getMRep();

    /** Returns a const reference to the VRep of this Vector. */
    const MRep &getMRep() const;

    /**
     * Gets value of adressed element.
     *
     *  Called by
     *$ ConstElemRef
     *  .
     *@see operator[] const
     */
    double get() const;

    /**
     * Sets value of adressed element.
     *
     *  Called by
     *$ ElemRef
     *  .
     *@see operator[]
     */
    void set(double d);
};


/**
 *         Matrix representation abstract base class.
 *
 *@author  Philipp Berndt
 *@version $Id: Matrix.h,v 1.2 2000/01/12 16:02:11 pumphutt Exp $
 *@see     Matrix
 */
struct Matrix::MRep : public Rep
{
public:

    /** Constructor */
    MRep(int r, int c) : rows(r), cols(c) {}

    /** Returns the number of rows of this matrix. */
    const int& getRows() const { return rows; }

    /** Returns the number of cols of this matrix. */
    const int& getCols() const { return cols; }

    /** Gets the value of one element. */
    virtual double get(int row, int col) const =0;

    /** Sets the value of one element. */
    virtual void set(int row, int col, double d) =0;

#ifdef PVM_SUPPORT
    /** Packs this MRep into the active PVM sendbuffer. */
    virtual void pvm_pack() const =0;
#endif

private:
    int rows, cols; // Dimensions
};


/** Stream output of a Vector */
inline ostream& operator<<(ostream& s, const Matrix& m)
{ m.print(s); return s; }


/*****************************************************************************\
*                         INLINE FUNCTION DEFINITIONS                         *
*               No classes or user interfaces below this line !               *
\*****************************************************************************/

// Class Matrix:
inline Matrix::Matrix(MRep *r)
    : Handle(r) {}

inline Matrix::Matrix(const Matrix& m)
    : Handle(m) {}

inline int Matrix::getRows() const 
{ return static_cast<MRep*>(rep)->getRows(); }

inline int Matrix::getCols() const 
{ return static_cast<MRep*>(rep)->getCols(); }

inline const ConstElemRef Matrix::operator() (int r, int c) const
{
    rowIndex = r;
    colIndex = c;
    return ConstElemRef(*this);
}

inline ElemRef Matrix::operator() (int r, int c)
{
    rowIndex = r;
    colIndex = c;
    return ElemRef(*this);
}

inline void Matrix::clear()
{
    for (int r=0; r<getRows(); ++r)
        for (int c=0; c<getCols(); ++c)
            (*this)(r,c) = 0.0; 
}

inline Matrix& Matrix::operator= (const Matrix& m)
{
    if (getCols() != m.getCols() || getRows() != m.getRows())
        throw TypeMismatchException();
  
    for (int r=0; r<getRows(); ++r)
        for (int c=0; c<getCols(); ++c)
            (*this)(r,c) = m(r,c); 
    return *this;
}

inline Matrix& Matrix::operator+=(const Matrix& m)
{
    if (getCols() != m.getCols() || getRows() != m.getRows())
        throw TypeMismatchException();
  
    for (int r=0; r<getRows(); ++r)
        for (int c=0; c<getCols(); ++c)
            (*this)(r,c) += m(r,c); 
    return *this;
}

inline Matrix& Matrix::operator-=(const Matrix& m)
{
    if (getCols() != m.getCols() || getRows() != m.getRows())
        throw TypeMismatchException();
  
    for (int r=0; r<getRows(); ++r)
        for (int c=0; c<getCols(); ++c)
            (*this)(r,c) -= m(r,c); 
    return *this;
}


inline Matrix& Matrix::operator*=(double d)
{
    for (int r=0; r<getRows(); ++r)
        for (int c=0; c<getCols(); ++c)
            (*this)(r,c) *= d; 
    return *this;
}

inline Matrix& Matrix::operator/=(double d)
{ return *this *= 1.0/d; }

inline bool Matrix::empty() const
{
    for (int r=0; r<getRows(); ++r)
        for (int c=0; c<getCols(); ++c)
            if ((*this)(r,c) != 0)
                return false;
    return true;
}
 
inline size_t Matrix::size() const
{
    size_t res=0;
    for (int r=0; r<getRows(); ++r)
        for (int c=0; c<getCols(); ++c)
            if ((*this)(r,c) != 0)
                ++res;
    return res;
}

inline void Matrix::print(ostream &os) const
{
    if (getCols()>20 || getRows()>20)
        {
        int s = size();
        os << s << " entries.\n";
        if (s>20)
            os << "Matrix too big for printing!\n";
        else
            {
            double d;
            for (int r=0; r<getRows(); ++r)
                for (int c=0; c<getCols(); ++c)
                    if ((d=(*this)(r,c)) != 0)
                        os <<'('<<r<<", "<<c<<")="<<d<<'\t';
            os << '\n';
            }
        }
    else
        for (int r = 0; r<getRows(); ++r)
            {
            for (int c = 0; c<getCols(); ++c)
                os << (*this)(r,c) << '\t';
            os << '\n';
            }
}

inline double Matrix::get() const
{ return getMRep().get(rowIndex, colIndex); }

inline void Matrix::set(double d)
{
    prepareWrite();
    getMRep().set(rowIndex, colIndex, d); 
}

#ifdef PVM_SUPPORT
inline void Matrix::pvm_pack() const
{ getMRep().pvm_pack(); }
#endif


// Class Matrix::MRep:

inline Matrix::MRep& Matrix::getMRep() 
{ return static_cast<MRep&>(*rep); }

inline const Matrix::MRep& Matrix::getMRep() const 
{ return static_cast<MRep&>(*rep); }

#endif // not MATRIX_H
