// File    SparseMatrix.h  (C++ class definition and inline implementation)
// Library VMMATH          (Library for Vector and Matrix Algebra)

#ifndef SPARSEMATRIX_H
#define SPARSEMATRIX_H

#include "Matrix.h"
#include "IndexList.h"

/**
 *         A class for sparsely populated matrices.
 *
 *         This class is optimized for sparse matrices
 *         (that contain many zero elements.)
 *
 *@author  Philipp Berndt
 *@version $Id: SparseMatrix.h,v 1.2 2000/01/12 16:02:12 pumphutt Exp $
 *@see     SMRep
 */
class SparseMatrix : public Matrix
{
    class SMRep;
public:
    /** Constructor */
    SparseMatrix(int r, int c);

    /** Copy constructor */
    SparseMatrix(const SparseMatrix &m); 

    /** Assignment operator */
    SparseMatrix& operator= (const SparseMatrix &);

    /** Converting assignment operator */
    SparseMatrix& operator= (const Matrix&);
    
    /** Sets all elements of this matrix to 0 */
    void clear();

    /** Returns true if this Matrix contains only zero elements. */
    bool empty() const;

    /** Returns the number of nonzero elements. */
    size_t size() const;

    // Matrix operations
    SparseMatrix& operator+= (const SparseMatrix &);
    SparseMatrix& operator-= (const SparseMatrix &);
    SparseMatrix& operator+= (const FilterMatrix &);
    SparseMatrix& operator+= (const Matrix &);
    SparseMatrix& operator-= (const Matrix &);
    SparseMatrix& operator*= (const Matrix &);
    SparseMatrix& operator*= (double);
    SparseMatrix& operator/= (double);

#ifdef PVM_SUPPORT
    /** Constructor that unpacks from active PVM receive buffer. */
    SparseMatrix(PVM_UNPACK_t);
#endif

    /** Returns a reference to the SMRep of this SparseMatrix. */
    SMRep &getSMRep();

    /** Returns a const reference to the SMRep of this SparseMatrix. */
    const SMRep &getSMRep() const;
};


/** Divides every matrix element by d. */
SparseMatrix operator/(const SparseMatrix &m, double d);

/** Adds two SparseMatrices. */
SparseMatrix operator+(const SparseMatrix &m, const Matrix& m2);


/**
 *         The representation of a SparseMatrix.
 *
 *         The values are stored in a
 *         single list of single lists (rows) of non-zero elements (cols).
 *<pre>         
 *@@       SMRep
 *@@         |
 *@@         v
 *@@        Row -> (Elem) -> (Elem) -> (Elem)
 *@@         |
 *@@         v
 *@@        Row -> (Elem) -> (Elem)
 *</pre>
 *@author  Philipp Berndt
 *@version $Id: SparseMatrix.h,v 1.2 2000/01/12 16:02:12 pumphutt Exp $
 *@see     SparseMatrix
 */
class SparseMatrix::SMRep : public Matrix::MRep,
                    public IndexList< IndexList<double> >
{
public:
    typedef IndexList<double> Row;
    typedef const_iterator      row_const_iterator;
    typedef iterator            row_iterator;
    typedef Row::const_iterator col_const_iterator;
    typedef Row::iterator       col_iterator;

    /** Constructor */
    SMRep(int r, int c) : MRep(r,c) {}

    /** Copy constructor */
    SMRep(const SMRep &x);

    /** Returns a new copy of this SMRep. */
    SMRep* clone() const;

    /** Gets the value of one element. */
    double get(int row, int col) const;

    /** Sets the value of one element. */
    void set(int row, int col, double d);
  
#ifdef PVM_SUPPORT
    /** Packs this SMRep into the active PVM send buffer. */
    void pvm_pack() const;

    /** Unpacks from active PVM receive buffer into this SMRep. */
    void pvm_unpack();
#endif

private:
    void find(int row, int col) const;
    bool rowex(int row) const;
    bool elmex(int row, int col) const;
  
    mutable IndexList<Row>::iterator rowit;
    mutable Row::iterator colit;
};


/*****************************************************************************\
*                         INLINE FUNCTION DEFINITIONS                         *
*               No classes or user interfaces below this line !               *
\*****************************************************************************/

// Class SparseMatrix:

inline SparseMatrix::SparseMatrix(int r, int c) 
    : Matrix(new SMRep(r,c)) {}

inline SparseMatrix::SparseMatrix(const SparseMatrix &m) 
    : Matrix(m) {}

inline SparseMatrix& SparseMatrix::operator= (const SparseMatrix &m) 
{
    Handle::operator= (m);
    return *this; 
}

inline void SparseMatrix::clear()
{ (*this) = SparseMatrix(getRows(), getCols()); }

inline bool SparseMatrix::empty() const
{ return getSMRep().empty(); }

inline size_t SparseMatrix::size() const
{
    SparseMatrix::SMRep::row_const_iterator ri;
    SparseMatrix::SMRep::col_const_iterator ci;
    size_t res=0;
    for (ri=getSMRep().begin(); ri!=getSMRep().end(); ++ri)
        for (ci=ri->begin(); ci!=ri->end(); ++ci)
            ++res;
    return res;
}

#ifdef PVM_SUPPORT
inline SparseMatrix::SparseMatrix(PVM_UNPACK_t)
    : Matrix(new SMRep(pvm_upkint(), pvm_upkint()))
{ getSMRep().pvm_unpack(); }
#endif

inline SparseMatrix::SMRep& SparseMatrix::getSMRep() 
{ return static_cast<SMRep&>(*rep); }

inline const SparseMatrix::SMRep& SparseMatrix::getSMRep() const 
{ return static_cast<SMRep&>(*rep); }

inline SparseMatrix& SparseMatrix::operator=(const Matrix& m)
{
    *this = SparseMatrix(m.getRows(), m.getCols());
    for (int r=0; r < getRows(); ++r)
        for (int c=0; c < getCols(); ++c)
            (*this)(r,c) = m(r,c);
    return *this;
}

inline SparseMatrix& SparseMatrix::operator*=(double d)
{
    prepareWrite();
    SparseMatrix::SMRep::row_iterator ri;
    SparseMatrix::SMRep::col_iterator ci;

    for (ri=getSMRep().begin(); ri != getSMRep().end(); ++ri)
        for (ci=ri->begin(); ci != ri->end(); ++ci)
            *ci *= d;
    return *this;
}

inline SparseMatrix& SparseMatrix::operator/=(double d)
{ 
    *this *= 1.0/d;
    return *this; 
}

inline SparseMatrix& SparseMatrix::operator+=(const Matrix& m)
{
    Matrix::operator+=(m);
    return *this;
}

inline SparseMatrix& SparseMatrix::operator-=(const Matrix& m)
{
    Matrix::operator-=(m);
    return *this;
}

inline SparseMatrix &SparseMatrix::operator+=(const SparseMatrix &m)
{
    SparseMatrix::SMRep::row_const_iterator ri;
    SparseMatrix::SMRep::col_const_iterator ci;

    for (ri=m.getSMRep().begin(); ri!=m.getSMRep().end(); ++ri)
        for (ci=ri->begin(); ci!=ri->end(); ++ci)
            (*this)(ri, ci) += *ci;
    return *this;
}

inline SparseMatrix &SparseMatrix::operator-=(const SparseMatrix &m)
{
    SparseMatrix::SMRep::row_const_iterator ri;
    SparseMatrix::SMRep::col_const_iterator ci;

    for (ri=m.getSMRep().begin(); ri!=m.getSMRep().end(); ++ri)
        for (ci=ri->begin(); ci!=ri->end(); ++ci)
            (*this)(ri, ci) -= *ci;
    return *this;
}


// Class SMRep:

inline SparseMatrix::SMRep::SMRep(const SMRep &r) 
    : MRep(r), IndexList<Row>(r) 
{
    rowit.invalidate();
    colit.invalidate(); 
}

inline SparseMatrix::SMRep* SparseMatrix::SMRep::clone() const
{ return new SMRep(*this); }

inline void SparseMatrix::SMRep::find(int row, int col) const
{
    SMRep& smr = const_cast<SMRep&>(*this);

    if (rowit.getIndex() != row)
        {
        if (rowit.getIndex() > row)
            rowit = smr.begin();
      
        while (rowit != end() && rowit.getIndex() < row)
            ++rowit;
      
        if (rowit.getIndex() != row)
            {
            colit.invalidate();
            return;
            }
        colit = rowit->begin();
        }
  
    if (colit.getIndex() > col)
        colit = rowit->begin();
  
    while (colit != rowit->end() && colit.getIndex() < col)
        ++colit;
}

inline bool SparseMatrix::SMRep::rowex(int row) const
{ return (rowit.getIndex() == row); }

inline bool SparseMatrix::SMRep::elmex(int row, int col) const
{ return (rowex(row) && colit.getIndex() == col); }

inline double SparseMatrix::SMRep::get(int row, int col) const 
{ 
    find(row,col);
    return elmex(row,col) ? *colit : 0.0;
}

inline void SparseMatrix::SMRep::set(int row, int col, double d)
{
    find(row, col);
    if (d==0.0)
        {
        if (elmex(row, col))
            {
            rowit->erase(colit);
            if (rowit->empty())
                erase(rowit);
            }
        }
    else
        {
        if (!elmex(row, col))
            {
            if (!rowex(row))
                {         
                insert(rowit, row);
                colit = rowit->begin();
                }
            rowit->insert(colit, col);    
            }
        *colit = d;
        }
}

#ifdef PVM_SUPPORT
inline void SparseMatrix::SMRep::pvm_pack() const
{
    SparseMatrix::SMRep::row_const_iterator ri;
    SparseMatrix::SMRep::col_const_iterator ci;
    
    pvm_pkint(getRows());
    pvm_pkint(getCols());
  
    for (ri = begin(); ri != end(); ++ri)
        {
        pvm_pkint(ri);
        for (ci = ri->begin(); ci != ri->end(); ++ci)
            {
            pvm_pkint(ci);
            pvm_pkdouble(*ci);
            }
        pvm_pkint(-1); 
        }
    pvm_pkint(-1); 
}

inline void SparseMatrix::SMRep::pvm_unpack()
{
    int row, col;
    double val;
    clear();

    SMRep::row_iterator ri = begin();
    while ((row=pvm_upkint()) != -1)
        {
        insert(ri, row);
        col_iterator ci = ri->begin();      
        while ((col = pvm_upkint()) != -1)
            {
            val = pvm_upkdouble();
            ri->insert(ci, col, val);
            ++ci;
            }
        ++ri;
        }
}
#endif // PVM_SUPPORT


// Non-member functions:

inline SparseMatrix operator/(const SparseMatrix& m, double d)
{
    SparseMatrix res(m);
    return res /= d;
}

inline SparseMatrix operator+(const SparseMatrix& m, const Matrix& m2)
{
    SparseMatrix res(m);
    return res += m2;
}

#endif // not SPARSEMATRIX_H
