// File    SparseVector.h  (C++ class definition and inline implementation)
// Library VMMATH          (Library for Vector and Matrix Algebra)

#ifndef SPARSEVECTOR_H
#define SPARSEVECTOR_H

#include "IndexList.h"
#include "Vector.h"

/**
 *         A class for sparsely populated vectors.
 *
 *         This class is optimized for sparse vectors
 *         (that contain many zero elements).
 *
 *@author  Philipp Berndt
 *@version $Id: SparseVector.h,v 1.2 2000/01/12 16:02:12 pumphutt Exp $
 *@see     SVRep
 */
class SparseVector : public Vector
{
    class SVRep;
public:
    /** Constructor */
    explicit SparseVector(int dim);

    /** Copy constructor */
    SparseVector(const SparseVector &v);

    /** Converting constructor */
    SparseVector(const Vector &v);

    /** Assignment operator */
    SparseVector& operator= (const SparseVector &v);

    /** Converting assignment operator */
    SparseVector& operator= (const Vector&);

    /** Sets all elements of this vector to 0 */
    void clear();

    /** Dot product */
    double operator*(const Vector& v2) const;
    
    /** Dot product */
    double operator*(const SparseVector& v2) const;

    SparseVector& operator+= (const SparseVector&);
    SparseVector& operator*= (double);
    SparseVector& operator*= (const FilterMatrix&);
    SparseVector& operator*= (const SparseMatrix&);
    SparseVector& operator*= (const DiagMatrix& d);

#ifdef PVM_SUPPORT
    /** Constructor that unpacks from active PVM receive buffer. */
    SparseVector(PVM_UNPACK_t);

    /** Sends this SparseVector using multiple messages. */
    void pvm_send_frag(int dest, int tag);
#endif

    /** Returns a reference to the SVRep of this SparseVector. */
    SVRep &getSVRep();

    /** Returns a const reference to the SVRep of this SparseVector. */
    const SVRep &getSVRep() const;
};


/**
 *         The representation of a SparseVector.
 *
 *         The values are stored in an single list of non-zero elements.
 *
 *@author  Philipp Berndt
 *@version $Id: SparseVector.h,v 1.2 2000/01/12 16:02:12 pumphutt Exp $
 *@see     SparseVector
 */
class SparseVector::SVRep :  public Vector::VRep, public IndexList<double>
{
public:
    /** Constructor */
    explicit SVRep(int dim) : VRep(dim) {}

    /** Copy constructor */
    SVRep(const SVRep& svr) : VRep(svr), IndexList<double>(svr)
        { it.invalidate(); }

    /** Returns a new copy of this VRep. */
    SVRep* clone() const { return new SVRep(*this); }

    /** Gets the value of one element. */
    double get(int col) const { find(col); return elemFound(col) ? *it : 0.0; }

    /** Sets the value of one element. */
    void set(int col, double d);

#ifdef PVM_SUPPORT
    /** Packs this SVRep into the active PVM send buffer. */
    void pvm_pack() const;

    /** Unpacks from active PVM receive buffer into this SVRep. */
    void pvm_unpack();
#endif

private:
    void find(int col) const;
    bool elemFound(int col) const { return (it.getIndex() == col); }
    mutable iterator it;
};

/** Adds two SparseVectors. */
inline SparseVector operator+(const SparseVector &v1, const SparseVector &v2)
{
    SparseVector res(v1);
    res += v2;
    return res;
}

/** S-Product */ 
inline SparseVector operator*(const SparseVector &v, double s)
{
    SparseVector res(v);
    return res *= s;
}


/*****************************************************************************\
*                         INLINE FUNCTION DEFINITIONS                         *
*               No classes or user interfaces below this line !               *
\*****************************************************************************/

// Class SparseVector:

inline SparseVector::SparseVector(int dim)
    : Vector(new SparseVector::SVRep(dim)) {}

inline SparseVector::SparseVector(const SparseVector &v) 
    : Vector(v) {}

inline SparseVector& SparseVector::operator= (const SparseVector &v) 
{
    Handle::operator= (v);
    return *this; 
}

inline SparseVector::SparseVector(const Vector &v) 
    : Vector(new SparseVector::SVRep(v.getCols()))
{ *this = v; }

inline SparseVector& SparseVector::operator= (const Vector& v)
{
    clear();
    double d;
    SVRep &svr = getSVRep();
    for (int i=0; i<getCols(); ++i)
        if ((d=v[i]) != 0.0)
            svr.set(i, d);
    return *this;
}

inline void SparseVector::clear()
{ (*this) = SparseVector(getCols()); }

inline SparseVector& SparseVector::operator+= (const SparseVector& v) 
{
    SparseVector::SVRep::const_iterator it;
    prepareWrite();
    SVRep& vr = getSVRep();
    for (it = v.getSVRep().begin(); it != v.getSVRep().end(); ++it)
        vr.set(it, vr.get(it) + *it);
    return *this;
}

inline SparseVector& SparseVector::operator*= (double d)
{
    prepareWrite();
    for (SVRep::iterator it = getSVRep().begin();
        it != getSVRep().end();
        ++it)
        *it *= d;
    return *this;
}

inline double SparseVector::operator*(const SparseVector& v2) const
{
    SVRep::const_iterator it1(getSVRep().begin()), it2(v2.getSVRep().begin());
    double ac=0.0;
    int x1, x2;
  
    while (it1 != getSVRep().end() && it2 != v2.getSVRep().end())
        {
        x1 = it1.getIndex(); x2 = it2.getIndex();
        if (x1 == x2)
            {
            ac += (*it1) * (*it2);
            ++it1; ++it2;
            continue;
            }
        if (x1 < x2)
            ++it1;
        else
            ++it2;
        }
    return ac;
}

inline double SparseVector::operator*(const Vector& v2) const
{
    double ac = 0.0;
    for (SVRep::const_iterator i=getSVRep().begin();
        i != getSVRep().end(); ++i)
        ac += *i * v2[i];

    return ac;
}

inline SparseVector::SVRep& SparseVector::getSVRep() 
{ return static_cast<SVRep&>(*rep); }

inline const SparseVector::SVRep& SparseVector::getSVRep() const 
{ return static_cast<SVRep&>(*rep); }

#ifdef PVM_SUPPORT
inline SparseVector::SparseVector(PVM_UNPACK_t)
    : Vector(new SVRep(pvm_upkint()))
{  getSVRep().pvm_unpack(); }
#endif


// Class SparseVector::SVRep:

inline void SparseVector::SVRep::find(int col) const
{
    SVRep& svr = const_cast<SVRep&>(*this);
    if (it.getIndex() > col)
        it = svr.begin();
    
    while (it != end()
          && it.getIndex() < col)
        ++it;
}                      

inline void SparseVector::SVRep::set(int col, double d)
{
    find(col);

    if (d == 0.0)
        {
        if (elemFound(col))
            erase(it);
        }
    else
        {
        if (!elemFound(col))
            insert(it, col);
        *it = d;
        }
}

#ifdef PVM_SUPPORT
inline void SparseVector::SVRep::pvm_pack() const
{
    pvm_pkint(getCols());
    for (const_iterator i = begin(); i != end(); ++i)
        {
        pvm_pkint(i);
        pvm_pkdouble(*i);
        }
    pvm_pkint(-1); 
}

inline void SparseVector::SVRep::pvm_unpack()
{
    int idx;
    double val;
    clear();
    iterator i = begin();

    while (true)
        {
	idx = pvm_upkint();
	if (idx < 0)
	    {
	    if (idx==-2)
		{
		int sender = pvm_upkint();
		int tag = pvm_upkint();
		pvm_recv(sender, tag);
		continue;
		}		
	    else break;
	    }

        val = pvm_upkdouble();
        insert(i, idx, val);
        ++i;
        }
}

inline void SparseVector::pvm_send_frag(int dest, int tag)
{
#ifdef PVM_FRAGMENT
    // 12 bytes per vector elem, 60 bytes header
    const int fragsize = (pvm_getopt(PvmFragSize) - 60) / 12;
    int cnt = 0;
#endif
    pvm_initsend(PvmDataRaw);
    pvm_pkint(getCols());
    
    for (SVRep::const_iterator i = getSVRep().begin();
	i != getSVRep().end(); ++i)
        {
#ifdef PVM_FRAGMENT
	if (++cnt == fragsize)
	    {
	    pvm_pkint(-2);
	    pvm_pkint(pvm_mytid());
	    pvm_pkint(tag);
	    pvm_send(dest, tag);
	    pvm_initsend(PvmDataRaw);
	    cnt = 0;
	    }	
#endif
        pvm_pkint(i);
        pvm_pkdouble(*i);
        }
    pvm_pkint(-1); 
    P3PP::pvm_send(dest, tag);
}

#endif // PVM_SUPPORT

#endif // not SPARSEVECTOR_H
