// File    Vector.h        (C++ class definition and inline implementation)
// Library VMMATH          (Library for Vector and Matrix Algebra)

#ifndef VECTOR_H
#define VECTOR_H

#include <cmath> // sqrt, fabs
#include "Elem.h"
#include "Handle.h"

#ifdef PVM_SUPPORT
#include "pvm3cpp.h"
#endif

//#define PRINT_AS_COL_VEC

// Defined elsewhere:
class FilterMatrix;
class SparseMatrix;
class DiagMatrix;
class SparseVector;
class DenseVector;

/**
 *         Vector base class.
 *
 *@author  Philipp Berndt
 *@version $Id: Vector.h,v 1.2 2000/01/12 16:02:13 pumphutt Exp $
 *@see     VRep
 */
class Vector : public Elem, public Handle
{
    class VRep;

public:
    /** Constructor */
    Vector(VRep *);

    /** Copy constructor */
    Vector(const Vector& v) : Handle(v) {}

    /** Destructor */
    virtual ~Vector() {}

    /** Assignment operator */
    virtual Vector& operator= (const Vector&);

    /** Returns the dimension of this column vector. */
    int getCols() const;

    /** Index operator.
     *  Adresses one element of this vector for reading. */
    const ConstElemRef operator[] (int i) const;

    /** Index operator.
     *  Adresses one element of this vector for reading and writing. */
    ElemRef operator[] (int i);

    /** Sets all elements of this vector to 0 */
    virtual void clear();
    
    // Vector operations
    virtual double operator* (const Vector&) const;       //     v . v2
    virtual double operator* (const SparseVector&) const; //     v . v2
//  virtual Vector& operator+= (const Vector&) =0;        // v = v + v2
    virtual Vector& operator+= (const SparseVector&) =0;  // v = v + sv
    virtual Vector& operator-= (const Vector&);           // v = v - v2
    virtual Vector& operator*= (double);                  // v = v * s
    virtual Vector& operator/= (double);                  // v = v * 1/s
//  virtual Vector& operator*= (const Matrix       &);
    virtual Vector& operator*= (const FilterMatrix &) = 0;
    virtual Vector& operator*= (const SparseMatrix &) = 0;
    virtual Vector& operator*= (const DiagMatrix   &);

#ifdef PVM_SUPPORT
    /** Packs this vector into the active PVM sendbuffer. */
    void pvm_pack() const;
#endif

    /** Prints this vector to a stream in human readable form. */
    void print(ostream &os) const;

    /** Parses this vector from a stream in human readable form. */
    void parse(istream &is);

    /** Writes this vector to a stream in binary form. */
    virtual void write(ostream &os) const;

    /** Reads this vector from a stream in binary form. */
    virtual void read(istream &is);

private:
    friend ElemRef;

    /** Returns a reference to the VRep of this Vector. */
    VRep &getVRep();
    
    /** Returns a const reference to the VRep of this Vector. */
    const VRep &getVRep() const;

    /** The index of the adressed element. */
    mutable int colIndex;

    /**
     *     Gets value of adressed element.
     *
     *     Called by
     *$    ConstElemRef
     *     .
     *@see operator[] const
     */
    double get() const;

    /**
     *     Sets value of adressed element.
     *
     *     Called by
     *$    ElemRef
     *     .
     *@see operator[]
     */
    void set(double d);
};


/** Stream output of a Vector */
inline ostream& operator<<(ostream& s, const Vector& v)
{ v.print(s); return s; }

/** Stream input of a Vector */
inline istream& operator>>(istream& s, Vector &v)
{ v.parse(s); return s; }

/** Returns the largest difference between the corresponding elements
    of v1 and v2. */
double maxDiff(const Vector &v1, const Vector &v2);

/** Checks whether the largest difference between the corresponding elements
    of v1 and v2 is less than eps. */
bool near(const Vector &v1, const Vector &v2, double eps);

/** Returns the absolute of a vector. */
double abs(const Vector&);



/**
 *         Vector representation abstract base class.
 *
 *@author  Philipp Berndt
 *@version $Id: Vector.h,v 1.2 2000/01/12 16:02:13 pumphutt Exp $
 *@see     Vector
 */
class Vector::VRep : public Rep
{
public:
    /** Constructor */
    explicit VRep(int c) : cols(c) {}
    
    /** Returns the dimension of this column vector. */
    const int& getCols() const { return cols; }

    /** Gets the value of one element. */
    virtual double get(int col) const =0;

    /** Sets the value of one element. */
    virtual void set(int col, double d) =0;

#ifdef PVM_SUPPORT
    /** Packs this VRep into the active PVM sendbuffer. */
    virtual void pvm_pack() const =0;
#endif

private:
    int cols;
};



/*****************************************************************************\
*                         INLINE FUNCTION DEFINITIONS                         *
*               No classes or user interfaces below this line !               *
\*****************************************************************************/

// Class Vector:
inline Vector::Vector(VRep *r)
    : Handle(r) {}

inline ElemRef Vector::operator[] (int i) 
{ 
    colIndex = i;
    return ElemRef(*this);
}  

inline const ConstElemRef Vector::operator[] (int i) const  
{
    colIndex = i;
    return ConstElemRef(*this);
}  

inline int Vector::getCols() const
{ return getVRep().getCols(); }

inline void Vector::clear()
{
    for (int i=0; i<getCols(); ++i)
        (*this)[i] = 0.0;
}

inline Vector& Vector::operator= (const Vector& v)
{
    for (int i=0; i<getCols(); ++i)
        (*this)[i] = v[i];
    return *this;
}

inline void Vector::write(ostream &os) const
{
    int cols = getCols();
    os.write(&cols, sizeof(int));
    for (int i=0; i<cols; ++i)
        {
        double d = (*this)[i];
        os.write(&d, sizeof(double));
        }
}

inline void Vector::read(istream &is)
{
    int cols;
    is.read(&cols, sizeof(int));
    for (int i=0; i<cols; ++i)
        {
        double d;
        is.read(&d, sizeof(double));
        (*this)[i]=d;
        }
}

inline void Vector::print(ostream &os) const
{
    os.setf(ios::fixed, ios::floatfield);
#ifdef PRINT_AS_COL_VEC
    if (getCols()>20)
        os << "Vector too big for printing!\n";
    else
        {
        os << '\n';
        int c = getCols()-1;
        for (int i = 0; i<c; i++)
            os << (*this)[i] << '\n';
        os << (*this)[c];
        }
#else
    if (getCols()>20)
        os << "Vector too big for printing!\n";
    else
        for (int i = 0; i<getCols(); i++)
            os << (*this)[i] << '\t';
#endif
    os.setf(0, ios::floatfield);
}

inline void Vector::parse(istream &is)
{
    double d;
    for (int i=0; i<getCols(); ++i)
        {
        is >> d;
        (*this)[i] = d;
        }
}

inline double Vector::get() const
{ return getVRep().get(colIndex); }

inline void Vector::set(double d)
{
    prepareWrite();
    getVRep().set(colIndex, d); 
}

#ifdef PVM_SUPPORT
inline void Vector::pvm_pack() const
{ getVRep().pvm_pack(); }
#endif

inline double Vector::operator* (const Vector& v) const
{
    if (getCols() != v.getCols())
        throw TypeMismatchException();
    double d = 0.0;

    for (int i = 0; i<getCols(); i++)
        d += (*this)[i] * v[i];

    return d;
}

// inline Vector& Vector::operator+= (const Vector& v)
// {
//     if (getCols() != v.getCols())
//      throw TypeMismatchException();

//     for (int i = 0; i<getCols(); i++)
//      (*this)[i] += v[i];

//     return *this;
// }

inline Vector& Vector::operator-= (const Vector& v)
{
    if (getCols() != v.getCols())
        throw TypeMismatchException();

    for (int i = 0; i<getCols(); i++)
        (*this)[i] -= v[i];

    return *this;
}

inline Vector& Vector::operator*= (double d)
{
    for (int i = 0; i<getCols(); i++)
        (*this)[i] *= d;

    return *this;
}

inline Vector& Vector::operator/= (double d)
{   return *this *= 1.0/d; }


// Class Vector::VRep:

inline Vector::VRep& Vector::getVRep() 
{ return static_cast<VRep&>(*rep); }

inline const Vector::VRep& Vector::getVRep() const 
{ return static_cast<VRep&>(*rep); }


// Non-member functions:

inline double abs(const Vector& v)
{
    double tmp, abs = 0.0;

    for (int i = 0; i < v.getCols(); ++i)
        if ((tmp = v[i]) != 0.0)
            abs += tmp * tmp;

    return sqrt(abs);
}

inline double maxDiff(const Vector &v, const Vector &w)
{
    double tmp, maxDiff = 0.0;

    for (int i = 0; i < v.getCols(); ++i)
        if ((tmp = fabs(v[i] - w[i])) > maxDiff)
            maxDiff = tmp;
  
    return maxDiff;
}

inline bool near(const Vector &v1, const Vector &v2, double eps)
{
    for (int i = 0; i < v1.getCols(); ++i)
        if (fabs(v1[i] - v2[i]) > eps)
            return false;
    return true;
}

#endif // not VECTOR_H
