// File    mixed.h         (C++ class definition and inline implementation)
// Module  mixed           (Misc. mixed algebra operators)
// Library VMMATH          (Library for Vector and Matrix Algebra)

#ifndef MIXED_H
#define MIXED_H

//  operator*  ////////////////////////////////////////////////////////////////

inline double Vector::operator* (const SparseVector& v) const
{ return v * (*this); }

inline double operator* (const Vector& v, const OnesVector& e)
{
    double sum = 0.0;

    for (int i = 0; i < v.getCols(); ++i)
        sum += v[i];
  
    return sum;
}

inline SparseVector operator* (const Vector &v, const FilterMatrix &f)
{
    SparseVector res(v.getCols());
    for (int c=0; (c = f.getFMRep().BitSet::NextSet(c)) != -1; c++)
        res[c] = v[c];
    return res;
}

inline DenseVector operator*(const DenseVector &v, const SparseMatrix &m)
{
    typedef SparseMatrix::SMRep::row_const_iterator rci;
    typedef SparseMatrix::SMRep::col_const_iterator cci;

    if (v.getCols() != m.getRows())
        throw TypeMismatchException();
    DenseVector res(m.getCols());
    for (rci ri= m.getSMRep().begin(); ri != m.getSMRep().end(); ++ri)
        for (cci ci= ri->begin(); ci != ri->end(); ++ci)
            res[ci] += v[ri] * (*ci);
    return res;
}

inline double operator* (const SparseVector& v, const OnesVector& e)
{
    double sum = 0.0;
    SparseVector::SVRep::const_iterator it;
    for (it = v.getSVRep().begin(); it != v.getSVRep().end(); ++it)
        sum += *it;
  
    return sum;
}

inline SparseVector operator*(const SparseVector& v, const SparseMatrix& m)
{
    typedef SparseMatrix::SMRep::row_const_iterator rci;
    typedef SparseMatrix::SMRep::col_const_iterator cci;
    SparseVector res(v.getCols());
    res.prepareWrite();
    const SparseVector::SVRep &v_svr = v.getSVRep();
    SparseVector::SVRep &res_svr = res.getSVRep();
    

    for (rci ri= m.getSMRep().begin(); ri != m.getSMRep().end(); ++ri)
        {
        double ve = v_svr.get(ri);
//      double ve = v[ri];
        for (cci ci= ri->begin(); ci != ri->end(); ++ci)
//          res[ci] += ve * (*ci);
            {
            int idx = ci.getIndex();
            res_svr.set(idx, res_svr.get(idx) + ve * (*ci));
            }
        }
    return res;
}

inline SparseMatrix operator*(const SparseMatrix& m, const FilterMatrix& f)
{
    typedef SparseMatrix::SMRep::row_const_iterator rit;
    typedef SparseMatrix::SMRep::col_const_iterator cit;
    SparseMatrix res(m.getRows(), m.getCols());

    for (rit ri = m.getSMRep().begin(); ri != m.getSMRep().end();++ri)
        for (cit ci = ri->begin(); ci != ri->end();++ci)
            if (f.diagElemIsOne(ci))
                res(ri,ci) = *ci;
    return res;
}

inline SparseMatrix operator* (const FilterMatrix &f, const SparseMatrix &m)
{
    SparseMatrix res(m.getRows(),f.getCols());
    SparseMatrix::SMRep::row_iterator ri = res.getSMRep().begin();
    SparseMatrix::SMRep::row_const_iterator mi = m.getSMRep().begin();
  
    for (;mi != m.getSMRep().end(); ++mi)
        if (f.diagElemIsOne(mi))
            {
            res.getSMRep().insert(ri,mi);
            *ri = *mi;
            ++ri;
            }
    return res;
}


//  operator *=  //////////////////////////////////////////////////////////////

inline Vector& Vector::operator*= (const DiagMatrix& d)
{
    for (int i=0; i<getCols(); ++i)
        (*this)[i] *= d.getDiagElem(i);
    return *this;
}

inline DenseVector& DenseVector::operator*=(const FilterMatrix &f)
{
    for (int c=0; (c = f.getFMRep().BitSet::NextUnSet(c)) != -1; c++)
        (*this)[c] = 0.0;
    return *this;
}

inline DenseVector& DenseVector::operator*=(const SparseMatrix &m)
{ return (*this) = (*this) * m; }

inline SparseVector& SparseVector::operator*= (const DiagMatrix& d)
{
    double s;
    prepareWrite();
    SVRep::iterator it = getSVRep().begin();
    while (it != getSVRep().end())
        if ((s = d.getDiagElem(it))!=0.0)
            *(it++) *= s;
        else
            getSVRep().erase(it);
    return *this;
}

inline SparseVector& SparseVector::operator*= (const FilterMatrix& f)
{
    prepareWrite();
    SVRep::iterator it = getSVRep().begin();
    while (it != getSVRep().end())
        if (!f.diagElemIsOne(it))
            getSVRep().erase(it);
        else
            ++it;
    return *this;
}

inline SparseVector& SparseVector::operator*=(const SparseMatrix& m)
{ return (*this) = (*this) * m; }

inline SparseMatrix& operator*= (SparseMatrix &m, const FilterMatrix &f)
{
    typedef SparseMatrix::SMRep::row_iterator rit;
    typedef SparseMatrix::SMRep::col_iterator cit;
    m.prepareWrite();

    for (rit ri = m.getSMRep().begin(); ri != m.getSMRep().end();)
        {
        for (cit ci = ri->begin(); ci != ri->end();)
            if (!f.diagElemIsOne(ci))
                ri->erase(ci);
            else
                ++ci;
        if (ri->empty())
            m.getSMRep().erase(ri);
        else
            ++ri;
        }
    return m;
}


//  operator+=  ///////////////////////////////////////////////////////////////

// inline Vector& Vector::operator+= (const SparseVector& v) 
// {
//     SparseVector::SVRep::const_iterator it;
//     for (it = v.getSVRep().begin(); it != v.getSVRep().end(); ++it)
//      (*this)[it] += *it;
//     return *this;
// }

inline DenseVector& DenseVector::operator+= (const SparseVector& v) 
{
    double *data = getData();
    SparseVector::SVRep::const_iterator it;
    for (it = v.getSVRep().begin(); it != v.getSVRep().end(); ++it)
        data[it] += *it;
    return *this;
}


inline SparseMatrix& SparseMatrix::operator+= (const FilterMatrix& f)
{
    for (int c=0; (c = f.getFMRep().BitSet::NextSet(c)) != -1; c++)
        (*this)(c,c) += 1.0;
    return *this;
}


// Misc.  /////////////////////////////////////////////////////////////////////

inline DenseVector::DenseVector(const SparseVector& v)
    : Vector(new DenseVector::DVRep(v.getCols()))
{
    SparseVector::SVRep::const_iterator it;
    double *data = getDVRep().data;
    for (it = v.getSVRep().begin(); it != v.getSVRep().end(); ++it)
        data[it] = *it;
}


#ifdef PVM_SUPPORT
inline void DenseVector::filterAndSend(const FilterMatrix &f,
				       int dest, int tag) const
{
#ifdef PVM_FRAGMENT
    // 12 bytes per vector elem, 60 bytes header
    const int fragsize = (pvm_getopt(PvmFragSize) - 60) / 12;
    int cnt = 0;
#endif
    const double *data = getDVRep().data;
    pvm_initsend(PvmDataRaw);
    pvm_pkint(getCols());
    for (int c=0; (c = f.getFMRep().BitSet::NextSet(c)) != -1; c++)
        {
#ifdef PVM_FRAGMENT
	if (++cnt == fragsize)
	    {
	    pvm_pkint(-2);
	    pvm_pkint(pvm_mytid());
	    pvm_pkint(tag);
	    pvm_send(dest, tag);
	    pvm_initsend(PvmDataRaw);
	    cnt = 0;
	    }	
#endif
        pvm_pkint(c);
        pvm_pkdouble(data[c]);
        }
    pvm_pkint(-1);
    pvm_send(dest, tag);
}

inline void DenseVector::pvm_add_recv_SparseVector()
{
    double *data = getDVRep().data;
    int dim = pvm_upkint();
    assertDim(dim == getCols());

    while (true)
        {
	int idx = pvm_upkint();
	if (idx < 0)
	    {
	    if (idx==-2)
		{
		int sender = pvm_upkint();
		int tag = pvm_upkint();
		pvm_recv(sender, tag);
		continue;
		}		
	    else break;
	    }

        data[idx] += pvm_upkdouble();
        }
}
#endif

#endif // not MIXED_H
