/******************************************************************************

File              :     analysis_str.c

Purpose           :     Statistical analysis of output data received from
			simulating the transient phase;
			check conditions for backend start and provide data
			for visualisation
			

Programmer        :     Ralf Putz

History           :     December 1995

exported function :     void expand_sample_space()    
			void Measure_data_out()
			Str_Stat *STR_RewardMeasure_Init()
			void Analysis()
			void reached_max_real_time()

local functions   :     char *sample_malloc()
			int calc_measure_precision()
                        void free_sample_space()
			double get_min_value()
			double get_max_value()
			void write_header_of_curves_file()
			void update_curves_file()
			void close_curves_file()
			bool send_conditions_reached()
			bool visu_conditions_reached()
			bool measure_reached_precision()
			void handle_visualisation()

******************************************************************************/

#include <math.h> 
#include <stdio.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "stat_utils.h"
#include "sim_master_types.h"
#include "analysis_str.h"

/* external variables */
extern  double          current_time;
extern  int             conf_level, no_slaves, no_runs;
extern  int             no_measures, no_sampling_points;
extern  double          eps_max, transient_time;
extern  meas_names_list *measure_names;
extern  int             StatSema;
extern  FILE            *stat_file, *temp_sim_curves_file, *sim_curves_file;
extern  FILE            *res_file;
extern  Str_Stat        **all_str_measures;
extern  char            model_name[FILE_NAME];
extern  bool            PercentRuleOn;

/* global variables */
static char             *tnethome;
static Backend_running = false;
static Start_Visualisation = false;

char *er_malloc ();
char *getenv();

/* void bedarf()
{
  long speicherkosten = 0;
  struct rusage ru;
 
  getrusage(RUSAGE_SELF,&ru);
  speicherkosten = ru.ru_maxrss * getpagesize();
  printf("\n Speicherbedarf: %ld",speicherkosten);
} */

/*************************************************************************

  function:     calc_measure_precision   

  parameters:   Measaure -> pointer to a reward measure  

  returns:      percentage of bases which have reached the predefined 
		accuaracy (rounded as integer)

  purpose:      calculate for each measure the percentage of bases which
		have reached the predefined accuracy

  author:       Ralf Putz

  date:         December 1995

*************************************************************************/

int calc_measure_precision(Measure)
Str_Stat *Measure;
{
  int counter = 0;
  int k;
  double *dummy;
  
  dummy = (double *)er_malloc(sizeof(double));
  for( k = 0; k < no_sampling_points; k++ )
    if( Measure->Epsilon[k] < eps_max )
       counter++;
  if( (modf( (double)(counter * 100)/ (double)no_sampling_points ,dummy))  > 0.5 )
    {
      /* round up */
      free(dummy);
      return (int)ceil(((double)counter/(double)no_sampling_points ) * 100.0) ;
    }
  else
    {
      /* round off */
      free(dummy);
      return (int)floor(((double)counter/(double)no_sampling_points) * 100.0) ;
    }
}

/****************************************************************************

  function:     Measure_data_out 

  parameters:   --   

  returns:      --   

  purpose:      write simulation results into the stat-file and the res-file

  author:       Ralf Putz

  date:         December 1995

****************************************************************************/
void Measure_data_out()
{
  int i,k,counter;
  double time, step, eps;
  FILE *stat, *res;
  Str_Stat *MeasureStat;

  counter = 0;
  stat = stat_file;
  res = res_file;
  step = transient_time / ( no_sampling_points - 1 );

  for( i = 0; i < no_measures; i++ )    
    {
      MeasureStat = all_str_measures[i];
      if(MeasureStat->SampleCount >= ANALYSIS_SIZE )
	counter++;
    }
  if( counter >= no_measures )
    {
    for ( i = 0; i < no_measures; i++ )
      {
	MeasureStat = all_str_measures[i];
	time = 0.0;
	fprintf(stat,"\nsamples: %d",
	      MeasureStat->SampleCount);
	fprintf(stat,"    %d percent of bases reached the predefined accuracy",
			  calc_measure_precision(MeasureStat));
	fprintf(stat,"\n#time:  mean:   c.-intervall:  epsilon: variance:  \n");
	for (k  = 0; k < no_sampling_points;k++)
	  {
	    fprintf(stat,"\n %.3f\t",time);
	    fprintf(stat,"%.3f\t",MeasureStat->AveragedMean[k]);
	    fprintf(stat,"[%.3f,%.3f]\t",
		       MeasureStat->AveragedMean[k] - MeasureStat->DeltaX[k],
		       MeasureStat->AveragedMean[k] + MeasureStat->DeltaX[k]);
	    fprintf(stat,"%.5f\t",MeasureStat->Epsilon[k]);
	    fprintf(stat,"%f",MeasureStat->Variance[k]); 
	    time += step;  
	  }
	 fprintf(stat,"\n");
	 fflush(stat);
      }
    fprintf(res,"Results for transient simulation at time %.2f\n\n",transient_time); 
    for ( i = 0; i < no_measures; i++ )
      {
	MeasureStat = all_str_measures[i];
        fprintf(res,"%s = %.5f\n",   
           MeasureStat->name,
           MeasureStat->AveragedMean[no_sampling_points-1]);
      }
    fflush(res); 
  }
  else
    {
      sleep(1);
      printf("\nNot enough samples: -> No results\n");
    }
} /* Measure_data_out  */

/*************************************************************************

  function:     get_min_value 

  parameters:   Measaure -> pointer to a reward measure  

  returns:      minimum value of all bases of one measure 

  purpose:      get the minimum value of all bases of one measure      

  author:       Ralf Putz

  date:         December 1995

*************************************************************************/

double get_min_value( Measure )
Str_Stat *Measure;
{
  int i;
  double min = Measure->AveragedMean[0] - Measure->DeltaX[0];

  for( i = 1; i < no_sampling_points; i++ )
    if( (Measure->AveragedMean[i] - Measure->DeltaX[i]) < min )
      min = Measure->AveragedMean[i] - Measure->DeltaX[i];

  return min;
}

/*************************************************************************

  function:     get_max_value

  parameters:   Measaure -> pointer to a reward measure

  returns:      maximum value of all bases of one measure 

  purpose:      get the maximum value of all bases of one measure      
  
  author:       Ralf Putz

  date:         December 1995

*************************************************************************/

double get_max_value( Measure )
Str_Stat *Measure;
{
  int i;
  double max = Measure->AveragedMean[0] + Measure->DeltaX[0];

  for( i = 1; i < no_sampling_points; i++ )
    if( (Measure->AveragedMean[i] + Measure->DeltaX[i]) >  max )
      max = Measure->AveragedMean[i] + Measure->DeltaX[i];

  return max;  
}

/*************************************************************************

  function:     write_header_of_curves_file   

  parameters:   pointer to curves-file

  returns:      --

  purpose:      write the header of the curves-file

  author:       Ralf Putz

  date:         December 1995

*************************************************************************/

void write_header_of_curves_file(curves_file) 
FILE *curves_file; 
{  
  int i,k;
  double time, step;
  Str_Stat *MeasureStat;

  /* header schreiben */
  fprintf(curves_file,"# file %s contains the simulation results of the transient phase\n#\n",curves_file);

  fprintf(curves_file,"# number of x values: a\n");
  fprintf(curves_file,"%d\n", no_sampling_points);

  fprintf(curves_file,"# number of defined reward measures: b\n");
  fprintf(curves_file,"%d\n", no_measures);

  fprintf(curves_file,"# list of reward measure names:\n");
  for( i = 0;i < no_measures; i++ )
    fprintf(curves_file,"%s ", all_str_measures[i]->name);
  fprintf(curves_file,"\n");

  fprintf(curves_file,"# last x value:\n");
  fprintf(curves_file,"%f \n", transient_time);

  fprintf(curves_file,"# maximum y value:\n");
  for( i = 0;i < no_measures; i++ )
  {
    fprintf(curves_file,"%f ", get_max_value(all_str_measures[i]) );
  }
  fprintf(curves_file,"\n");

  fprintf(curves_file,"# minimum y value:\n");
  for( i = 0; i < no_measures ; i++ )
  {
    fprintf(curves_file,"%f ", get_min_value(all_str_measures[i]) );
  }
  fprintf(curves_file,"\n");

  fprintf(curves_file,"# in case of an experiment, number of runs: c\n");
  fprintf(curves_file,"%d\n", 1 );    /* wird vorerst nicht unterstuetzt! */

  fprintf(curves_file,"# results: represented as a x 2b+1 matrices\n");
  fflush(curves_file);
}

/*************************************************************************

  function:     update_curves_file    

  parameters:   pointer to curves-file

  returns:      --

  purpose:      write data of all measures into curves-file for the next
		visualisation step 

  author:       Ralf Putz

  date:         December 1995

*************************************************************************/
void update_curves_file(curves_file)
FILE *curves_file;
{
  int i,k;
  double time, step;
  Str_Stat *MeasureStat;

  step = transient_time / ( no_sampling_points - 1 );
  time = 0.0;
      
	
  for( k = 0; k < no_sampling_points; k++ )
    {
      fprintf(curves_file,"%f ",time);
      for( i = 0; i < no_measures; i++ )
	{
	  MeasureStat = all_str_measures[i];
	  if( MeasureStat->StopSendedToBackend == false )
	    {
	      fprintf(curves_file,"%.6f ",MeasureStat->AveragedMean[k]);
	      fprintf(curves_file,"%.6f ",MeasureStat->DeltaX[k]);
	    }

	}
	time += step;
	fprintf(curves_file,"\n");
     }

  for( i = 0; i < no_measures; i++ )
    {
       MeasureStat = all_str_measures[i];
       if( (MeasureStat->StopSimulation == true)  &&
	   (MeasureStat->StopSendedToBackend == false) )
	 {
	   fprintf(curves_file,"&%d Measure %s stopped\n",i,MeasureStat->name);
	   MeasureStat->StopSendedToBackend = true;
	 }
     }


  fflush(curves_file); 
}

/*************************************************************************

  function:     close_curves_file    

  parameters:   curve_file -> pointer to curves-file
		modus -> RESULT, if data should be written into curves-file  

  returns:      --

  purpose:      write simulation stop message into curves-file respectivly
		data

  author:       Ralf Putz

  date:         December 1995

*************************************************************************/

void close_curves_file(curves_file, modus)
FILE *curves_file;
int modus;
{  
  double time, step;
  int k, i;
  Str_Stat *MeasureStat;

  if( modus == RESULT)
    {
      step = transient_time / ( no_sampling_points - 1 );
      time = 0.0;
	
      for( k = 0; k < no_sampling_points; k++ )
	{
	  fprintf(curves_file,"%f ",time);
	  for( i = 0; i < no_measures; i++ )
	    {

	      MeasureStat = all_str_measures[i];
	      fprintf(curves_file,"%.6f ",MeasureStat->AveragedMean[k]);
	      fprintf(curves_file,"%.6f ",MeasureStat->DeltaX[k]);
	    }
	    time += step;
	    fprintf(curves_file,"\n");
	 }
    }
  fprintf(curves_file,"# Results of Simulation\n");
  fflush(curves_file);
  fclose(curves_file); 
}

/*************************************************************************

  function:     sample_malloc     

  parameters:   number of bytes   

  returns:      pointer to the allocated memory block

  purpose:      allocation of memory for samples      

  author:       Ralf Putz

  date:         December 1995

*************************************************************************/

char *sample_malloc (bytes)
     int bytes;
{
  char *pointer;
  if (!bytes)
    return (NULL);
  pointer = (char *)malloc(bytes);
  if (pointer == NULL)
    {
      printf ("\nNot enough memory for %d bytes\n", bytes);
      printf ("Restart the Simulation with less measures or bases\n");
      close_curves_file(temp_sim_curves_file, TEMP);
      Measure_data_out();
      exit (FAIL);
    }
  return pointer;
}

/*************************************************************************

  function:     free_sample_space  

  parameters:   Measaure -> pointer to a reward measure

  returns:      --

  purpose:      free the sample block of a measure

  author:       Ralf Putz

  date:         December 1995

*************************************************************************/

void free_sample_space(Measure, size)
Str_Stat *Measure;
int size;
{
  int i;

  for( i = 0; i < size; i++ )  
    free( Measure->SampleBlock[i]);
  free(Measure->SampleBlock);
}

/*************************************************************************

  function:     expand_sample_space  

  parameters:   Measaure -> pointer to a reward measure

  returns:      --

  purpose:      expand the sample block of a measure

  author:       Ralf Putz

  date:         December 1995

*************************************************************************/

void expand_sample_space(Measure)
Str_Stat *Measure;
{
  int i, j, size;
  str_sample_type **HelpSampleBlock;
  double *test_double;

/*  printf("\n expand_sample_space:");
  printf("\n MeasureName: %s MemoryCounter: %d",Measure->name,Measure->MemoryCounter); */
  Measure->MemoryCounter++;

  size = SAMPLE_MALLOC_SIZE * Measure->MemoryCounter;
  HelpSampleBlock = (str_sample_type **)sample_malloc(size * 
				 sizeof(str_sample_type *) );
  for( i = 0; i < size; i++)
     HelpSampleBlock[i] = (str_sample_type *)sample_malloc(no_sampling_points *
						   sizeof(str_sample_type) );
  size = SAMPLE_MALLOC_SIZE * (Measure->MemoryCounter - 1); 
  for( i = 0; i < size; i++ )
    memcpy(HelpSampleBlock[i],Measure->SampleBlock[i],
				no_sampling_points * sizeof(str_sample_type)); 
  free_sample_space(Measure, size);
  Measure->SampleBlock = HelpSampleBlock;
}

/*************************************************************************

  function:     STR_RewardMeasure_Init

  parameters:   index to list of measure names

  returns:      pointer to the new defined statistic

  purpose:      defines a new reward measure statistic 

  author:       Ralf Putz

  date:         September 1995

*************************************************************************/

Str_Stat *STR_RewardMeasure_Init(index)
int index;
{
  Str_Stat  *MeasureStat;
  int  i, counter;
  meas_names_list *help_measure_names;

  MeasureStat = (Str_Stat *) er_malloc (sizeof (Str_Stat));
  MeasureStat->AveragedMean = 
		    (double *) er_malloc(no_sampling_points * sizeof(double));
  MeasureStat->Variance = 
		    (double *) er_malloc(no_sampling_points * sizeof(double));
  MeasureStat->DeltaX = 
		    (double *) er_malloc(no_sampling_points * sizeof(double));
  MeasureStat->Epsilon= 
		    (double *) er_malloc(no_sampling_points * sizeof(double));
  StatSema++;

  MeasureStat->StopSimulation = false;
  MeasureStat->StopSendedToSlave = false;
  MeasureStat->StopSendedToBackend = false;
  MeasureStat->MemoryCounter = 1;
  MeasureStat->SendFactor = 1; 
  MeasureStat->SampleCount = 0;
  MeasureStat->EndTime = 0.0; 
  MeasureStat->PrecPortion = 1.0;
  for (i=0; i < no_sampling_points; i++)
    {
      MeasureStat->AveragedMean[i] = 0.0;
      MeasureStat->Variance[i] = 0.0; 
      MeasureStat->Epsilon[i] = 0.0;
      MeasureStat->DeltaX[i] = 0.0;
    }     
  help_measure_names = measure_names;
  counter = 0;
  while( counter < index )
    {
      help_measure_names = help_measure_names->next;
      counter++;
    }
  strcpy(MeasureStat->name,help_measure_names->name);
  return (MeasureStat);
} /*RewardMeasure_Init*/

/*************************************************************************

  function:     reached_max_real_time   

  parameters:   --  

  returns:      --   

  purpose:      finaly write data into sim_curves- and temp_sim_curves-file    
		if real time limit has been reached

  author:       Ralf Putz

  date:         December 1995

*************************************************************************/

void reached_max_real_time()
{
  int i;
  Str_Stat *Measure;
  int counter = 0;

  for( i = 0; i < no_measures; i++ )
  { 
    Measure = all_str_measures[i];
    if(Measure->StopSimulation != true)
      {
        Measure->StopSimulation = true;
        StatSema--;
      }
    Measure->StopSendedToSlave = true;
  }
  for( i = 0; i < no_measures; i++ )    
    {
      Measure = all_str_measures[i];
      if(Measure->SampleCount >= ANALYSIS_SIZE )
	counter++;
    }
  if( counter >= no_measures )
    {
       close_curves_file(temp_sim_curves_file, TEMP);
       write_header_of_curves_file(sim_curves_file); 
       close_curves_file(sim_curves_file, RESULT);
     }
  else
    close_curves_file(temp_sim_curves_file, TEMP);
}

/*************************************************************************

  function:     send_conditions_reached    

  parameters:   --

  returns:      true: all measures have collected enough data for writing   
		      them into curves-file

  purpose:      check whether enough data have been collected for the next
		visualisation step

  author:       Ralf Putz

  date:         December 1995

*************************************************************************/

bool send_conditions_reached()
{
  int i, counter;
  Str_Stat *Measure;

  counter = 0;
  for( i = 0; i < no_measures; i++)
    {
      Measure = all_str_measures[i];
      if( (Measure->SampleCount >= (Measure->SendFactor * ANALYSIS_SIZE))  || 
	  (Measure->StopSimulation == true) )
	counter++;
    }
    if( counter >= no_measures )
      {
	for( i = 0; i < no_measures; i++)
	  {
	    Measure = all_str_measures[i];
	    Measure->SendFactor++;
	  }
	return true;
      }
    else
      return false;
}

/*************************************************************************

  function:     visu_conditions_reached  

  parameters:   --  

  returns:      true: all measures have collected enough data to start backend

  purpose:      check whether enough datas have been collected to start backend
       
  author:       Ralf Putz

  date:         December 1995

*************************************************************************/

bool visu_conditions_reached()
{
  int i, counter;
  Str_Stat *Measure;

  counter = 0;
  for( i = 0; i < no_measures; i++)
    {
      Measure = all_str_measures[i];
      if( (Measure->SampleCount >= ANALYSIS_SIZE)  || 
	  (Measure->StopSimulation == true) )
	counter++;
    }
    if( counter >= no_measures )
      return true;
    else
      return false;
}

/*************************************************************************

  function:     measure_reached_precision  

  parameters:   Measaure -> pointer to a reward measure

  returns:      true: measure has reached its predefined accuracy

  purpose:      checks whether a measure has reached its predefined accuracy
		(using percentage-rule if PercentRuleOn = true)

  author:       Ralf Putz

  date:         December 1995

*************************************************************************/
      
bool measure_reached_precision(Measure)
Str_Stat *Measure;
{
  int counter = 0;
  int k, no_allowed_deviaters;
  double new_prec_portion;

    if(PercentRuleOn == true) 
      if( Measure->SampleCount  >= PREC_LIMIT )
	{
	  new_prec_portion = 1.0 - 
		( ((double)( floor((double)(Measure->SampleCount  
					  / PREC_LIMIT)) )) / 10.0 );
	  Measure->PrecPortion = (new_prec_portion > 0.0 ? 
					  new_prec_portion: 0.0 );

	}
  no_allowed_deviaters = floor(no_sampling_points - 
                               (no_sampling_points * Measure->PrecPortion));
/*  printf("\n deviaters: %d    ",no_allowed_deviaters);  */
  for( k = 0; k < no_sampling_points; k++ )
    if( Measure->Epsilon[k] >= eps_max )
       counter++;
/*  printf("\n counter: %d    ",counter); */
  return ( (counter <= no_allowed_deviaters ) ? true : false );
} 

/*************************************************************************

  function:     handle_visualisation 

  parameters:   -- 

  returns:      --     

  purpose:      if enough simulation results has been achieved the backend
		is started. If the backend is running data for the next
		visualisation step is written into the curves-file

  author:       Ralf Putz

  date:         December 1995

*************************************************************************/

void handle_visualisation()
{
  char command[1000];
  
  if( Start_Visualisation == false )
    Start_Visualisation = visu_conditions_reached();

  if( (Backend_running == false) && (Start_Visualisation == true) )
    {
     write_header_of_curves_file(temp_sim_curves_file);
     switch( fork() )
       {
       case -1: perror("Can't start Backend"); 
		exit(1);
       case 0:
	      {
		if(  !(tnethome = getenv("TNETHOME")) )
		  {
		    fprintf(stderr,"TNETHOME not set -> can't start Backend\n");
		    exit(-1);
		   }
		 sprintf(command,"%s/Simulation/bin/Backend %s.temp_sim_curves -name 'Measure Display' -geometry 600x300+200+100",tnethome,model_name);
		 system(command);    
		 exit(0);
	       }
       }
       Backend_running = true;
   }
   if(Backend_running)
     {
	if( StatSema > 0 )
	 {
	   if(send_conditions_reached())
	     update_curves_file( temp_sim_curves_file );
	  }
	else
	  {
	    update_curves_file( temp_sim_curves_file );  
	    close_curves_file( temp_sim_curves_file, TEMP ); 

	    write_header_of_curves_file(sim_curves_file); 
	    close_curves_file(sim_curves_file, RESULT); 
	  }
      }
}

/*************************************************************************

  function:     Analysis   

  parameters:   Measaure -> pointer to a reward measure

  returns:      --

  purpose:      performs statistical analysis of measure data

  author:       Ralf Putz

  date:         December 1995

*************************************************************************/

void Analysis(Measure)
Str_Stat *Measure;

{
  double help1, help2;
  double alpha = ALPHAT_DEF / 2; 
  char command[1000];
  int i, k, size;
  for(k = 0; k < no_sampling_points; k++)
    {
      help1 = 0.0; 
      for (i = 0; i < Measure->SampleCount; i++)
	help1 += Measure->SampleBlock[i][k];
      Measure->AveragedMean[k] = help1 / Measure->SampleCount;  
    }
  for(k = 0; k < no_sampling_points; k++)
    {
      help2 = 0.0;
      for (i = 0; i < Measure->SampleCount; i++)
	help2 += pow(Measure->SampleBlock[i][k] - 
			Measure->AveragedMean[k],2.0);

      Measure->Variance[k] = help2 / (Measure->SampleCount -1);
      Measure->DeltaX[k] = student_distr((Measure->SampleCount)- 1,1.0 - alpha)
			    * sqrt(Measure->Variance[k]) /  
			      sqrt((double)Measure->SampleCount); 
      
      if (Measure->AveragedMean[k] == 0.0)  /* Only possible if all values */
	 Measure->Epsilon[k] = 0.0;         /* are zeros */
      else 
	 Measure->Epsilon[k] = Measure->DeltaX[k] / Measure->AveragedMean[k];
    } 


  if( measure_reached_precision(Measure) == true ) 
    if( Measure->StopSimulation == false)
      {
	Measure->StopSimulation = true;
        size = SAMPLE_MALLOC_SIZE * Measure->MemoryCounter; 
        free_sample_space(Measure,size); 
	StatSema--;
       }
  handle_visualisation();   
} 



