/*NEUE_VERSION (BM)*/
/*************************************************************************

  filename:	batch_cv.c

  purpose:	It implements functions to perform an statistical analysis 
                of simulation output data using variance reduction techniques.
		The reduction technique applied is known as control variates.
		The function for testing the required precision uses the
		method for independent replications i.e. analysis of 
		uncorrelated means coolected as output from several 
		replications combinded with the 
		application of control variates.

		See 
		1.[K. Pawlikowsky]: "Steady-State Simulation of Queueing
		  Processes: Problems ans Solutions",
		  ACM Computing Surveys 1990, p. 123.
		2.[Lavenberg, Welch]: "A Perspectiveon the use of control 
		  variables to increase the efficiency of Monte-Carlo-
		  Simulations",
		  Managment Science, Vol.27, No.3, March1981.

  author:	Christian Kelling
                Technical University of Berlin
		Computer Science Department
		Institute for Technical Computer Science

  date:		April 1994

  exported functions:

			     	      CVR: control varites with replications
  void	CVR_Analysis_Init    	      initialize the statistic package 
  void	CVR_RewardMeasure_Init        initialize each reward measure statistic 
  int	CVR_Analysis  		      record each sample 
  void	CVR_RewardMeasure_data_out    outputs of each reward measure statistic 
  void	CVR_Globals_out  	      outputs of the statistic package 
  double CVR_Mean_out		      return the mean value
  double CVR_Lower_out		      return the lower bound of the conf. int.
  double CVR_Upper_out		      return the upper bound of the conf. int.

  local functions:
  
  void	ListConvert
  int   TestCorrelation
  int   CVEstimation

  Datei im Ramen des PDV-Projekts im WS 95/96 gendert. (BM)

  nderung:	SEND_BLOCK_DEF wird durch STAT_BLOCK_DEF ersetzt.
                SEND_BLOCK_DEF war die statische Definition der Lnge von 
		bertragenen Sample-Paketen. Die empfangenen Paket wurden als
		Block abgelegt. Da diese Pakete jetzt eine variable Lnge haben
		mssen sie nun jeweils direkt hinter das letzte empfangene
		Sample-Paket im Speicher gelegt werden, um Speicherverschnitt
		zu vermeiden. Durch die Variable Durch STAT_BLOCK_DEF wird nun
		die kleinste Einheit von Samples, also der krzeste Sample-
		Block, definiert.
		Beide Konstanten sind in in Master/stat_utils.h definiert.


  Autoren:	Mathias Neumann und Benjamin Hohnhuser 

  Datum:   	April 1996

*************************************************************************/


#include <math.h>
#include <stdio.h>

#include "num_utils.h"
#include "stat_utils.h"
#include "sim_master_types.h"
#include "batch_cv.h"

/* external functions */
#ifdef __STDC__
extern double uniPdfAreaErr(char   *measureName,
                            double lowerBorder,
                            double upperBorder,
                            int n,
                            double meanValue,
                            double variance,
                            double alpha);    /* Irrtumswahrscheinlichkeit */
#else
extern double uniPdfAreaErr();
#endif

/* external variables */
extern  double  current_time;
extern 	int	conf_level;
extern	double	eps_max;
extern  double  errPercent;
extern  meas_names_list *measure_names;
extern	int	StatSema;
extern  int     NoPrecCount;
extern  bool    min_flag;
extern  int     time_exceeded;
extern  bool    trans_detection;
extern	FILE	*stat_file;

/* global variables used by the modul */

static long 		CVR_n_max;

static int		paw_stattotal	= 0;

static SampSeq		*TempDetSeq	= NULL;

char *er_malloc ();

/* end of variables */


/*************************************************************************

  functions:	CVR_Mean_out CVR_Lower_out CVR_Upper_out 

  parameters:	MeasureStat -> pointer to a reward measure statistic

  return:	the mean of the measure and the bounds of the conf. interval

  author:	Christian Kelling
		Technical University of Berlin

  date:		May 1993

*************************************************************************/

double CVR_Mean_out(MeasureStat)
CvR_Stat *MeasureStat;
{
  return (MeasureStat->SampleCount )  * LOCAL_BATCH_SIZE_DEF 
    *MeasureStat->Least_Sq_Mean/MeasureStat->EndTime ; 
}

double CVR_Lower_out(MeasureStat)
CvR_Stat *MeasureStat;
{
  return (MeasureStat->SampleCount ) * LOCAL_BATCH_SIZE_DEF*
    (MeasureStat->Least_Sq_Mean - MeasureStat->DeltaX)/MeasureStat->EndTime;;
}

double CVR_Upper_out(MeasureStat)
CvR_Stat *MeasureStat;
{
  return (MeasureStat->SampleCount  ) * LOCAL_BATCH_SIZE_DEF*
    (MeasureStat->Least_Sq_Mean + MeasureStat->DeltaX)/MeasureStat->EndTime;;
}


/*************************************************************************

  function:	CVR_RewardMeasure_data_out

  parameters:	MeasureStat -> pointer to a reward measure statistic

  purpose:	performs the output of results of a statistic

  author:	Christian Kelling
		Technical University of Berlin

  date:		May 1993

*************************************************************************/

void CVR_RewardMeasure_data_out(MeasureStat)
CvR_Stat *MeasureStat;
{
  FILE *stat;
  stat = stat_file;

  if (trans_detection)
    {
      fprintf(stat, "\nTransient Analysis at slaves (maximum values):\n");
      fprintf(stat, "==============================================\n\n");
      fprintf(stat, "Length: %d samples\t",MeasureStat->InitLength);
      fprintf(stat, "TransTime: %f\n",MeasureStat->TransTime);
    }
  else
      fprintf(stat, "\nTransient Analysis was turned off.\n");
  fprintf(stat, "\nCVR Staedy State Analysis:\n");
  fprintf(stat, "==========================\n\n");
  fprintf(stat, "Epsilon: %f\t",MeasureStat->Epsilon);
  fprintf(stat, "Samples: %lu\t",(MeasureStat->SampleCount + 1)*LOCAL_BATCH_SIZE_DEF);
  fprintf(stat, "VirtTime: %f\n",MeasureStat->EndTime);
  fprintf(stat, "AvMean: %f\t",MeasureStat->AveragedMean );
  fprintf(stat, "L_Sq_Mean: %f\t",MeasureStat->Least_Sq_Mean );
  fprintf(stat, "Variance: %1.12f\n",MeasureStat->Variance);
  fprintf(stat, "LowInterv: %f\t",
	 (MeasureStat->Least_Sq_Mean - MeasureStat->DeltaX) );
  fprintf(stat, "UpInterv: %f\n",
	 (MeasureStat->Least_Sq_Mean + MeasureStat->DeltaX) );
  fprintf(stat, "\nSeqCount: %d\t",MeasureStat->SequenceCount);
  fprintf(stat, "Batchings: %d\t",MeasureStat->BatchingCount);
  fprintf(stat, "Analysed Samples: %d\n",MeasureStat->NoAnalysedSamples);
  fprintf(stat, "StopSimul: %d\t", MeasureStat->StopSimulation);
  fprintf(stat, "NoPrecision: %d\n\n", MeasureStat->NoPrecision);
} /* RewardMeasure_data_out */


/*************************************************************************

  function:	CVR_Globals_out

  parameters:	

  purpose:	performs the output of global results of the statistical
		analysis for a simulation run.

  author:	Christian Kelling
		Technical University of Berlin

  date:		May 1993

*************************************************************************/

void CVR_Globals_out()
{
  FILE *stat;
  stat = stat_file;

  fprintf(stat, "\n\n\nVariance Analysis with conrol variates\n");
  fprintf(stat, "======================================\n\n");
  fprintf(stat, "ConfidenceLevel: %d\n",conf_level);
  fprintf(stat, "Statistics: %d\t\t",paw_stattotal);
  fprintf(stat, "NoPrecision: %d\n",NoPrecCount);
} /* CVR_globals_out */


/*************************************************************************

  function:	CVR_Analysis_Init

  parameters:	maxsamples -> the maximum allowed length of the simulation
			run measured in the number of recorded observations.
			The calling function uses the number of messages
			to transmit.

  purpose:	init local data of this modul and global variables used
                by all statistics

  author:	Christian Kelling
		Technical University of Berlin

  date:		May 1993

*************************************************************************/

void CVR_Analysis_Init (maxsamples)
long maxsamples;

{
  CVR_n_max = maxsamples;
} /* CVR_Analysis_Init */


/*************************************************************************

  function:	CVR_RewardMeasure_Init

  parameters:	number of control variates

  returns:      pointer to the new defined statistic

  purpose:      Defines a new reward measure statistic 

  author:	Christian Kelling
		Technical University of Berlin

  date:		May 1993

*************************************************************************/


CvR_Stat *CVR_RewardMeasure_Init(index,controls)

int  index,controls;
{
  CvR_Stat  *MeasureStat;
  ContrSeqList  *TempSeqPointer;
  int  i, counter;
  meas_names_list *help_measure_names;

  MeasureStat = (CvR_Stat *) er_malloc (sizeof (CvR_Stat));
  MeasureStat->ControlVariates = controls;
  StatSema++;
  NoPrecCount ++;

  MeasureStat->kindOfMeasure=OTHER_MEAS;
  MeasureStat->StopSimulation = false;
  MeasureStat->StopSendedToBackend = false;
  MeasureStat->NoPrecision = true;
  MeasureStat->Epsilon = 1.0 - eps_max;
  MeasureStat->StatStatus = STEADY_STATE;
  MeasureStat->NoAnalysedSamples = 1;
  MeasureStat->SampleCount = 1;
  MeasureStat->SendFactor =1;
  MeasureStat->SequenceCount = 1;
  MeasureStat->BatchingCount = 0;
  MeasureStat->AveragedMean = 0.0;
  MeasureStat->Least_Sq_Mean = 0.0;
  MeasureStat->SampSeq_first = NULL;
  MeasureStat->SampSeq_last = NULL;
  MeasureStat->CVList_first = NULL;
  MeasureStat->CVList_last = NULL;

  i = MeasureStat->ControlVariates;
  while (i--)
    {
      TempSeqPointer = (ContrSeqList *) er_malloc(sizeof(ContrSeqList));
      TempSeqPointer->Next = MeasureStat->CVList_first;
      TempSeqPointer->Seq_first = NULL;
      TempSeqPointer->Seq_last = NULL;
      MeasureStat->CVList_first = TempSeqPointer;
    }
  paw_stattotal++;
  help_measure_names = measure_names;
  counter = 0;
  while( counter < index )
    {
      help_measure_names = help_measure_names->next;
      counter++;
    }
  if (help_measure_names->name!=NULL)
   strcpy(MeasureStat->name,help_measure_names->name);
  else strcpy(MeasureStat->name,"NoName");
 
  return (MeasureStat);
} /*RewardMeasure_Init*/


/*************************************************************************

  function:	ListConvert

  parameters:	First -> Pointer to Pointer to first node of linked list
                SampSeq
		Last -> Pointer to Pointer to last node of linked list
                SampSeq
		Size          -> BatchSize (b)
		SequenceCount -> Number of sequences (s)

  purpose:	converts the SampSeq from s * b samples X1(m),X2(m),.. 
		into b batch means Y1(s*m),Y2(s*m),.. .

  author:	Christian Kelling
		Technical University of Berlin

  date:		May 1993

*************************************************************************/

static void ListConvert(First, Last, FinalSize, SequenceCount)
SampSeq **First, **Last;
int FinalSize, SequenceCount;
{
  SampSeq *TempSeq, *OldSeq, *NewSeq,*hilfSeq;
  SampSeq  *NewList_first, *NewList_last;
  double sum;
  int i, r,hilp;

  NewList_first = NULL;
 
  TempSeq = *First;
  for (i=1; i<= FinalSize; i++)
    {
      sum = 0.0;
      for (r=1; r<=SequenceCount; r++)
	{
	  sum += TempSeq->Value;
	  OldSeq = TempSeq;
/*test printouts */
/*	  printf ("old %d=%f\n", r, TempSeq->Value);*/
	  TempSeq = TempSeq->Next;
 	  free(OldSeq);    /* to do ||| einmal free einsparen */
	}
      NewSeq = (SampSeq *) er_malloc(sizeof(SampSeq));
      NewSeq->Value = sum / SequenceCount;
/*test printouts */
/*      printf ("new =%f\n", NewSeq->Value);*/
      NewSeq->Next = NULL;
      if (NewList_first == NULL)
	NewList_first = NewSeq;
      else
	NewList_last->Next = NewSeq;
      NewList_last = NewSeq;
    }
  *First = NewList_first;
  TempSeq = *First;
  *Last = NewList_last;
} /* BatchMeans_ListConvert */


/*************************************************************************

  function:	CVEstimation

  parameters:	MeasureStat -> Pointer the a reward measure statistic
                sample -> new sampled observation
  
  returns:	true  -> the required precision has been reached.
		
false -> the required precision has not been reached; too short
			 simulation run is assumed.

  purpose:	calculates estimates and tests their precision until a required
		precision is reached.
		The boolean result will be assigned to the CvR_Stat-flag
		StopSimulation too.

  author:	Christian Kelling
		Technical University of Berlin

  date:		May 1993

*************************************************************************/

static int CVEstimation (MeasureStat, sample)
CvR_Stat *MeasureStat;
double sample;
{
  static double **A, **X, **Xt, **help_matrix;
  static double *B, *beta;
  ContrSeqList *CVList, *TempSeq;
  SampSeq  *CVSeq, *MeasureSeq,*hilfSeq;
  double sigma1, sigma2, sigma, variance, help, mean, mean_ls;
  double **dmatrix(), *dvector();
  double alpha = ALPHAT_DEF/2.0;
  int hilp, i,j,k, dim1, dim2, i_controls;

/* tests ||| */
  double test_sigma1   = 0.0;
  double test_cv_corr = 0.0;
  double var_y = 0.0;
  
  if (((MeasureStat->SampleCount-1) % (/*TWO_TIMES **/ CV_MIN_SAMPLES)) == 0)
    {  /* 40 new samples have been collect, the sum is now 80 samples,
	 batch 80 samples to 40 */
/* test printouts */
     /*  puts("batching");
      printf("bei %d samples\n", MeasureStat->SampleCount); */
      MeasureStat->BatchingCount ++;
      
      i_controls = MeasureStat->ControlVariates;
      TempSeq = MeasureStat->CVList_first;
      while (i_controls --)
	{
          ListConvert(&(TempSeq->Seq_first), 
		      &(TempSeq->Seq_last),
		      CV_MIN_SAMPLES,TWO_TIMES);
	  TempSeq =  TempSeq->Next ;
	}
      ListConvert(&(MeasureStat->SampSeq_first), 
		  &(MeasureStat->SampSeq_last),
		  CV_MIN_SAMPLES,TWO_TIMES);
      MeasureStat->NoAnalysedSamples = CV_MIN_SAMPLES;
    }
  else 
    MeasureStat->NoAnalysedSamples++;
  
  dim1 = MeasureStat->ControlVariates+1;
  dim2 = MeasureStat->NoAnalysedSamples;
/* test printouts */
/*  printf ("samples to analyze: %d\n", dim2);  */

  A = dmatrix (1,dim1,1,dim1);
  B = dvector(1,dim1);
  Xt = dmatrix (1,dim2,1,dim1);

  MeasureSeq = MeasureStat->SampSeq_first;
  CVList = MeasureStat->CVList_first;
  
/* initilize the matrices */

  for (i=1;i<=dim1;i++)
    {
      B[i] = 0.0;  /* initialize the right hand side vector B */
      for (j=1;j<=dim1;j++)
	A[i][j] = 0.0;  /* initialize the A matrix */
      if (i > 1)      /* initialize the current position pointer of the */
	{               /* CV list to the first element*/
	  CVList->CurrentPosition = CVList->Seq_first;
	  CVList = CVList->Next;
	}
    }
  for (i=1;i<=dim2;i++)
    Xt[i][1] = 1.0;
  A[1][1] = MeasureStat->NoAnalysedSamples;
  sigma1 = 0.0;
  
/* calculate all sums used in the LGS */

  for (k=1;k<=dim2;k++)
    {
      CVList = MeasureStat->CVList_first;
      if  (CVList != NULL)
	CVSeq = CVList->CurrentPosition;
      j = 2;
      while (CVList != NULL)
	{  /* compose the first term of the elements of the upper right 
	     triangular A-matrix see Goehler, p.113 */
	  CVSeq = CVList->CurrentPosition;
	  A[1][j] += CVSeq->Value;
	  Xt[k][j] = CVSeq->Value;

/* tests */
/*	  test_sigma1 += (MeasureSeq->Value)*(MeasureSeq->Value);
	  test_cv_corr += CVSeq->Value*CVSeq->Value;
	  var_y += 
*/
	  j ++;
	  CVList = CVList->Next;  /* next control variate */
	}
      B[1] += MeasureSeq->Value;

/* test printouts ||| */
      
/*      if (k > KBO)*/
/*	printf ("sample %d = %f, B[1]= %f\n", k, MeasureSeq->Value, B[1]);*/

      sigma1 += MeasureSeq->Value*MeasureSeq->Value;
      for (i=2;i<=dim1;i++)
	{
	  CVList = MeasureStat->CVList_first;
	  for(j=1;j<(i-1);j++)
	    CVList = CVList->Next;  /* first element is element 
	                               with number of column */
	  CVSeq = CVList->CurrentPosition;  /* start with element k*/
	  CVList->CurrentPosition = CVSeq->Next;  /* new position,
					     dont need old one longer */
	  help = CVSeq->Value;  /* factor for sum elements of whole row */
/* test printouts */
/*	  printf("help: %f\n", help);*/
	  
	  /* compose the right hand side B-vector see Goehler, p.113 */
	  B[i] += help*MeasureSeq->Value; 
	  /* prepare the computation of the variance, sigma1 is the first term
	     and denotes the (sum)Y[k]^2 */
	  A[i][i] += help*help; /* diagonal element is simple square */
	  j = i+1;
	  CVList = CVList->Next;
	  while (CVList != NULL)
	  /* over columns, get the 2nd factor of sum 
	   elements of the upper right triangular A-matrix see Goehler, p.113*/
	    { 
	      CVSeq = CVList->CurrentPosition;
	      A[i][j] += help*CVSeq->Value;
	      j ++;
	      CVList = CVList->Next;
	    }
	}
      MeasureSeq = MeasureSeq->Next;
    }

/* test printouts */
/*      printf ("test_sigma1 = %f\n",test_sigma1);
      printf ("test_cv_corr = %f\n",39/test_cv_corr);
      printf ("var_y = %f\n",var_y/39); */

  mean = B[1]/dim2; /* first estimation of the mean (averaged batch means) */
/* test printouts */
/*  printf ("mean over samples = %f, NoAnalysedSamples: %d\n",mean, dim2);*/

  for (i=1;i<=dim1;i++)
    for (j=1;j<=dim1;j++)
      if (i != j)  /* copy the the elements of the upper right
                  triangular A-matrix to the appropriate elements of the lower
		  left triangular A-matrix (off diagonal elements only) */
	A[j][i] = A[i][j];

/* test printouts */
/*  for(j=1;j<=MeasureStat->ControlVariates+1;j++)
    {
      for(i=1;i<=MeasureStat->ControlVariates+1;i++)
	printf("A[%d,%d] = %f\t", j,i,A[j][i]);
      printf("\n");
    }
  printf("\n"); 
  for(i=1;i<=MeasureStat->ControlVariates+1;i++)
    printf("B[%d] = %f\t",i,B[i]);
  printf("\n");*/



/* debug 
  printf("vor solve\n");*/
  solve_LES (A, B, dim1);
/* debug 
  printf("nach solve\n");*/
    /* solve the set of equations Ax=B, the solution vector x will be 
       returned in B , the matrix A will be destroyed */
  
  mean_ls = B[1]; /* least square estimation of the mean */
  if (mean_ls <= 0.0)
    {
      printf ("ERROR: least square fit gives no positive mean value\n");
      printf ("mean is: %f, averaged mean is: %f\n", mean_ls, mean);
      exit (FAIL);
    }
  CVList = MeasureStat->CVList_first;
  for (j=1;j<dim1;j++)
    {
      CVList->CurrentPosition = CVList->Seq_first;
      CVList = CVList->Next;
    }

  sigma2 = 0.0;
  for (i=1;i<=dim2;i++)  /* over all CVs and samples */
    {  /* compute sigma2 as the second sum of the variance calculation  */
      help = 0.0;
      CVList = MeasureStat->CVList_first;  
        /* list starts each time at first CV */
      j=2;
      while (CVList != NULL)
	{  /* compute the vector product beta*(cv-my) */
	  CVSeq = CVList->CurrentPosition;  
	  	/* get the current position according to i */
	  help += B[j]*CVSeq->Value;

/* test printouts ||| */
/*	  printf ("vector products %f\n", B[j]*CVSeq->Value);*/

	  j ++;
	  CVList->CurrentPosition = CVSeq->Next;  
	  	/* prepare current position for next i */
	  CVList = CVList->Next;
	}
      sigma2 += (mean_ls + help)*(mean_ls + help);
    }
  sigma = (sigma1-sigma2)/(MeasureStat->NoAnalysedSamples
			   -MeasureStat->ControlVariates-1);

/* test printouts */
/*  printf ("real sigma2 = %f\n", sigma2);
  help = 0.0; mean = 0.0;
  for(j=1;j<=MeasureStat->NoAnalysedSamples;j++)
    {
      for(i=2;i<=MeasureStat->ControlVariates+1;i++)
	{
	  help += (B[1]+B[2]*Xt[j][i])*(B[1]+B[2]*Xt[j][i]);
	  printf("test products %f\n", B[2]*Xt[j][i]);
	  mean += Xt[j][i]*Xt[j][i];
	printf("Xt[%d,%d] = %f\t", j,i,Xt[j][i]);
	}
      printf("\n");
    }
  printf ("test_sigma2 = %f\n", help);*/

/* debug 
  printf("vor transpose \n");*/
  X = transpose (Xt,dim2,dim1);
/* debug 
  printf("transpose ok\n");*/
  free_dmatrix (A,1,dim1,1,dim1);
  help_matrix = mult_matr (X, Xt, dim1, dim2, dim1);
/* debug 
  printf("mult ok\n");*/
  A = inverse (help_matrix, dim1);
/* debug 
  printf("inverse ok\n");*/
  free_dmatrix (Xt, 1,dim2, 1,dim1);
  free_dmatrix (help_matrix, 1, dim1, 1, dim1);
  free_dmatrix (X, 1,dim1, 1,dim2);

/* test printouts */
/*  for (j=1;j<=MeasureStat->ControlVariates+1;j++)
      printf("B[%d] = %f\n", j, B[j]);*/
  
  free_dvector (B,1,dim1); 
/* test printouts */
/*  printf("transpose and inverse mit dimension %d und %d\n", dim1, dim2);*/
  variance = A[1][1]*sigma;
/* test printouts */
/*  printf("variance = %1.16f\n", variance);*/

  if (fabs(variance) < 1.0e-15) variance = 0.0;
  if (variance < 0.0)
    {
      printf ("ERROR: variance is no positive value\n");
      printf ("A[1][1] is: %f, sigma is: %1.15f\n", A[1][1], sigma);
      exit (FAIL);
    }
  free_dmatrix (A, 1,dim1,1, dim1);

  MeasureStat->Epsilon = 
    student_distr((MeasureStat->NoAnalysedSamples
		   -MeasureStat->ControlVariates-1),1.0-alpha);
/* test printouts */
/*  printf("eps1 = %f\n", MeasureStat->Epsilon);*/
  MeasureStat->Epsilon *= (sqrt(variance) / mean_ls);
/* test printouts */
/* force no ssamples = 10000 */
/* if (MeasureStat->SampleCount >= 10001)*/
  MeasureStat->AveragedMean = mean;
  MeasureStat->Least_Sq_Mean = mean_ls;
  MeasureStat->Variance = variance;
  MeasureStat->DeltaX = 
    student_distr((MeasureStat->NoAnalysedSamples - 
		   MeasureStat->ControlVariates-1), 
		   1.0-alpha) * sqrt(variance);
  if((MeasureStat->Epsilon < eps_max) && min_flag &&
     (((MeasureStat->kindOfMeasure==PROB_MEAS) &&   /* it's a probabilty */
             (uniPdfAreaErr(MeasureStat->name,
                            CVR_Lower_out(MeasureStat),
                            CVR_Upper_out(MeasureStat),
                            STEPS,
                            CVR_Mean_out(MeasureStat),
                            MeasureStat->NoAnalysedSamples*variance/MeasureStat->EndTime,
                            ALPHAT_DEF) <= errPercent)
            ) ||
            MeasureStat->kindOfMeasure==OTHER_MEAS      /* is not a p. */
           )
          )
    {
      MeasureStat->StopSimulation = true;
      MeasureStat->NoPrecision = false;
    }
  return(MeasureStat->StopSimulation);
} /* CVEstimation */


/*************************************************************************

  function:	CVR_Analysis

  parameters:	MeasureStat -> Pointer the a reward measure statistic
		sample -> new sampled observation

  returns:	true  -> The reward measure statistic has been reached the 
                         required precision. This marks the maximum length
			 of simulation run. 
		false -> The statistic has an unacceptable precision.

  purpose:	performs the main function of the batch means analysis using 
                 control variates.
		After determining the final batch size it estimates the 
		cotrolled variance until a defined precision is reached.
		The return value true means the statistcs have get the
		required precision. At this point the simulation run could be
		terminated by the calling function or an upper function level.


  author:	Christian Kelling
		Technical University of Berlin

  date:		May 1993

*************************************************************************/

bool CVR_Analysis (MeasureStat, sample, cv, offset)
CvR_Stat *MeasureStat;
int offset;
double sample, *cv; /*cv is a pointer to the first sample in a sequ. of cvs,
		      the actual sample is determined by cv[offset],
		      the single cv-samples are spaced with 
		      STAT_BLOCK_DEF */
{
  SampSeq *TempSeq;
  ContrSeqList *CVList;
  int i,hilp;

  if (MeasureStat == NULL)
    {
      printf("Statistic Analysis: Null-Pointer for statistic\n");
      exit(FAIL);
    }
  if (MeasureStat->StopSimulation == true)
      return(true);

  TempSeq = (SampSeq *) er_malloc(sizeof(SampSeq));
  TempSeq->Value = sample;
  TempSeq->Next = NULL;
  if (MeasureStat->SampSeq_first == NULL)
    MeasureStat->SampSeq_first = TempSeq;
  else
    MeasureStat->SampSeq_last->Next = TempSeq;
  MeasureStat->SampSeq_last = TempSeq;
  MeasureStat->SampleCount++;

  CVList = MeasureStat->CVList_first;
        


  for (i=0; i<MeasureStat->ControlVariates; i++)
    {
        TempSeq = (SampSeq*) er_malloc (sizeof(SampSeq));
      if (CVList == NULL)
	{
	  printf("ERROR: Insufficent number of control variates defined\n");
	  exit (FAIL);
	}
      if (CVList->Seq_first == NULL)
        CVList->Seq_first = TempSeq;
      else
	CVList->Seq_last->Next = TempSeq;
      CVList->Seq_last = TempSeq;
      TempSeq->Value = cv[offset+i*STAT_BLOCK_DEF];		/*(BM)*/
      TempSeq->Next = NULL;
      CVList = CVList->Next;
    }
  
  if (MeasureStat->SampleCount <= CV_START_SAMPLES)
    return (false);
  if (MeasureStat->SampleCount == CV_START_SAMPLES+1)
    MeasureStat->NoAnalysedSamples = CV_START_SAMPLES-1;
/*  if (!min_flag)
    printf (" no min flag \n");*/
  if (CVEstimation (MeasureStat, sample))
    {
      MeasureStat->SampleCount -= 2; /* to correct the sample counter */
      StatSema--;
      NoPrecCount--;
    }
  if (MeasureStat->StopSimulation != true)
    {
      if (MeasureStat->SampleCount > CVR_n_max)
	{
	  printf(
		 "\nNOTE: A measure has exceeded the maximum number of samples\n");
	  printf(
		 "      Simulation stops without reaching the predefined accuracy!\n");
	  MeasureStat->StopSimulation = true;
	  MeasureStat->NoPrecision = true;
	  StatSema--;
	}
      if (time_exceeded)
	{
          printf(
		 "\nNOTE: The simulation has exceeded the specified model time limit.\n");
          printf(
		 "      Simulation stops without reaching the predefined accuracy!\n");
          MeasureStat->StopSimulation = true;
          MeasureStat->NoPrecision = true;
	  StatSema--;
        }
    }
  return (MeasureStat->StopSimulation);
} /* CVR_Analysis */


