// File     Parspace.h      (C++ class definition and inline implementation)
// Program  msolve          (Stationary Analysis Experiments)
/**
 *@author  Philipp Berndt
 *@version $Id: Parspace.h,v 1.2 2000/01/12 16:02:22 pumphutt Exp $
 */

//#define DEBUG
#ifndef PARSPACE_H
#define PARSPACE_H

#include <iostream>
#include <vector>
#include <list>
#include <algorithm>
#include <functional>
#include <cassert>
#include <limits.h>

/**
 * Administration of (one-dimensional) parameter space for experiments.
 *
 *@author  Philipp Berndt
 *@version $Id: Parspace.h,v 1.2 2000/01/12 16:02:22 pumphutt Exp $
 */
class Parspace
{
    struct Par;
    struct Interval;
public:
    enum stat { UNDONE, INPROGRESS, DONE };
  
    /**
     * Returns a hint about what parameter index to use for the next analysis.
     * The returned index is chosen using the following algorithm:
     * <pre>
     * IF     possible take the beginning of the biggest UNDONE interval,
     * ELSEIF possible take the       end of the biggest UNDONE interval
     * ELSEIF possible take the   middle  of the biggest UNDONE interval
     * ELSE signal END
     * </pre>
     *@see getPar
     *@see setStatus
     */
    int getNext() const;

    /** Sets the status of parameter with given index. */
    void setStatus(int index, stat newstatus);

    /** Returns the parameter value for given parameter index. */
    double getPar(int) const;

    /** Adds parameter value to this parspace. */
    void addPar(double p);

    /**
     * Prints out the stati of this parspace.
     *
     * The symbols have the following meaning: <pre>
     * - UNDONE
     * * INPROGRESS
     * X DONE
     * </pre>
     */ 
    void print(ostream &s) const;

    /** Returns index of parameter with status DONE which value is nearest
     *  to the value of par with given index. */
    int nearest(int n) const;

    /** Returns true iff all parameters have status DONE. */
    bool done() const;

private:
    vector<Par>    pars;
    list<Interval> ints;
};

/** One parameter with value and status. */
struct Parspace::Par
{
    Par(double _par) : par(_par), status(UNDONE) {}
    double par;
    stat status;
};

/** One Intervall of not DONE parameters. */  
struct Parspace::Interval
{
    Interval(int s, int e) : start(s), end(e) {}
    int start;
    int end;
    static bool start_clear(const Interval &I);
    static bool end_clear(const Interval &I);
    int getRange() const
	{ return end - start +1 ; }

    bool operator< (const Interval& I)
	{ return getRange() > I.getRange(); } 
    void print(ostream &os) const;
};

/** Attribute: Has the first element of Interval status UNDONE ? */
struct start_clear 
    : public binary_function<Parspace::Interval, vector<Parspace::Par>, bool>
{
    bool operator() (const Parspace::Interval& I,
		     const vector<Parspace::Par>& pars) const
	{ return pars[I.start].status == Parspace::UNDONE; }
};

/** Attribute: Has the last element of Interval status UNDONE ? */
struct end_clear 
    : public binary_function<Parspace::Interval, vector<Parspace::Par>, bool>
{
    bool operator() (const Parspace::Interval& I,
		     const vector<Parspace::Par>& pars) const
	{ return pars[I.end].status == Parspace::UNDONE; }
};

/** Attribute: Is the index contained in Interval ? */
struct contains : public binary_function<Parspace::Interval, int, bool>
{
    bool operator() (const Parspace::Interval& I, int n) const
	{ return I.start <= n && n <= I.end; }
};


/*****************************************************************************\
*                         INLINE FUNCTION DEFINITIONS                         *
*               No classes or user interfaces below this line !               *
\*****************************************************************************/

inline ostream& operator<<(ostream& s, const Parspace& ps)
{ ps.print(s); return s; }

inline ostream& operator<<(ostream& s, const Parspace::Interval& I)
{ I.print(s); return s; }

inline void Parspace::Interval::print(ostream &os) const
{
    os << "[" << start << "; " << end << "] ";     
}

inline bool Parspace::done() const
{ return ints.empty(); }

inline int Parspace::getNext() const
{
    list<Interval>::const_iterator i;
    int r;

    if(ints.empty())
	return -2;

    if((i = find_if(ints.begin(), ints.end(),
		    bind2nd(start_clear(), pars) )) != ints.end())
	return i->start;
  
    if((i = find_if(ints.begin(), ints.end(),
		    bind2nd(end_clear(), pars) )) != ints.end())
	return i->end;
  
    if((i = ints.begin()) != ints.end())
	if((r = i->getRange()) > 2)
	    return i->start + r/2;

    return -1;
}

inline void Parspace::print(ostream &s) const
{
    typedef vector<Par>::const_iterator it;
    for(it i = pars.begin(); i != pars.end(); ++i)
	{
	char c = '?';
	switch(i->status)
	    {
	    case UNDONE:
		c = '-';
		break;
	    case INPROGRESS:
		c = '*';
		break;
	    case DONE:
		c = 'X';
	    }
	cout << c;
	}

#ifdef DEBUG
    typedef list<Interval>::const_iterator lit;
    for(lit i = ints.begin(); i != ints.end(); ++i)
	i.print(cout); cout << endl;
#endif
}

inline int Parspace::nearest(int n) const
{
    double npar = pars[n].par;
    int imin = -1;
    double dmin = DBL_MAX;
    for(int i = 0; i < pars.size(); ++i)
	{
	if(pars[i].status != DONE)
	    continue;

	double d = fabs(pars[i].par - npar);
	if(d < dmin)
	    {
	    imin = i;
	    dmin = d;
	    }
	}
    return imin;
}

inline void Parspace::setStatus(int n, stat newstat)
{
    typedef list<Interval>::iterator it;
    int oldstat = pars[n].status;

    if(newstat == oldstat)
	return;

    pars[n].status = newstat;

    switch(newstat)
	{
	case UNDONE:
	    {
	    assert(oldstat == INPROGRESS);

	    it i1, i2;
	    i2 = i1 = find_if(ints.begin(), ints.end(),
			      bind2nd(contains(), n));
	         i2 = find_if(++i2,         ints.end(),
			      bind2nd(contains(), n));

	    if(i2 != ints.end())
		{
		cout << *i1 << " - " << *i2 << endl;
	  
		if(i2->end <= i1->start)
		    {
		    it tmp = i1;
		    i1 = i2;
		    i2 = tmp;
		    }
	  
		cout << *i1 << " - " << *i2 << endl;
	  
		assert(i1->end == n && i2->start == n);
	  
		i1->end = i2->end;
		ints.erase(i2);
		ints.sort();
		}
	    break;

	    }

	case INPROGRESS:
	    {
	    assert(oldstat == UNDONE);
	    it i = find_if(ints.begin(), ints.end(), bind2nd(contains(), n));
	    if(n != i->start && n != i->end)
		{
		Interval I2(n, i->end);
		i->end = n;
		ints.push_back(I2);
		ints.sort();
		}
	    break;
	    }

	case DONE:
	    {
	    assert(oldstat == INPROGRESS);

	    for(it i = ints.begin(); i != ints.end(); ++i)
		{
		if(i->start == n && i->end == n)
		    {
		    it i2 = i--;
		    ints.erase(i2);
		    }
		else if(i->end == n)
		    i->end--;
		else if(i->start == n)
		    i->start++;
		}
	    ints.sort();
	    }
	}
}

inline double Parspace::getPar(int n) const
{ return pars[n].par; }

inline void Parspace::addPar(double p)
{
    int n = pars.size();
    pars.push_back(Par(p));

    if(ints.empty())
	ints.push_back(Interval(n,n));
    else
	{
	assert(ints.size() == 1);
	Interval &I=ints.front();
	assert(I.end == n-1);
	I.end++;
	}
}

#endif // not PARSPACE_H
