/******************************************************************************
Programm	:	erand.c

Parameter	:	argv[1]: Host, on which the execute_erand-Process
				 is running
			argv[2]: socket portnumber of execute_erand-Process
			argv[3]: desired precision of calculation
			argv[4]: name of the DSPN
			argv[5]: desired bits for arbitrary precision
			argv[6]: desired truncation error of calculation

Purpose		: 	Implementation of the randomization algorithm.

Project Director:	Christoph Lindemann TU Berlin

Programmer	:	Enrik Baumann, Tobias Bading

History		:	Release 1: Enrik Baumann, October 1991
			Release 2: Tobias Bading, September 1992
			Release 3: Ulrich Hertlein, November 1993
                        Release 4: Patrick Drescher, June 1995
                        Release 5: Oliver Hesse, June 1999
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <time.h>

#include <string.h>
#include <math.h>

#include <signal.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/resource.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#include "portab.h"
#include "erand.h"
#include "matrices_Q_delta.h"
#include "matrices_P_C.h"
#include "parsdfunc.h"

#include "solve_factors.h"      /* calculate factors with arbitrary prec. */
#include "TimeWrapper.h"        /* time and memory usage */   

#define True    1
#define False   0

#define MAXREAL	1.0e35
#define MINREAL 1.0e-100


// #define SHOW_SPECIAL_FORMAT  /* show time and mem. usage in spec. format */
/* #define CONSTHASHTABLE */

#define INITIAL          2
#define COMPETITIVE      4
#define CONCURRENT       8

#ifdef CONSTHASHTABLE
#define HASHTABLESIZE    1024
#endif

#define BUFFERENLARGEMENT 1024


#define OUT_OF_MEMORY {\
   fprintf(stderr, "Error (%s,%u): Out of memory\n", __FILE__,__LINE__);\
   my_exit(-1);\
}

#define SOCKET_READ_ERROR {\
   fprintf(stderr,"Error (%s,%u): Can't read socket\n", __FILE__,__LINE__);\
   my_exit(-1);\
}

#define FILE_OPEN_ERROR(fn) {\
   fprintf(stderr,"Error (%s,%u): Can't open file '%s'\n", __FILE__,__LINE__,fn);\
   my_exit(-1);\
}

#define FILE_WRITE_ERROR(fn) {\
   fprintf(stderr,"Error (%s,%u): Can't write file '%s' !\n", __FILE__,__LINE__,fn); \
   my_exit(-1);\
}


#ifdef ultrix
static int isinf(x)
double x;
{
   return (((x) == HUGE_VAL) || ((x) == -HUGE_VAL));
}
#endif


/*
 * Type definition
 */
typedef struct state_to_ind_entry {
  unsigned state;
  unsigned index;
  struct state_to_ind_entry *next;
} state_to_ind_entry;

typedef struct list_entry {
  double value;
  unsigned index;
  struct list_entry *next;
} list_entry;


/*
 * Function prototyping
 */
void my_exit();
static void sigill_handler(), sigfpe_handler(), sigbus_handler(),
     sigsegv_handler(), sigpipe_handler();

static void init();
static unsigned read_info();
static void read_matrices();

static list_entry *new_item();
static void delete_vector();

static list_entry *add_to_vector();
void add_to_matrix();

unsigned hash();
unsigned get_index();
void put_index();
void enlarge_ind_to_state();

static void calculate_A_from_Q ();
static list_entry *calculate_phi_k ();
static void trunc_random_gen();

double epsilon;	         /* precision */
unsigned digits;         /* number of digits to display */


/*
 * Local variables
 */
static unsigned MC_number;    /* Number of local Markov Chain */
static unsigned toptan;       /* Number of elements in Mdet() */
static unsigned exptoptan = 0;

static list_entry **matrix_Q; /* Generator matrix Q[] */
static list_entry **matrix_A; /* Normalized generator matrix Q[] */
static list_entry **matrix_delta;

static unsigned eta_k;   /* Anzahl der Eintraege der Generatormatrix Q<k>
                            (ohne Diagonalelemente) */
static double max_qii;   /* Absolutwert des kleinsten Diagonaleintrages der
                            Generatormatrix Q<k> */
static double delta;	      /* 1.0 / (1.02 * max_qii) */
static unsigned *ind_to_state;/* Array, welches neue Zustandsnummern auf alte
			         Zustandsnummern abgebildet */
static unsigned char *marking_flags;
     /* Flags zu jedem andauernden Zustand, z.Z.:
	Bit 0: ungenutzt
	Bit 1: Initialzustand
	Bit 2: Es existieren konkurierende exp. Transitionen
	Bit 3: Es existieren verdraengende exp. Transitionen
	Momentan werden zwar alle Flags per Socket uebertragen,
	aber nur das Flag INITIAL verwendet. */
static state_to_ind_entry **state_to_ind;
     /* Array fuer's Hashing. Damit werden alte
	Zustandsnummern auf neue Zustandsnummern
	abgebildet */
static unsigned buffersize;	/* aktuelle Groesse von ind_to_state[] */
static int socket_fd;
     /* Filedescriptor fuer die Socket-Verbindung zum Server */
static FILE *socket_fp;		/* Filepointer fuer die Socket */

static char *server_host, *server_port, *filename;  /* from argv */

static char *conv_filename,*emc_filename,*probmark_filename;
static FILE *conv_fp, *emc_fp;	/* Filepointer fuer obige Files */

static time_t starttime;

static char *transition_name = NULL;
static time_l *expoly;


/*
 * Private exit routine
 */
void my_exit (exit_code)
int exit_code;
{
  write (socket_fd, "ready.\n", 7);     /* To tell the server we're ready */
  exit (exit_code);
}

/*
 * Signal handlers
 */
static void sigill_handler()
{
  fprintf(stderr, "Error in %s: Illegal Instruction\n", __FILE__);
  my_exit(-1);
}

static void sigfpe_handler()
{
  fprintf(stderr, "Error in %s: Floatingpoint exception\n", __FILE__);
  my_exit(-1);
}

static void sigbus_handler()
{
  fprintf(stderr, "Error in %s: Bus error\n", __FILE__);
  my_exit(-1);
}

static void sigsegv_handler()
{
  fprintf(stderr, "Error in %s: Segmentation fault\n", __FILE__);
  my_exit(-1);
}

static void sigpipe_handler()
{
  exit(-1);    /* lost server connection, so no `ready' message */
}


/*
 * Memory management
 */
#define MEM_BUFFERS      1000
#define MEM_ITEMS        1000

static list_entry *first_free_item;

/******************************************************************************
 Funktion:	new_item

 Parameter:	---

 Rueckgabewert:	Zeiger auf eine list_entry-Struktur

 Zweck:		Erzeugen eines neuen Elementes eines Vektors oder
		einer Matrix.
******************************************************************************/
static list_entry *new_item ()
{
  static list_entry **buffers;
  static unsigned number_of_buffers, last_used_buffer;

  if (!first_free_item)
  {
    number_of_buffers = MEM_BUFFERS;
    if (!(buffers = (list_entry **)malloc (MEM_BUFFERS * sizeof (list_entry *))))
      OUT_OF_MEMORY;
    if (!(buffers[0] = (list_entry *)calloc (MEM_ITEMS, sizeof (list_entry))))
      OUT_OF_MEMORY;
    last_used_buffer = 0;
    first_free_item = buffers[0] + 1;
    return buffers[0];
  }
  {
    register list_entry *returnvalue = first_free_item;

    if (first_free_item->next)
      first_free_item = first_free_item->next;
    else
      if (++first_free_item - buffers[last_used_buffer] == MEM_ITEMS)
      {
	if (++last_used_buffer == number_of_buffers)
	{
	  number_of_buffers += MEM_BUFFERS;
	  if (!(buffers = (list_entry **)realloc (buffers, number_of_buffers * sizeof (list_entry *))))
	    OUT_OF_MEMORY;
	}
	if (!(buffers[last_used_buffer] = (list_entry *)calloc (MEM_ITEMS, sizeof (list_entry))))
	  OUT_OF_MEMORY;
	first_free_item = buffers[last_used_buffer];
      }
    return returnvalue;
  }
}


#define free_item(item) { \
  item->next = first_free_item; \
  first_free_item = item; \
}


/******************************************************************************
 Funktion:	add_to_matrix

 Parameter:	matrix, zeilenindex, spaltenindex, wert

 Rueckgabewert:	---

 Zweck:		Fuegt einen Wert an der Position (zeilenindex, spaltenindex)
		in eine Matrix ein. An dieser Position darf noch kein Wert
		gespeichert sein.
******************************************************************************/
void add_to_matrix (matrix, zeilenindex, spaltenindex, wert)
  list_entry **matrix;
  unsigned zeilenindex, spaltenindex;
  double wert;
{
  register list_entry	*eintrag = matrix[zeilenindex],	*vorheriger_eintrag = NULL;

  while (eintrag)
    if (eintrag->index < spaltenindex)
    {
      vorheriger_eintrag = eintrag;
      eintrag = eintrag->next;
    }
    else
      if (eintrag->index == spaltenindex)
      {
	printf ("Error: matrix entry at %u,%u already exists!\n",
		zeilenindex, spaltenindex);

#ifdef ACCOUNT
	eintrag->value += wert;

	return;
#else
	my_exit (1);
#endif
      }
      else
	break;

  {
    register list_entry *neuer_eintrag = new_item ();

    neuer_eintrag->index = spaltenindex;
    neuer_eintrag->value = wert;
    neuer_eintrag->next = eintrag;
    if (vorheriger_eintrag)
      vorheriger_eintrag->next = neuer_eintrag;
    else
      matrix[zeilenindex] = neuer_eintrag;
  }
}


/******************************************************************************
 Funktion:	hash

 Parameter:	alte Zustandsnummer

 Rueckgabewert:	Index fuer das Array state_to_ind

 Zweck:		Ermittlung oben genannten Index
******************************************************************************/
unsigned hash (state)
unsigned state;
{
#ifdef CONSTHASHTABLE
    return (state % HASHTABLESIZE);
#else
    return (state % toptan);
#endif
}


/******************************************************************************
 Funktion:	get_index

 Parameter:	alte Zustandsnummer

 Rueckgabewert:	neue Zustandsnummer (Index)

 Zweck:		Abbilden einer alten Zustandsnummer auf eine
		neue Zustandsnummer.
******************************************************************************/
unsigned get_index (state)
unsigned state;
{
  register state_to_ind_entry *entry;

  for (entry = state_to_ind[hash (state)]; entry; entry = entry->next)
    if (entry->state == state)
      return (entry->index);

  return (0);
}


/******************************************************************************
 Funktion:	put_index

 Parameter:	index:	Neue Zustandsnummer
		state:	Alte Zustandsnummer

 Rueckgabewert:	---

 Zweck:		Ordnet der alten Zustandsnummer eine neue Zustandsnummer
		(Index) zu und traegt die entsprechenden Werte in der
		Hashtabelle 'state_to_ind' ein.
*******************************************************************************/
void put_index (index, state)
unsigned index, state;
{
  unsigned pos;
  state_to_ind_entry *entry;

  pos = hash (state);
  if (!(entry = (state_to_ind_entry *) malloc (sizeof (state_to_ind_entry))))
    OUT_OF_MEMORY;
  if (state_to_ind[pos])
    entry->next = state_to_ind[pos];
  else
    entry->next = NULL;
  state_to_ind[pos] = entry;
  entry->state = state;
  entry->index = index;
}


/******************************************************************************
 Funktion:	enlarge_ind_to_state

 Parameter:	---

 Rueckgabewert:	---

 Zweck:		Vergroessern des Arrays 'ind_to_state'
******************************************************************************/
void enlarge_ind_to_state ()
{
  register unsigned old_buffersize = buffersize;

  buffersize += BUFFERENLARGEMENT;

  if (!(ind_to_state = (unsigned *)realloc (ind_to_state, buffersize * sizeof (unsigned))))
    OUT_OF_MEMORY;
  if (!(marking_flags = (unsigned char *)
        realloc (marking_flags, buffersize * sizeof (unsigned char))))
    OUT_OF_MEMORY;

  memset (ind_to_state + old_buffersize, 0, BUFFERENLARGEMENT * sizeof (unsigned));
}


/******************************************************************************
 Funktion:	init

 Parameter:	--

 Rueckgabewert:	---

 Zweck:		- Aufbauen der Verbindung zum Server-Prozess 'exec-rands'
		- Umlenken von stdout und stderr in die Socket
******************************************************************************/
static void init(argv)
char **argv;
{
   struct sockaddr_in server;
   register struct hostent *hp;
   unsigned tries;

   /* Verbindung zum Server (exec_rands.c) aufnehmen: */
   if( (socket_fd = socket(AF_INET, SOCK_STREAM, 0)) < 0 )
      exit( -1 );
   if( !(hp = gethostbyname(server_host)) )
      exit( -1 );

   memcpy( &server.sin_addr, hp->h_addr, hp->h_length);
   server.sin_family = AF_INET;
   server.sin_port = htons(atoi(server_port));
   for( tries = 0; tries < 60; tries++ )
      if( connect(socket_fd, (struct sockaddr *)&server, sizeof (server)) != -1 )
         break;
      else
         sleep(3);

   if( tries == 60 )
   {
      fprintf(stderr,"Abort after 60 tries\n");
      exit( -1 );
   }

   /* create file pointer for socket */
   if( !(socket_fp = fdopen (socket_fd, "r+")) )
      exit( -1 );

   /* divert stdout & stderr to socket */
   if( dup2(socket_fd, 1) < 0 )
      exit( -1 );
   if( dup2(socket_fd, 2) < 0 )
      exit( -1 );

   /* flush output immediately (no buffering) */
   setvbuf(stdout,NULL,_IOLBF,BUFSIZ);
   setvbuf(stderr,NULL,_IOLBF,BUFSIZ);

   /* install signal handlers */
   signal(SIGILL, sigill_handler);
   signal(SIGFPE, sigfpe_handler);
   signal(SIGBUS, sigbus_handler);
   signal(SIGSEGV, sigsegv_handler);
   signal(SIGPIPE, sigpipe_handler);
}

/******************************************************************************
 Funktion:	read_info

 Parameter:	---

 Rueckgabewert:	Anzahl der Elemente der Menge Mdet<k>

 Zweck:		Lesen einiger Daten des gewuenschten Loesungsverfahrens
		von der Socketverbindung zum Server:
		  - Nummer der Markoff-Kette
		  - Zustaende der Menge Mdet(MC_number)

                Ermittlung von expoly durch Aufruf von parse_pmf()
******************************************************************************/
static unsigned read_info(argv)
char **argv;
{
   unsigned index;
   unsigned name_len;

   unsigned elem = 0;
   unsigned state_num;

   /* read number of Markov-Chain */
   if (fread(&index, sizeof (int), 1, socket_fp) != 1)
      SOCKET_READ_ERROR;

#ifdef DEBUG
   printf("index = %d\n", index);
#endif

   MC_number=index;

   /* read length of transition name */
   if( fread(&name_len,sizeof(unsigned),1,socket_fp) != 1 )
     SOCKET_READ_ERROR;

   if( name_len )
     {
     char buf[256];

     if( (transition_name = (char *)malloc(name_len + 1)) == NULL )
        OUT_OF_MEMORY;

     /* read transition name */
     if( fread(transition_name,name_len,1,socket_fp) != 1 )
       SOCKET_READ_ERROR;
     transition_name[name_len] = '\0';
#ifdef DEBUG
     printf("I am transition '%s' of type general\n", transition_name);
#endif

     /* build filename for pmf */
     sprintf(buf,"%s.pmf", filename);

     /* get expoly */
     if( (expoly = parse_pmf(buf,transition_name)) == NULL )
       fprintf(stderr,"No pmf defined for transition '%s'\n", transition_name);
#ifdef DEBUG
     else
       print_expoly(expoly);
#endif
     }
   else
     {
     fprintf(stderr, "Error: name length of general transition is zero\n");
     my_exit(-1);
     }

   buffersize = BUFFERENLARGEMENT;
   if (!(ind_to_state = (unsigned *)calloc(buffersize, sizeof(unsigned))))
      OUT_OF_MEMORY;
   if(!(marking_flags = (unsigned char *)malloc(buffersize * sizeof(unsigned char))))
      OUT_OF_MEMORY;

   /* read state numbers & flags (state_num == 0 marks end-of-list) */
   do
   {
      if (fread (&state_num, sizeof (state_num), 1, socket_fp) != 1)
         SOCKET_READ_ERROR;
/* #ifdef DEBUG
      printf("state_num: %d\n",state_num);
#endif */
      /* Jeweils das hoechste Bit einer State_Num gibt an, ob es sich
         um einen Initialzustand handelt (Bit 31 gesetzt = Initialzustand) */
      if (state_num)
      {
         if (++elem == buffersize)
            enlarge_ind_to_state ();
         /* read state flag */
         if (fread (&marking_flags[elem], sizeof (char), 1, socket_fp) != 1)
            SOCKET_READ_ERROR;
/* #ifdef DEBUG
      printf("marking_flags: %d\n",marking_flags[elem]);
#endif */
         ind_to_state[elem] = state_num;
      }
   }
   while (state_num);

   return elem;
}

/******************************************************************************
 Funktion:	read_matrices

 Parameter:	argv

 Rueckgabewert:	---

 Zweck:		- Lesen von Epsilon (???)
		- Lesen der Generatormatrix Q
		- Lesen der Delta-Matrix
******************************************************************************/
static void read_matrices(argv)
char **argv;
{
  double neg_summe;
  unsigned zustandsnummer, neue_zustandsnummer, zustandsuebergaenge;
  unsigned ziel_zustandsnummer;
  double uebergangsrate;
  unsigned det_state_number;
  unsigned i, j;
  unsigned anzahl_uebergaenge;

  starttime = time (NULL);

  toptan = read_info(argv); // side effects: read expoly, filename, ...


  if (!(conv_filename = malloc (strlen (filename)+ 8))) OUT_OF_MEMORY;
  if (!(emc_filename = malloc (strlen (filename) + 7))) OUT_OF_MEMORY;

  sprintf (conv_filename, "%s.CONV%02u", filename, MC_number);
  sprintf (emc_filename, "%s.EMC%02u", filename, MC_number);

#ifdef CONSTHASHTABLE
  if (!(state_to_ind = (state_to_ind_entry **)
	calloc (HASHTABLESIZE, sizeof (state_to_ind_entry *))))
#else
  if (!(state_to_ind = (state_to_ind_entry **)
	calloc (toptan, sizeof (state_to_ind_entry *))))
#endif
    OUT_OF_MEMORY;

  /* Zustaende aus Mdet(k) bzw. Mexp in Hashtabelle eintragen: */
  for (i = 1; i <= toptan; ++i)
    put_index (i, ind_to_state[i]);

  if (!(matrix_Q = (list_entry **)calloc (toptan + 1, sizeof (list_entry *))))
    OUT_OF_MEMORY;
  if (!(matrix_delta = (list_entry **)calloc (toptan + 1, sizeof (list_entry *))))
    OUT_OF_MEMORY;

/**************************************/
/*** Generatormatrix Q<k> einlesen: ***/
/**************************************/
  eta_k = 0;
  max_qii = -42.0;

  for (det_state_number = 1; det_state_number <= toptan ; ++det_state_number)
  {
    /* Header eines Datenfeldes einlesen: */
    if (fread (&Q_row_header, SIZEOF_Q_row_header, 1, socket_fp) != 1)
      SOCKET_READ_ERROR;
    zustandsuebergaenge = Q_row_header.entries_in_row;
    zustandsnummer = Q_row_header.row_index;
    neg_summe = Q_row_header.neg_sum;

/* #ifdef DEBUG  ! fuehrt bei zu grosser Matrix irgendwann zum Stau !
    printf("index:%d no_ueb:%d neg:%f\n",zustandsnummer,zustandsuebergaenge,neg_summe);
# endif */

    eta_k += zustandsuebergaenge;

    if (det_state_number != get_index (zustandsnummer))
    {
      fprintf (stderr, "Error in %s: wrong statenumber in matrix Q<%u> !\n", __FILE__, MC_number);
      my_exit (1);
    }

    if (zustandsuebergaenge)
    {
      if (max_qii < -neg_summe)
        max_qii = -neg_summe;

      /* Diagonalelement der Generatormatrix speichern: */
      add_to_matrix (matrix_Q, det_state_number, det_state_number, neg_summe);

      /* Zustandsuebergaenge aus aktuellem Zustand (= det_state_number) bearbeiten: */
      for (j = 1; j <= zustandsuebergaenge; ++j)
      {
	if (fread (&Q_row_entry, SIZEOF_Q_row_entry, 1, socket_fp) != 1)
          SOCKET_READ_ERROR;

	ziel_zustandsnummer = Q_row_entry.column_index;

	uebergangsrate = Q_row_entry.rate;

/* #ifdef DEBUG
    printf("column:%d rate:%f\n",ziel_zustandsnummer,uebergangsrate);
# endif */

	if (!(neue_zustandsnummer = get_index (ziel_zustandsnummer)))
	{
	  /* Der erreichbare Zustand ist nicht Element der Menge Mdet(k),
	     kann aber von dort durch exp. Schalten erreicht werden. */
	  neue_zustandsnummer = toptan + ++exptoptan;
	  if (neue_zustandsnummer == buffersize)
            enlarge_ind_to_state ();
	  ind_to_state[neue_zustandsnummer] = ziel_zustandsnummer;
	  put_index (neue_zustandsnummer, ziel_zustandsnummer);
	}

	/* exponentiellen Uebergang M(det_state_number) -> M(neue_zustandsnummer) speichern: */
	if (uebergangsrate != 0.0)
	   add_to_matrix (matrix_Q, det_state_number, neue_zustandsnummer, uebergangsrate);
      }
    }
  }

  if (max_qii > 0.0)
    delta = 1.0 / (1.02 * max_qii);
  else
  {
    max_qii = 0.0;
    delta = MAXREAL;
  }

/********************************************************************/
/*** Matrix Delta<k> der Uebergangswahrscheinlichkeiten einlesen: ***/
/********************************************************************/
    for (det_state_number = 1; det_state_number <= toptan; ++det_state_number)
    {
      if (fread (&delta_row_header, SIZEOF_delta_row_header, 1, socket_fp) != 1)
        SOCKET_READ_ERROR;

      if (det_state_number != get_index (delta_row_header.row_index))
      {
        fprintf (stderr, "Error in %s: wrong rowindex in matrix delta<%u>\n", __FILE__, MC_number);
        my_exit (1);
      }

      anzahl_uebergaenge = delta_row_header.entries_in_row;

/* #ifdef DEBUG
    printf("index:%d no_ueb:%d\n",delta_row_header.row_index,anzahl_uebergaenge);
# endif */

      for (j = 0; j < anzahl_uebergaenge; ++j)
      {
        if (fread (&delta_row_entry, SIZEOF_delta_row_entry, 1, socket_fp) != 1)
	  SOCKET_READ_ERROR;
        ziel_zustandsnummer = delta_row_entry.column_index;

/* #ifdef DEBUG
    printf("column:%d prob:%f\n",ziel_zustandsnummer,delta_row_entry.probability);
# endif */

        if (!(neue_zustandsnummer = get_index (ziel_zustandsnummer)))
        {
	  /* Der erreichbare Zustand ist nicht Element der Menge Mdet(k),
	     kann aber durch Schalten der betrachteten det. Transition k erreicht werden. */
          neue_zustandsnummer = toptan + ++exptoptan;
          if (neue_zustandsnummer == buffersize)
	    enlarge_ind_to_state ();
          ind_to_state[neue_zustandsnummer] = ziel_zustandsnummer;
          put_index (neue_zustandsnummer, ziel_zustandsnummer);
	}
	if (delta_row_entry.probability != 0.0)
	  add_to_matrix (matrix_delta, det_state_number, neue_zustandsnummer,
			 delta_row_entry.probability);
      }
    }
}


/******************************************************************************
 Funktion:	calculate_A_from_Q

 Parameter:	---

 Rueckgabewert:	---

 Zweck:		Normalisierung der Generatormatrix Q
******************************************************************************/
static void calculate_A_from_Q ()
{
  register unsigned statenumber;
  register list_entry *item;
  register short diag_item;

  matrix_A = matrix_Q;
  matrix_Q = (list_entry **)0;
  for (statenumber = 1; statenumber <= toptan;  statenumber++)
  {
    diag_item = False;
    for (item = matrix_A[statenumber]; item; item = item->next)
    {
      if (item->index == statenumber)
      {
	diag_item = True;
	item->value = item->value * delta + 1.0;
      }
      else
	item->value *= delta;
    }
    if (!diag_item)
      add_to_matrix (matrix_A, statenumber, statenumber, 1.0);
  }
}


/******************************************************************************
 Funktion:	add_to_vector

 Parameter:	vector:    Adresse eines Zeigers auf das erste Element
			   eines Vektors
		prev_item: Zeiger auf ein Element des Vektors, welches
			   irgendwo vor dem einzufuegenden Element (index,
			   value) steht. Dieser Zeiger kann NULL sein und
			   dient zur Performace-Erhoehung, da der Vektor
			   nicht von Beginn an durchsucht zu werden braucht.
		index:	   Index des einzutragenden Elementes
		value:	   Wert des Elementes

 Rueckgabewert:	Zeiger auf das eingetragene Element. Dieser Zeiger kann im
		naechsten Aufruf dieser Funktion als prev_item-Zeiger dienen,
		falls Elemente in aufsteigender Reihenfolge eingefuegt
		werden solle.
		Der Parameter vector wird geaendert, falls das einzutragende
		Element am Anfang eingetragen werden musste.

 Zweck:		Eintragen eines neuen Wertes in einen Vektor. Besitzt der
		Vektor an dieser Position schon einen Eintrag, so wird der
		neue Wert value zu dem alten addiert.
******************************************************************************/
static list_entry *add_to_vector (vector, prev_item, index, value)
list_entry **vector;
register list_entry *prev_item;
register unsigned index;
double value;
{
  register list_entry *prev_entry = NULL;
  register list_entry *entry = *vector;

  if (prev_item)
    entry = prev_item->next;

  while (entry)
    if (entry->index < index)
    {
      prev_entry = entry;
      entry = entry->next;
    }
    else
      if (entry->index == index)
      {
	entry->value += value;
	return entry;
      }
      else
	break;

  {
    register list_entry *new_entry = new_item ();

    new_entry->index = index;
    new_entry->value = value;
    new_entry->next = entry;
    if (prev_entry)
      prev_entry->next = new_entry;
    else
      if (prev_item)
	prev_item->next = new_entry;
      else
	*vector = new_entry;
    return new_entry;
  }
}


/******************************************************************************
 Funktion:	delete_vector

 Parameter:	Adresse eines Zeigers auf das erste Element eines Vektors

 Rueckgabewert:	Der Parameter vector wird auf NULL gesetzt

 Zweck:		Loeschen eines Vektors und Freigeben des von ihm benutzten
		Speichers
******************************************************************************/
static void delete_vector (vector)
list_entry **vector;
{
  register list_entry *new_first = *vector;
  register list_entry *entry = *vector;

  if (!entry)
    return;

  while (entry->next)
    entry = entry->next;

  entry->next = first_free_item;
  first_free_item = new_first;
  *vector = (list_entry *)0;
}


/******************************************************************************
 Funktion:	calculate_phi_k

 Parameter:	k >= 0:	     spezifiziert den neu zu berechnenden Vektor
		statenumber: Zustandsnummer, fuer welche phi(k) berechnet werden soll

 Rueckgabewert:	Vektor phi(k) fuer den Zustand statenumber
		Achtung: Dieser Vektor darf nicht modifiziert werden, da er
			 beim naechsten Aufruf dieser Funktion wieder verwendet
			 wird. (Weil phi(k) = phi(k-1) * matrix_A)

 Zweck:		Berechnet phi(k) fuer einen Zustand statenumber.
******************************************************************************/
static list_entry *calculate_phi_k (k, statenumber)
unsigned k, statenumber;
{
  static unsigned prev_k, prev_statenumber;
  static list_entry *prev_phi;
  list_entry *result_phi = NULL;
  unsigned i,j;

  if (statenumber != prev_statenumber)  /*** nur beim ersten mal ***/
  {
    delete_vector (&prev_phi);          /*** ???? ***/
    add_to_vector (&prev_phi, (list_entry *)0, statenumber, 1.0);  /*** phi,k(0) Initial! ***/
    prev_statenumber = statenumber;
    prev_k = 0;
    if (!k)   
      return prev_phi;
  }
  else
    if (k <= prev_k)
    {
      fprintf (stderr, "Error in %s: k <= prev_k in calculate_phi_k\n", __FILE__);
      my_exit (1);
    }

  /*** result_phi = prev_phi * matrix_A ^ (k-prev_k)  ausrechnen: ***/
  for (i = prev_k + 1; i <= k; ++i)
  { 
    register list_entry *phi_eintrag;

    /*** gehe ueber alle prev_phi-Vektoreintraege ***/
    for (phi_eintrag = prev_phi; phi_eintrag;phi_eintrag = phi_eintrag->next)
    {
      register list_entry *matrix_eintrag;
      register list_entry *prev_entry = NULL; /*** nur performance Zwecke ***/
      register double phi_factor = phi_eintrag->value;  /*** ebenso ***/

      /*** gehe ueber matrix_A-Eintraege in Zeile prev_phi-Index ***/
      if (phi_eintrag->index <= toptan)
      {
	for (matrix_eintrag = matrix_A[phi_eintrag->index]; matrix_eintrag;
	     matrix_eintrag = matrix_eintrag->next)
	  prev_entry = add_to_vector (&result_phi, prev_entry, matrix_eintrag->index,
				      matrix_eintrag->value * phi_factor);
      }
      else  /*** matrix_delta Diagonalelement aus S'E gleich 1 ***/
      {
	do   /*** alle weiteren phi_eintraege sind aus S'E, da vektor indizes aufsteigend sortiert sind ***/
	{
	  add_to_vector (&result_phi, NULL, phi_eintrag->index, phi_eintrag->value);
	  phi_eintrag = phi_eintrag->next;
	}
	while (phi_eintrag);
	break;
      }
    }

    /* result_phi zu prev_phi machen, dann result_phi loeschen */
    delete_vector (&prev_phi);
    prev_phi = result_phi;
    result_phi = (list_entry *)0;
  }
  prev_k = k;
  return prev_phi;
}


/******************************************************************************
 Funktion:	trunc_random_gen
******************************************************************************/
static void trunc_random_gen(arb_bits, trunc_error)
long    arb_bits;
double  trunc_error;
{
   char *conv_buffer, *emc_buffer;

   double *omega_f, *psi_f;
   double omega_err = 0.0, psi_err = 0.0;
   unsigned R = 0;

   list_entry *phi = NULL, *tmp_vector = NULL,
              *c_vector = NULL, *p_vector = NULL;
   unsigned state_num;


#  ifdef DEBUG
   printf("trunc_random_gen()\n");
#  endif

   /* open CONV & EMC files */
   if (!(conv_fp = fopen(conv_filename, "w")))
      FILE_OPEN_ERROR(conv_filename);

   if (!(emc_fp = fopen(emc_filename, "w")))
      FILE_OPEN_ERROR(emc_filename);


   /* calculate omega & psi factors */
   calculate_factors_C(arb_bits, expoly, trunc_error, 1.0/delta,
                       &omega_f, &psi_f, &R, &omega_err, &psi_err);


#  ifdef DEBUG
   printf("Right truncation = %u\n", R);
   printf("Omega error = %.*g, Psi error = %.*g\n",
           digits,omega_err,digits,psi_err);
#  endif


   calculate_A_from_Q();

   for(state_num = 1; state_num <= toptan; ++state_num)
   {
      unsigned k;
      register double factor2 = 1.0;

      if (!(marking_flags[state_num] & INITIAL))
      {
         /* Not an initial state */
         P_row_header.row_index = ind_to_state[state_num];
         P_row_header.entries_in_row = 0;
         if(fwrite (&P_row_header, SIZEOF_P_row_header, 1, emc_fp) != 1)
            FILE_WRITE_ERROR(emc_filename);

         C_row_header.row_index = ind_to_state[state_num];
         C_row_header.entries_in_row = 0;
         if(fwrite (&C_row_header, SIZEOF_C_row_header, 1, conv_fp) != 1)
            FILE_WRITE_ERROR (conv_filename);
         continue;
      }

   delete_vector (&tmp_vector);
   delete_vector (&c_vector);

   for (k = 0; k <= R; ++k)
   {
      phi = calculate_phi_k (k, state_num);

      /* phi(k) * omega_factor[k] zu tmp_vector summieren und
	 phi(k) * psi_factor[k] fuer n = L..k) zu c_vector summieren: */
      {
         register list_entry *phi_eintrag, *prev_tmp_entry = NULL, *prev_c_entry = NULL;

         for (phi_eintrag = phi; phi_eintrag; phi_eintrag = phi_eintrag->next)
         {
            prev_tmp_entry = add_to_vector(&tmp_vector, prev_tmp_entry,
                                           phi_eintrag->index,
                                           phi_eintrag->value * omega_f[k]);
            prev_c_entry = add_to_vector(&c_vector, prev_c_entry,
                                         phi_eintrag->index,
                                         phi_eintrag->value * psi_f[k]);
         }
      }
   }

    delete_vector (&p_vector);
    /* tmp_vector * matrix_delta berechnen: */
    {
      register list_entry *tmp_vector_eintrag;

      for (tmp_vector_eintrag = tmp_vector; tmp_vector_eintrag;
	   tmp_vector_eintrag = tmp_vector_eintrag->next)
      {
	if (tmp_vector_eintrag->index <= toptan)
	{
	  register list_entry *matrix_eintrag, *prev_entry = NULL;
	  register double tmp_vector_factor = tmp_vector_eintrag->value;

	  for (matrix_eintrag = matrix_delta[tmp_vector_eintrag->index]; matrix_eintrag;
	       matrix_eintrag = matrix_eintrag->next)
	    prev_entry = add_to_vector (&p_vector, prev_entry, matrix_eintrag->index,
					matrix_eintrag->value * tmp_vector_factor);
	}
	else
	{
	  register list_entry *prev_entry = NULL;

	  do
	  {
	    prev_entry = add_to_vector (&p_vector, prev_entry, tmp_vector_eintrag->index,
					tmp_vector_eintrag->value);
	    tmp_vector_eintrag = tmp_vector_eintrag->next;
	  }
	  while (tmp_vector_eintrag);
	  break;
	}
      }
    }

    /*** Ergebnis-Vektor (Zeile) der P-Matrix ins File <netname>.EMC<MC_number> schreiben: ***/
    {
      register unsigned anzahl_uebergaenge = 0;
      register list_entry *eintrag;

      for (eintrag = p_vector; eintrag; eintrag = eintrag->next)
	if (eintrag->value > MINREAL)
	  ++anzahl_uebergaenge;

      P_row_header.row_index = ind_to_state[state_num];
      P_row_header.entries_in_row = anzahl_uebergaenge;
      if (fwrite (&P_row_header, SIZEOF_P_row_header, 1, emc_fp) != 1)
	FILE_WRITE_ERROR (emc_filename);

      for (eintrag = p_vector; eintrag; eintrag = eintrag->next)
	if (eintrag->value > MINREAL)
	{
	  P_row_entry.column_index = ind_to_state[eintrag->index];
	  P_row_entry.probability = eintrag->value;
	  if (fwrite (&P_row_entry, SIZEOF_P_row_entry, 1, emc_fp) != 1)
	    FILE_WRITE_ERROR (emc_filename);
	}
    }

    /*** Ergebnis-Vektor (Zeile) der C-Matrix ins File
	 <netname>.CONV<MC_number> schreiben: ***/
    {
      register unsigned anzahl_uebergaenge = 0;
      register list_entry *eintrag;

      for (eintrag = c_vector; eintrag && eintrag->index <= toptan; eintrag = eintrag->next)
      {
         if ( eintrag->value > MINREAL )
            anzahl_uebergaenge++;
      }

      C_row_header.row_index = ind_to_state[state_num];
      C_row_header.entries_in_row = anzahl_uebergaenge;
      if (fwrite (&C_row_header, SIZEOF_C_row_header, 1, conv_fp) != 1)
	FILE_WRITE_ERROR (conv_filename);

      for (eintrag = c_vector; eintrag && eintrag->index <= toptan; eintrag = eintrag->next)
	if (eintrag->value > MINREAL)
	{
	  C_row_entry.column_index = ind_to_state[eintrag->index];
	  C_row_entry.conversion_factor = eintrag->value;
	  if (fwrite (&C_row_entry, SIZEOF_C_row_entry, 1, conv_fp) != 1)
	    FILE_WRITE_ERROR (conv_filename);
	}
    }
  }

  fclose (emc_fp);
  fclose (conv_fp);
}

/******************************************************************************
  main:
******************************************************************************/
main (argc, argv)
int    argc;
char **argv;
{
   double   p;
   time_t   endtime;
   char     hostname[30];
   long     arb_bits;
   double   trunc_error;

   void    *cpu_startTimeErand;    // start time of erand process
   void    *cpu_endTimeErand;      // end time of erand process
   void    *day_startTimeErand;    // start time of erand process
   void    *day_endTimeErand;      // end time of erand process

   long     mem_atStartErand;      // memory at start of erand process
   long     mem_atEndErand;        // memory at end of erand process

   clock(); /* Prozessuhr starten.. ( fuer show_rusage() )*/
   
   /* get start times and memory  */
   cpu_startTimeErand = getCPUTime_C();
   day_startTimeErand = getTimeOfDay_C();
   mem_atStartErand = getUsedMem_C();

   /* close stdout and stderr
    * we have to keep stdin open at this point to prevent the socket_fp to
    * become just that; this is because yyin is stdin from the beginning and
    * it is closed by a freopen() call in parsdfunc.y; if we closed stdin now,
    * along with yyin, our socket would be closed ...
    *
    * (confused? call 1-800-DONT-PANIC)
    */
   close(1);
   close(2);

#ifndef ultrix
   if(!gethostname(hostname, 29))
      if(strcmp(argv[1], hostname))
         /* process is running remote, fool rsh */
         switch( fork() )
         {
	    case 0:  break;
	    case -1: exit( -1 );
            default: exit( 0 );
         }
#endif



   /* check command line args */
   if(argc < 6)
      exit( -1 );

   /* determine precisions & arbitrary prec. bits & filename */
   server_host = argv[1];
   server_port = argv[2];
   epsilon     = (double) atof(argv[3]);
   filename    = argv[4];
   arb_bits    = (long)   atof(argv[5]);
   trunc_error = (double) atof(argv[6]);

#ifdef DEBUG
   printf("Server_Host      = %s", server_host);
         fflush(stdout);
   printf("Server_Port      = %s\n", server_port);
   printf("Precision        = %s (%g)\n", argv[3], epsilon);
   printf("Filename         = %s\n",      filename);
   printf("Arb. prec. bits  = %s (%d)", argv[5], arb_bits);
         fflush(stdout);
   printf("Truncation error = %s (%g)", argv[6], trunc_error);
         fflush(stdout);
#endif


   /* determine number of digits */
   for( digits = 0, p = epsilon; p < 1.0; digits++, p *= 10.0 )
      ;

   /* initialize server connection, read info */
   init(argv);

   /* read matrices from server */
   read_matrices(argv);

   /* start solution */
   trunc_random_gen(arb_bits, trunc_error);

   printf("delta = %.*g, eta = %u\n", digits,max_qii, eta_k);

   /* print statistics */
   endtime = time(NULL);
   //show_rusage();

   /* get times and memory usage */
   cpu_endTimeErand = getCPUTime_C();
   day_endTimeErand = getTimeOfDay_C();
   mem_atEndErand = getUsedMem_C();

   /* show statistics about erand process */
   printf("\n");
   printf("Erand process : Elapsed CPU time  : ");
   printTimeDiff_C(cpu_endTimeErand, cpu_startTimeErand);
   printf("Erand process : Elapsed real time : ");
   printTimeDiff_C(day_endTimeErand, day_startTimeErand);
   printf("Erand process : Required memory : ");
   printMemDiff_C(mem_atEndErand, mem_atStartErand);
   printf("\n");

#  ifdef SHOW_SPECIAL_FORMAT
      printTimeDiffSec_C(cpu_endTimeErand, cpu_startTimeErand);
      printTimeDiffSec_C(day_endTimeErand, day_startTimeErand);
      printf("%ld", mem_atEndErand - mem_atStartErand);
      printf("\n");
#  endif

   /* delete time pointers */
   freeTime_C(cpu_startTimeErand);
   freeTime_C(day_startTimeErand);
   freeTime_C(cpu_endTimeErand);
   freeTime_C(day_endTimeErand);

   /* close files */
   fclose(stdin);
   fclose(stdout);
   fclose(stderr);

printf("done\n");
   my_exit(0);
}
