/******************************************************************************
 Program:           graph

 component:         execute_erand.c

 Parameters:        The complete path name of the EDSPN to be solved and epsilon

 Purpose:           start a couple of rand processes on the local machine. 
                    Send all necessary information to them. 
                    Wait for all started rand processes to
                    finish their operations.

 Project Director:  Reinhard German

 Programmer:        Patrick Drescher

 History:           Release 1: June 1995
******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <malloc.h>

#include <signal.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>

#include "portab.h"

#ifdef SOL2
#include <sys/filio.h>
#endif

#include "mem.h"
#include "execute_erand.h"
#include "graph_types.h"
#include "csmc_types.h"
#include "matrices_Q_delta.h"

#define TNETHOME_ENV "TNETHOME"
#define BIN_DIR "StatAnalysis/bin"

extern SMC **ESMCs; /* list of SMCs of the general transitions */

/*
 * local variables
 */

static connect_t *con;        /* list of connections */
static int running = 0;       /* number of started rand processes */

static int serv_fd;           /* server socket */
static int port_no;           /* server port number */

FILE *wait_connect();

char *local_hostname;              /* local host name */
double ru_sum = (double )0.0;      /* sum of rusage_time */

/*
 * Function:        execute_erand_processes
 *                  set-up server for rand process execution; start one rand
 *                  process per general transition; 
 *                  wait for processes to respond and collect
 *                  their replies
 *
 * Arguments:       model filename,
 *                  epsilon
 *                  arbitrary precision bits
 *                  truncation error
 *
 * Result:          none
 */
int execute_erand_processes (filename, epsilon, arb_bits, trunc_error)
char   *filename;
double  epsilon;
long    arb_bits;
double  trunc_error;
{
   static char *dspnhome;
   char rand_path[200];
   int i;
   FILE *fp;

   time_t start_time;
   long hrs = 0, min = 0, sec = 0;

   fprintf(stdout,"\nSTARTING ERAND-PROCESS EXECUTION ...\n\n");

   /* fetch environment */
   if( (local_hostname = (char *)getenv("HOST")) == (char *)NULL )
   {
      fprintf(stderr,"Warning: Environment variable HOST not set\n");

      /* fetch local hostname */
      local_hostname = (char *)Malloc(16);
      if( gethostname(local_hostname, 16) )
      {
         fprintf(stderr,"Unable to get hostname\n");
         exit(1);
      }
   }

  /* Pfad fuer das Executable erand bestimmen: */
  if (!(dspnhome = getenv (TNETHOME_ENV)))
  {
     fprintf (stderr, "Environment variable '%s' not set!\n",TNETHOME_ENV);
     exit(1);
  }
  sprintf (rand_path, "%s/%s/erand", dspnhome,BIN_DIR);

   /* initialize connection */
   init_connect(no_gen_trans);

   /* initialize server socket, start timer */
   create_server();
   start_time = time(NULL);

   /* start rand processes */
   for( i = 0 ; i < no_gen_trans; i++ )
      {
         /* start process and wait for connection */
         start_rand(filename, i, rand_path, epsilon, arb_bits, trunc_error);
         if( (fp = wait_connect(0)) == (FILE *)NULL )
	 {
            fprintf(stderr,"Unable to establish connection\n");
            exit(1);
	 }

         /* send markings, Q matrix, and Delta matrix to process */
         send_MDET(fp,i);
         send_Q(fp,i);
         send_DELTA(fp,i);
      }

   /* collect responses */
   msg_receive(WAIT_ALL);

   /* free connection */
   free_connect();

   /* stop timer, calculate run-time */
   sec = time(NULL) - start_time;
   if( sec > 59 )
   {
      min = sec / 60;
      sec %= 60;
   }
   if( min > 59 )
   {
      hrs = min / 60;
      min %= 60;
   }
   fflush(stdout);
   fprintf(stdout,"\nTotal CPU time: %.2lf sec.\n", ru_sum);
   fprintf(stdout,"Total overall time: %02lu:%02lu:%02lu\n", hrs,min,sec);
}

/*
 * Function:        chld_handler
 *                  release zombie children from proc-table
 *
 * Arguments:       none
 *
 * Result:          none
 */
void chld_handler()
{
   int status;

   waitpid(-1,&status,WNOHANG);
}

/*
 * Function:        Fwrite
 *                  write a buffer to a file pointer; if the write fails
 *                  partially, try again; if it fails completely, abort
 *                  program execution
 *
 * Arguments:       a file pointer
 *                  pointer to data buffer
 *                  length of data buffer
 *
 * Result:          length of buffer written
 */
static int Fwrite(fp,data,len)
FILE *fp;
char *data;
unsigned len;
{
   int rt;

again:
   errno = 0;

   if( (rt = write((int )fileno(fp),data,len)) != len )
   {
      if( rt >= 0 )
      {
         /* partial write */
         data += rt;
         len -= rt;
         goto again;
      }

      fprintf(stderr,"Fwrite: Pipe broken; can't write to socket\n");
      exit(1);
   }

   return( rt );
}


/*
 * Function:        init_connect
 *                  allocate memory for connections with rand processes and
 *                  for server
 *
 * Arguments:       number of connections to allocate memory for (usually
 *                  the number of transitions)
 *
 * Result:          none
 */
int init_connect(amount)
int amount;
{
   int i;

#ifdef DEBUG
      fprintf(stdout,"init_connect: initializing %d transitions\n", amount);
#endif

   /* allocate room for file descriptors */
   con = (connect_t *)Calloc(amount, sizeof(connect_t));

   for( i = 0; i < amount; i++ )
   {
      con[i].index = -1;
      con[i].fd = -1;
      con[i].fp = NULL;
   }

   /* catch SIGCHLD */
   if( (void *)signal(SIGCHLD,chld_handler) == (void *)-1L )
   {
      fprintf(stderr,"Unable to install catcher for SIGCHLD\n");
      exit(1);
   }
}


/*
 * Function:        create_server
 *                  create server socket and bind it to an address, get port
 *                  number of the connection, listen for calls
 *
 * Arguments:       none
 *
 * Result:          none
 */
int create_server()
{
   struct sockaddr_in server;
   unsigned length = sizeof(server);

   /* create & bind server socket */
   if( (serv_fd = socket(AF_INET,SOCK_STREAM,0)) < 0 )
   {
      fprintf(stderr,"Unable to create server socket\n");
      exit(1);
   }
   memset(&server,0,sizeof(server));
   server.sin_family = AF_INET;
   server.sin_port = 0;
   if( bind(serv_fd,(struct sockaddr *)&server,sizeof(server)) )
   {
      fprintf(stderr,"Unable to bind socket to address\n");
      exit(1);
   }

   /* retrieve port number */
   getsockname(serv_fd,(struct sockaddr *)&server,&length);
   if( length == sizeof(server) )
   {
      port_no = ntohs(server.sin_port);
#ifdef DEBUG
         fprintf(stdout,"create_server: port_no = %d\n", port_no);
#endif
   }
   else
   {
      fprintf(stderr,"Unable to get server port number\n");
      exit(1);
   }

   /* listen for incoming calls */
   if( listen(serv_fd, BACKLOG) )
   {
      fprintf(stderr,"Unable to listen for potential clients\n");
      exit(1);
   }
}


/*
 * Function:        start_rand
 *                  fork process and execute a rand process on the local host
 *
 * Arguments:       filename of the EDSPN, without extension,
 *                  index in names list of the started transition,
 *                  filename of the rand process to execute
 *
 * Result:          none
 */
int start_rand(filename, index, rand_path, epsilon, arb_bits, trunc_error)
char  *filename;
int    index;
char   rand_path[200];
double epsilon;
long   arb_bits;
double trunc_error;
{
char ups_string[300];

   /* start rand process */
   switch( fork() )
   {
   case -1:
      fprintf(stderr,"Unable to fork() rand process\n");
      exit(1);
   case 0:
      {
         char port_str[8];
         char prec_str[16];
         char bits_str[16];
         char trunc_str[16];

         /* build command strings */
         sprintf(port_str,  "%u", port_no);
         sprintf(prec_str,  "%g", epsilon);
         sprintf(bits_str,  "%d", arb_bits);
         sprintf(trunc_str, "%g", trunc_error);

         /* start rand process */
         execl(rand_path, "erand", local_hostname, port_str,
               prec_str, filename, bits_str, trunc_str, NULL);
 
	 /*
         strcpy(ups_string,local_hostname); ups-version zum debuggen
 	 strcat(ups_string," ");
         strcat(ups_string,port_str);
         strcat(ups_string," ");
         strcat(ups_string,prec_str);
         strcat(ups_string," ");
         strcat(ups_string,filename);
         printf("ups_string:%s\n",ups_string);
         execl("/home/common/X11/R5/sun4/bin/ups","ups",rand_path,"-a",ups_string,NULL); */

         /* exec() must not return */
         fprintf(stderr,"Exec() failed!\n");
         exit(1);
      }
#ifdef DEBUG
   default:
         fprintf(stdout,"start_rand: fork()\n");
#endif
   }

   con[running].index = index;
   con[running].fd = -1;
   con[running].fp = NULL;
   running++;
}


/*
 * Function:        wait_connect
 *                  waits on the server socket for replies from a rand process,
 *                  it then accepts the connection and creates an individual
 *                  link with that process
 *
 * Arguments:       none
 *
 * Result:          file pointer for a rand process, or NULL if the connection
 *                  failed
 */
FILE *wait_connect(mode)
int mode;
{
   fd_set fd_rset;
   struct timeval t_out;
   int sel_ret;

   struct sockaddr_in remote_addr;
   int remote_addrsize = sizeof(remote_addr);
   struct hostent *remote_host;

   int fd, i;

   while( 1 )
   {
      /* listen on server socket for a max. of 20 sec */
      FD_ZERO(&fd_rset);
      FD_SET(serv_fd, &fd_rset);
      t_out.tv_sec = 20L;
      t_out.tv_usec = 0L;

      /* wait for rand process to connect */
      do
      {
         sel_ret = select(FD_SETSIZE,&fd_rset,NULL,NULL,&t_out);
      } while( sel_ret < 0 && errno == EINTR );
      if( sel_ret < 0 )
      {
         fprintf(stderr,"Select() failed\n");
         exit(1);
      }
      if( !sel_ret )
         return( (FILE *)NULL );

      /* connect with rand process */
      memset(&remote_addr,0,sizeof(remote_addr));
      do
      {
         fd = accept(serv_fd,(struct sockaddr *)&remote_addr,&remote_addrsize);
      } while( fd < 0 && errno == EINTR );
      if( fd < 0 )
      {
         fprintf(stderr,"Accept() failed\n");
         exit(1);
      }
#ifdef DEBUG
         fprintf(stdout,"wait_connect: accepted (fd = %d)\n", fd);
#endif

      /* determine remote host */
#ifdef NETWORKED
      if( !(remote_host = gethostbyaddr(&remote_addr.sin_addr.s_addr,4,AF_INET)) )
      {
         fprintf(stderr,"Couldn't get host address of %s\n", remote_host->h_name);
         exit(1);
      }
#endif

      /* find index in host list */
      for( i = 0; i < no_gen_trans; i++ )
         if( con[i].fd < 0 )
         {
#ifdef DEBUG
               fprintf(stdout,"wait_connect: running rand (i = %d)\n", i);
#endif

            /* we want buffered input/output! */
            con[i].fd = fd;
            if( (con[i].fp = fdopen(fd,"w")) == (FILE *)NULL )
            {
               fprintf(stderr,"Unable to re-open file pointer\n");
               exit(1);
            }

            return( con[i].fp );
         }
   }
}


/*
 * Function:        send_MDET
 *                  sends basic information (Markov-Chain number) to the erand-process,
 *                  followed by the number of markings; 
 *                  the marking numbers must be in the same order as the
 *                  rows of the Q and delta matrices
 *
 * Arguments:       a file pointer,
 *                  index of the ESMC to be send
 *
 * Result:          none
 */
int send_MDET(fp, index)
FILE *fp;
int index;
{
   unsigned      i, state_nr;
   unsigned      len;
   unsigned char flags;
   int           chainnumber;             /* number of the EMC & CONV file */
   tang_marking  *state_walk;
   SMC           *esmc;

   chainnumber = last_class_id + 1 + index;
    
   esmc = ESMCs[index];

#ifdef DEBUG
      fprintf(stdout,"send_MDET: index of chain = %d\n", chainnumber);
#endif

   /* send number of Markov-Chain */
   Fwrite(fp,&chainnumber,sizeof(chainnumber));
   
   /* send length of name */
   len = strlen(esmc->trans->name);
   Fwrite(fp,&len,sizeof(len));

   /* send name */
   if( len )
       Fwrite(fp,esmc->trans->name,len);
      
   /* write active markings list */
   state_walk = esmc->states;
   for( i = 0; i < esmc->no_states; i++ )
   {
      state_nr = (unsigned)(state_walk->number);
      Fwrite(fp,&state_nr,sizeof(state_nr));

      /* write flags for deterministic transitions */
      flags = (unsigned char)(state_walk->det);
      Fwrite(fp,&flags,sizeof(flags));

      state_walk = state_walk->next; 
   }

   /* write end marker */
   state_nr = (unsigned )0;
   Fwrite(fp,&state_nr,sizeof(state_nr));
}


/*
 * Function:        send_Q
 *                  sends the Q matrix of a general transition to a rand process;
 *                  the row header includes the row index, negative row sum, and
 *                  the number of columns
 *
 * Arguments:       a file pointer,
 *                  index of the ESMC
 *
 * Result:          none
 */
int send_Q(fp, index)
FILE *fp;
int index;
{
   int          r;
   int          path;
   tang_marking *mark;
   SMC          *esmc;

#ifdef DEBUG
      fprintf(stdout,"send_Q: index of ESMC-list = %d\n", index);
#endif

   /* fetch rows */
   esmc = ESMCs[index];
   mark = esmc->states;

   for( r = 0; r < esmc->no_states; r++ )
   {
      /* write row header */
      Q_row_header.neg_sum = -(double)mark->sum_of_rates;
      Q_row_header.row_index = mark->number;
      Q_row_header.entries_in_row = mark->no_exp_paths;
      Fwrite(fp,&Q_row_header,SIZEOF_Q_row_header);
     
      /* write columns */
      for (path = 0; path < mark->no_exp_paths; ++path)
      {
	  Q_row_entry.rate = (double)mark->exp_paths[path].prob;
	  Q_row_entry.column_index = ((tang_marking *)
	   			     mark->exp_paths[path].marking)->number;
          Fwrite(fp,&Q_row_entry,SIZEOF_Q_row_entry);
      }

      mark = mark->next;
   }
}


/*
 * Function:        send_DELTA
 *                  sends the delta matrix of a general transition to its
 *                  rand process; each protocol element consists of
 *                  a row header, containing the row index and the number of
 *                  column entries in that row; it is followed by a number of
 *                  column entries, each with column index and the value of that entry
 *
 * Arguments:       a file pointer to a connection (socket or file),
 *                  index of the ESMC
 *
 * Result:          none
 */
int send_DELTA(fp, index)
FILE *fp;
int index;
{
   int          r;
   int          path;
   tang_marking *mark;
   SMC          *esmc;

#ifdef DEBUG
      fprintf(stdout,"send_DELTA: index of ESMC-list = %d\n", index);
#endif

   /* fetch rows */
   esmc = ESMCs[index];
   mark = esmc->states;

   for( r = 0; r < esmc->no_states; r++ )
   {
      /* write rows */
      delta_row_header.row_index = mark->number;
      delta_row_header.entries_in_row = mark->no_det_paths;
      Fwrite(fp,&delta_row_header,SIZEOF_delta_row_header);

      /* write columns */
      for (path = 0; path < mark->no_det_paths; ++path)
      {
	  delta_row_entry.probability = (double)mark->det_paths[path].prob;
	  delta_row_entry.column_index = ((tang_marking *)mark->det_paths[
					 path].marking)->number;
          Fwrite(fp,&delta_row_entry,SIZEOF_delta_row_entry);
      }

      mark = mark->next;
   }
}


/*
 * Function:        msg_receive
 *                  checks the open connections for incoming messages, the
 *                  messages of each connection are collected in a buffer and
 *                  printed when the process finishes ("ready."); the buffer is
 *                  scanned for keywords such as the run-time of that process,
 *                  the overall run-time is calculated
 *
 * Arguments:       receive mode: check for a message, receive just one, or
 *                  receive all messages
 *
 * Result:          none
 */
int msg_receive(mode)
int mode;
{
   int i,j;

   fd_set fdset;

   static char **output;
   char msg_buf[256], *msg;

   short ready, again, rand_abort;
   int select_ret, read_ret;

   double ru_time;

   /* allocate list */
   if( !output )
      output = (char **)Calloc(no_gen_trans, sizeof(char*));

   /* receive messages */
   do
   {
      /* create fd-readset */
      FD_ZERO(&fdset);
      for( ready = 1, i = 0; i < no_gen_trans; i++ )
      {
         if( con[i].fd >= 0 )
	 {
            FD_SET(con[i].fd, &fdset);
            ready = 0;
	 }
      }

      /* all rand processes finished? */
      if( ready )
         return;

      /* check pending messages */
      again = 0;
      do
      {
            select_ret = select(FD_SETSIZE,&fdset,NULL,NULL,NULL);
      } while( select_ret < 0 && errno == EINTR );
      if( select_ret < 0 )
      {
         fprintf(stderr,"Message select failed\n");
         exit(1);
      }

      /* check received messages */
      if( select_ret > 0 )
         for( i = 0; i < no_gen_trans; i++ )
            if( con[i].fd >= 0 && FD_ISSET(con[i].fd,&fdset) )
	    {
               /* clear entry in read set */
               FD_CLR(con[i].fd, &fdset);

               /* read message by chars */
               msg = msg_buf - 1;
               rand_abort = 0;
               do
	       {
                  if( (read_ret = read(con[i].fd,++msg,1)) < 0 )
		  {
                     fprintf(stderr,"Read on socket failed for %d\n", i);
                     exit(1);
		  }
                  if( !read_ret )
		  {
                     rand_abort = 1;
                     break;
		  }
	       } while( *msg && *msg != '\n' && (msg - msg_buf) < 199 );
               *msg = '\n';
               *++msg = '\0';

#ifdef DEBUG
                  fprintf(stdout,"msg_receive: msg_buf = %s\n", msg_buf);
#endif

               /* rand process #`i' calculation done? */
               if( rand_abort || !strncmp(msg_buf,"ready.",6) )
	       {
                  /* print collected messages */
                  fprintf(stdout,"Rand process for ");
                  fprintf(stdout,"transition '%s'", ESMCs[con[i].index]->trans->name);
                  fprintf(stdout," terminated:\n");
                  fputs(output[i],stdout);
                  fprintf(stdout,"\n");
                  fflush(stdout);

                  /* close socket, free message buffer */
                  fclose(con[i].fp);
                  con[i].fp = (FILE *)NULL;
                  con[i].fd = -1;

                  free(output[i]);
                  output[i] = (char *)NULL;

                  /* wait for more? */
                  if( mode == WAIT_ONE )
                     return;
	       }
               else
	       {
                  if( !strncmp(msg_buf,"Elapsed CPU time",16) )
		  {
                     /* calculate rusage sum */
                     sscanf(msg_buf + 18,"%lf", &ru_time);
                     ru_sum += ru_time;
		  }

                  /* buffer messages */
                  if( !output[i] )
		  {
                     if( !(output[i] = (char *)strdup(msg_buf)) )
		     {
                        fprintf(stderr,"Unable to duplicate string\n");
                        exit(1);
		     }
		  }
                  else
		  {
                     /* concatenate strings */
                     output[i] = (char *)Realloc(output[i],strlen(output[i]) + strlen(msg_buf) + 1);
                     strcat(output[i],msg_buf);
		  }
	       }
               again = 1;
	    }
   } while( again );
}


/*
 * Function:        free_connect
 *                  frees all connection dependent data
 *
 * Arguments:       none
 *
 * Result:          none
 */
int free_connect() 
{
   /* close server */
   close(serv_fd);

   /* free connection fd & host */
   free(con);
}
