// File      msolve.cpp         (C++ Implementation)
// Program   msolve             (Stationary Analysis Experiments)
/**
 *@author  Philipp Berndt
 *@version $Id: msolve.cpp,v 1.5 2001/05/25 12:01:43 azi Exp $
 */

//#define DEBUG
#include <string>
#include <iostream>
#include <strstream>
#include <fstream>
#include <vector>
#include <map>
#include <cmath>
#include <cerrno>
#include <stdio.h>	// old C for FILE, remove

#ifdef PVM_SUPPORT
#include "pvm3cpp.h"
#endif
#include "Time.h"
#include "Parspace.h"


namespace
{
    const double SLAVELOAD = 1.0;
    const double MASTERLOAD = 0.0;
    string modelpart1;
    string modelpart2;
    Parspace ps;  
    map<int,int> work;
    map<int, string> hosts;
    enum PvmTags
    {
	TAG_DONE,
	TAG_SLAVE_DIED
    };
};


/***************************** Utility functions *****************************/

#include <unistd.h> // vfork, execvp, close
#include <fcntl.h>  // open
#include <wait.h>   // wait

/** Reads one line from a stream and puts it in a string. */
string readLine(istream &is)
{
    const int BUFSIZE = 100;
    char buffer[BUFSIZE];
    string res;
  
    do  {
        is.getline(buffer, BUFSIZE, '\n');
        res += buffer;
        } while (is.gcount() == BUFSIZE);

    return res;
}

/** Copies a file. */
bool copy(string src, string dst)
{
    ifstream is(src.c_str());
    if (!is)
	return false;

    ofstream os(dst.c_str());
    if (!os)
	return false;

    char c;
    while(is.get(c)) os.put(c);
    if (!is.eof() || !os)
	return false;
}

/** fork()-exec()-wait() */
int execute(const string &path, const vector<string>& argv, bool mute=false)
{
    const char **argvec = new const char*[argv.size()+2];
    const char **argp = argvec;
    *(argp++) = path.c_str();
    for (vector<string>::const_iterator i= argv.begin(); i != argv.end() ;)
	*(argp++) = (i++)->c_str();
    *argp = 0;

    if (!vfork())
	{
	if (mute)
	    {  // Redirect stdout to /dev/null
	    close(STDOUT_FILENO);
	    open("/dev/null", O_WRONLY, 0); // should be STDOUT 
	    }
	execvp(path.c_str(), const_cast<char*const*>(argvec));
	perror(("Error executing "+path).c_str());
	return -1;
	}
    int status;
    wait(&status);
    return status;
}

/** Returns string representation of integer. */
string itos(int i)
{
    ostrstream ost;
    ost << i << ends;
    string res(ost.str());
    ost.freeze(0);
    return res;
}


/**************************** Argument processing ****************************/

namespace Args
{
    enum  partype_t  { MARKING, DELAY };
    enum  execmode_t { PARALLEL, SEQUENTIAL };
    class InvArgExc  {};

    map<char,string> argss;
    const string     modeldir((string(getenv("MODELDIR")))+"/");
    const string     tnethome((string(getenv("TNETHOME")))+"/");
    
    string    parname;
    partype_t partype;
    double    from;
    double    to;
    double    step;
    int       base;
    string    nicelevel;

    string     netname;
    execmode_t execmode;
    string     epsilon;
    string     trncerr;

    string    getParname() { return parname; }
    partype_t getPartype() { return partype; }
    double    getFrom()    { return from;    }
    double    getTo()      { return to;      }
    double    getStep()    { return step;    }
    int       getBase()    { return base;    }

    string getModelName()      { return netname;                      }
    string getModelName(int n) { return netname+"#"+itos(n);          }
    string getModelPath()      { return modeldir+netname;             }
    string getModelPath(int n) { return modeldir+netname+"#"+itos(n); }

    vector<string> build(int n, bool load_u);
    void parse(int argc, char **argv);
}

/** Parses and stores command line arguments. */
void Args::parse(int argc, char **argv)
{
    char c;
    while((c = getopt(argc, argv, "EI:psn:di:ulr:a:D:M:B:")) != EOF)
	argss[c] = (optarg) ? string(optarg) : "";

    if (argc - optind != 6)
	throw InvArgExc();
    argv = &argv[optind];
      
    netname = argv[0];
    epsilon = argv[1];
    trncerr = argv[2];
    (istrstream(argv[3])) >> from >> throw_if_fail;
    (istrstream(argv[4])) >> to   >> throw_if_fail;
    (istrstream(argv[5])) >> step >> throw_if_fail;

    map<char,string>::iterator i;
    if ((i=argss.find('D'))!=argss.end())
	{
	partype = DELAY;
	parname = i->second;
	argss.erase(i);
	if ((i=argss.find('I'))!=argss.end()) // In case of FIA...
	    argss['v'] = "";                  // ...force saving of u.
	}
    else if ((i=argss.find('M'))!=argss.end())
	{
	partype = MARKING;
	parname = i->second;
	argss.erase(i);
	}
    else
	{
	cout << "You must specify a parameter with -D or -M !" << endl;
	throw InvArgExc();
	}
    cout <<"par:"<<parname<<endl;

    if ((i=argss.find('s'))!=argss.end())
	{
	execmode = SEQUENTIAL;
	}
    else if ((i=argss.find('p'))!=argss.end())
	{
	argss.erase(i);
	argss['s'] = "";
	execmode = PARALLEL;
	}
    else
	{
	cout << "You must specify -s or -p !" << endl;
	throw InvArgExc();
	}

    if ((i=argss.find('n'))!=argss.end())
	{
	nicelevel = i->second;
	argss.erase(i);
	}
    else
	nicelevel = "0";

    if ((i=argss.find('B'))!=argss.end())
	{
	(istrstream(i->second.c_str())) >> base >> throw_if_fail;
	argss.erase(i);
	}
    else base = 0;
}

/** Builds argument vector for starting analysis. */
vector<string> Args::build(int n, bool load_u)
{
    const string minus("-");
    map<char, string>::const_iterator i;
    vector<string> argstrv;
    argstrv.push_back(Args::tnethome+"StatAnalysis/scripts/SOLVE");
    argstrv.push_back(Args::nicelevel);
    argstrv.push_back(Args::getModelPath(n)+".logg");
    for (i=argss.begin(); i!=argss.end(); ++i)
	{
	char c = i->first;
	if (load_u && (c=='u' || c=='r'))
	    c = 'l';
	argstrv.push_back(minus + c);
	if (!i->second.empty())
	    argstrv.push_back(i->second.c_str());
	}
    argstrv.push_back(getModelName(n));
    argstrv.push_back(epsilon);
    argstrv.push_back(trncerr);
  
    return argstrv;
}


/************************* Shared Highlevel Functions ************************/

/** Prints out small help text. */
void usage()
{
cout
<<"\nUsage:"
<<"\nMSOLVE -E         <explicite-options> <experiment-options> <model-args>"
<<"\n -or-"
<<"\nMSOLVE -I maxiter <iterative-options> <experiment-options> <model-args>"
<<"\n"
<<"\n explicite-options := {-d|-i maxiter}    [-a bits]"
<<"\n iterative-options := {-u|-l|-r seedval} [-a bits]"
<<"\nexperiment-options := [-B log-base] {-s|-p} {-D delaypar|-M markingpar}"
<<"\n                      {-n nice-level}"
<<"\n     model-args    := netname epsilon truncerror from to step "
<<"\n"
<<"\nE: EMC explicite solution meth.  I: avoiding fill-in solution meth."
<<"\ns: sequential execution          p: distributed execution"
<<"\nd: direct solve of LGS           i: iterative solve of LGS"
<<"\nu: uniform initial vector        r: random initial vector"
<<"\nl: load file for initial vector  v: save stationary vector"
<<"\na: use arbitrary precision"
<<"\n"<<endl;
}

/** Reads in model description of a net. */
bool readModel()
{
    string modeldesc;
    string filename = Args::getModelPath() + ".TN";
    ifstream is(filename.c_str());
    if (!is)
	{
	cout << "Unable to open model " << filename << " !" << endl;
	return false;
	}

    for (char c; is.get(c);)
	modeldesc += c;

    if (!is.eof())
	{
	cout << "Error while reading " << filename << " !" << endl;
	return false;
	}
  
    string key;
    cout << "Argpar:"<<Args::getParname()<<endl;
    switch(Args::getPartype())
	{
	case Args::MARKING:
	    key = "MARKPAR "+Args::getParname()+" ";
	    break;
	case Args::DELAY:
	    key = "DELAYPAR "+Args::getParname()+" ";
	}

    int p1, p2;
    if ((p1 = modeldesc.find(key)) == modeldesc.npos)
	{
	cout << key << "not defined!" << endl;
	return false;
	}
    p1 += key.size();
  
    if ((p2 = modeldesc.find(" ",p1)) == modeldesc.npos)
	{
	cout << key << "Unexpected end of model description!" << endl;
	return false;
	}
  
    modelpart1 = modeldesc.substr(0,p1);
    modelpart2 = modeldesc.substr(p2);
    return true;
}

/** Writes model description of net number n with parameter par. */
void writeModel(int n, double par)
{
    ofstream os((Args::getModelPath(n)+".TN").c_str());
    os << modelpart1 << par << modelpart2;
}

/** Deletes model description and temporary files. */
void deleteModel(int n)
{ 
    string name = Args::getModelPath(n)+".TN";
    if (remove(name.c_str()))
	{
	cerr << "Warning: Could not delete temporary model description "
	     << name << " ! (" << errno << ")" << endl;
	perror("remove");
	}

    execute(Args::tnethome+"Shared/scripts/CLEANUP",
	    vector<string>(1, Args::getModelName(n)), true);
}

/** Generates Parspace according to Experiment options. */
void buildParspace(Parspace &ps)
{
    for (double par = Args::getFrom(); par <= Args::getTo();)
	{
	ps.addPar(par);
      
	par = (Args::getBase())
	    ? par * pow(Args::getBase(), Args::getStep())
	    : par + Args::getStep();

	if (Args::getPartype() == Args::MARKING)
	    par = static_cast<int>(par);
	}
}

/** Prepares starting of an analysis. */
vector<string> prepareSolve(int n)
{
    bool load_u = false;
    if (Args::getPartype() == Args::DELAY)
	{
	int np = ps.nearest(n);
	if (np < 0)
	    load_u = copy(Args::getModelPath()+".u",
			  Args::getModelPath(n)+".u");
	else
	    load_u = copy(Args::getModelPath(np)+".u",
			  Args::getModelPath(n)+".u");
	}

    double par = ps.getPar(n);
    writeModel(n, par);

    ps.setStatus(n, Parspace::INPROGRESS);
    cout << ps << endl;

    return Args::build(n, load_u);
}

/** Extracts results of an analysis and appends them to .val_tab */
bool extractResults(int n)
{
    static bool first_time = true;
    ofstream idtab, valtab;

    ifstream results((Args::getModelPath(n)+".RESULTS").c_str());
    if (!results)
	return false;

    valtab.open((Args::getModelPath()+".val_tab").c_str(),
		(first_time) ? ios::trunc|ios::out : ios::app);

    if (first_time)
	{
	valtab << "# Data for gnuplot\n";

	idtab.open((Args::getModelPath()+".id_tab").c_str());
	idtab << Args::getParname() << '\n';
	}

    valtab << ps.getPar(n);

    while(!results.eof())
	{
	string line = readLine(results);
	int p;
	if ((p = line.find("=")) == line.npos)
	    continue;
      
	if (first_time)
	    idtab << line.substr(0, p-1) << '\n';

	valtab << line.substr(p+1);      
	}

    valtab << '\n';

    first_time = false;
    return true;
}


/**************************** Sequential Execution ***************************/

/** Performs experiment using in sequential execution. */
void run_sequential()
{
    for (int n=0; !ps.done(); ++n)
	{
	vector<string> argv = prepareSolve(n);
	execute((Args::tnethome+"Shared/scripts/nice_wrapper").c_str(),
		argv);
	if (extractResults(n))
	    ps.setStatus(n, Parspace::DONE);
	else 
	    ps.setStatus(n, Parspace::UNDONE);
	cout << ps << endl;
	}
}


/***************************** Parallel Execution ***************************/
#ifdef PVM_SUPPORT  

/** Starts analysis n on specified host. */
bool spawnSlave(int n, string host)
{
    vector<string> argv = prepareSolve(n);

    cout << "Starting " << Args::getModelName(n) 
	 << " with " << Args::getParname() << "=" <<  ps.getPar(n) 
	 << " on " << host << "..." << flush;
    try {
    int tid=pvm_spawn((Args::tnethome+"Shared/scripts/nice_wrapper").c_str(),
		      argv,
		      PvmTaskHost, host.c_str()); 

    // Ask for notification when slave dies
    pvm_notify(PvmTaskExit, TAG_SLAVE_DIED, 1, &tid);
    
    hosts[tid] = host;
    work[tid] = n;

    cout << "OK" << endl;

    }
    catch(PvmException)
	{
	cout << "failed" << endl;
	deleteModel(n);
	return false;
	}
    return true;
}

/** Performs appropriate actions when a slave has terminated (un)successful. */
void slaveTerminated(int tid)
{
    map<int, int>::iterator workpar = work.find(tid);
    int n = workpar->second;
    work.erase(workpar);

    map<int,string>::iterator host = hosts.find(tid);
    if (host != hosts.end())
	{
	string hostname = host->second;
	hosts.erase(host);

	HostConfig::addVirtLoad(hostname, -SLAVELOAD);

	if (extractResults(n))
	    ps.setStatus(n, Parspace::DONE);
	else 
	    ps.setStatus(n, Parspace::UNDONE);
	}
    else // host was deleted
	ps.setStatus(n, Parspace::UNDONE);
    
    deleteModel(n);
    HostConfig::update();
}

/** Sorts results in .val_tab */
void sortResults()
{
    cout << "Sorting results..." << flush;
    string val_tab_name = Args::getModelPath() + ".val_tab";
    vector<string> argv;
    argv.push_back("-n");
    argv.push_back("-o");
    argv.push_back(val_tab_name);
    argv.push_back(val_tab_name);
    if (execute("sort", argv))
	cout << "Failed." << endl;
    else
	cout << "OK" << endl;
}

/** Performs experiment using in sequential execution. */
void run_parallel()
{
    try {
    initPvm();

    // Mark this host as BUSY so no slave is started here.
    // msolve doesn't eat performance at all, but if this host stays
    // idle, nobody will boot it. ;->
    HostConfig::addVirtLoad(HostConfig::gethostname(), MASTERLOAD);  

    cout << "Entering main service loop" << endl;

    while(!ps.done())
	{
	while(true)
	    {
	    int n;

	    if (pvm_probe(-1, -1))
		break;

	    if (!HostConfig::idleHosts())
		break;

	    if ((n = ps.getNext()) < 0)
		break;

	    string hostname = HostConfig::getHost();
	    if (spawnSlave(n, hostname))
                HostConfig::addVirtLoad(hostname, SLAVELOAD);
	    }

	int bufid, bytes, msgtag, tid;
	bufid = pvm_recv(-1, -1);
	pvm_bufinfo(bufid, &bytes, &msgtag, &tid);

	if (checkSystemMessage(msgtag))
	    continue;

	if (msgtag == TAG_SLAVE_DIED)
	    {
	    tid = pvm_upkint();
	    slaveTerminated(tid);
	    cout << ps << endl;
		}
	else
	    {
	    cerr << "Ignoring unknown Message " << msgtag
		 <<" from " << hex<<tid<<dec << endl;
	    }
	}
    }
    catch(PvmException e)
	{
	e.perror("main()");
	cerr << "Aborting." << endl;
	exit(1);
	}
    catch(InterruptedException e)
	{
	cout << "User Interrupt. Aborting" << endl;
	}
    shutdownPvm();
    sortResults();
}
#endif

/************************************ Main ***********************************/

int main(int argc, char **argv)
{
    try
	{
	Args::parse(argc, argv);
	}
    catch(...)
	{
	usage();
	exit(1);
	}

    if (!readModel())
	exit(1);

    buildParspace(ps);

    Time beginning = Time::getTimeOfDay();

    cout << "STARTING EXPERIMENT for net " << Args::getModelName() << endl
	 << "\nVarying " << ((Args::partype == Args::DELAY)? "delay":"marking")
	 << " parameter " << Args::parname
	 << " from " << Args::from << " to " << Args::to << endl;
    if (Args::base>0)
	cout << " with log-base " << Args::base 
	     << " and exponent " << Args::step << endl;
    else
	cout << "with stepsize " << Args::step << endl;

    if (Args::execmode == Args::PARALLEL)
	{
#ifdef PVM_SUPPORT  
	run_parallel();
#else
	cout << "Cannot run parallel." << endl;
	cout << "Set PVM_ROOT to compile with PVM_SUPPORT" << endl;
	exit(1);
#endif
	}
    else
	{
	run_sequential();
	}
    Time now = Time::getTimeOfDay();
    cout << "\nEXPERIMENT DONE."
	 << "\nElapsed overall real time : " << now-beginning << endl;
    cout << "Results are stored in " 
	 << (Args::getModelPath()+".val_tab").c_str() << endl;
    cout << "Use Evaluation/Results/gnuplot to display them graphically." << endl;
}
