/****************************************************************************
** Evaluation : STATIONARY ANALYSIS                                        **
*****************************************************************************
** Program    : solve                                                      **
** Component  : solve_base                                                 **
** File       : solve_base.cpp (C++ source)                                **
**                                                                         **
** Purpose    : implementations of member functions of the class SolveBase **
**              which are used by the iterative solution method with       **
**              fill-in avoidance                                          **
**                                                                         **
** Master's T.: realis. of an iterative solution method for not marc. SPNs **
**                                                                         **
** Programmer : Oliver Hesse                                               **
**                                                                         **
** Date       : june of 1999                                               **
****************************************************************************/

/****************************************************************************
** LIST OF IMPORTS                                                         **
****************************************************************************/

#include <iostream>
#include <fstream>
#include <stdio.h>

#include "solve_base.h"

/****************************************************************************
** DEFINITIONS                                                             **
****************************************************************************/

//#define DEBUG                     // show some informations
//#define DEBUG_U_INIT              // show initial iteration vector
#define SHOW_RESULTS                // show state probs. and firing freq.
#define SHOW_ALL_TIMES

/****************************************************************************
** FUNCTION : get 'u_init'                                                 **
*****************************************************************************
** Purpose  : function generates the initial iteration vector u(0), which  **
**            could be uniform, randomized or loaded from file             **
**                                                                         **
** Input    : nothing                                                      **
**                                                                         **
** Output   : u_init          initial iteration vector u(0)                **
****************************************************************************/

DenseVector SolveBase::get_u_init() const
{
   DenseVector   u_init(dim);       // initial iteration vector
   Rand          rand(SeedValue);   // initializise random generator
   OnesVector    e(dim);            // vector of ones
  
   if (uSource == "uniform")
   {
      cout << "Generation of initial u(0) with uniform distribution ..."
           << endl;

      u_init = e;
   }
   else if (uSource == "random") 
   {
      cout << "Ramdomisation of initial u(0) ..." << endl;

      for(int i = 0; i < u_init.getCols(); ++i)
         u_init[i] = rand.getDouble();
   }
   else if (uSource == "load")
   {
      cout << "Load initial u(0) from file ..." << endl;

      try
      {
         u_init = read_u();
      }
      catch(...)
      {
         cout << endl;
         cout << "Warning: Could not load initial iteration vector"
	      << "from file " << Netname << ".u !" << endl; 
	 cout << "Defaulting to uniform distribution." << endl;

	 u_init = e;
      }
   }
  
  /*** normalisation sum of elements of u_iter_init = 1 ***/
  u_init /= u_init * e;

#  ifdef DEBUG_U_INIT
      cout << endl;
      cout << "u_init :" << endl;
      cout << u_init << endl;
#  endif

   return u_init;
}


/****************************************************************************
** FUNCTION : read 'u_init'                                                **
*****************************************************************************
** Purpose  : read the initial iteration vector u(0) from file <model>.u   **
**                                                                         **
** Input    : Netname (hidden)  name of the corresponding DSPN             **
**                                                                         **
** Output   : u_init            initial iteration vector u(0)              **
****************************************************************************/

DenseVector SolveBase::read_u() const
{
   DenseVector u(dim);

   string filename = Netname + ".u";

   ifstream is(filename.c_str());

   u.read(is);

   if (is.fail())
      throw 1; // use ios::failure when it becomes available !

   return u;
}


/****************************************************************************
** FUNCTION : write 'u_stat'                                               **
*****************************************************************************
** Purpose  : write the stationary iteration vector u(M) to file <model>.u **
**                                                                         **
** Input    : u_stat            stationary iteration vector u(M)           **
**            Netname (hidden)  name of the corresponding DSPN             **
**                                                                         **
** Output   : nothing                                                      **
****************************************************************************/

void SolveBase::write_u(const Vector& u) const
{
   string filename = Netname + ".u";

   ofstream os(filename.c_str());
   
   u.write(os);
   if (os.fail())
      throw 1; // use ios::failure when it becomes available !
}


/****************************************************************************
** FUNCTION : delete 'u_stat' file                                         **
*****************************************************************************
** Purpose  : delete the file <model>.u if it exists                       **
**                                                                         **
** Input    : Netname (hidden)  name of the corresponding DSPN             **
**                                                                         **
** Output   : nothing                                                      **
****************************************************************************/

void SolveBase::delete_u_file() const
{
   string filename = Netname + ".u";

   remove(filename.c_str());
}


/****************************************************************************
** FUNCTION : calculate 'pi' and 'phi'                                     **
*****************************************************************************
** Purpose  : convertion of the stationary iteration vector u(m)           **
**            calculation of the state probability vector pi and the       **
**            firing frequencies vector phi                                **
**            and call functions to write the results to files             **
**                                                                         **
** Formulas : c   = u(M) * Exp.C * e + Gen_b(M) * e                        **
**                                                                         **
**            pi  = 1/c * u(M) * Exp.C + 1/c * Gen_b(M)                    **
**            phi = 1/c * Gen_a(M)                                         **
**                                                                         **
** Input    : u_stat       stationary iteration vector u(M)                **
**            Gen_a        Gen_a(M) = Sum of [gen_a(M)] (jensens method)   **
**            Gen_b        Gen_b(M) = Sum of [gen_b(M)] (jensens method)   **
**                                                                         **
** Output   : nothing                                                      **
****************************************************************************/

void SolveBase::calculate_pi_phi(const DenseVector& u,
		                 const DenseVector& a,
		                 const DenseVector& b) const
{
#  ifdef DEBUG
      cout << endl << "Calculation of pi and phi ..." << endl;
#  endif

   double         factor_c;              // conversion factor c
   DenseVector    pi(dim);               // vector of state probabilities
   DenseVector    phi(dim);              // vector of firing frequencies
   OnesVector     e(dim);                // vector of ones

   /*** compute pi and c (most efficient way) ***/
   pi = u;
   pi *= Exp.C;
   factor_c = pi * e + b * e;
   pi += b;
   pi /= factor_c;

   /*** compute phi (most efficient way) ***/
   phi = a;
   phi /= factor_c;

#  ifdef SHOW_RESULTS
      cout << endl;
      cout.setf(ios::scientific, ios::floatfield);
      cout << "State probabilities pi :" << endl << pi  << endl;
      cout << "Firing frequencies phi :" << endl << phi << endl;
      cout.setf(0, ios::floatfield);
      cout << endl;
#  endif  

   /*** save vector pi of state probabilities ***/
   write_probmark(pi);
  
   /*** save vector phi of firing frequencies ***/
   write_firefreq(phi);
}


/****************************************************************************
** FUNCTION : write probmark 'pi'                                          **
*****************************************************************************
** Purpose  : write state probabilities to <model>.PROBMARK file           **
**                                                                         **
** Input    : pi                vector of state probabilities              **
**            Netname (hidden)  name of the corresponding DSPN             **
**                                                                         **
** Output   : nothing                                                      **
****************************************************************************/

void SolveBase::write_probmark(SparseVector pi) const
{
   FILE     *mpdf;
   double    mpdf_buf;
   string    probmark_filename = Netname + ".PROBMARK";

   double    prec; // unspecified but necessary

#  ifdef DEBUG
      cout << "saving state probabilities ..." << endl;
#  endif

   /*** open file for write access ***/
   if ((mpdf = fopen(probmark_filename.c_str(), "w")) == 0)
   {
      perror("fopen: .PROBMARK ");
      exit(1);
   }

   /*** write header ***/
   mpdf_buf = -prec;
   if (fwrite((char*) &mpdf_buf, sizeof(double), 1, mpdf) != 1)
      if (ferror(mpdf))
      {
         perror("fwrite: .PROBMARK ");
         exit(1);
      }

   mpdf_buf = -1.0;
   if (fwrite((char*) &mpdf_buf, sizeof(double), 1, mpdf) != 1)
      if (ferror(mpdf))
      {
         perror("fwrite: .PROBMARK ");
         exit(1);
      }

   /*** write pi ***/
   for(int i = 0; i < pi.getCols(); ++i)
   {
      mpdf_buf = pi[i];
      if (fwrite((char*) &mpdf_buf, sizeof(double), 1, mpdf) != 1)
         if (ferror(mpdf))
         {
            perror("fwrite: .PROBMARK ");
            exit(1);
         }
   }
   fclose(mpdf);
}


/****************************************************************************
** FUNCTION : write firefreq 'phi'                                         **
*****************************************************************************
** Purpose  : write firing frequencies to fole <model>.FIREFREQ file       **
**                                                                         **
** Input    : phi               vector of firing frequencies for general   **
**                              transitions                                **
**            Netname (hidden)  name of the corresponding DSPN             **
**                                                                         **
** Output   : nothing                                                      **
****************************************************************************/

void SolveBase::write_firefreq(SparseVector phi) const
{
   FILE     *mpdf;
   double    mpdf_buf;
   string    firefreq_filename = Netname + ".FIREFREQ";

   double    prec; // unspecified but necessary

#  ifdef DEBUG
      cout << "saving firing frequencies ..." << endl;
#  endif

   /*** open file for write access ***/
   if ((mpdf = fopen(firefreq_filename.c_str(), "w")) == 0)
   {
      perror("fopen: .FIREFREQ ");
      exit(1);
   }

   /*** write header ***/
   mpdf_buf = -prec;
   if (fwrite((char*) &mpdf_buf, sizeof(double), 1, mpdf) != 1)
      if (ferror(mpdf))
      {
         perror("fwrite: .FIREFREQ ");
         exit(1);
      }

   mpdf_buf = -1.0;
   if (fwrite((char*) &mpdf_buf, sizeof(double), 1, mpdf) != 1)
      if (ferror(mpdf))
      {
         perror("fwrite: .FIREFREQ ");
         exit(1);
      }

   /*** write phi ***/
   for(int i = 0; i < phi.getCols(); ++i)
   {
      mpdf_buf = phi[i];
      if (fwrite((char*) &mpdf_buf, sizeof(double), 1, mpdf) != 1)
         if (ferror(mpdf))
         {
            perror("fwrite: .FIREFREQ ");
            exit(1);
         }
   }
   fclose(mpdf);
}

