/****************************************************************************
** Evaluation : STATIONARY ANALYSIS                                        **
*****************************************************************************
** Program    : solve                                                      **
** Component  : solve_factors                                              **
** File       : solve_factors.cpp (C++ source)                             **
**                                                                         **
** Purpose    : computing the alpha_f, alpha_F_bar factors and errors      **
**              used by the EMC explicit sol. method (called from erand.c) **
**              and the fill-in avoidance SM (called from solve_gen.cpp)   **
**                                                                         **
** Master's T.: realis. of an iterative solution method for not marc. SPNs **
**                                                                         **
** Programmer : Oliver Hesse                                               **
**                                                                         **
** Date       : june of 1999                                               **
****************************************************************************/

/****************************************************************************
** LIST OF IMPORTS                                                         **
****************************************************************************/

#include "RR.h"                  // arbitrary precision class RR

#include <iostream>
#include <vector>

#include "solve_factors.h"
#include "expoly_types.h"        // struct. of the expolynomial distribution

extern "C" {
#include "expoly.h"              // calculate factors with prec. of double
}

/****************************************************************************
** DEFINITIONS                                                             **
****************************************************************************/

#define MAXFACTORS 4000000       // maximal number of factors
#define OUTBITS    16            // output precision


//#define DEBUG                  // show factors and some informations
//#define DEBUG_ERRORS           // show current errors of calculation
//#define DEBUG_EXPOLY           // show the expoly distribution
//#define DEBUG_EXACT            // show calculation of exact sums
//#define DEBUG_0                // show calculation of factor  k = 0
//#define DEBUG_K                // show calculation of factors k > 0


/****************************************************************************
** FUNCTION : calculate factors RR                                         **
*****************************************************************************
** Purpose  : calculation of alpha factors with arbitrary precision        **
**            class RR                                                     **
**                                                                         **
** Input    : ArbBits               bits for arbitrary precision           **
**            expoly_init           initial representation of the          **
**                                  expolynomial distribution              **
**	      TruncError            truncation error                       **
**	      qii_max               renewal rate (max. diag. entry of Q)   **
**                                                                         **
** Output   : alpha_f_factors       vector of factors with transient probs.**
**	      alpha_F_bar_factors   vector of factors with sojourn times   **
**  	      R                     right truncation point                 **
**            alpha_f_error         calculation error of alpha_f factors   **
**	      alpha_F_bar_error     calculation error of alpha_F_bar fact. **
****************************************************************************/

void calculate_factors_RR(long ArbBits, time_l *expoly_init,
                          double TruncError, double qii_max,
                          vector<double> &alpha_f_factors,
		          vector<double> &alpha_F_bar_factors, unsigned &R,
		          double &alpha_f_error, double &alpha_F_bar_error)
{

   /*** set precisions ***/
   RR::SetPrecision(ArbBits);
   RR::SetOutputPrecision(OUTBITS);

   /*** declarations ***/
   RR         alpha_f_factor_RR;      // current factor with arbitrary prec.
   RR         alpha_F_bar_factor_RR;  // current factor with arbitrary prec.

   double     alpha_f_factor;         // current factor with prec. of double
   double     alpha_F_bar_factor;     // current factor with prec. of double

   RR         alpha_f_error_RR;       // errors with arbitrary precision
   RR         alpha_F_bar_error_RR;   // errors with arbitrary precision

   RR         alpha_f_exact_sum;      // exact sum of transient probabilities
   RR         alpha_F_bar_exact_sum;  // exact expected firing time
   RR         alpha_f_sum;            // current sum of trans. probabilities

   time_l     *expoly;                // copy of initial expolynomial distr.
   power_m    *pow_m;                 // power building blocks
   expo_n     *exp_n;                 // exponential building blocks

   int        mass_cnt;               // counter of purely point masses
   int        blck_cnt;               // counter of purely expoly blocks

   vector<RR> beta1_mass_k;           // iter. var. for each point mass
   vector<RR> beta1_k;                // iter. var. for each expoly block
   vector<RR> beta2_k;                // iter. var. for each expoly block
   vector<RR> gamma1_k;               // iter. var. for each expoly block
   vector<RR> gamma2_k;               // iter. var. for each expoly block
   vector<RR> eta1_k;                 // iter. var. for each expoly block
   vector<RR> eta2_k;                 // iter. var. for each expoly block

   unsigned   m;                      // exponent m with prec. of double
   RR         mm;                     // exponent m with arbitrary prec.
   unsigned   h, k;
   RR         hh, kk;

   RR         x1, x2,                 // entries of expoly. distribution
              p, lambda,
              b, c,
           
              qpl,                    // useful calculations
              qplx1, qplx2,
              lx1, lx2,
              exp_qx1,
              exp_qplx1, exp_qplx2,
              pow_x1_m, pow_x2_m,

              beta1, beta2,           // temporary variables
              beta1_sum, beta2_sum,
              gamma1, gamma2,
              exact1, exact2,
              tmp, diff;


   RR q = to_RR(qii_max);             // renewal rate (RR)


   /**********************************************
   *** calcucation of factors (k = 0)          ***
   *** calculation of the expected firing time ***
   *** initialisation of errors                ***
   *** initialisation of iteration variables   ***
   **********************************************/

#  ifdef DEBUG_0
   cout << "calculating the expected firing time" << endl;
   cout << "and factors for k = 0 ..." << endl;
#  endif

   /*** initialisations ***/
   alpha_f_exact_sum     = 1.0;       // sum of all transient probs. is one
   alpha_F_bar_exact_sum = 0.0;       // initial expected firing time is zero

   alpha_f_factor_RR = 0.0;           // initial factor[0] is zero

   expoly = expoly_init;              // get initial expoly distribution
   mass_cnt = -1;                     // initial counter of point masses
   blck_cnt = -1;                     // initial counter of expoly blocks

   /*** for all time intervals of the expolynomial distribution ... ***/
   for( int interval = 0; expoly; expoly = expoly->next, interval++ )
   {
      x1 = to_RR(expoly->x1);         // begin of current interval (RR)
      if (expoly->next)
         x2 = to_RR(expoly->x2);      // end of current interval (RR)

      p = to_RR(expoly->p);           // point mass at x1 (RR)

#     ifdef DEBUG_EXPOLY
         cout << endl;
         cout << "[interval " << interval << "]" << endl;
         cout << "[x1 = " << x1;
         if(expoly->next)
            cout << ", x2 = " << x2 << "]" << endl;
         else
            cout << ", x2 = infinity]" << endl;
         cout << "[p  = " << p << "]" << endl;
#     endif

      if((expoly->p != 0.0) && (qii_max != 0.0))
      {
	 /**********************************************
         *** initialisation of the iterator variable ***
	 *** corresponding to point mass             ***
         **********************************************/

         mass_cnt += 1;

	 /*** check exponent for overflow ***/
	 if((q * x1) > 1e07)
	 {
	    cout << "ERROR: exponent too large." << endl;
	    cout << "Computing of alpha factors failed." << endl;
            exit(1);
	 }

  	 /*** initial poisson prob. ***/
	 exp_qx1 = exp(-q * x1);

         beta1_mass_k.push_back(exp_qx1);

	 /*********************************************
         *** compute part of alpha_f_factor (k = 0) ***
	 *** corresponding to point mass            ***
         *********************************************/

         alpha_f_factor_RR += p * beta1_mass_k[mass_cnt];

#        ifdef DEBUG_0
            cout << "[mass_cnt = " << mass_cnt << "]" << endl;
            cout << "<0> <mass> add to alpha_f[k = 0] : "
                 << p * beta1_mass_k[mass_cnt] << endl;
#        endif
      }

      if(expoly->p != 0.0)
      {
	 /*******************************************
         *** compute part of expected firing time ***
	 *** corresponding to point mass          ***
         *******************************************/

	 alpha_F_bar_exact_sum += p * (x2 - x1);

#        ifdef DEBUG_EXACT
	    cout << "<exact> add to alpha_F_bar_exact_sum : "
	         << p * (x2 - x1) << endl;
#        endif
      }

      if( !expoly->next )
         break;

      /*** for all power terms ... ***/
      for( pow_m = expoly->power; pow_m; pow_m = pow_m->next )
      {
         m = pow_m->m;                        // exponent (unsigned)
         mm = to_RR((long)(m));               // exponent (RR)

	 /*** for all exponential terms ... ***/
         for( exp_n = pow_m->expo; exp_n; exp_n = exp_n->next )
         {
            lambda = to_RR(exp_n->lambda);    // rate (RR)
            b      = to_RR(exp_n->b);         // const. corresp. to 1-PDF (RR)
            c      = to_RR(exp_n->c);         // const. corresp. to pmf (RR)

#           ifdef DEBUG_EXPOLY
               cout << "[m = " << m << ", lambda = " << lambda
                    << ", b = " << b << ", c = " << c << "]" << endl;
#           endif

            /*** useful calculations ***/
            qpl   = q + lambda;
            qplx1 = qpl * x1;
            qplx2 = qpl * x2;
            lx1   = lambda * x1;
            lx2   = lambda * x2;

            pow_x1_m = power(x1, (long)(m));
            pow_x2_m = power(x2, (long)(m));

            if((exp_n->c != 0.0) && (qii_max != 0.0))
            {

	       /*** check exponents for overflow ***/
	       if((qplx1 > 1e08) || (qplx2 > 1e08))
	       {
	          cout << "ERROR: exponent too large." << endl;
	          cout << "Computing of alpha factors failed." << endl;
                  exit(1);
	       }

  	       exp_qplx1 = exp(-qplx1);
	       exp_qplx2 = exp(-qplx2);

   	       /*********************************************
	       *** compute part of alpha_f_factor (k = 0) ***
	       *** corresponding to expolynom. distribut. ***
	       *********************************************/

   	       /*** initial poisson probs. ***/ 
	       beta1 = exp_qplx1;
	       beta2 = exp_qplx2;

	       /*** initial sums ***/ 
	       beta1_sum = beta1;
	       beta2_sum = beta2;

	       /*** initial temporary factor ***/
               tmp = 1.0 / qpl;

               for( h = 1; h <= m; h++ )
               {
   	          hh = to_RR((long)(h));

  	          /*** current temporary factor ***/
	          tmp *= hh / qpl;

	          /*** current poisson probs. ***/ 
	          beta1 *= qplx1 / hh;
	          beta2 *= qplx2 / hh;

	          /*** current sums ***/ 
	          beta1_sum += beta1;
	          beta2_sum += beta2;
               }

               /*** initial gammas ***/
               gamma1 = tmp * (1.0 - beta1_sum);
               gamma2 = tmp * (1.0 - beta2_sum);

               diff = gamma2 - gamma1;

	       /*** add to factor ***/
               alpha_f_factor_RR += c * diff;

#              ifdef DEBUG_0
	          cout << "<0> <k> diff                  : "
		       << diff << endl;
     	          cout << "<0> <k> add to alpha_f[k = 0] : "
		       << c * diff << endl;
#              endif

	       /*********************************************
               *** initialisation of iterator variables   ***
	       *** corresponding to expolynom. distribut. ***
               *********************************************/

               blck_cnt += 1;

   	       /*** initial poisson probs. ***/ 
	       beta1_k.push_back(exp_qplx1);
	       beta2_k.push_back(exp_qplx2);

	       /*** initial etas ***/
	       eta1_k.push_back(pow_x1_m / qpl);
	       eta2_k.push_back(pow_x2_m / qpl);

               /*** initial gammas ***/
               gamma1_k.push_back(gamma1);
               gamma2_k.push_back(gamma2);

#              ifdef DEBUG_0
                  cout << "[blck_cnt = " << blck_cnt << "]" << endl;
	          cout << "<init> beta1  : "
		       << beta1_k[blck_cnt] << endl;
	          cout << "<init> beta2  : "
		       << beta2_k[blck_cnt] << endl;
	          cout << "<init> eta1   : "
		       << eta1_k[blck_cnt] << endl;
	          cout << "<init> eta2   : "
		       << eta2_k[blck_cnt] << endl;
     	          cout << "<init> gamma1 : "
		       << gamma1_k[blck_cnt] << endl;
	          cout << "<init> gamma2 : "
		       << gamma2_k[blck_cnt] << endl;
#              endif
	    }

	    /*********************************************
            *** compute part of expected firing time   ***
	    *** corresponding to expolynom. distribut. ***
            *********************************************/

            if(exp_n->lambda == 0.0)
            {
	       /*** exact sums ***/
	       exact1 = pow_x1_m * x1 / (mm + 1.0);
	       exact2 = pow_x2_m * x2 / (mm + 1.0);
            }
            else
            {
	       /*** check exponents for overflow ***/
	       if((lx1 > 1e08) || (lx2 > 1e08))
     	       {
	          cout << "ERROR: exponent too large." << endl;
	          cout << "Computing of alpha factors failed." << endl;
                  exit(1);
	       }

	       /*** initial poisson probs. ***/ 
	       beta1 = exp(-lx1);
	       beta2 = exp(-lx2);

	       /*** initial sums ***/ 
	       beta1_sum = beta1;
	       beta2_sum = beta2;

	       /*** initial temporary factor ***/
               tmp = 1.0 / lambda;

               for( h = 1; h <= m; h++ )
               {
   	          hh = to_RR((long)(h));

  		  /*** current temporary factor ***/
                  tmp *= hh / lambda;

		  /*** current poisson probs. ***/ 
		  beta1 *= (lx1) / hh;
		  beta2 *= (lx2) / hh;

		  /*** current sums ***/ 
		  beta1_sum += beta1;
		  beta2_sum += beta2;
               }

	       /*** exact sums ***/
               exact1 = tmp * (1.0 - beta1_sum);
               exact2 = tmp * (1.0 - beta2_sum);
            }

	    /*** add to expected firing time ***/
	    alpha_F_bar_exact_sum += b * (exact2 - exact1);

#           ifdef DEBUG_EXACT
	       cout << "<exact> add to alpha_F_bar_exact_sum : "
		    << b * (exact2 - exact1) << endl;
#           endif

	 } // for ... exp_n
      } // for ... pow_m
   } // for ... expoly

#  ifdef DEBUG_EXACT
      cout << "alpha_f_exact_sum     : " << alpha_f_exact_sum << endl;
      cout << "alpha_F_bar_exact_sum : " << alpha_F_bar_exact_sum << endl;
#  endif


   if(expoly_init == NULL)
   {
      cout << "ERROR: Missing expolynomial distribution." << endl;
      cout << endl;

      /*** factors in case of no distribution ***/
      alpha_f_factor_RR     = 0.0;
      alpha_F_bar_factor_RR = 0.0;

      /*** errors in case of no distribution ***/
      alpha_f_error_RR     = 1.0;
      alpha_F_bar_error_RR = 0.0;
   }
   else
   {
      if(qii_max == 0.0)
      {
         /*** factors in case q is zero ***/
	 alpha_f_factor_RR     = alpha_f_exact_sum;
	 alpha_F_bar_factor_RR = alpha_F_bar_exact_sum;

	 /*** errors in case q is zero ***/
	 alpha_f_error_RR     = 0.0;
	 alpha_F_bar_error_RR = 0.0;
      }
      else
      {
         /*** initial sum of alpha_f_factors ***/
	 alpha_f_sum = alpha_f_factor_RR;

	 /*** initial alpha_F_bar_factor ***/
	 alpha_F_bar_factor_RR = (1.0 - alpha_f_sum) / q;

	 /*** initial errors ***/
	 alpha_f_error_RR     = alpha_f_exact_sum     - alpha_f_factor_RR;
	 alpha_F_bar_error_RR = alpha_F_bar_exact_sum - alpha_F_bar_factor_RR;
      }
   }

#  ifdef DEBUG_ERRORS
      cout << "<init> alpha_f_error     : " << alpha_f_error_RR << endl;
      cout << "<init> alpha_F_bar_error : " << alpha_F_bar_error_RR << endl;
#  endif

   /*** convertion of initial factors ***/
   conv(alpha_f_factor,     alpha_f_factor_RR);
   conv(alpha_F_bar_factor, alpha_F_bar_factor_RR);

   /*** push initial factors to empty vectors of factors ***/
   alpha_f_factors.push_back(alpha_f_factor);
   alpha_F_bar_factors.push_back(alpha_F_bar_factor);



   /**************************************
   *** calcucation of factors (k > 0)  ***
   *** calculation of current errors   ***
   **************************************/

#  ifdef DEBUG_K
      if (qii_max != 0.0)
         cout << "calculating factors for k > 0 ..." << endl;
#  endif

   /*** calculate factors until the precision is reached ... ***/
   k = 1;
   while((qii_max != 0.0) && ((alpha_f_error_RR > TruncError) ||
			      (alpha_F_bar_error_RR > TruncError)) &&
	                      (alpha_f_error_RR > 0.0) &&
	                      (alpha_F_bar_error_RR > 0.0) && (k < MAXFACTORS))
   {
      kk = to_RR((long)(k));

#     ifdef DEBUG_K
         cout << endl;
         cout << endl;
         cout << "[k = " << k << "]" << endl;
#     endif


      /*** initialisations ***/
      alpha_f_factor_RR = 0.0;        // initial factor[k] is zero

      expoly = expoly_init;           // get initial expoly distribution
      mass_cnt = -1;                  // initial point mass counter
      blck_cnt = -1;                  // initial expoly blocks counter

      /*** for all time intervals of the expolynomial distribution ... ***/
      for(int interval = 0; expoly; expoly = expoly->next, interval++ )
      {
         x1 = to_RR(expoly->x1);      // begin of current interval (RR)
         if (expoly->next)
	    x2 = to_RR(expoly->x2);   // end of current interval (RR)

	 p = to_RR(expoly->p);        // point mass at x1 (RR)

#        ifdef DEBUG_EXPOLY
            cout << endl;
            cout << "[interval " << interval << "]" << endl;
            cout << "[x1 = " << x1;
            if(expoly->next)
               cout << ", x2 = " << x2 << "]" << endl;
            else
               cout << ", x2 = infinity]" << endl;
            cout << "[p  = " << p << "]" << endl;
#        endif

         if(expoly->p != 0.0)
         {
	    /*********************************************
            *** compute part of alpha_f_factor (k > 0) ***
	    *** corresponding to point mass            ***
            *********************************************/

  	    mass_cnt += 1;

#           ifdef DEBUG_K
               cout << "[mass_cnt = " << mass_cnt << "]" << endl;
     	       cout << "<mass> <old> beta1_mass_k : "
		    << beta1_mass_k[mass_cnt] << endl;
#           endif

  	    /*** current poisson prob. ***/
            beta1_mass_k[mass_cnt] *= (q * x1) / kk;

            /*** add to current factor **/
            alpha_f_factor_RR += p * beta1_mass_k[mass_cnt];

#           ifdef DEBUG_K
     	       cout << "<mass> <new> beta1_mass_k : "
		    << beta1_mass_k[mass_cnt] << endl;
     	       cout << "<mass> add to alpha_f[k = " << k << "] : "
                    << p * beta1_mass_k[mass_cnt] << endl;
#           endif
         }

         if( !expoly->next )
            break;

	 /*** for all power terms ... ***/
         for( pow_m = expoly->power; pow_m; pow_m = pow_m->next )
         {
            m = pow_m->m;                     // exponent (unsigned)
            mm = to_RR((long)(m));            // exponent (RR)

	    /*** for all exponential terms ... ***/
            for( exp_n = pow_m->expo; exp_n; exp_n = exp_n->next )
            {
  	       lambda = to_RR(exp_n->lambda); // rate (RR)
               b      = to_RR(exp_n->b);      // const. corresp. to 1-PDF (RR)
               c      = to_RR(exp_n->c);      // const. corresp. to pmf (RR)

#              ifdef DEBUG_EXPOLY
                  cout << "[m = " << m << ", lambda = " << lambda
                       << ", b = " << b << ", c = " << c << "]" << endl;
#              endif

	       if(exp_n->c != 0.0)
	       {
   	          /*********************************************
	          *** compute part of alpha_f_factor (k > 0) ***
	          *** corresponding to expolynom. distribut. ***
	          *********************************************/

 		  blck_cnt += 1;

#                 ifdef DEBUG_K
                     cout << "[blck_cnt = " << blck_cnt << "]" << endl;
	             cout << "<k> <old> current beta1  : "
		          << beta1_k[blck_cnt] << endl;
	             cout << "<k> <old> current beta2  : "
		          << beta2_k[blck_cnt] << endl;
	             cout << "<k> <old> current eta1   : "
		          << eta1_k[blck_cnt] << endl;
	             cout << "<k> <old> current eta2   : "
		          << eta2_k[blck_cnt] << endl;
		     cout << "<k> <old> current gamma1 : "
		          << gamma1_k[blck_cnt] << endl;
	             cout << "<k> <old> current gamma2 : "
		          << gamma2_k[blck_cnt] << endl;
#                 endif

                  /*** useful calculations ***/
                  qpl = q + lambda;
                  qplx1 = qpl * x1;
                  qplx2 = qpl * x2;

                  tmp = (kk + mm) / kk;
                  tmp *= q / qpl;

	          /*** current poisson probs. ***/
	          beta1_k[blck_cnt] *= (qplx1) / kk;
	          beta2_k[blck_cnt] *= (qplx2) / kk;

	          /*** current etas ***/
	          eta1_k[blck_cnt] *= q / qpl;
	          eta2_k[blck_cnt] *= q / qpl;

	          /*** current gammas ***/
	          gamma1_k[blck_cnt] *= tmp;
	          gamma2_k[blck_cnt] *= tmp;

	          gamma1_k[blck_cnt] -= eta1_k[blck_cnt] * beta1_k[blck_cnt];
	          gamma2_k[blck_cnt] -= eta2_k[blck_cnt] * beta2_k[blck_cnt];

                  diff = gamma2_k[blck_cnt] - gamma1_k[blck_cnt];

	          /*** add to current factor ***/
                  alpha_f_factor_RR += c * diff;

#                 ifdef DEBUG_K
	             cout << "<k> <new> current beta1  : "
		          << beta1_k[blck_cnt] << endl;
	             cout << "<k> <new> current beta2  : "
		          << beta2_k[blck_cnt] << endl;
	             cout << "<k> <new> current eta1   : "
		          << eta1_k[blck_cnt] << endl;
	             cout << "<k> <new> current eta2   : "
		          << eta2_k[blck_cnt] << endl;
		     cout << "<k> <new> current gamma1 : "
		          << gamma1_k[blck_cnt] << endl;
	             cout << "<k> <new> current gamma2 : "
		          << gamma2_k[blck_cnt] << endl;
	             cout << "<k> current gamma diff   : "
		          << diff << endl;
     	             cout << "<k> add to alpha_f[k = " << k << "] : "
                          << c * diff << endl;
#                 endif
	       }

	    } // for ... exp_n
         } // for ... pow_m
      } // for ... expoly


      /*** current sum of alpha_f_factors ***/
      alpha_f_sum += alpha_f_factor_RR;

      /*** current alpha_F_bar_factor ***/
      alpha_F_bar_factor_RR = (1.0 - alpha_f_sum) / q;

      /*** current errors ***/
      alpha_f_error_RR     -= alpha_f_factor_RR;
      alpha_F_bar_error_RR -= alpha_F_bar_factor_RR;

#     ifdef DEBUG_ERRORS
         cout << "<k> current alpha_f_error_RR     : " 
	      << alpha_f_error_RR << endl;
         cout << "<k> current alpha_F_bar_error_RR : "
	      << alpha_F_bar_error_RR << endl;
#     endif

      /*** convertion of current factors ***/
      conv(alpha_f_factor,     alpha_f_factor_RR);
      conv(alpha_F_bar_factor, alpha_F_bar_factor_RR);

      /*** push current factors to vectors of factors ***/
      alpha_f_factors.push_back(alpha_f_factor);
      alpha_F_bar_factors.push_back(alpha_F_bar_factor);

      k++;
   }


   /*** convertion of errors ***/
   conv(alpha_f_error,     alpha_f_error_RR);
   conv(alpha_F_bar_error, alpha_F_bar_error_RR);

   /*** right truncation point ***/
   R = k - 1;
}


/****************************************************************************
** FUNCTION : print factors (C++ vectors)                                  **
*****************************************************************************
** Purpose  : print the computed alpha factors (C++ vectors) and errors    **
**            compute the approximations of alpha factors                  **
**                                                                         **
** Input    : alpha_f_factors       vector of factors with transient probs.**
**	      alpha_F_bar_factors   vector of factors with sojourn times   **
**  	      R                     right truncation point                 **
**            alpha_f_error         error of alpha_f_factors               **
**	      alpha_F_bar_error     error of alpha_F_bar_factors           **
**                                                                         **
** Output   : nothing                                                      **
****************************************************************************/

void print_factors_CPP(vector<double> alpha_f_factors,
		       vector<double> alpha_F_bar_factors,
		       unsigned R,
		       double alpha_f_error, double alpha_F_bar_error)
{
   double alpha_f_approx = 0.0,
          alpha_F_bar_approx = 0.0;

   /*** get initial output precision ***/
   streamsize InitBits = cout.precision();

   /*** set output precision ***/
   cout.precision(OUTBITS);

   cout << "------------------------------------------------------------------"
        << endl;
   cout << "*** print alpha_f and alpha_F_bar factors, R = " << R << " ***"
        << endl;
   cout << "alpha_f_error = " << alpha_f_error
        << "  alpha_F_bar_error = " << alpha_F_bar_error << endl;
   cout << "------------------------------------------------------------------"
        << endl;

   cout.setf(ios::fixed, ios::floatfield);

   for(unsigned k = 0; k <= R; k++)
   {
      alpha_f_approx     += alpha_f_factors[k];
      alpha_F_bar_approx += alpha_F_bar_factors[k];

      cout << "k = " << k
           << "  alpha_f_factor = " << alpha_f_factors[k]
           << "  alpha_F_bar_factor = " << alpha_F_bar_factors[k] << endl;
   }

   cout.setf(0, ios::floatfield);

   cout << "------------------------------------------------------------------"
        << endl;
   cout << "alpha_f_approx = " << alpha_f_approx
        << "  alpha_F_bar_approx = " << alpha_F_bar_approx << endl;
   cout << "------------------------------------------------------------------"
        << endl;
   cout << endl;

   /*** set default output precision ***/
   cout.precision(InitBits);
}


/****************************************************************************
** FUNCTION : calculate factors (C arrays)                                 **
*****************************************************************************
** Purpose  : interface for calculation of alpha factors needed for each   **
**            erand process                                                **
**            convert factors to array of doubles in C format              **
**                                                                         **
** Input    : ArbBits               Bits for arbitrary precision           **
**            expoly                initial internal representation of     **
**                                  expolynomial distribution              **
**	      TruncError            truncation error for factors           **
**	      q                     max. diag. entry of generator matrix   **
**                                                                         **
** Output   : alpha_f_factors       array of factors with transient probs. **
**	      alpha_F_bar_factors   array of factors with sojourn times    **
**  	      R                     right truncation point                 **
**            alpha_f_error         calculation error of alpha_f factors   **
**	      alpha_F_bar_error     calculation error of alpha_F_bar fact. **
****************************************************************************/

void calculate_factors_C(long ArbBits, time_l *expoly,
                         double TruncError, double q,
                         double **alpha_f_factors,
                         double **alpha_F_bar_factors, unsigned *R,
                         double *alpha_f_error, double *alpha_F_bar_error)
{
#  ifdef DEBUG
      print_expoly(expoly);
      cout << endl;
#  endif

   if ( ArbBits == 0 )
   {
#     ifdef DEBUG
         cout << "Calculate factors with precision of double ..." << endl;
#     endif

      /*** compute arrays of alpha factors with precision of double ***/
      calculate_factors(expoly, TruncError, q,
	  	        alpha_f_factors, alpha_F_bar_factors, R,
	                alpha_f_error, alpha_F_bar_error);
   }
   else
   {
#     ifdef DEBUG
         cout << "Calculate factors with arbitrary precision ..." << endl;
#     endif

      /*** tmp. vectors ***/
      vector<double> alpha_f_tmp;
      vector<double> alpha_F_bar_tmp;

      double *alpha_f_ptr;
      double *alpha_F_bar_ptr;

      /*** compute tmp. vectors of alpha factors with arb. prec. ArbBits ***/
      calculate_factors_RR(ArbBits, expoly, TruncError, q,
	                   alpha_f_tmp, alpha_F_bar_tmp,
	                   *R, *alpha_f_error, *alpha_F_bar_error);

      /*** convert tmp. vectors to arrays ***/
      alpha_f_ptr = new double[*R + 1];
      alpha_F_bar_ptr = new double[*R + 1];

      for(unsigned k = 0; k <= *R; ++k)
      {
         alpha_f_ptr[k] = alpha_f_tmp[k];
         alpha_F_bar_ptr[k] = alpha_F_bar_tmp[k];
      }
      *alpha_f_factors = alpha_f_ptr; 
      *alpha_F_bar_factors = alpha_F_bar_ptr; 
   }

#  ifdef DEBUG
      print_factors(*alpha_f_factors, *alpha_F_bar_factors, *R,
	            *alpha_f_error, *alpha_F_bar_error);
#  else
      cout << "Calulated alpha factors for current transition ..." << endl;
      cout << "Right truncation point : "  << *R << ", ";
      cout << "alpha_f_error : "     << *alpha_f_error << ", ";
      cout << "alpha_F_bar_error : " << *alpha_F_bar_error << endl;
      cout << endl;
#  endif
}



/****************************************************************************
** FUNCTION : calculate factors (C++ vectors)                              **
*****************************************************************************
** Purpose  : interface for calculation of alpha factors needed for each   **
**            inner iteration (jensen's method)                            ** 
**            convert factors to vectors of doubles in C++ format          **
**                                                                         **
** Input    : ArbBits               Bits for arbitrary precision           **
**            expoly                initial internal representation of     **
**                                  expolynomial distribution              **
**	      TruncError            truncation error for factors           **
**	      q                     max. diag. entry of generator matrix   **
**                                                                         **
** Output   : alpha_f_factors       vector of factors with transient probs.**
**	      alpha_F_bar_factors   vector of factors with sojourn times   **
**  	      R                     right truncation point                 **
**            alpha_f_error         calculation error of alpha_f factors   **
**	      alpha_F_bar_error     calculation error of alpha_F_bar fact. **
****************************************************************************/

void calculate_factors_CPP(long ArbBits, time_l *expoly,
                           double TruncError, double q,
                           vector<double> &alpha_f_factors,
		           vector<double> &alpha_F_bar_factors, unsigned &R,
		           double &alpha_f_error, double &alpha_F_bar_error)
{
#  ifdef DEBUG
      print_expoly(expoly);
      cout << endl;
#  endif

   if ( ArbBits == 0 )
   {
#     ifdef DEBUG
         cout << "Calculate factors with precision of double ..." << endl;
#     endif

      /*** tmp. array ***/
      double *alpha_f_tmp;
      double *alpha_F_bar_tmp;

      /*** compute tmp. arrays of alpha factors with prec. of double ***/
      calculate_factors(expoly, TruncError, q,
	  	        &alpha_f_tmp, &alpha_F_bar_tmp, &R,
	                &alpha_f_error, &alpha_F_bar_error);

      /*** convert tmp. arrays to vectors ***/
      alpha_f_factors = vector<double>(&alpha_f_tmp[0], &alpha_f_tmp[R + 1]);
      alpha_F_bar_factors = vector<double>(&alpha_F_bar_tmp[0],
				           &alpha_F_bar_tmp[R + 1]);

      /*** clear tmp. arrays ***/
      delete alpha_f_tmp;
      delete alpha_F_bar_tmp;
   }
   else
   {
#     ifdef DEBUG
         cout << "Calculate factors with arbitrary precision ..." << endl;
#     endif

      /*** compute vectors of alpha factors with arbitrary prec. ArbBits ***/
      calculate_factors_RR(ArbBits, expoly, TruncError, q,
	 	           alpha_f_factors, alpha_F_bar_factors, R,
                           alpha_f_error, alpha_F_bar_error);
   }

#  ifdef DEBUG
      print_factors_CPP(alpha_f_factors, alpha_F_bar_factors, R,
                        alpha_f_error, alpha_F_bar_error);
#  else
      cout << "Calulated alpha factors for current transition  ..." << endl;
      cout << "Right truncation point : "  << R << ", ";
      cout << "alpha_f_error : "     << alpha_f_error << ", ";
      cout << "alpha_F_bar_error : " << alpha_F_bar_error << endl;
      cout << endl;
#  endif
}
