//#define DEBUG_QA
//#define CONV_CHECK 10
/****************************************************************************
** Evaluation : STATIONARY ANALYSIS                                        **
*****************************************************************************
** Program    : solve                                                      **
** Component  : solve_gen                                                  **
** File       : solve_gen.cpp (C++ source)                                 **
**                                                                         **
** Purpose    : implementations of member functions of the structure       **
**              GenMarkingsSmc for the SMCs of each general transition     **
**              which are used by the iterative solution method with       **
**              fill-in avoidance                                          **
**                                                                         **
** Master's T.: realis. of an iterative solution method for not marc. SPNs **
**                                                                         **
** Programmer : Oliver Hesse                                               **
**                                                                         **
** Date       : june of 1999                                               **
****************************************************************************/

/****************************************************************************
** LIST OF IMPORTS                                                         **
****************************************************************************/

#include <vector>

#include "solve_gen.h"            // struct. GenMarkingsSmc and GenMarkings
#include "solve_factors.h"        // calculate factors with arbitrary prec.
#include "vmmath.h"               // vector and matrics representations
                                  // and operations
#include "Time.h"                 // time and memory usage

extern "C" {
#include "parsdfunc.h"            // calling of parse_pmf_def()
}


/****************************************************************************
** DEFINITIONS                                                             **
****************************************************************************/

//#define DEBUG_QA                // show calc. of q and A
//#define DEBUG_ITER              // show number of inner iter.
//#define DEBUG_DIFF              // show diff. of phi(k), phi(k-1)


/****************************************************************************
** GLOBAL VARIABLE 'GEN'                                                   **
****************************************************************************/

vector<GenMarkingsSmc> Gen;        // 'Gen' is a set of 'gen's


/****************************************************************************
** CONSTRUCTOR OF 'GEN MARKINGS SMC'                                       **
****************************************************************************/

GenMarkingsSmc::GenMarkingsSmc(int dim)
   : I(dim), D_tau(SparseVector(dim)), Q(dim,dim), Q_bar(dim,dim), A(dim,dim),
     DELTA(dim,dim), alpha_f(0), alpha_F_bar(0), R(0) {}


/****************************************************************************
** MEMBER FUNCTIONS OF 'GEN MARKINGS SMC'                                  **
****************************************************************************/

#ifdef PVM_SUPPORT

void GenMarkingsSmc::estimateEffort()
{
    effort = (A.size()+Q_bar.size()) * (q + 50); // Experience value
    cout << "A:"<<A.size() << "\tQbar:"<<Q_bar.size() << "\tq:"<<q 
	 << "\tEst. effort:" << effort << endl;
}

bool GenMarkingsSmc::operator< (const GenMarkingsSmc& g) const
{
  return effort > g.effort;
}

GenMarkingsSmc::GenMarkingsSmc(PVM_UNPACK_t)
  : name(pvm_upkstring()),
    pmf_def(pvm_upkstring()),
    I(PVM_UNPACK),
    D_tau(SparseVector(PVM_UNPACK)),
    Q(PVM_UNPACK),
    Q_bar(PVM_UNPACK),
    A(PVM_UNPACK),
    DELTA(PVM_UNPACK),
    R(pvm_upkuint())
{ q = pvm_upkdouble(); }

void GenMarkingsSmc::pvm_pack()
{
   pvm_pkstring(name);
   pvm_pkstring(pmf_def);
   I.pvm_pack();
   SparseVector tmp = D_tau.getDiagVector(); // Ensure format
   tmp.pvm_pack();                           // DiagMatrix(SparseVector)
   Q.pvm_pack();
   Q_bar.pvm_pack();
   A.pvm_pack();
   DELTA.pvm_pack();
   pvm_pkuint(R);
   pvm_pkdouble(q);
}

GenMarkingsSmc GenMarkingsSmc::pvm_unpack()
{ return GenMarkingsSmc(PVM_UNPACK); }

#endif


/****************************************************************************
** FUNCTION : get 'dim'                                                    **
*****************************************************************************
** Purpose  : get the dimension for a vector(dim) or a matrix(dim, dim)    **
**            independence of the declared dim in class SolveBase          **
**            (needed for distributed execution of the solve process)      **
**                                                                         **
** Member   : GenMarkingsSmc ('gen')                                       **
**                                                                         **
** Input    : gen.Q (hidden)  any declared matrix                          **
**                                                                         **
** Output   : dim             the dimens. is equal to the number of states **
****************************************************************************/

int GenMarkingsSmc::getDim() const
{
   return Q.getCols();
}


/****************************************************************************
** FUNCTION : calculate 'maxDiag'                                          **
*****************************************************************************
** Purpose  : calculation of the maximal value 'max' of all diagonal       **
**            entries of any matrix 'M'                                    **
**                                                                         **
** Input    : M             any matrix M                                   **
**                                                                         **
** Output   : max           maximal value of all diagonal entries qii      **
****************************************************************************/

double maxDiag(const Matrix &M)
{
   double tmp, max = 0.0;

   /*** for each row ***/
   for(int i = 0; i < M.getRows(); ++i)
   {
      if((tmp = fabs(M(i,i))) > max)
         max = tmp;
   }

   return max;
}


/****************************************************************************
** FUNCTION : calculate 'q' and 'A'                                        **
*****************************************************************************
** Purpose  : calculation of the normalisized matrix A and the maximal     **
**            diagonal entry q (renewal rate) of the generator matrix Q    **
**                                                                         **
** Member   : GenMarkingsSmc ('gen')                                       **
**                                                                         **
** Formula  : gen.A = gen.Q / gen.q + gen.I (gen.q != 0.0)                 **
**            gen.A = gen.I                 (gen.q = 0.0)                  **
**                                                                         **
** Input    : gen.Q (hidden)    generator Matrix Q                         **
**            gen.I (hidden)    filter matrix I                            **
**                                                                         **
** Output   : gen.q (hidden)    renewal rate q >= |qii max|                **
**            gen.A (hidden)    normalisized generator matix A             **
**            gen.Q (hidden)    cleared because not used anymore           **
****************************************************************************/

void GenMarkingsSmc::calculate_q_and_A()
{
   if ((q = maxDiag(Q) * 1.02) == 0.0)
      A = I;
   else
   {
      A = Q;
      Q.clear();
      A /= q;
      A += I;
   }   

#  ifdef DEBUG_QA
      cout << "maximal diagonal entry q : " << q << endl;
      cout << "normalized generatormatrix A : " << endl << A << endl;
#  endif
}


/****************************************************************************
** FUNCTION : calculate 'alphas'                                           **
*****************************************************************************
** Purpose  : call of calculation of alpha - factors by using the pre-     **
**            specified truncation error and the internal representation   **
**            of the expolynomial distribution                             **
**            separation is necessary for the parallel solve process       **
**                                                                         **
** Member   : GenMarkingsSmc ('gen')                                       **
**                                                                         **
** Input    : TruncError                  prespecified truncation error    **
**            ArbBits                     prespecified number of bits      **
**            gen.pmf_def     (hidden)    def. of probab. mass function    **
**                                                                         **
** Output   : gen.alpha_f     (hidden)    vector of alpha_f factors        **
**            gen.alpha_F_bar (hidden)    vector of alpha_F_bar factors    **
**            gen.R           (hidden)    right truncation point           **
****************************************************************************/

void GenMarkingsSmc::calculate_alphas(double TruncError, long ArbBits)
{
  time_l     *expoly;                   
  double      alpha_f_error = 0.0;    
  double      alpha_F_bar_error = 0.0;

  /*** get the expolynomial distribution of gen.pmf_def ***/
  expoly = parse_pmf_def(pmf_def.c_str()); 

  /*** compute vectors of alpha factors ***/
  calculate_factors_CPP(ArbBits, expoly, TruncError, q,
	                alpha_f, alpha_F_bar, R,
                        alpha_f_error, alpha_F_bar_error);
}


/****************************************************************************
** FUNCTION : jensens method                                               **
*****************************************************************************
** Purpose  : Jensen's method (inner iteration) calculates a(m) and b(m)   **
**            a special algorithm is used if phi(k-1) is nearly equal to   **
**            phi(k) (reached convergence)                                 **
**                                                                         **
** Member   : GenMarkingsSmc ('gen')                                       **
**                                                                         **
** Formulas : gen_phi(k) = gen_u(m) * (gen.A) ^ k                          **
**            gen_a(m) = Sum of [gen_phi(k) * gen.alpha_f(k,q)]            **
**            gen_b(m) = Sum of [gen_phi(k) * gen.alpha_F_bar(k,q)]        **
**            (0 <= k <= R)                                                **
**                                                                         **
** Input    : gen_u                       current vector gen_u(m)          **
**            Epsilon                     accuracy of vector equality      **
**            gen.A           (hidden)    norm. generator matrix A         **
**            gen.alpha_f     (hidden)    vector of alpha_f factors        **
**            gen.alpha_F_bar (hidden)    vector of alpha_F_bar factors    **
**            gen.R           (hidden)    right truncation point           **
**                                                                         **
** Output   : gen_a                       current vector gen_a(m)          **
**            gen_b                       current vector gen_b(m)          **
****************************************************************************/

void GenMarkingsSmc::jensens_method_a_b(const SparseVector& gen_u,
				              SparseVector &gen_a,
                                              SparseVector &gen_b)
{
   SparseVector gen_phi_k_new(gen_u); 

   /*** initial a(m) and b(m) ***/   
   gen_a = gen_phi_k_new * alpha_f[0];
   gen_b = gen_phi_k_new * alpha_F_bar[0];

   /*** iterate until max. number of alpha factors reached ***/
   for (unsigned k = 1; k <= R; ++k)
   {
         /*** current phi(k) ***/
         gen_phi_k_new *= A;

	 /*** current a(m) and b(m) ***/   
	 gen_a += gen_phi_k_new * alpha_f[k];
	 gen_b += gen_phi_k_new * alpha_F_bar[k];
   }
}


void GenMarkingsSmc::jensens_method_a(const SparseVector& gen_u,
				            SparseVector &gen_a)
{
   SparseVector gen_phi_k_new(gen_u); 

   /*** initial a(m) ***/   
   gen_a = gen_phi_k_new * alpha_f[0];

   /*** iterate until max. number of alpha factors reached ***/
   for (unsigned k = 1; k <= R; ++k)
   {
         /*** current phi(k) ***/
         gen_phi_k_new *= A;

	 /*** current a(m) ***/   
	 gen_a += gen_phi_k_new * alpha_f[k];
   }
}



#if 0
/****************************************************************************
** FUNCTION : jensens method DYNAMIC                                       **
****************************************************************************/

void GenMarkingsSmc::jensens_method_dynamic(const SparseVector& gen_u,
                                            double EpsInner,
                                            SparseVector &gen_a,
                                            SparseVector &gen_b)
{
   SparseVector   gen_phi_k_new(getDim()); // gen_phi(k) of current iter.
   SparseVector   gen_phi_k_old(getDim()); // gen_phi(k-1) of last iter.

   double         alpha_f_sum = 0.0;
   double         alpha_F_bar_sum = 0.0;

   unsigned       k;

   /*** initial phi(0) ***/
   gen_phi_k_new = gen_u;                  // start with phi(0) = gen_u(m)
   
   /*** initial a(m) and b(m) ***/   
   gen_a = gen_phi_k_new * alpha_f[0];
   gen_b = gen_phi_k_new * alpha_F_bar[0];


   unsigned min = (k_min < R) ? k_min : R;

   /*** iterate until k_min reached without test of convergence ***/
   for (k = 1; k <= min; ++k)
   {
         /*** hold phi(k-1) ***/
         gen_phi_k_old = gen_phi_k_new;

         /*** current phi(k) ***/
         gen_phi_k_new *= A;

	 /*** current a(m) and b(m) ***/   
	 gen_a += gen_phi_k_new * alpha_f[k];
	 gen_b += gen_phi_k_new * alpha_F_bar[k];
   }


   /*** iterate from k_min until max. number of alpha factors reached ***/
   for (; k <= R; ++k)
   {
      /*** check if convergence between phi(k) and phi(k-1) is reached ***/
      if(near(gen_phi_k_old, gen_phi_k_new, EpsInner))
      {
	 //cout << "convergence reached, at k = " << k << endl;
	 break;
      }

      /*** hold phi(k-1) ***/
      gen_phi_k_old = gen_phi_k_new;

      /*** current phi(k) ***/
      gen_phi_k_new *= A;

      /*** current a(m) and b(m) ***/   
      gen_a += gen_phi_k_new * alpha_f[k];
      gen_b += gen_phi_k_new * alpha_F_bar[k];
   }

   k_min = k - 1;


   /*** sum up "left over" factors with converted phi(k) ***/
   for (; k <= R; ++k)
   {
      /*** current a(m) and b(m) ***/   
      gen_a += gen_phi_k_new * alpha_f[k];
      gen_b += gen_phi_k_new * alpha_F_bar[k];
   }

   //cout << "R = " << R << endl;
}



/****************************************************************************
** FUNCTION : jensens method ORIGINAL                                      **
****************************************************************************/

void GenMarkingsSmc::jensens_method_orig(const SparseVector& gen_u,
                                         double Epsilon,
                                         SparseVector &gen_a,
                                         SparseVector &gen_b)
{
   SparseVector   gen_phi_k_new(getDim()); // gen_phi(k) of current iter.

#ifdef CONV_CHECK
   SparseVector   gen_phi_k_old(getDim()); // gen_phi(k-1) of last iter.

   double         alpha_f_sum = 0.0;
   double         alpha_F_bar_sum = 0.0;

   const unsigned stepSize = CONV_CHECK;
   unsigned       checkPoint = stepSize;   // next converg. check while iter.
   bool           conv_reached = false;    // is phi(k-1) ~= phi(k)
#endif

   /*** initial phi(0) ***/
   gen_phi_k_new = gen_u;                 // start with phi(0) = gen_u(m)
   
   /*** initial a(m) and b(m) ***/   
   gen_a = gen_phi_k_new * alpha_f[0];
   gen_b = gen_phi_k_new * alpha_F_bar[0];

#  ifdef DEBUG_ITER
      cout << endl;
      cout << "current transition : " << name  << endl;
      cout << endl;
      cout << "===== inner iteration (0) =====" << endl;
      cout << "<inner> phi(0) = gen_u(m)       : " << gen_phi_k_new << endl;
      cout << "<inner> phi(0) * alpha_f(0)     : " << gen_a << endl;
      cout << "<inner> phi(0) * alpha_F_bar(0) : " << gen_b << endl;
#  endif

   /*** iterate until max. number of alpha factors reached ***/
   for (unsigned k = 1; k <= R; ++k)
   {
#  ifdef DEBUG_ITER
      cout << "===== inner iteration (" << k << ") =====" << endl;
#  endif

#  ifdef CONV_CHECK
      /*** check if convergence between phi(k) and phi(k-1) is reached ***/
      if (k >= checkPoint)
      {
         if(near(gen_phi_k_old, gen_phi_k_new, Epsilon))
	 {
            conv_reached = true;
            checkPoint = R + 1;

            //cout << "convergence reached, at k = " << k << endl;

	 }
         else
            checkPoint += stepSize;
      }

      /*** do next iteration in case of no convergence ***/
      if (!conv_reached)
      {
         /*** hold phi(k-1) ***/
         gen_phi_k_old = gen_phi_k_new;
#  endif

         /*** current phi(k) ***/
         gen_phi_k_new *= A;

	 /*** current a(m) and b(m) ***/   
	 gen_a += gen_phi_k_new * alpha_f[k];
	 gen_b += gen_phi_k_new * alpha_F_bar[k];

#        ifdef DEBUG_ITER
            cout << "<inner> phi(k) *= A             : "
                 << gen_phi_k_new << endl;
            cout << "<inner> phi(k) * alpha_f(k)     : "
                 << gen_phi_k_new * alpha_f[k] << endl;
            cout << "<inner> phi(k) * alpha_F_bar(k) : "
                 << gen_phi_k_new * alpha_F_bar[k] << endl;
#        endif

#  ifdef CONV_CHECK
      }
      else
      {
         /*** sum up "left over" factors in case of reached convergence ***/   
         alpha_f_sum     += alpha_f[k];
         alpha_F_bar_sum += alpha_F_bar[k];

#        ifdef DEBUG_ITER
            cout << "<inner> phi(k) convergented     : "
                 << gen_phi_k_new << endl;
#        endif
      }
#     ifdef DEBUG_DIFF
         cout << "<inner> max. diff. of phi(" << k << ") and phi(last) : "
              << maxDiff(gen_phi_k_old, gen_phi_k_new) << endl;
#     endif
#  endif
   }

#ifdef CONV_CHECK
   /*** sum up "left over" iterations in case of convergence ***/
   if (conv_reached)
   {
      gen_a += gen_phi_k_new * alpha_f_sum;
      gen_b += gen_phi_k_new * alpha_F_bar_sum;
   }
#endif

   //cout << "R = " << R << endl;
}

#endif
