// File      solve_parallel.cpp (C++ Implementation)
// Component solve_parallel     (Master-side of parallel FIA Analysis)
// Program   graph              (Stationary Analysis of SPN)
/**
 *@author  Philipp Berndt
 *@version $Id: solve_parallel.cpp,v 1.2 2000/01/12 16:02:31 pumphutt Exp $
 */

//#define DEBUG
//#define DEBUG_U
//#define SLOMO
//#define PROGRESSIND

#include <string>
#include <iostream>
#include <iomanip>
#include <cassert>
#include <vector>
#include <algorithm>  // sort
#include <unistd.h>   // close, dup

#include "pvm3cpp.h"
#include "Time.h"
#include "solve_parallel.h"

#include "solve_exp.h"
#include "solve_gen.h"
#include "solve_base.h"

namespace
{
    const int RESPAWN_LIMIT = 100; // Maximum number of times for restarting
    const double MASTERLOAD = 1.5; // Virtual load of the master process
    const double SLAVELOAD  = 0.7; // Virtual load of one slave  process

    vector<int>    slavetids;   // PVM Task ID array for each slave
    vector<string> hosts;       // Hostname          for each slave
    vector<int>    deaths;      // Needed respawns   for each slave 
    vector<int>    position;    // Position table of slaves
                            // (receive order of results in the last iteration)
    SolveBase *sb;       // All we need to know is here.
    DenseVector u(0);    // Iteration Vector
    int noslaves;        // Number of independent SMC Components
    class Exception {};
    bool iteration_done = false; 
    int dim;             // Dimension of all Vectors and Matrices
    Time outer_time;     // Outer iteration real time
    Time outer_time_cpu; // Outer iteration  CPU time
    int iter = 0;        // Iteration counter
}


/***************************** Utility functions *****************************/

/** Discards all pending TAG_U_NEW messages from incoming message queue. */
void clearQueue()
{
    while(pvm_nrecv(-1, TAG_U_NEW)) 
        cout << "Discarding stale messages from queue." << endl;
}

/** Sends the slave g its filtered part of the u vector. */
inline void send_u(int g)
{
    int tag = (iteration_done) ? TAG_U_STAT : TAG_U;
    u.filterAndSend(sb->Gen[g].I, slavetids[g], tag);
}

/**
 * Sends every slave its filtered part of the u vector.
 *
 * The sending order is the reverse receiving order of the results. 
 * This way the calculation which needs the most time is started first.
 */
inline void distrib_u()
{
    for (int i=noslaves-1; i>=0; --i)
        send_u(position[i]);
}

/**
 * Returns the Group Instance Number 
 * of the task with given PVM Task IDentifier (TID).
 */
int getgin(int tid)
{
    for (int gin=0; gin < noslaves; ++gin)
        if (slavetids[gin] == tid)
            return gin;
    return -1; // Unknown slave.
}

/** Returns true iff there are at least num file descriptors left. */
bool fdsAvail(int num)
{
    vector<int> fds(num);
    int i = 0;
    while ((fds[i] = dup(STDOUT_FILENO)) != -1)
        if (++i == num)
            break;
    
    for (int j=i; j > 0;)
        close(fds[--j]);
    return (i == num);
}


/**************************** Highlevel Functions ****************************/

/**
 * Spawns one slave and 
 * sends it it's specific initial data (not u^g).
 */
void spawnSlaveGen(int g)
{
    int slavetid=0;  // PVM Task ID of the new slave
    string host;   // Host to spawn slave on

    do {
        // Use Task-to-Task TCP-Links (faster) iff at least 2 fds left
        int newmode = fdsAvail(2) ? PvmRouteDirect : PvmDontRoute;
        int oldmode = pvm_setopt(PvmRoute, newmode);
        try {
            if (!HostConfig::idleHosts())
                HostConfig::addHost();
            host = HostConfig::getHost();

            cout << "Spawning slaveGen " << g 
                 << " " << sb->Gen[g].name << "\ton " << host 
                 << "..." << flush;
            
            vector<string> argv;

            if (pvm_getopt(PvmRoute) == PvmRouteDirect)
                cout << " direct " << flush;

            slavetid = pvm_spawn(
		(string(getenv("TNETHOME"))
		 +"/StatAnalysis/bin/solve_slave").c_str(), argv,
		PvmTaskHost, host);

            // Ask for notification when slave dies
            pvm_notify(PvmTaskExit, TAG_SLAVE_DIED, 1, &slavetid);
            
            pvm_initsend(PvmDataDefault);       // Initialize sendbuffer
            sb->Gen[g].pvm_pack();
            pvm_pkint(sb->ArbBits);
            pvm_pkdouble(sb->TruncError);
            pvm_pkdouble(sb->aperiodicFact);
            pvm_send(slavetid, TAG_INITGEN);    // Send data

            int status;
            if ((status = pvm_pstat(slavetid))<0)
                {
                cout << "pvm_pstat() = " << status << endl;
                throw PvmException("Slave died right after spawning");
                }
            }
        catch(PvmException e)
            {
            e.perror("spawnSlaveGen()");
            if (slavetid)
                {
                PVM3::pvm_kill(slavetid);
                slavetid = 0;
                }
            HostConfig::addVirtLoad(host, 1.0); // Penalty for host
            
            if (pvm_mstat(host.c_str()) < 0)
                HostConfig::update();
            }
        pvm_setopt(PvmRoute, oldmode);      // Restore routing mode
        } while (slavetid == 0);
        
    slavetids[g] = slavetid;
    hosts[g] = host;
    HostConfig::addVirtLoad(host, SLAVELOAD);
    cout << "\tOK. TID=" << hex<<slavetid<<dec << endl;
}

/**
 * Takes appropriate action when a slave has died.
 *
 * Usually respawns the slave and sends it the current iteration vector.
 */
void slaveDied()
{
    static int cnt=0;
    if (++cnt > RESPAWN_LIMIT)
        {
        cout << "Respawn limit exceded. Exiting." << endl;
        throw Exception();
        }

    // Find out who died
    int deadtid = pvm_upkint();
    int deadgin = getgin(deadtid); 
    cout << "Slave TID:" << hex<<deadtid<<dec 
         << ", GIN:" << deadgin 
         << " died." << endl;

    if (deadgin < 0) // Unknown slave ?
        return;

    // Update configuration
    ++deaths[deadgin];
    string host = hosts[deadgin];
    hosts[deadgin] = "";
    HostConfig::addVirtLoad(host, -SLAVELOAD);
    slavetids[deadgin] = 0;
    HostConfig::addVirtLoad(host, 0.1); // Penalty for host

    HostConfig::update();
    HostConfig::print(cout);

    spawnSlaveGen(deadgin); // Respawn.

    send_u(deadgin); // Send current iteration vector.
}

/** Collects TAG_CPU_TIME-Messages and prints statistics table. */
void collectCpuTimes()
{
    // Collect Slave CPU-Time messages
    cout << "Collecting CPU-Time messages\n" << endl;
    bool ack[noslaves];
    vector<Time> cpuTime(noslaves);   // Elapsed CPU-Time for each slave
    vector<long> usedMem(noslaves);   // Used memory      for each slave
    for (int i = noslaves;i;)
        {
        int bufid, bytes, msgtag, tid;
        bufid = pvm_recv(-1, -1);
        pvm_bufinfo(bufid, &bytes, &msgtag, &tid);
        if (checkSystemMessage(msgtag))
            continue;
            
        switch(msgtag)
            {
            case TAG_SLAVE_DIED:
                tid = pvm_upkint();
                if (!ack[getgin(tid)])
                    { // Slave died quietly
                    cout <<"GIN: "<< getgin(tid) << "died quietly." <<endl;
                    --i;
                    }
                break;
            case TAG_CPU_TIME:
                {
                --i;
                ack[getgin(tid)] = true;
                int gin = getgin(tid);
                if (gin >= 0)
		    {
                    double cnt = static_cast<double>(pvm_upkint());
                    cpuTime[gin] = Time(PVM_UNPACK)/(cnt/iter);
                    usedMem[gin] = pvm_upklong();
		    }
                break;
                }
            default:
                cerr << "Ignoring unknown Message " << msgtag
                     <<" from " << hex<<tid<<dec << endl;
            }
        }

    // print CPU times
    cout.fill(' ');
    cout << '\n'
         << "Process          CPU-Time      Memory       TID      Host\n"
         << "------------------------------------------------------------\n";
    Time time_sum = outer_time_cpu;
    long mem_sum = getUsedMem() - loadMem(sb->Netname, "atStartSol");
    cout << "Master        " << time_sum << "   " 
         << setw(6) << mem_sum <<" KB   "
         << setw(8) << hex << pvm_mytid() << dec << "   "
         << HostConfig::gethostname() << endl;

    for (int gin=0; gin <noslaves; ++gin)
        {
        cout << "SlaveGen" << setiosflags(ios::left) << setw(3) << gin
             << resetiosflags(ios::left)
             << "   " << cpuTime[gin]
             << "   " << setw(6) << usedMem[gin] << " KB   "
             << setw(8) << hex << slavetids[gin] << dec << "   "
             << hosts[gin] << endl;
        time_sum += cpuTime[gin];
        mem_sum += usedMem[gin];
        }
    cout << "-----------------------------------------------------------\n"
         << "Sum           " << time_sum << "   "
         << setw(6) << mem_sum << " KB\n" << endl;

    cout << "Real time:    " << outer_time << endl; 
    Time per_iteration = outer_time / iter; 
    cout << "Avg. real time per iteration: " << per_iteration << endl << endl;

    const int BYTES_PER_MBIT = 125000;
    int data     = (dim - sb->Exp.I.getOnesCount()) * 12;
    int overhead = noslaves * 48; // at least 48 (experience)
    double load = (overhead + data)* 2  /per_iteration.to_secs() 
        /BYTES_PER_MBIT;
    cout << "Avg. arith. network load: " << load << " MBit/s\n" << endl; 
}


/************************************ Main ***********************************/

void solve_parallel(SolveBase &SolveBase)
{
    //////////////////////////////////////////////////////// 1) INITIALIZATION

    sb       = &SolveBase;
    noslaves = sb->Gen.size();
    dim      = sb->Exp.P.getCols();

    // Redirect stderr to stdout: 2>&1
    close(STDERR_FILENO);
    dup(STDOUT_FILENO);

    Time qnA_start     = Time::getTimeOfDay();
    Time qnA_start_cpu = Time::getCPUTime();

    // Estimate effort and sort
    for (GenMarkings::iterator g = sb->Gen.begin(); g != sb->Gen.end(); ++g)
        {
        g->calculate_q_and_A();
        g->estimateEffort();
        }
    cout << "Sorting." << endl;
    sort(sb->Gen.begin(), sb->Gen.end());

    // Allocate vectors
    slavetids = vector<int>(noslaves);
    hosts     = vector<string>(noslaves);
    deaths    = vector<int>(noslaves);
    position  = vector<int>(noslaves);

    // Set starting positions according to est. effort (only for 1st iteration)
    for (int i=0;i<noslaves;++i)
        position[i] = noslaves-i -1;
 
    u = sb->get_u_init();    // Initialize iteration vector

    Time pvm_start     = Time::getTimeOfDay();
    Time pvm_start_cpu = Time::getCPUTime();

    cout << "Calculation of q and A : Elapsed CPU time  : "
	 << pvm_start_cpu - qnA_start_cpu << endl;
    cout << "Calculation of q and A : Elapsed real time : "
	 << pvm_start - qnA_start << endl;
    try {
        initPvm();

#ifdef DEBUG
        pvm_catchout(stdout);        // Display output from Slaves...
      //pvm_setopt(PvmShowTids,0);   // ...but not with prepended TID.
#else
        pvm_setopt(PvmOutputTid, 0); // Don't trace output
        pvm_setopt(PvmTraceTid, 0);
#endif

        // Mark this host as BUSY so no slave is started here.
        HostConfig::addVirtLoad(HostConfig::gethostname(), MASTERLOAD);  

        // Start all slaves
        for (int i=0;i<noslaves; ++i)
            spawnSlaveGen(i);

        Time outer_start = Time::getTimeOfDay();
        Time outer_start_cpu = Time::getCPUTime();
	cout << "Starting of slave processes : Elapsed CPU time  : "
	     << outer_start_cpu - pvm_start_cpu << endl;
	cout << "Starting of slave processes : Elapsed real time : "
	     << outer_start - pvm_start << endl;

        
	/////////////////////////////////////////////// 2) MAIN ITERATION LOOP

        cout << "\nEntering main iteration loop" << endl;
        do {
            clearQueue();

            // Send every slave its part of the u vector
            distrib_u();

#ifdef DEBUG_U
            cout << "Sent u:\t" << u <<endl;
#endif

            DenseVector u_new = sb->aperiodicFact * u * sb->Exp.P;
            u_new += (1 - sb->aperiodicFact) * (u * sb->Exp.I);

            // Collect u results from slaves
            for (int i=0; i<noslaves;)
                {
                int bufid, bytes, msgtag, sender;
                bufid = pvm_recv(-1, -1);
                pvm_bufinfo(bufid, &bytes, &msgtag, &sender);
                if (checkSystemMessage(msgtag))
                    continue;

                switch(msgtag)
                    {
                    case TAG_U_NEW:
			u_new.pvm_add_recv_SparseVector();
                        position[i] = getgin(sender);
                        ++i;
                        break;
                    case TAG_SLAVE_DIED:
                        slaveDied();
                        break;
                    }
                }

            // Check termination conditions
	    ++iter;
            if (near(u_new, u, sb->Epsilon))
                {
                cout << "Convergence after " 
                     << iter << " iterations." << endl;
                iteration_done = true;
                }
            else if (iter > sb->MaxIter)
                {
                cout << "No convergence after " 
                     << iter << " iterations." << endl;
                iteration_done = true;
                }
            if (iter % 100 == 0)
                {
                cout << "_" << iter << flush;
                if (iter % 1000 == 0)
		    {
                    sb->write_u(u);
		    cout << endl;
		    }
                }

#ifdef SLOMO
            sleep(1); // Give xpvm a chance to catch up.
#endif

            u = u_new; // Next iteration
            } while (!iteration_done);

        outer_time     = Time::getTimeOfDay() - outer_start;
        outer_time_cpu = Time::getCPUTime()   - outer_start_cpu;
    
        cout << "u_stat:\t"<< u << endl;


	////////////////////////////////////////////// 3) CALCULATION OF a & b

        clearQueue();
        distrib_u();
    
        // Collect a and b results from slaves
        DenseVector a(dim), b(dim);
        for (int i=0; i<noslaves;)
            {
            int bufid, bytes, msgtag, sender;
            bufid = pvm_recv(-1, -1);
            pvm_bufinfo(bufid, &bytes, &msgtag, &sender);
            if (checkSystemMessage(msgtag))
                continue;
            switch(msgtag)
                {
                case TAG_A_B:
                    a += SparseVector(PVM_UNPACK);
                    b += SparseVector(PVM_UNPACK);
                    ++i;
                    break;
                case TAG_SLAVE_DIED:
                    slaveDied();
                    break;
                default:
                    cerr << "Ignoring unknown Message " << msgtag
                         <<" from " << hex<<sender<<dec << endl;
                }
            }


	/////////////////////////////////////////////////////// 4) TERMINATION

        // Terminate slave processes
        cout << "Terminating slave processes" << endl;
        for (int i=0;i<noslaves; ++i)
            pvm_psend(slavetids[i], TAG_TERMINATE, 0, 0, PVM_INT);

        // Save final iteration vector
        if (sb->uSave)
            sb->write_u(u);
        else
	    sb->delete_u_file();

        // Calculate and save pi and phi
        sb->calculate_pi_phi(u, a, b);

        // Collect and print resource usage
        collectCpuTimes();
        }
    catch(PvmException e)
        {
        cerr << "Caught PvmException:";
        e.perror("solve_parallel()");
        cerr << "Aborting." << endl;
        sb->write_u(u);
        }
    catch(InterruptedException e)
        {
        cerr << "User Interrupt. Aborting" << endl;
        sb->write_u(u);
        }
    catch(Exception e)
        {
        cerr << "Caught Exception in solve_parallel(). Aborting." << endl;
        sb->write_u(u);
        }
    HostConfig::addVirtLoad(HostConfig::gethostname(), -MASTERLOAD);  
    shutdownPvm();
}
