/****************************************************************************
** Evaluation : STATIONARY ANALYSIS                                        **
*****************************************************************************
** Program    : solve                                                      **
** Component  : solve_sequential                                           **
** File       : solve_sequential.cpp (C++ source)                          **
**                                                                         **
** Purpose    : implementations of functions of the sequential iterative   **
**              solution method with fill-in avoidance                     **
**                                                                         **
** Master's T.: realis. of an iterative solution method for not marc. SPNs **
**                                                                         **
** Programmer : Oliver Hesse                                               **
**                                                                         **
** Date       : june of 1999                                               **
****************************************************************************/

/****************************************************************************
** LIST OF IMPORTS                                                         **
****************************************************************************/

#include <iostream>

#include "solve_base.h"           // class SolveBase
#include "solve_gen.h"            // struct. GenMarkinsSmc and GenMarkings
#include "solve_exp.h"            // struct. ExpMarkings
#include "Time.h"                 // time and memory usage


/****************************************************************************
** DEFINITIONS                                                             **
****************************************************************************/

//#define DEBUG_U                 // show current u(m)
#define DEBUG_U_STAT            // show stationary u(M)  
//#define DEBUG_DIFF              // show max. differeces of u(m), u(m - 1)
//#define DEBUG_EXP               // show exp. part of u(m)
//#define DEBUG_GEN               // show gen. parts of u(m)
//#define DEBUG_GEN_STAT          // show gen. parts of u(M)
#define SHOW_ALL_TIMES          // show all used times and memory usages


/****************************************************************************
** FUNCTION : outer iteration                                              **
*****************************************************************************
** Purpose  : sequential calculation of the stationary iteration vector    **
**            u(M) by using the power method                               **
**            calling of jensen's method for inner iterations              **
**            the aperiodic factor prevents alterniering behaviour of u(m) **
**                                                                         **
** Formula  : u(m+1) = u(m) * P                                            **
**            u(m+1) = u(m) * Exp.P + Gen.a(m) * Gen.DELTA                 **
**                                  + Gen.b(m) * Gen.Q_bar;                **
**                                                                         **
** Input    : SB            class SolveBase                                **
**                                                                         **
** Output   : u_stat        stationary iteration vector u(M)               **
****************************************************************************/
 
DenseVector outer_iteration(SolveBase& SB)
{
#  ifdef DEBUG_JM_TIME
      Time   cpu_startTime;                    // start of CPU time
      Time   day_startTime;                    // start of real time
#  endif


   DenseVector    u_old(SB.dim);         // u(m-1) of the last iteration
   DenseVector    u_new(SB.dim);         // u(m) of the current iteration
   SparseVector   gen_a(SB.dim);         // jensen's method vector gen_a(m)
   SparseVector   gen_b(SB.dim);         // jensen's method vector gen_b(m) 
   long           iter = 0;              // counter of iterations

   long           stepSize = 100;
   long           nextOutput = stepSize; // next output of current iter.
   unsigned       conv_reached = 0;      // is u(m-1) ~= u(m)

#  if 0
   double         EpsInner = 0.0;
   //double         EpsInner = 1.0;
   long           stepSizeEps = 1;
   long           nextEpsInner = stepSizeEps;
#  endif

#  ifdef DEBUG_GEN
      SparseVector  gen_a_sum(SB.dim);   // sum of all gen_a(m)
      SparseVector  gen_b_sum(SB.dim);   // sum of all gen_b(m)
#  endif

   /*****************************
   *** initialisation of u(0) ***
   *****************************/

   u_new = SB.get_u_init();              // start with u(m) = u(0)
   u_old = OnesVector(SB.dim);           // u(m-1) != u(m) at first

   cout << endl;
   cout << "Iterative calculation of u(M) ..." << endl;

   /*** do until precision or max. number of iterations is reached ***/
   while ((!conv_reached) && (iter < SB.MaxIter))
   {
      /*******************
      *** current u(m) ***
      *******************/ 

#     ifdef DEBUG_U
         cout << "===== outer iteration (" << iter << ") =====" << endl;
         cout << "<outer> current u(m) : " << u_new << endl;
         cout << "<outer> sum of all elements : "
              << u_new * OnesVector(SB.dim) << endl;
#     endif

#     ifdef DEBUG_DIFF
         cout << "<outer> max. diff. of u(" << iter << ") and u(last) : "
              << maxDiff(u_old, u_new) << endl;
#     endif

      /*** hold u(m-1) ***/
      u_old = u_new;

      /*****************************************
      *** calculate exponential part of u(m) ***
      *****************************************/ 

      u_new = SB.aperiodicFact * u_old * SB.Exp.P;
      u_new += (1 - SB.aperiodicFact) * u_old * SB.Exp.I;

#     ifdef DEBUG_EXP
         cout << "<outer> exp. part of u(m) : " << u_new << endl;
#     endif

      /**************************************
      *** calculate general parts of u(m) ***
      **************************************/ 

#     ifdef DEBUG_GEN
         gen_a_sum.clear();
         gen_b_sum.clear();
#     endif

      /*** for each general transition gen of Gen do ***/
      for(GenMarkings::iterator i = SB.Gen.begin(); i != SB.Gen.end(); ++i)
      {
         /*** current transition ***/
         GenMarkingsSmc  &gen = *i;


#        ifdef DEBUG_JM_TIME
            cpu_startTime = Time::getCPUTime();
            day_startTime = Time::getTimeOfDay();
#        endif

         if(gen.Q_bar.empty())
	 {
            /*** call jensen's method ***/
            gen.jensens_method_a((u_old * gen.I), gen_a);

            u_new += SB.aperiodicFact * gen_a * gen.DELTA;
            u_new += (1 - SB.aperiodicFact) * u_old * gen.I;
         }
         else
	 {
            /*** call jensen's method ***/
            gen.jensens_method_a_b((u_old*gen.I), gen_a, gen_b);
            //gen.jensens_method_dynamic((u_old*gen.I), EpsInner,gen_a,gen_b);

            u_new += SB.aperiodicFact * gen_a * gen.DELTA;
            u_new += SB.aperiodicFact * gen_b * gen.Q_bar;
            u_new += (1 - SB.aperiodicFact) * u_old * gen.I;

         }

#        ifdef DEBUG_JM_TIME
            cout << "Jensens Method : Elapsed CPU time  : "
                 << Time::getCPUTime() - cpu_startTime
                 << " / Elapsed real time : "
                 << Time::getTimeOfDay() - day_startTime << endl;
#        endif


#        ifdef DEBUG_GEN
            gen_a_sum += gen_a;
            gen_b_sum += gen_b;

            cout << "Current transition : " << gen.name << endl;
	    cout << "<outer> gen_a(m)         : " << gen_a << endl;
	    cout << "<outer> gen_b(m)         : " << gen_b << endl;
	    cout << "<outer> gen_a(m) * gen.DELTA : "
                 << gen_a * gen.DELTA << endl;
	    cout << "<outer> gen_b(m) * gen.Q_bar : "
                 << gen_b * gen.Q_bar << endl;
#        endif
      }

#     ifdef DEBUG_GEN
         cout << "<outer> sum of all gen_a(m) : " << gen_a_sum << endl;
         cout << "<outer> sum of all gen_b(m) : " << gen_b_sum << endl;
#     endif

      iter++;
      
      /******************
      *** checkpoints ***
      ******************/ 

      if (iter >= nextOutput)
      {
  	 cout << "_" << iter << flush;
         nextOutput += stepSize;
      }

#     if 0
      if (iter >= nextEpsInner)
      {
	 EpsInner = SB.Epsilon;
         //EpsInner = maxDiff(u_old, u_new)/10.0;

	 //cout << "EpsInner = " << EpsInner << endl;
         nextEpsInner += stepSizeEps;
      }
      else
      {
         /*** efficient check if u(m-1) ~= u(m) ***/
         if(near(u_old, u_new, SB.Epsilon))
            conv_reached = 1;
      }
#     else
      /*** efficient check if u(m-1) ~= u(m) ***/
      if(near(u_old, u_new, SB.Epsilon))
         conv_reached = 1;
#     endif
   }

   /************************
   *** show informations ***
   ************************/

   cout << endl;
   cout << endl;

   if (conv_reached)
      cout << "Convergence reached after " << iter << " iterations" << endl;
   else
      cout << "No convergence reached after " << iter << " iterations" << endl;

   cout << "Reached precision : " << maxDiff(u_old, u_new) << endl;

#  ifdef DEBUG_U_STAT
      cout << endl;
      cout.setf(ios::scientific, ios::floatfield);
      cout << "u_stat : " << endl << u_new << endl;
      cout.setf(0, ios::floatfield);
      cout << "Sum of all elements       : "
	   << u_new * OnesVector(SB.dim) << endl;
#  endif

   /***************************
   *** save stationary u(M) ***
   ***************************/

   if(SB.uSave)
      SB.write_u(u_new);

   /*** return u(M) = u_stat ***/
   return u_new;
}



/****************************************************************************
** FUNCTION : solve sequential                                             **
*****************************************************************************
** Purpose  : main function of the sequential iterative solution method    **
**            with fill-in avoidance                                       **
**                                                                         **
** Input    : SB            class SolveBase                                **
**                                                                         **
** Output   : Nothing                                                      **
****************************************************************************/

void solve_sequential(SolveBase& SB)
{
   DenseVector     u_stat(SB.dim);          // stationary vector u(M)
   SparseVector    gen_a_stat(SB.dim);      // vector gen_a(M)
   SparseVector    gen_b_stat(SB.dim);      // vector gen_b(M)
   DenseVector     Gen_a_stat(SB.dim);      // vector Gen_a(M)
   DenseVector     Gen_b_stat(SB.dim);      // vector Gen_b(M)

   Time   cpu_startTime;                    // start of CPU time
   Time   day_startTime;                    // start of real time
   long   mem_atStart;                      // used memory at start

   Time   cpu_diffTime;                     // difference of CPU time
   Time   day_diffTime;                     // defference of real time
   long   mem_diff;                         // difference of used memory

   /**********************
   *** initialisations ***
   **********************/

#  ifdef SHOW_ALL_TIMES
      Time   cpu_sumOfTimesQA = 0.0;
      Time   day_sumOfTimesQA = 0.0;
      long   mem_sumQA = 0;
#  endif

   Time   cpu_sumOfTimes = 0.0;
   Time   day_sumOfTimes = 0.0;
   long   mem_sum = 0;


   /*** for each general transition ***/
   for(GenMarkings::iterator i = SB.Gen.begin(); i != SB.Gen.end(); ++i)
   {
      /*** current transition ***/
      GenMarkingsSmc  &gen = *i;

#     ifdef SHOW_ALL_TIMES
         cpu_startTime = Time::getCPUTime();
         day_startTime = Time::getTimeOfDay();
         mem_atStart = getUsedMem();
#     endif


      /*** calculate gen.q and gen.A ***/
      gen.calculate_q_and_A();


#     ifdef SHOW_ALL_TIMES
         cpu_sumOfTimesQA += Time::getCPUTime() - cpu_startTime;
         day_sumOfTimesQA += Time::getTimeOfDay() - day_startTime;
         mem_sumQA += getUsedMem() - mem_atStart;
#     endif

      cpu_startTime = Time::getCPUTime();
      day_startTime = Time::getTimeOfDay();
      mem_atStart = getUsedMem();


      /*** calculate gen.alpha_f and gen.alpha_F_bar factors ***/
      gen.calculate_alphas(SB.TruncError, SB.ArbBits);
      gen.k_min = 10; 

      cpu_sumOfTimes += Time::getCPUTime() - cpu_startTime;
      day_sumOfTimes += Time::getTimeOfDay() - day_startTime;
      mem_sum += getUsedMem() - mem_atStart;
   }

#  ifdef SHOW_ALL_TIMES
      cout << endl;
      cout << "Calculation of q and A : Elapsed CPU time  : "
           << cpu_sumOfTimesQA << endl;
      cout << "Calculation of q and A : Elapsed real time : "
           << day_sumOfTimesQA << endl;
      cout << "Calculation of q and A : Required memory : "
           << mem_sumQA << " KB" << endl;
      cout << endl;
#  endif

   cout << "Calculation of all alpha factors : Elapsed CPU time  : "
	<< cpu_sumOfTimes << endl;
   cout << "Calculation of all alpha factors : Elapsed real time : "
	<< day_sumOfTimes << endl;
   cout << "Calculation of all alpha factors : Required memory : "
	<< mem_sum << " KB" << endl;
   cout << endl;


   /***********************************************************
   *** call calculation of stationary iteration vector u(M) ***
   ***********************************************************/

   /*** get start time and memory ***/
   cpu_startTime = Time::getCPUTime();
   day_startTime = Time::getTimeOfDay();
   mem_atStart = getUsedMem();


   u_stat = outer_iteration(SB);


   /*** get used time and memory ***/
   cpu_diffTime = Time::getCPUTime() - cpu_startTime;
   day_diffTime = Time::getTimeOfDay() - day_startTime;
   mem_diff = getUsedMem() - mem_atStart;

   cout << endl;
   cout << "Calculation of u(M) : Elapsed CPU time  : "
	<< cpu_diffTime << endl;
   cout << "Calculation of u(M) : Elapsed real time : "
	<< day_diffTime << endl;
   cout << "Calculation of u(M) : Required memory : "
	<< mem_diff << " KB" << endl;
   cout << endl;


   /********************************************************
   *** calculation of sums of stationary a(M)s and b(M)s ***
   ********************************************************/

#  ifdef SHOW_ALL_TIMES
      cpu_startTime = Time::getCPUTime();
      day_startTime = Time::getTimeOfDay();
      mem_atStart = getUsedMem();
#  endif

   /*** for each general transition ***/
   for(GenMarkings::iterator i = SB.Gen.begin(); i != SB.Gen.end(); ++i)
   {
      /*** current transition ***/
      GenMarkingsSmc  &gen = *i;

      /*** call jensen's method ***/
      //gen.jensens_method_a_b((u_stat * gen.I), SB.Epsilon,
      //                        gen_a_stat, gen_b_stat);
      gen.jensens_method_a_b((u_stat * gen.I), gen_a_stat, gen_b_stat);

      /*** renormalize tau dependency ***/
      gen_b_stat *= gen.D_tau;

      /*** get sum of all gen_a(M) ***/
      Gen_a_stat += gen_a_stat;

      /*** get sum of all gen_b(M) ***/
      Gen_b_stat += gen_b_stat;

#     ifdef DEBUG_GEN_STAT
         cout << "Current transition : " << gen.name << endl;
         cout << "<stat> gen_a_stat(M) : " << gen_a_stat << endl;
         cout << "<stat> gen_b_stat(M) : " << gen_b_stat << endl;
#     endif
   }

#  ifdef DEBUG_GEN_STAT
      cout << "<stat> sum of all gen_a_stat(M) :" << Gen_a_stat << endl;
      cout << "<stat> sum of all gen_b_stat(M) :" << Gen_b_stat << endl;
#  endif


   /***********************************************************
   *** calculate state probabilities and firing frequencies ***
   ***********************************************************/

   SB.calculate_pi_phi(u_stat, Gen_a_stat, Gen_b_stat);

#  ifdef SHOW_ALL_TIMES
      cpu_diffTime = Time::getCPUTime() - cpu_startTime;
      day_diffTime = Time::getTimeOfDay() - day_startTime;
      mem_diff = getUsedMem() - mem_atStart;

      cout << endl;
      cout << "Calculation of a(M), b(M), pi and phi : Elapsed CPU time  : "
           << cpu_diffTime << endl;
      cout << "Calculation of a(M), b(M), pi and phi : Elapsed real time : "
           << day_diffTime << endl;
      cout << "Calculation of a(M), b(M), pi and phi : Required memory : "
           << mem_diff << " KB" << endl;
      cout << endl;
#  endif

}



