// File    solve_slave.cpp   (C++ Implementation)
// Program slaveGen          (Slave-Program of parallel FIA Analysis)
/**
 *@author  Philipp Berndt
 *@version $Id: solve_slave.cpp,v 1.2 2000/01/12 16:02:33 pumphutt Exp $
 */

//#define DEBUG
//#define DEBUG_U

#include <iostream>
#include <unistd.h> // close, dup 

#include "pvm3cpp.h"
#include "Time.h"
#include "solve_parallel.h"
#include "solve_gen.h"

// The following variables are initialized from PVM receivebuffer,
// so don't mess with the line order!
namespace
{
    const long     mem_at_start  = getUsedMem();    
    const int      mytid         = pvm_mytid();   // Enroll in pvm
    const int      master        = pvm_parent();  // TID of master
//  const int      oldroute      = pvm_setopt(PvmRoute, PvmDontRoute);
    int            bufid         = pvm_recv(master, TAG_INITGEN);
    GenMarkingsSmc g             = GenMarkingsSmc::pvm_unpack();
    const int      ArbBits       = pvm_upkint();
    const double   TruncError    = pvm_upkdouble();
    const double   mu            = pvm_upkdouble(); // AperiodicFact
    const double   mu_bar        = 1.0 - mu;
    const int      dim           = g.getDim();
    Time           outer_start_cpu;
    Time           outer_time_cpu;
    int            cnt           = 0;               // Iteration counter
}

/** Function to be called by PVM when master dies unexpectedly. */
int masterDiedMessageHandler(int mid)
{
    cerr << "Master died. Bailing out." << endl;
    exit(1);
}

/**
 * Calculates u^g'(m+1) and sends it to master.
 * Use this if Q_bar is not empty.
 */
void handle_u_with_b()
{
    // Unpack the u-vector from receive buffer
    SparseVector u(PVM_UNPACK);       // Receive our part
	    
#ifdef DEBUG_U
    cout << "Received u:\t" << u << endl;
#endif

    // Calculate vector gen.a(m) 
    SparseVector a(dim), b(dim);

    g.jensens_method_a_b(u, a, b);

    // Calculate our part for the new global u
    SparseVector u_new  = a * g.DELTA;
                 u_new += b * g.Q_bar;

                 u_new *= mu;
                 u_new += (u *= mu_bar);

    u_new.pvm_send_frag(master, TAG_U_NEW);		 
}

/**
 * Calculates u^g'(m+1) and sends it to master.
 * Optimized for empty Q_bar.
 */
void handle_u_without_b()
{
    // Unpack the u-vector from receive buffer
    SparseVector u(PVM_UNPACK);       // Receive our part
	    
#ifdef DEBUG_U
    cout << "Received u:\t" << u << endl;
#endif

    // Calculate vector gen.a(m) 
    SparseVector a(dim);

    g.jensens_method_a(u, a);

 		// Calculate our part for the new global u

    SparseVector u_new  = a * g.DELTA;

                 u_new *= mu;
                 u_new += (u *= mu_bar);

    u_new.pvm_send_frag(master, TAG_U_NEW);		 
}

/** Computes a^g(M) and b^g(M) and sends it to master. */
void handle_u_stat()
{
#ifdef DEBUG
    cout << "Received U_STAT message." << endl;
#endif
    outer_time_cpu = Time::getCPUTime() - outer_start_cpu;
	    
    // Unpack the u^g from receive buffer
    SparseVector u_stat(PVM_UNPACK);

    // Calculate a and b
    SparseVector a_stat(dim), b_stat(dim);
    g.jensens_method_a_b(u_stat, a_stat, b_stat);

    b_stat *= g.D_tau;    // Renormalize tau dependency 

    pvm_initsend(PvmDataRaw);
    a_stat.pvm_pack();
    b_stat.pvm_pack();
    pvm_send(master, TAG_A_B);
}

/** Sends CPU_TIME-message to master and exits. */
void handle_terminate()
{
#ifdef DEBUG
    cout << "Received TERMINATE message." << endl;
#endif
    // Report CPU-Time
    pvm_initsend(PvmDataDefault);
    pvm_pkint(cnt);
    outer_time_cpu.pvm_pack();
    long data_mem = getUsedMem() - mem_at_start;
    pvm_pklong(data_mem);
    pvm_send(master, TAG_CPU_TIME);

    // Leave PVM
    pvm_exit();
	    
    // Exit gracefully
    exit(0);
}


/************************************ Main ***********************************/

int main()
{
    // Redirect stderr to stdout: 2>&1
    close(STDERR_FILENO);
    dup(STDOUT_FILENO);

#ifdef DEBUG
    cout << "SlaveGen started" << endl;
    cout << "Name: " << g.name << endl;
    cout << "pmf: " << g.pmf_def << endl;
//    cout << "I:\n" << g.I << endl;
//    cout << "A:\n" << g.A << endl;
//    cout << "Q_bar:\n" << g.Q_bar << endl;
//    cout << "DELTA:\n" << g.DELTA << endl;
#endif

    // Terminate whenever Master dies unexpectly
    pvm_addmhf(-1, TAG_MASTER_DIED, PvmBaseContext, masterDiedMessageHandler);
    pvm_notify(PvmTaskExit, TAG_MASTER_DIED, 1, &master);
      
    // Calculate alpha-factors and errors
    g.calculate_alphas(TruncError, ArbBits);

    void (*handle_u)(void);
    if (g.Q_bar.empty())
	handle_u = handle_u_without_b;
    else
	handle_u = handle_u_with_b;

    // Main iteration loop
#ifdef DEBUG
    cout << "Entering main iteration loop" << endl;
#endif
    outer_start_cpu = Time::getCPUTime();
    while (true)
	{
	// Receive next message. If more than one message has arrived
	// discard all but the last one.
	int bufid, bytes, msgtag, tid;
	do {
	    bufid = pvm_recv(master, -1);

	    pvm_bufinfo(bufid, &bytes, &msgtag, &tid);
	    } while(pvm_probe(master, -1) && msgtag != TAG_TERMINATE);

	switch (msgtag)
	    {
	    case TAG_U:
		handle_u();
		cnt++;
		break;
	  
	    case TAG_U_STAT:
		handle_u_stat();
		break;

	    case TAG_TERMINATE:
		handle_terminate();
	    } // switch
	} // while
}


