/******************************************************************************
 Program:           erandfilt.c

 Parameters:        The complete pathname of a DSPN.

 Purpose:           This program reads in the files written by execute_rand
                    when called with the '-f' option. These files contain the
                    Markov Chains, Delta- and Q-matrices of the transitions of
                    the DSPN. This data is formated for better readability.
                    The files read are .mdetXX, .qxx, and .deltaXX where XX
                    is the number of the transition that file represents.
                    File number 00 represents the exponential transitions.

 Project Director:  Reinhard German

 Programmer:        Ulrich Hertlein

 History:           Release 1: June 17, 1993
                    Release 2: January 5, 1994
******************************************************************************/


#include <stdio.h>
#include <fcntl.h>

#include "socket_service.h"


/*
 * Function:        usage
 *                  print program usage
 *
 * Arguments:       none
 *
 * Result:          none
 */
static int usage()
{
   fprintf(stderr,"Usage: erandfilt [-i] [-h] filename ...\n");
   fprintf(stderr,"\t-i\tignore files that don't exist\n");
   fprintf(stderr,"\t-h\tthis help message\n");

   exit( 1 );
}


/*
 * Function:        read_MDET
 *                  read the MDET set of a transition from a file
 *
 * Arguments:       a file descriptor,
 *                  number of the transition (zero = exp.transitions)
 *
 * Result:          none
 */
static int read_MDET(fd, num)
int fd, num;
{
   unsigned ver, mc;
   long i;
   unsigned char flag;

   read(fd,&ver,sizeof(ver));
   printf("Protocol version = %d\n", ver);

   read(fd,&mc,sizeof(mc));
   if( mc != num )
   {
      printf("File number doesn't match Markov Chain number\n");
      exit( 1 );
   }

   if( mc )
   {
      double tau;

      if( ver >= 2 )
      {
         char type;
         unsigned len;
         char nam[256];

	 read(fd,&type,sizeof(type));
	 printf("Type = %c\n", type);
	 
         read(fd,&len,sizeof(len));
         if( len )
	 {
            read(fd,nam,len);
            nam[len] = '\0';

            printf("Name = %s (%d)\n", nam,len);
	 }
      }

      read(fd,&tau,sizeof(tau));
      printf("Tau = %.4lf\n", tau);
   }

   printf("( ");
   do
   {
      read(fd,&i,sizeof(long));

      if( i )
      {
         printf("%d ", i);

         if( mc )
            read(fd,&flag,sizeof(flag));
      }
   } while( i );
   printf(")\n\n");

   return( 0 );
}


/*
 * Function:        read_Q
 *                  read the contents of the Q matrix from a file
 *
 * Arguments:       a file descriptor,
 *                  number of the transition
 *
 * Result:          none
 */
static int read_Q(fd, num)
int fd, num;
{
   int ok = 1,n,i;

   Q_tau_row_t qtrow;
   Q_row_t qrow;
   Q_column_t qcol;

   while( ok )
   {
      if( num )
      {
         ok = read(fd,&qtrow,Q_TAU_ROW_SIZE);

         if( ok )
            printf("Row index = %d, Neg-Sum = %.4lf, Tau = %.4lf, Entries = %d\n", qtrow.row_index,qtrow.neg_sum,qtrow.tau,qtrow.ncolumn);
         n = qtrow.ncolumn;
      }
      else
      {
         ok = read(fd,&qrow,Q_ROW_SIZE);

         if( ok )
            printf("Row index = %d, Neg-Sum = %.4lf, Entries = %d:\n", qrow.row_index,qrow.neg_sum,qrow.ncolumn);
         n = qrow.ncolumn;
      }

      for( i = 0; ok && i < n; i++ )
      {
         read(fd,&qcol,Q_COLUMN_SIZE);
         printf("(%d, %.4lf) ", qcol.column_index,qcol.rate);
      }

      printf("\n");
   }

   return( 0 );
}


/*
 * Function:        read_DELTA
 *                  read the contents of the delta matrix of a transition
 *                  from a file
 *
 * Arguments:       file descriptor
 *
 * Result:          none
 */
static int read_DELTA(fd)
int fd;
{
   int i;

   delta_head_t dhead;
   delta_row_t drow;
   delta_column_t dcol;

/*
   read(fd,&dhead,sizeof(dhead));
   printf("Tau = %.4lf\n", dhead.tau);
*/

   while( read(fd,&drow,DELTA_ROW_SIZE) )
   {
      printf("Row index = %d: \n", drow.row_index);

      for( i = 0; i < drow.ncolumn; i++ )
      {
         read(fd,&dcol,DELTA_COLUMN_SIZE);
         printf("(%d, %.4lf) ", dcol.column_index,(double )dcol.prob);
      }
      printf("\n");
   }

   return( 0 );
}


/*
 * Function:        main
 *
 * Arguments:       base filename of a model (e.g. ~/Models/Fcs/fcs2)
 *
 * Result:          none
 */
int main(argc, argv)
int argc;
char **argv;
{
   int c;
   extern char *optarg;
   extern int optind;

   int errflag = 0, ignore = 0;

   int num,ok;
   char mdet_file[256], q_file[256], delta_file[256];
   int fd;

   /* parse arguments */
   while( (c = getopt(argc,argv,"ih")) != -1 )
      switch( c )
      {
      case 'i': if( ignore )
                   errflag++;
                else
                   ignore++;
                break;
      case 'h':
      case '?': errflag++;
      }
   if( optind >= argc || errflag )
      usage();

   /* parse file list */
   num = 0;
   for(; optind < argc; optind++ )
   {
      ok = 1;

      printf("CONTENTS OF DSPN %s\n\n", argv[optind]);

      do
      {
         /* build file names */
         sprintf(mdet_file,"%s.mdet%02u", argv[optind],num);
         sprintf(q_file,"%s.q%02u", argv[optind],num);
         sprintf(delta_file,"%s.delta%02u", argv[optind],num);

         /* read Markov Chain */
         if( (fd = open(mdet_file,O_RDONLY)) < 0 )
         {
            if( !ignore && num )
               ok = 0;
         }
         else
	 {
            printf("Markov Chain of %d:\n", num);
            read_MDET(fd);
            close(fd);
	 }

         /* read Q matrix */
         if( (fd = open(q_file,O_RDONLY)) < 0 )
         {
            if( !ignore && num )
               ok = 0;
         }
         else
	 {
            printf("Q Matrix of %d\n", num);
            read_Q(fd);
            close(fd);
	 }

         /* read Delta matrix */
         if( num )
         {
            if( (fd = open(delta_file,O_RDONLY)) < 0 )
            {
               if( !ignore )
                  ok = 0;
            }
            else
	    {
               printf("Delta Matrix of %d:\n", num);
               read_DELTA(fd);
               close(fd);
	    }
         }
         printf("\n");

         num++;
      }
      while( ok );
   }

   return( 0 );
}
